#!/usr/local/mascot/perl64/bin/perl
##############################################################################
# Create a Mascot master results report                                      #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 1998-2019 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################
#    $Archive:: /www/cgi/master_results.pl                                 $ #
#     $Author: robertog@matrixscience.com $ #
#       $Date: 2022-05-16 14:27:02 +0100 $ #
#   $Revision: 2248db7aed94d66543e99125f32fa6d43c1e70b4 | MASCOT_REL_2_8_1-3-gcae025d431 $ #
# $NoKeywords::                                                            $ #
##############################################################################
 use strict;                                                                #
 use warnings;
##############################################################################

# The following (case insensitive) URL arguments are recognised:
#
# reptype                   - 'protein' or 'peptide' or 'archive' or 'concise' or 'select' or 'unassigned'
# file                      - relative path to result file for all reports except archive
# intermediate              - relative path to result file for archive report
# include                   - list of hit numbers for archive report
# report                    - max number of hits to be reported, (0 = AUTO)
# unigene                   - UniGene index species to be used to cluster hits
# _showsubsets              - display protein hits that are missing up to this fraction of the protein score of the main hit
# _requireboldred           - 1 to report protein hits only if they include at least one bold, red peptide match
# _showallfromerrortolerant - 1 to display all matches from an error tolerant search, including garbage
# _onlyerrortolerant        - 1 to display only error tolerant matches from a new style error tolerant search
# _noerrortolerant          - 1 to display no error tolerant matches from a new style error tolerant search
# _sigthreshold             - probability significance threshold, default 0.05
# _sortunassigned           - sort order for unassigned: 'scoredown' or 'queryup' or 'intdown'
# _ignoreionsscorebelow     - cut off on MS/MS ions scores (-1 for "automatic" setting)
# _showpopups               - 'TRUE' to force appearance of yellow 'top 10' pop-ups, 'FALSE' to suppress
# _alwaysgettitle           - 1 to always fetch titles from ms-getseq for non-primary hits
# _show_decoy_report        - 1 to display report for the integrated decoy search (Mascot 2.2. & later)
# _server_mudpit_switch     - if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
# _querylist                - which queries to re-search: 'all' or 'unassigned' or 'homology' or 'identity' 
# percolate                 - 1 to display scores and expect values based on Percolator PEPs  
# percolate_rt              - 1 to include retention time feature when running Percolator
# _prefertaxonomy           - integer index > 0 to ../config/taxonomy for selecting a preferred taxonomy
#
# 'Global' variables
  my(
    @accession_list,           # list of primary hit accession strings, index is hit number
    %et_accession_list,        # list of accession strings used to create JavaScript array for error tolerant search
    @errorTolDiv,              # spooled DIV elements for error tolerant pop-ups
    $fileIn,                   # relative path to result file, passed as required URL argument
    @headerAccession,          # list of significant protein hit accessions for report header, index is hit number
    @headerDescription,        # list of significant protein hit descriptions for report header, index is hit number
    @headerQuant,              # quantitation summary information for report header, index is hit number
    @headerText,               # top scoring protein hit for protein summary report header
    $ignoreIonsScoreBelow,     # cut off on MS/MS ions scores
    $ignoreIonsScoreBelowDefined, # if present in the URL
    @includeHit,               # $includeHit[$hitNum] set to 1 if $hitNum was in @includeList
    @includeList,              # list of hit numbers for archive report
    @indexText,                # text for protein summary 'index' block, index is hit number
    $massDP,                   # number of decimal places in formatted mass values
    $minPepLen,                # peptides shorter than $minPepLen are ignored when generating Peptide summary 
    $msresFlags,               # flags used to create Mascot Parser summary object
    $msresFlags2,              # flags2 used to create Mascot Parser summary object
    $mudpitSwitch,             # if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
    $selectSwitch,             # if number of queries greater than this value, switch to a Select summary, default 1000
    $myName,                   # file name of this script for links to self (not complete path)
    $numHits,                  # max number of hits to be reported
    $objDatFile,               # Mascot parser mascot.dat file object
    $objMascotOptions,         # Mascot parser configuration options object
    $objQuantMethod,           # Mascot parser quant_method object
    $quantProtocolType,        # REPORTER, MULTIPLEX, PRECURSOR
    @reportRatioNames,         # If quantitation is in use, list of report ratio names
    $objParams,                # Mascot parser searchparams object
    $objResFile,               # Mascot parser result file object
    $objSummary,               # Mascot parser result summary object (may be protein or peptide)
    $objMS2Quant,              # Mascot parser quantitation data object (or undef if file is not Reporter or Multiplex)
    $objUnigeneOptions,        # Mascot parser configuration unigene object
    %pepRankByHQ,              # Used by overview table: keys are hn_qm strings, values are peptide ranks
    %pepSeqByHQ,               # Used by overview table: keys are hn_qm strings, values are peptide sequences
    $percolate,                # 1 to display scores and expect values based on Percolator PEPs
    $percolate_rt,             # 1 to include retention time feature when running Percolator
    $preferTaxonomy,           # index > 0 to select a preferred taxonomy
    @quantFormatControlText,   # contains the HTML text for the quant format controls
    @quantWarnings,            # contains the HTML text for any warnings returned by quant routines
    @query_title_list,         # list of query titles, HTML escaped, index is query number
    $reportType,               # 'protein' or 'peptide' or 'archive' or 'concise' or 'select' or 'unassigned'
    $RequireBoldRed,           # 1 to report protein hits only if they include at least one bold, red peptide match
    $sec_peptides,             # SEC_PEPTIDES or SEC_DECOYPEPTIDES 
    $sec_summary,              # SEC_SUMMARY or SEC_DECOYSUMMARY 
    @selectList,               # element set to 1 if this query appears in select report, used by &toolText()
    $session,                  # Mascot security session object
    $sessionID,                # Mascot security sessionID
    $shipper,                  # flag set to 'TRUE' if this is Mascot for intranet, 'FALSE' for public internet site
    $ShowDecoyReport,          # 1 to display report for the integrated decoy search (Mascot 2.2. & later)
    $ShowAllFromErrorTolerant, # 1 to display all hits from an error tolerant search, including garbage
    $OnlyErrorTolerant,        # 1 to display only error tolerant matches from a new style error tolerant search
    $NoErrorTolerant,          # 1 to display only standard matches from a new style error tolerant search
    $OldStyleErrTolReport,     # 1 to suppress certain parts of the report as per old-style error tolerant search 
    $OmitErrTolThresholds,     # 1 to suppress displaying thresholds for ET queries in results files from before 2.8
    $showOverview,             # boolean flag to indicate whether overview table is required
    $ShowSubSets,              # display protein hits that are missing up to this fraction of the protein score of the main hit
    $showToolTips,             # 1 to display yellow tooltip pop-ups showing top ranking peptide matches
    $sigThreshold,             # probability significance threshold (OneInXprobRnd)
    $sortUnassigned,           # sort order for unassigned: 'scoredown', 'queryup', or 'intdown'
    @spoolText,                # spool file for main body of summary report
    $thisScript,               # CGI object
    $UniGeneFile,              # path to a UniGene index file
    %urlParams,                # keys are URL parameter names in lower case, values are names in original case
    $anyPepSumMatch,           # $objResFile->anyPeptideSummaryMatches($sec_peptides)
    $anyLibraryMatch,          # $objResFile->anyPeptideSummaryMatches(SEC_LIBRARYPEPTIDES)
    @peptideViewFromTo,        # 'from' and 'to' arguments to peptide_view.pl
    %dupe_state_cache,         # cache for duplicate_group_main()
  );
  
# Constants
  my $pepViewLink = "peptide_view.pl";
  my $protViewLink = "protein_view.pl";
  my $scoreGifLink = "score_gif.pl";
  my @colourKey;               # for a protein summary of a PMF mixture, components are colour coded
  $colourKey[1] = "#0000ff";   # blue
  $colourKey[2] = "#ff0000";   # red
  $colourKey[3] = "#008080";   # teal
  $colourKey[4] = "#800080";   # purple
  $colourKey[5] = "#ff00ff";   # fuchsia
  $colourKey[6] = "#008000";   # green
  
# set current directory to script location for FastTrack on NT
# chdir also handles a Windows drive letter change
  if ($ENV{'WINDIR'} && $0 =~ /(.*)[\\\/]/) {
    chdir($1);
  }

 # required Perl packages
  use CGI;
  use Scalar::Util qw(looks_like_number);
  use HTTP::Request::Common;
  use LWP::UserAgent;
  use msparser; 
  use Util::URI qw(uri_unescape);
  use Util::Peptides qw(possible_etmod_assignments etmodified_termini_and_residues);
  use PeptideSummary::Backend qw(duplicate_group_main);
  use MasterResults::ProteinFamily::ReportFunc qw(taxonomy_choice);
  use PeptideSummary::Taxonomy;
  use PeptideSummary::Util;
  use MascotResults;
  use ExportDat::ProteinData;
  use Util::IPC qw(pipefork);

  $| = 1; # set autoflush on STDOUT

# print HTTP header so that the browser will display fatal error messages
  $thisScript = new CGI;
  print $thisScript->header( -charset => "utf-8" );
# IE 9 has a bug in quirks mode: if the window width is below a certain unknown
# threshold, the browser may show a blank white page, or it may disable the
# scrollbar. This only happens with certain files. The easiest workaround is to
# define a valid HTML 4.01 DOCTYPE, which will put the page in at least IE
# 7 rendering mode. Printing a DOCTYPE cannot have any adverse effect on
# any browser, since they will have to assume it is HTML 4.01 in any case.
  print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">
EOF
 
# pull in some common subroutines
  require "./common_subs.pl";
  require "./quant_subs.pl";
  
# www.pl contains the code to decrypt result file names
# its existence indicates that this script is running on the public web site
  if (-e "./www.pl") {
    do "./www.pl";
    $shipper = 'FALSE';
  } else {
    $shipper = 'TRUE';
  }

# get the name of this script for creating URL links to self
#  $myName = $thisScript->url(-relative=>1);
  ($myName) = $0 =~ /([^\\\/]*)$/;

# taint mode requires path to be set explicitly
# Unix path may be required to find decompression utilities in &common_subs::decompress()
  if ($ENV{'WINDIR'}){
    $ENV{"PATH"} = "";
  } else {
    $ENV{"PATH"} = "/usr/local/sbin:/usr/sbin:/sbin:/usr/local/bin:/usr/bin:/bin";
  }
# to bypass taint mode check on the path, comment out above lines and uncomment following line
# ($ENV{"PATH"}) = $ENV{"PATH"} =~ /(.*)/;

# create hash for URL arguments, so that we can always use lower case keys
  foreach ($thisScript->param) {
    if ($urlParams{lc($_)} && $urlParams{lc($_)} ne $_) {
      &common_subs::fatal("URL parameter names must use consistent case", __LINE__, $myName, "");
    }
    $urlParams{lc($_)} = $_;
  }

# check report type requested on URL first, because this determines where to look for data file name
  if (lc($thisScript->param($urlParams{reptype}) || '') eq 'archive') {
    $fileIn = $thisScript->param($urlParams{intermediate}) 
      || &common_subs::fatal("No intermediate result file name supplied", __LINE__, $myName, "");
  # an archive report must include at least one hit - die if it doesn't
  # Bug 12292: Perl 5.20 and later require using CGI::multi_param() in list
  # context, but this method doesn't exist in Perl 5.16 or earlier. So
  # we need to detect it first.
    if ($thisScript->can('multi_param')) {
      @includeList = $thisScript->multi_param($urlParams{include});
    } else {
      @includeList = $thisScript->param($urlParams{include});
    }
    if (@includeList) {
      foreach (@includeList) {
        $includeHit[$_] = 1;
      }
    } else {
      &common_subs::fatal("No hits specified for archive report", __LINE__, $myName, "");
    }
  } else {
    $fileIn = $thisScript->param($urlParams{file})
      || &common_subs::fatal("No result file name supplied", __LINE__, $myName, "");
  }

# verify the result file exists, decompress if necessary
# untaint the file path. This regex may need fine tuning
  ($fileIn) = $fileIn =~ /([\/\w\-\+.$%&()[\]:]+)/;
  if ($fileIn) {
    unless (-e &common_subs::decompress($fileIn)) {
      &common_subs::fatal("result file does not exist", __LINE__, $myName, $fileIn);
    }
  } else {
    &common_subs::fatal("result filename could not be untainted", __LINE__, $myName, "");
  }

# create an configuration options object from mascot.dat
  $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objDatFile->isValid) {
    common_subs::fatal(common_subs::checkErrorHandler($objDatFile), __LINE__, $myName, "");
  }
  $objMascotOptions = $objDatFile->getMascotOptions;
  $objUnigeneOptions = $objDatFile->getUniGeneOptions;

# see whether to use ms_mascotresfile::RESFILE_USE_CACHE flag
  my $resfileFlags;
  if (common_subs::useResfileCache($objMascotOptions, $myName)) {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_USE_CACHE;
  } else {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_NOFLAG;
  }
  my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;

# create Mascot Parser result file object
#  print "<!-- begin create result file object at " . (times)[0] . " seconds -->\n"; ### TIMING ###
  $objResFile = new msparser::ms_mascotresfile(&common_subs::decompress($fileIn), 
                                                5, 
                                                "<!-- %d seconds -->\n", 
                                                $resfileFlags, 
                                                $resfileCacheDirectory);
#  print "<!-- end create result file object at " . (times)[0] . " seconds -->\n"; ### TIMING ###

# abort on fatal error. Note that isValid() remains true if there are only warnings
  unless ($objResFile->isValid) {
    # Ignore warnings about old unsupported results.
    my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
    common_subs::fatal($errorList, __LINE__, $myName, $fileIn);
  }
  
# delete the flag file, if it exists, to show that user has seen result report
  unlink common_subs::decompress($fileIn) . ".flg";
# also delete the batch log file
  unlink common_subs::decompress($fileIn) . ".log";

# Bug 12182: Redirect to master_results_2.pl if there are any spectral library
# matches. Ideally, this redirect should be done using a proper HTTP header,
# but we can't do so; the ms_mascotresfile constructor leaks keepalive 
# messages, so an HTTP header must have been printed already. We can only
# redirect using Javascript.
#
# Bug 13134: Same treatment with crosslinked searches.
  if ($objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES)
      || $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_CROSSLINK_PEPTIDES)) {
    my $master_results_2_url = 'master_results_2.pl?'.$thisScript->query_string;
    print <<EOF;
<HTML>
<HEAD><TITLE>Redirect...</TITLE></HEAD>
<SCRIPT LANGUAGE="JavaScript">
<!-- Begin hiding Javascript from old browsers.
window.location = "$master_results_2_url";
// End hiding Javascript from old browsers. -->
</SCRIPT>
<BODY>
<P>Please view the <A HREF="$master_results_2_url">Protein Family Summary</A>.
</BODY>
</HTML>
EOF
    exit;
  }

# create a Mascot Parser searchparams object
  $objParams = new msparser::ms_searchparams($objResFile);
  
# Security
  $sessionID = "" unless ($sessionID = $thisScript->param($urlParams{sessionid}));
  $session = new msparser::ms_session($sessionID);
  if ($session->isValid) {
    unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
      print common_subs::notAuthorised($session->getUserName);
      exit 1;
    }
    $sessionID = $session->getID();
  } else {
    &common_subs::printHeader1("Matrix Science - Not logged in");
    &common_subs::printHeader2("", "Not logged in", 1);
    &common_subs::go2login($thisScript->self_url, $session->getLastError, $session->getLastErrorString);
    &common_subs::printFooter;
    print "</BODY>\n";
    print "</HTML>\n";
    exit 1;
  }

# validate a variety of mascot.dat and URL parameters
  &validateParams();
  
# if this is an old-style error tolerant search, may need to decompress the parent
  if ($objResFile->isErrorTolerant()) {
    if ($objParams->getErrTolParentFilename()) {
      &common_subs::decompress($objParams->getErrTolParentFilename());
    }

    # create indexes to display possible mod assignments
    my $mass_type = $msparser::MASS_TYPE_MONO;
    if (lc($objParams->getMASS()) eq "average") {
      $mass_type = $msparser::MASS_TYPE_AVE;
    }
    eval {
      Util::Peptides::init_etmod_index($mass_type, $sessionID);
    };

    if ($@) {
      common_subs::fatal(''.$@, __LINE__, $myName, $fileIn);
    }
  # set value for $OldStyleErrTolReport
    if ($ShowAllFromErrorTolerant || $OnlyErrorTolerant) {
      $OldStyleErrTolReport = 1;
    } elsif (!$objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      $OldStyleErrTolReport = 1;
    } else {
      $OldStyleErrTolReport = 0;
    }
  } else {
    $OldStyleErrTolReport = 0;
  }

# Bug 11198: If _sigthreshold is not defined as a URL parameter, and
# results file has first_pass_threshold, set $sigThreshold to it.
#
# FIXME: Use get_ms_mascotresults_params(), which sets it automatically.
  do {
    my $str = $thisScript->param($urlParams{'_sigthreshold'});
    if (not defined $str or $str eq '') {
      my $first_pass_threshold = $objResFile->getSectionValueDouble($msparser::ms_mascotresfile::SEC_HEADER, 'first_pass_threshold');
      if ($first_pass_threshold > 0) {
        # validateParams() has set $sigThreshold to a default,
        # which we can override here.
        $sigThreshold = $first_pass_threshold;
      }
    }
  };

# determine whether to display tooltip pop-ups
# can force either way using URL argument _showpopups
# if running as CGI application, require at least version 4 browser
  if ($thisScript->param($urlParams{_showpopups})
    && $thisScript->param($urlParams{_showpopups}) =~ /TRUE/i) {
    $showToolTips = 1;
  } elsif ($thisScript->param($urlParams{_showpopups})
    && $thisScript->param($urlParams{_showpopups}) =~ /FALSE/i) {
    $showToolTips = 0;
  } else {
    if (&testBrowser() || !$ENV{'SERVER_NAME'}) {
      $showToolTips = 1;
    } else {
      $showToolTips = 0;
    }
  }
  
# all reports except concise PMF require complete set of query titles
  unless ($reportType eq 'concise' && $objResFile->isPMF) {
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
      $query_title_list[$queryNum] = $objResFile->getQuerySectionValueStr($queryNum, 'title');
    }
  }

# should this report include additional quantitation info?
# if so, $objQuantMethod will encapsulate the quantitation method
# report can only handle multiplex and reporter protocols
  if ($objParams->getQUANTITATION() 
      and lc($objParams->getQUANTITATION()) ne "none") {
    my $method = msparser::ms_quant_method->new();

    if ($objResFile->getQuantitationMethod($method)) {
      $quantProtocolType = uc($method->getProtocol->getType());
      if ($quantProtocolType eq "MULTIPLEX" or $quantProtocolType eq "REPORTER") {
        $objQuantMethod = $method;
      # default report settings to those in the quantitation method
        quant_subs::quantFormatOverRide($objQuantMethod, $thisScript, \%urlParams, \$mudpitSwitch, \$sigThreshold, \$ShowSubSets, \$RequireBoldRed);
      # read any URL arguments and create text for format controls 
        quant_subs::quantFormatControls($objQuantMethod, \$thisScript, \%urlParams, \@quantFormatControlText);
      # warn if trying to use unsupported features
        quant_subs::quantWarnings($objQuantMethod, \@quantWarnings);
      }
      if ($quantProtocolType eq 'REPORTER') {
        my ($min, $max) = (1e99, -1);
        my $mass_type = lc($objParams->getMASS());
        # The 'from' and 'to' arguments to peptide_view.pl should be the
        # minimum and maximum repoter ion intensities.
        for my $i (0 .. $objQuantMethod->getNumberOfComponents()-1) {
          my $comp = $objQuantMethod->getComponentByNumber($i);
          next unless $comp->haveMoverz();
          my $m = $mass_type eq 'monoisotopic' ? $comp->getMoverz->getMonoisotopic + 0 : $comp->getMoverz->getAverage + 0;
          next unless $m > 0;
          if ($m < $min) {
            $min = $m;
          }
          if ($m > $max) {
            $max = $m;
          }
        }
        if ($min < $max) { # also handles the case $min == $max
          @peptideViewFromTo = ($min, $max);
        }
      }
    } else {
    # abort on fatal error. Note that isValid() remains true if there are only warnings
      if (not $objResFile->isValid()) {
        my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
        common_subs::fatal($errorList, __LINE__, $myName, $fileIn);
      }
    }
  }

  $quantProtocolType = '' if not defined $quantProtocolType;

  my %params_for_flags;

  if (common_subs::useResultsCache($objMascotOptions, $myName)) {
    $params_for_flags{'_usecache'} = 1;
  } else {
    $params_for_flags{'_usecache'} = 0;
  }

# create Mascot Parser summary object of the required type
#  print "<!-- begin create summary object at " . (times)[0] . " seconds -->\n"; ### TIMING ###
  if ($reportType eq 'protein') {
    $msresFlags = $msparser::ms_mascotresults::MSRES_MAXHITS_OVERRIDES_MINPROB;
    if ($objParams->getDECOY() > 0
      && $ShowDecoyReport) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_DECOY;
    }
    # don't use the new code at all
    %params_for_flags = (':override' => { msresflags => $msresFlags });
  } elsif ($reportType eq 'concise') {
    $params_for_flags{':override'}{'grouping'} = 'standard';
    $params_for_flags{':override'}{'scoring'} = 'standard';
    $params_for_flags{':override'}{'showsubsets'} = 1;
    $params_for_flags{'_threshold_type'} = 'identity';
    if ($objParams->getDECOY() > 0) {
      $params_for_flags{'_show_decoy_report'} = $ShowDecoyReport;
    }
  } elsif ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    # Note: don't use the ":override" hash in this branch. It's only meant
    # for legacy use.
    #
    # Also be careful about undefined values.
    # make_ms_mascotresults_params() assumes undef means "use default
    # from get_ms_mascotresults_params()", so make sure all parameters
    # set here are defined.
    $params_for_flags{'_showsubsets'} = $ShowSubSets;
    $params_for_flags{'group_family'} = 0;
    $params_for_flags{'_requireboldred'} = $RequireBoldRed;
    $params_for_flags{'_threshold_type'} = 'identity';
    if ($objResFile->isErrorTolerant()) {
      if ($objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      # new-style integrated ET search
        if ($OnlyErrorTolerant) {
          $params_for_flags{'et'} = 'onlyet';
        } elsif ($NoErrorTolerant) {
          $params_for_flags{'et'} = 'onlynormal';
        } else {
          $params_for_flags{'et'} = 'integrated';
        }
      }
      if ($ShowAllFromErrorTolerant) {
        $params_for_flags{'_showallfromerrortolerant'} = 1;
      }
    } elsif ($anyLibraryMatch) {
      #$params_for_flags{'library'} = $libraryMode;
    }
    if ($objResFile->getNumSeqsAfterTax > 0 and $objResFile->getNumQueries / $objResFile->getNumSeqsAfterTax > $mudpitSwitch) {
      $params_for_flags{'_scoring'} = 'mudpit';
    } else {
      $params_for_flags{'_scoring'} = 'standard';
    }
    if ($objParams->getDECOY() > 0) {
      $params_for_flags{'_show_decoy_report'} = $ShowDecoyReport;
    }
    $params_for_flags{'percolate'} = $percolate;
    if ($percolate == 1) {
      $objMascotOptions->setPercolatorUseRT($percolate_rt);
      $objResFile->setPercolatorFeatures(
          $objMascotOptions,
          '',
      );
      if (MascotResults::no_percolator_files($objResFile)) {
        _percolate($objResFile, 1, 0, $percolate_rt);
      }
    }
  } else {
    common_subs::fatal("Unrecognised report type: ".common_subs::htmlEscape($reportType), __LINE__, $myName, $fileIn);
  }

  do {
    my $pepsum_params = PeptideSummary::Util::make_ms_mascotresults_params($objResFile, %params_for_flags);
    $msresFlags = $pepsum_params->getFlags();
    $msresFlags2 = $pepsum_params->getFlags2();
  };

  if ($reportType eq 'protein') {
    $objSummary = new msparser::ms_proteinsummary(
      $objResFile, 
      $msresFlags, 
      $sigThreshold,
      $numHits
    );
  } elsif ($reportType eq 'concise') {
    $objSummary = new msparser::ms_proteinsummary(
      $objResFile, 
      $msresFlags, 
      $sigThreshold,
      $numHits
    );
  } else { #($reportType eq 'peptide' || $reportType eq 'archive' || $reportType eq 'select' || $reportType eq 'unassigned')
    $objSummary = new msparser::ms_peptidesummary(
      $objResFile, 
      $msresFlags,
      $sigThreshold, 
      $numHits, 
      $UniGeneFile,
      $ignoreIonsScoreBelow,
      $minPepLen,
      "",
      $msresFlags2
    );

    # Bug 13542: If results file is from 2.7 or earlier, or bug 12447
    # isn't fixed, use old behaviour: italicise ET peptide rows in the
    # query popup and don't show expect values.
    if ($objResFile->isErrorTolerant and $objResFile->versionGreaterOrEqual(2, 7, 101)
        and $objSummary->isDataCached($msparser::ms_peptidesummary::BUGFIX_12447)) {
      $OmitErrTolThresholds = 0;
    } else {
      $OmitErrTolThresholds = 1;
    }
  }
#  print "<!-- end create summary object at " . (times)[0] . " seconds -->\n"; ### TIMING ###
  
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objResFile->isValid) {
    my ($should_repeat, $errorList) = common_subs::checkResfileErrors($objResFile);
    if ($should_repeat) {
    # special treatment for old and unsupported result file 
      &pleaseRepeat();
      exit 0;
    }
    common_subs::fatal($errorList, __LINE__, $myName, $fileIn);
  }

  if ($objQuantMethod) {
    $objMS2Quant = msparser::ms_ms2quantitation->new($objSummary, $objQuantMethod);

    if (not $objMS2Quant->isValid) {
      common_subs::fatal(common_subs::checkErrorHandler($objMS2Quant), __LINE__, $myName, $fileIn);
    }

    if ($objMS2Quant->hasIntensityNormalisation) {
      $objMS2Quant->normaliseIntensities();
    }
    if ($objMS2Quant->hasPeptideRatioNormalisation) {
      $objMS2Quant->normalisePeptideRatios();
    }

    @reportRatioNames = @{ $objMS2Quant->getCombinedReportRatioNames() };
  }

  my $objUniGene;

  if ($UniGeneFile) {
    $objUniGene = msparser::ms_unigene->new($objResFile, $UniGeneFile);

    if (not $objResFile->isValid) {
      common_subs::fatal("Cannot create ms_unigene object: ".$objResFile->getLastErrorString(), __LINE__, $myName, $fileIn);
    }
  }

  my $preferredTaxonomyAvailable;
  
  if ($objParams->getDECOY() > 0) {
    if ($ShowDecoyReport) {
      $preferredTaxonomyAvailable = 0;
    } 
  }
  
  if (not defined $preferredTaxonomyAvailable) {
    my $tax_ids_available = PeptideSummary::Taxonomy::taxids_available(
      $objResFile, $objSummary, 
    );

    if ($tax_ids_available) {
      $preferredTaxonomyAvailable = 1;
    } else {
      my $status = PeptideSummary::Util::database_online_status($objParams, $objResFile, $sessionID);
      my $num_online = 0;
      for (keys %$status) {
        if ($$status{$_}) {
          ++$num_online;
        }
      }
      $preferredTaxonomyAvailable = $num_online > 0;
    }
  }

  my$protData = ExportDat::ProteinData->new(objSummary => $objSummary,
					    objResFile => $objResFile,
					    objDatFile => $objDatFile,
					    objUniGene => $objUniGene,
					    sessionID => $sessionID,
					    ShowSameSets => 1,
					    ShowSubSets => ($msresFlags & $msparser::ms_mascotresults::MSRES_SHOW_SUBSETS ? 1 : 0),
					    keepalive => 1,
					    include_every_protein => 0,
					    MIME_boundary => undef,
					    objParams => $objParams,
					    fastaTitles => undef,
					    populated => undef,
					    shipper => $shipper,

					    do_fetch => $objSummary->anyEmPAI(),
					    prot_len => 1,
					    prot_pi => 0,
					    prot_seq => 0,
					    needProteinDescriptionCache => 1,
					    needGetSeqCache => 1,
					    includeFrame => 0,

					    preferTaxonomy => ($preferredTaxonomyAvailable && $preferTaxonomy),
					    prot_tax_id => 1,
					    prot_tax_str => undef,
  );

  my $length_cache = $protData->getseq_cache;
  my $sameset_ordering = $protData->sameset_ordering;

# start of MAIN LOOP to enumerate the protein hits
  my $averageThreshold = $objSummary->getAvePeptideIdentityThreshold(1 / $sigThreshold);

  for my $thisHit (1 .. $objSummary->getNumberOfHits()) {
    last if $reportType eq 'unassigned';

  # skip proteins not selected for an archive report
    if ($reportType eq 'archive' && !$includeHit[$thisHit]) {
      next;
    }

    my $objHit = $objSummary->getHit($thisHit);
    next if not $objHit;
    
    my @sameset = common_subs::compileSameset($objSummary, $objHit, 1, $$sameset_ordering{$thisHit});
    my $objProtein = shift @sameset;

    my $preferredAcc = $objProtein->getAccession();
  
  # a list of accession numbers is needed to create a JavaScript array for the overview table
    if (not $objProtein->isPMFMixture()) {
      $preferredAcc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());
      $accession_list[$thisHit] = $preferredAcc;
    }
    
  # spool 'top hit' line for protein or concise summary header block
    if (($reportType eq 'protein' or $reportType eq 'concise')
      && $thisHit == 1) {
      push @headerText, "<B>Top Score       : "
        . sprintf("%.0f",$objProtein->getScore())
        . " for <FONT COLOR=#FF0000>"
        . common_subs::htmlEscape(accession_plus_prefix($objProtein))
        . "</FONT>, ";
      if ($objProtein->isPMFMixture()) {
      # PMF mixture description string contains list of component accession strings
        push @headerText, common_subs::htmlEscape(parse_PMF_mixture_accessions($objSummary, $objProtein));
      } else {
	push @headerText, common_subs::htmlEscape($protData->mustGetProteinDescription($preferredAcc, $objProtein->getDB, scalar $thisScript->param($urlParams{_alwaysgettitle})));
      }
      push @headerText, "</B>\n";
    } 
  
    if ($reportType ne 'protein' and $reportType ne 'concise') {
  # spool hit list / index for peptide or archive summary header block
      push @headerAccession, accession_plus_prefix($objProtein);
      push @headerDescription, $protData->mustGetProteinDescription($preferredAcc, $objProtein->getDB, scalar $thisScript->param($urlParams{_alwaysgettitle}));
    }

  # spool text for protein summary 'index' block
    if ($reportType eq 'protein') {
      push @indexText, "<TR><TD NOWRAP ALIGN=RIGHT><TT><B>$thisHit.</B></TT></TD>";
      push @indexText, "<TD NOWRAP><TT><B>&nbsp;<A HREF=\"#Hit$thisHit\">" 
        . common_subs::htmlEscape(accession_plus_prefix($objProtein))
        . "</A></B></TT></TD>";
      if ((my $m = $objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB)) > 0) {
        push @indexText, "<TD NOWRAP><TT>&nbsp;&nbsp;" 
          . sprintf("%.0f", $m)
          . "</TT></TD>";
      } else {
        push @indexText, "<TD NOWRAP>&nbsp;</TD>";
      }
      push @indexText, "<TD NOWRAP ALIGN=RIGHT><TT>"
        . sprintf("%.0f", $objProtein->getScore())
        . "</TT></TD>";       

      if ($objProtein->isPMFMixture()) {
      # PMF mixture description string contains list of component accession strings
        push @indexText, "<TD NOWRAP><TT>&nbsp;&nbsp;" . common_subs::htmlEscape(parse_PMF_mixture_accessions($objSummary, $objProtein)) . "</TT></TD></TR>\n";
      } else {
      # get rid of any leading space on description strings because it spoils alignment
	my $tempString = $protData->mustGetProteinDescription($preferredAcc, $objProtein->getDB, $thisScript->param($urlParams{_alwaysgettitle}));
        $tempString =~ s/^\s+//g;
        push @indexText, "<TD NOWRAP><TT>&nbsp;&nbsp;" . common_subs::htmlEscape($tempString) . "</TT></TD></TR>\n";
      }
    }

  # spool text for main body of report
    spoolProtMatch($objProtein, $thisHit, $length_cache, \@sameset, $$sameset_ordering{$thisHit});
  }
# end of MAIN LOOP to enumerate the protein hits

# Peptide summary may have an unassigned list
# Select summary also needs to call createUnassignedList to create query list for repeat search
  if (($reportType eq 'peptide' || $reportType eq 'unassigned' || $reportType eq 'select')
    && !$OldStyleErrTolReport) {
    if ($sortUnassigned eq "queryup"){
      $objSummary->createUnassignedList($msparser::ms_mascotresults::QUERY);
    } elsif ($sortUnassigned eq "intdown"){
      $objSummary->createUnassignedList($msparser::ms_mascotresults::INTENSITY);
    } else {
      $objSummary->createUnassignedList($msparser::ms_mascotresults::SCORE);
    }
  
  # if so, print it out
  # user can choose a sort order
    if ($reportType ne 'select' && $objSummary->getNumberOfUnassigned() > 0) {
      push @spoolText,  
        "<TT><B>Peptide matches not assigned to protein hits:</B> (no details means no match)</TT><BR>\n<BR>\n";
      
      tableHeader(undef);
    # enumerate the peptides in the unassigned list
      for (my $i = 1; $i <= $objSummary->getNumberOfUnassigned(); $i++) {
        my $objPeptide = $objSummary->getUnassigned($i);
        if (defined($objPeptide)) {
          push @spoolText, "<TR>";
          tableBody(undef, 0, $objPeptide->getQuery(), $i, $objPeptide, 0, 0);
          push @spoolText, "</TR>\n";
          $selectList[$objPeptide->getQuery()] = 1;
        } else {
          common_subs::fatal("Undefined peptide object", __LINE__, $myName, $fileIn);
        }
      }
      push @spoolText, "</TABLE>\n";
      unless ($reportType eq 'select' || $reportType eq 'unassigned') {
        push @spoolText, "\n<HR>\n";  
      }
    }
  }

#  print "<!-- processing complete at " . (times)[0] . " seconds -->\n"; ### TIMING ###
  
# print the HTML header block & Javascript
  print <<EOF;
<HTML>
<HEAD>
<META HTTP-EQUIV="X-UA-Compatible" CONTENT="IE=edge">
<LINK REL="shortcut icon" HREF="../favicon.ico">
EOF
  print "<TITLE>", ucfirst($reportType), " Summary Report (";
  if ($objParams->getCOM()) {
    print common_subs::htmlEscape($objParams->getCOM());
  } else {
    print common_subs::htmlEscape($fileIn);
  }
  print ")</TITLE>\n";
  outputJavaScript();
  print "</HEAD>\n\n";

# print first few HTML <BODY> statements
  outputBody();
  
# display decoy search stats if applicable
  if ($objParams->getDECOY() > 0
    && !$ShowDecoyReport
    && $reportType ne 'archive') {
    outputDecoy();        
  }

  if ($objResFile->isErrorTolerant and $objResFile->versionGreaterOrEqual(2, 7, 101)) {
    if ($anyPepSumMatch or $anyLibraryMatch) {
      print "<P STYLE='color: red'><B>Warning: The automatic error tolerant search uses the homology threshold and results should be viewed in Protein Family Summary.</B></P>\n";
    }
  }

# print link to display score histogram
  unless ($reportType eq 'select' || $reportType eq 'unassigned') {
    print "<H3>Mascot Score Histogram</H3>\n";
    if ($reportType eq 'protein' || $reportType eq 'concise') {
      print "Protein score is -10*Log(P), where P is the probability that ";
    } else {
      print "Ions score is -10*Log(P), where P is the probability that ";
    }
    print "the observed match is a random event.<BR>";
    if ($reportType eq 'protein' || $reportType eq 'concise') {
      print "Protein scores greater than "
        . int($objSummary->getProteinThreshold(1 / $sigThreshold))
        . " are significant (p&lt;$sigThreshold).<BR>";
    } elsif ($averageThreshold > 0
      && !$OldStyleErrTolReport) {
      if ($objSummary->getHomologyThresholdForHistogram(1 / $sigThreshold) > 0) {
        print "Individual ions scores &gt; "
          . int($objSummary->getHomologyThresholdForHistogram(1 / $sigThreshold))
          . " indicate peptides with significant homology.<BR>";
      }
      print "Individual ions scores &gt; "
        . int($averageThreshold)
        . " indicate identity or extensive homology";
      if ($objResFile->isErrorTolerant) {
        print ".<BR>"
      } else {
        print "(p&lt;$sigThreshold).<BR>"
      }
    }
    if ($anyPepSumMatch or $anyLibraryMatch) {
      print "Protein scores are derived from ions scores as a non-probabilistic basis for ranking protein hits.";
    }

  # Create a session object to save data required by score_gif.pl
  # Time to live is 5 mins, in case score_gif.pl fails to run
    my $StorageSession = new msparser::ms_session(300, "score_gif");
    if ($StorageSession->isValid) {
      $StorageSession->saveStringParam("formver", $objParams->getFORMVER());
      if ($reportType eq 'protein' || $reportType eq 'concise') {
        $StorageSession->saveIntParam("sigscore", int($objSummary->getProteinThreshold(1 / $sigThreshold)));
        $StorageSession->saveStringParam("xLegend", "Protein Score");
        $StorageSession->saveStringParam("yLegend", "Number of Hits");
      } else {
        $StorageSession->saveIntParam("sigscore", int($averageThreshold));
        $StorageSession->saveIntParam("plughole", int($objSummary->getHomologyThresholdForHistogram(1 / $sigThreshold)));
        $StorageSession->saveStringParam("xLegend", "Protein Score");
        $StorageSession->saveStringParam("yLegend", "Number of Hits");
      }
    # create list of scores for score_gif.pl
      my $scoreList = sprintf("%.2f", $objSummary->getProteinScoreForHistogram(1));
      my $i = 2;
      while ($i <= 50) {
        $scoreList .= sprintf(" %.2f", $objSummary->getProteinScoreForHistogram($i));
        $i++;
      }
      $StorageSession->saveStringParam("scores", '' . $scoreList);
      unless ($StorageSession->isValid) {
        common_subs::fatal("Unable to write to private session object", __LINE__, $myName, "");
      }
    } else {
      common_subs::fatal("Unable to create private session object", __LINE__, $myName, "");
    }

    print "<P><IMG SRC=\"$scoreGifLink?"
      . $StorageSession->getID
      . "\" WIDTH=400 HEIGHT=200 ALT=\"Score Distribution\">\n";
  }

# Print controls to re-format the report
  if ($reportType eq 'archive') {
    print "<H3>Archive Report of Selected Matches</H3>\n";
  } else {
    formatControls($preferredTaxonomyAvailable);
  }

# Print link for importing result into MI
  if (-e "./mi_manual_import.pl") {
    print "<P><A HREF=\"mi_manual_import.pl?datfile=", common_subs::urlEscape($fileIn), "&getSampleIdFromUser=true\">Import results into MI</A>";
  }

# begin form for repeating search
  print "<FORM ACTION=\"./search_form.pl\" NAME=\"Results\"";
  if (($reportType eq 'peptide')
    && ($anyPepSumMatch || $anyLibraryMatch)
    && &atLeastVersion(1,7,15)) {
    print " onSubmit=\"return makeAccessionString(this)\"";
  } elsif ($reportType eq 'select') {
    print " onSubmit=\"return chooseQueryList(this)\"";
  }
  print " METHOD=\"POST\">\n";

# optional all-singing, all-dancing overview table
  if ($showOverview && $objSummary->getNumberOfHits()){
  # overview table includes list of matched and missed queries for hit 1
    overView();
  } elsif ($reportType eq 'protein' || $reportType eq 'concise') {
  # if no overview, repeat search requires list of matched and missed queries for hit 1 as hidden fields
    my $objProtein = $objSummary->getHit(1);
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
      my $tempString = common_subs::htmlEscape($objResFile->getRepeatSearchString($queryNum));
      print "  <INPUT TYPE=\"hidden\" NAME=\"QUE\" VALUE=\"$tempString\">\n";
      unless ($objProtein->anyMatchToQuery($queryNum)) {
        print "  <INPUT TYPE=\"hidden\" NAME=\"MISSES\" VALUE=\"$tempString\">\n";
      }
    }
  }

# stuff in all the search parameters as hidden values
  searchForm();

  if ($objSummary->getNumberOfHits() == 0 
    && ($reportType eq 'protein' || $reportType eq 'concise')) {
  # handle case of protein summary with no significant hits
    print "</FORM>\n";
    print "<H3>No significant hits to report</H3>\n";

  } elsif ($reportType eq 'protein') {
  # Protein summary with hits to be displayed
    print "</FORM>\n";

  # print index
    print "<H3>Index</H3>\n";
    print "<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>\n";
    print "<TR><TD>&nbsp;</TD><TD><TT><B>&nbsp;Accession</B></TT></TD><TD><TT><B>&nbsp;&nbsp;Mass</B></TT></TD><TD><TT><B>&nbsp;&nbsp;Score</B></TT></TD><TD><TT><B>&nbsp;&nbsp;Description</B></TT></TD></TR>\n";
    foreach (@indexText){
      print $_;
    }
    print "</TABLE>\n";
    
  # loop through & print hit list
    print "<H3>Results List</H3>\n";
    print @spoolText;

  } else {
    if ($objSummary->getNumberOfHits() == 0) {
    # handle case of peptide or archive or concise summary with no significant hits
      print "<H3>No significant hits to report</H3>\n";
    } else {
      print "<P>";
    }
    print @spoolText;
  # end the form after the main body of the results
    print "</FORM>\n";

  }

# search parameter summary
  unless ($reportType eq 'select' || $reportType eq 'unassigned') {
    outputParams();  
  }

# copyright, etc.
  common_subs::simpleFooter();
  
# JavaScript for tooltips
  if ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    toolTips();
  }
  
# Tooltips <DIV> blocks
  if ($showToolTips 
    && ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned')) {
#    print "<!-- begin toolText at " . (times)[0] . " seconds -->\n"; ### TIMING ###
    toolText();
#    print "<!-- end toolText at " . (times)[0] . " seconds -->\n"; ### TIMING ###
  # Pop-ups for error tolerant search
    print @errorTolDiv;
  }

# print $thisScript->end_html;
  print "</BODY>\n</HTML>\n\n";

# Good Night
  exit 0;
  
###############################################################################
# &validateParams()
# validate a variety of mascot.dat and URL parameters
# globals:
# my($massDP, $minPepLen, $ignoreIonsScoreBelow, $reportType,
# $objParams, $objResFile, $showOverview, $thisScript, %urlParams, $sigThreshold,
# $numHits, $ShowSubSets, $RequireBoldRed, $mudpitSwitch,
# $selectSwitch, $ShowAllFromErrorTolerant, $UniGeneFile, $sortUnassigned);
###############################################################################

sub validateParams {

  my $tempString;

# $massDP determines precision on peptide mass values
# default is 2 decimal places for peptides
  $massDP = 0 unless ($massDP = $objMascotOptions->getMassDecimalPlaces);
  $massDP = int($massDP);
  if ($massDP < 1 || $massDP > 6) {
    $massDP = 2;
  }
  
# peptides shorter than $minPepLen are not interesting
# default is to ignore anything shorter than 6 residues
  if (defined($thisScript->param($urlParams{'_minpeplen'}))
    && $thisScript->param($urlParams{'_minpeplen'}) >= 0) {
    $minPepLen = $thisScript->param($urlParams{'_minpeplen'}) + 0;
  } else {
    $minPepLen = 0 unless ($minPepLen = $objMascotOptions->getMinPepLenInPepSummary );
    $minPepLen = int($minPepLen); 
    if ($minPepLen < 1 || $minPepLen > 9) {
      $minPepLen = 6;
    }
  }
  
# ignoreIonsScoreBelow is a threshold on MS/MS ions scores
# a value of -1 ultimately sets it to _sigthreshold (bug 12007)
# a value of 0 specifies that all peptides are included (default)
# a value greater than 1 indicates an absolute score cutoff 
# a value less than 1 is treated as a probability threshold 
  $ignoreIonsScoreBelowDefined = 0;
  if (defined($thisScript->param($urlParams{'_ignoreionsscorebelow'}))) {
    $ignoreIonsScoreBelowDefined = 1;
    if ($thisScript->param($urlParams{'_ignoreionsscorebelow'}) < 0) {
      $ignoreIonsScoreBelow = -1;
    } else { 
      $ignoreIonsScoreBelow = $thisScript->param($urlParams{'_ignoreionsscorebelow'}) + 0;
    }
  } else {
    my $d = $objMascotOptions->getDisplayNonSignificantMatches;
    if ($d == 2) {
      $ignoreIonsScoreBelow = 0 unless ($ignoreIonsScoreBelow = $objMascotOptions->getIgnoreIonsScoreBelow);
      if ($ignoreIonsScoreBelow < 0) {
        $ignoreIonsScoreBelow = 0;
      }
    } elsif ($d == 1) {
      $ignoreIonsScoreBelow = 0;
    } elsif ($d == 0) {
      $ignoreIonsScoreBelow = -1;
    }
  }

# if queries / entries greater than $mudpitSwitch, default 0.001, switch to Mudpit scoring
  if (defined($thisScript->param($urlParams{'_server_mudpit_switch'}))
    && $thisScript->param($urlParams{'_server_mudpit_switch'}) > 0) {
    $mudpitSwitch = $thisScript->param($urlParams{'_server_mudpit_switch'}) + 0;
  } else {
    $mudpitSwitch = 0.001;
    my $tempString;
    $tempString = 0 unless ($tempString = $objMascotOptions->getMudpitSwitch);
    if ($tempString > 0) {
      $mudpitSwitch = $tempString + 0;
    }
  }

# if number of queries greater than $selectSwitch, default 1000, switch to a Select summary
  $selectSwitch = 1000;
  $tempString = 0 unless ($tempString = $objMascotOptions->getSelectSwitch);
  if ($tempString > 0) {
    $selectSwitch = $tempString + 0;
  } else {
    $tempString = 0 unless ($tempString = $objMascotOptions->getMudpit);
    if ($tempString > 0) {
      $selectSwitch = $tempString + 0;
    }
  }

# If $ShowDecoyReport is true, a report for the integrated decoy search is shown
  if (defined($thisScript->param($urlParams{'_show_decoy_report'}))) {
    $ShowDecoyReport = $thisScript->param($urlParams{'_show_decoy_report'}) + 0;
  }

# set $section parameters to point to the correct sections
  $sec_peptides = $msparser::ms_mascotresfile::SEC_PEPTIDES;
  $sec_summary = $msparser::ms_mascotresfile::SEC_SUMMARY;
  if ($objParams->getDECOY() > 0
    && $ShowDecoyReport) {
    $sec_peptides = $msparser::ms_mascotresfile::SEC_DECOYPEPTIDES;
    $sec_summary = $msparser::ms_mascotresfile::SEC_DECOYSUMMARY;
  }
# $objResFile->anyPeptideSummaryMatches($sec_peptides) is used constantly
  $anyPepSumMatch = $objResFile->anyPeptideSummaryMatches($sec_peptides);
  $anyLibraryMatch = $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES);

# try to honour URL request for a particular report type
  $reportType = lc($thisScript->param($urlParams{reptype}) || '');
# if nothing specified on URL, check the result file
  unless ($reportType) {
    $reportType = lc($objParams->getREPTYPE());
  # change peptide to select for large reports
    if ($reportType eq 'peptide' && $objResFile->getNumQueries() > $selectSwitch) {
      $reportType = 'select';
    }
  }
# unless valid report type specified, set a sensible default
  unless ($reportType eq 'protein' 
       || $reportType eq 'peptide'
       || $reportType eq 'archive'
       || $reportType eq 'concise'
       || $reportType eq 'select'
       || $reportType eq 'unassigned') {
    $reportType = 'concise';
    if ($anyPepSumMatch or $anyLibraryMatch) {
      if ($objResFile->getNumQueries() > $selectSwitch) {
        $reportType = 'select';
      } else {
        $reportType = 'peptide';
      }
    }
  }

# catch case where MS/MS report requested, but there are no peptide matches
  if (($reportType eq 'peptide' || $reportType eq 'select' 
    || $reportType eq 'unassigned' || $reportType eq 'archive') 
    && (!$anyPepSumMatch && !$anyLibraryMatch)) {
    $reportType = 'concise';
  }
  
# never permit protein or concise summary for error tolerant or library search
  if (($objResFile->isErrorTolerant() || $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES))
      && ($reportType eq 'protein' || $reportType eq 'concise')) {
    if ($objResFile->getNumQueries() > $selectSwitch) {
      $reportType = 'select';
    } else {
      $reportType = 'peptide';
    }
  }

# if protein or concise requested for large search, will get Wot no results, so force to select
  if (($reportType eq 'protein' || $reportType eq 'concise') && $objResFile->getNumHits($sec_summary) < 1) {
    if ($anyPepSumMatch or $anyLibraryMatch) {
      $reportType = 'select';
    }
  }

# 'protein' or 'concise' with no hits means search failed
  if (($reportType eq 'protein' || $reportType eq 'concise') && $objResFile->getNumHits($sec_summary) < 1) {
    &failed();
    exit 0;
  }

# $showOverview is boolean flag to indicate whether overview table is required
  if ($reportType eq 'concise' || $reportType eq 'select' 
    || $reportType eq 'unassigned' || !$objParams->getOVERVIEW()) {
    $showOverview = 0;
  } else {
    $showOverview = 1;
  }
  
# set $sigThreshold to probability significance threshold, default 0.05
  $sigThreshold = common_subs::validate_sigThreshold(
    scalar $thisScript->param($urlParams{'_sigthreshold'}),
    $objMascotOptions,
    $anyLibraryMatch, # should really be SL_ONLY or SL_INTEGRATED test
  );
  
# set number of hits to display & make sure it is reasonable
# if the value of REPORT is 'AUTO', $numHits is set to 0
  if (defined($thisScript->param($urlParams{report}))) {
    if ($thisScript->param($urlParams{report}) =~ /auto/i) {
      $numHits = 0;
    } else {
      $numHits = int($thisScript->param($urlParams{report}));
    }
  } else {
    $numHits = int($objParams->getREPORT());
  }
  if ($numHits < 0){
    $numHits = 0;
  }
  
# Display a sub-set protein hit if it is missing less than $ShowSubSets fraction of the protein score of the main hit
  if (defined($thisScript->param($urlParams{'_showsubsets'}))) {
    $ShowSubSets = $thisScript->param($urlParams{'_showsubsets'}) + 0;
  } else {
    $ShowSubSets = 0 unless ($ShowSubSets = $objMascotOptions->getShowSubsets);
  }

# If RequireBoldRed is true, only display protein hits that include at least one bold, red peptide match
  if (defined($thisScript->param($urlParams{'_requireboldred'}))) {
    $RequireBoldRed = $thisScript->param($urlParams{'_requireboldred'}) + 0;
  } else {
    $RequireBoldRed = 0 unless ($RequireBoldRed = $objMascotOptions->isRequireBoldRed);
  }

# If $ShowAllFromErrorTolerant is true, all hits from an error tolerant search are shown
  if (defined($thisScript->param($urlParams{'_showallfromerrortolerant'}))) {
    $ShowAllFromErrorTolerant = $thisScript->param($urlParams{'_showallfromerrortolerant'}) + 0;
  }  
# If $OnlyErrorTolerant is true, display only error tolerant matches from a new style error tolerant search
  if (defined($thisScript->param($urlParams{'_onlyerrortolerant'}))) {
    $OnlyErrorTolerant = $thisScript->param($urlParams{'_onlyerrortolerant'}) + 0;
  }
# If $NoErrorTolerant is true, display only standard matches from a new style error tolerant search
  if (defined($thisScript->param($urlParams{'_noerrortolerant'}))) {
    $NoErrorTolerant = $thisScript->param($urlParams{'_noerrortolerant'}) + 0;
  }
# pick up illegal combinations
  if ($NoErrorTolerant) {
    $ShowAllFromErrorTolerant = 0;
  }
  if ($OnlyErrorTolerant && $NoErrorTolerant) {
    &common_subs::fatal("Meaningless to specify both _onlyerrortolerant and _noerrortolerant", __LINE__, $myName, "");
  }

# $UniGeneFile is the path to the Unigene index to be used for clustering EST matches
  if (my $species = $thisScript->param($urlParams{'unigene'})) {
    if ($species eq "None") {
      $UniGeneFile = "";
    } else {
      $UniGeneFile = $objUnigeneOptions->getPropValStringByName($species)
        || &common_subs::fatal("cannot find UniGene ".common_subs::htmlEscape($species)." in mascot.dat", __LINE__, $myName, $fileIn);
      $UniGeneFile =~ s/\s*$//;
    }
  } else {
    $UniGeneFile = "";
  }
  
# Default sort order for unassigned matches
  unless ($sortUnassigned = $thisScript->param($urlParams{'_sortunassigned'})) {
    $sortUnassigned = $objMascotOptions->getSortUnassigned;
  }
  $sortUnassigned = 'scoredown' unless ($sortUnassigned);

# Whether to display scores and expect values based on Percolator PEPs
  if (defined($percolate = $thisScript->param($urlParams{'percolate'}))) {
    $percolate = $thisScript->param($urlParams{'percolate'});
  } else {
    $percolate = $objMascotOptions->isPercolator;
  }
  $percolate += 0;
  my($PercolatorMinQueries, $PercolatorMinSequences);
  $PercolatorMinQueries = 100 unless ($PercolatorMinQueries = $objMascotOptions->getPercolatorMinQueries);
  $PercolatorMinSequences = 100 unless ($PercolatorMinSequences = $objMascotOptions->getPercolatorMinSequences);
  if ($reportType eq 'protein' 
    || $reportType eq 'concise'
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_PEPTIDES)
    || $objResFile->getNumQueries < $PercolatorMinQueries
    || $objResFile->getNumSeqsAfterTax < $PercolatorMinSequences
    || $objResFile->isErrorTolerant
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_DECOYPEPTIDES)) {
    $percolate = -1;
  }
  if ($percolate == 1) {
    if (defined($thisScript->param($urlParams{'percolate_rt'}))) {
      $percolate_rt = $thisScript->param($urlParams{'percolate_rt'});
    } else {
      $percolate_rt = $objMascotOptions->isPercolatorUseRT;
    }
    $percolate_rt += 0;
  } else {
    $percolate_rt = 0;
  }

  $preferTaxonomy = 0;
  if (defined(my $p = $thisScript->param($urlParams{'_prefertaxonomy'}))) {
    if (looks_like_number($p) and $p > 0) {
      $preferTaxonomy = int($p);
    }
  }
}

###############################################################################
# &failed()
# Outputs a special form to advise that search has failed with option to repeat
# globals:
# my($shipper, $fileIn, $objResFile, $objParams);
###############################################################################

sub failed {

  common_subs::printHeader1("Matrix Science - Wot? No Results");
  common_subs::printHeader2("", "Wot? No Results", 1, $session);

  print <<'end_of_static_HTML_text_block';

<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>
<TR><TD COLSPAN=2>    
<H1>Wot? No Results</H1>

<P>Sorry ... but this is the page you get to see when your search produces no results.

<P>Here are some possible reasons:

<UL>
  <LI>Sequence Query: None of the specified sequences occur in
  the database being searched. <I>Sequence qualifiers requires an exact match, 
  unlike a Blast search which pulls out the most homologous sequences. If you suspect this is 
  the problem, delete the suspect sequence qualifier and re-submit the search.</I>
  <LI>All searches: Match worse than random. <I>Mascot suppresses matches which are worse than 
  random. This is unlikely to happen with poor quality, noisy data. The 
  databases are so large that spurious matches tend to occur, even when the mass
  tolerances are very narrow. A more likely cause is 
  very narrow tolerances combined with precise but inaccurate mass values. Try relaxing 
  the tolerances.</I>
  <LI>All searches: If a taxonomy filter has been specified, maybe there are very few entries
  in the selected taxonomy, (or none at all if the required taxonomy indexes are out of date or missing).
  <LI>All searches: Are the peptide and / or fragment ion mass values less than 16 kDa? 
</UL>

end_of_static_HTML_text_block

  print "<FORM ACTION=\"./search_form.pl\" NAME=\"Results\" METHOD=\"POST\">\n";
  &searchForm();
  if ($objParams->getSEARCH ne "MIS") {
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
      print "  <INPUT TYPE=\"hidden\" NAME=\"QUE\" VALUE=\"";
      print common_subs::htmlEscape($objResFile->getRepeatSearchString($queryNum));
      print "\">\n";
    }    
  } else {
    print "  <INPUT TYPE=\"hidden\" NAME=\"QUERYLIST\" VALUE=\"All\">\n";
  }
  print "</FORM>\n";
  print "</TD></TR>\n";
  print "</TABLE>\n";

  unless ($shipper eq 'TRUE') {

    print <<'end_of_static_HTML_text_block';
<FORM ACTION="./matrix_formmail.pl" METHOD="POST">  
<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>
<TR><TD COLSPAN=2>    

<P>If you believe there may be a problem with Mascot, then please email us with details
using the form below. Please do not include large quantities of raw MS data, we will 
contact you if more information is required to diagnose the problem.  

<P>&nbsp;

</TD></TR>

end_of_static_HTML_text_block

    print "<input type='hidden' name='recipient' value='info\@matrixscience.com'>\n";
    print "<TR><TD><B>Subject:</B>&nbsp;</TD><TD>\n";
    print "<input type='text' size=40 name='subject' value='Problem: No results from Mascot'>\n";
    print "</TD></TR><TR><TD NOWRAP><B>From name:</B>&nbsp;</TD><TD>\n";
    print "<input type='text' size=40 name='name' value='" . common_subs::htmlEscape($objParams->getUSERNAME()) . "'>\n";
    print "</TD></TR><TR><TD NOWRAP><B>From email:</B>&nbsp;</TD><TD>\n";
    print "<input type='text' size=40 name='email' value='" . common_subs::htmlEscape($objParams->getUSEREMAIL()) . "'>\n";
    print "</TD></TR><TR><TD COLSPAN=2>&nbsp;</TD>\n";
    print "<input type='hidden' name='redirect' value='../thankyou.html'>\n";
    print "<input type='hidden' name='env_report' value='REMOTE_HOST,HTTP_USER_AGENT'>\n";
    print "<INPUT TYPE='hidden' name='data' VALUE='", common_subs::htmlEscape($fileIn), "'>\n";
    print "</TR><TR><TD COLSPAN=2><B>Description of problem:</B>&nbsp;</TD></TR><TR><TD COLSPAN=2>\n";
    print "<TEXTAREA NAME='comments' ROWS=6 COLS=48></TEXTAREA>\n";
    print "</TD></TR><TR><TD COLSPAN=2>&nbsp;</TD>\n";
    print "</TR><TR><TD COLSPAN=2>\n";
    print "<INPUT TYPE='submit' VALUE='Send message'>\n";
    print "</TD></TR>\n";
    print "</TABLE>\n";
    print "</FORM>\n";

  }

  &common_subs::printFooter();
  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &pleaseRepeat()
# Result file obsolete, must repeat
# globals:
# my($objResFile, $objParams);
###############################################################################

sub pleaseRepeat {

  common_subs::printHeader1("Matrix Science - Obsolete Result File");
  common_subs::printHeader2("", "Obsolete Result File", 1, $session);

  print <<'end_of_static_HTML_text_block';

<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>
<TR><TD COLSPAN=2>    

<H1>Obsolete Result File</H1>

<P>Sorry, this is a very old result file (pre Mascot 1.02) and the search must be repeated before
you can view the results.

end_of_static_HTML_text_block

  print "<FORM ACTION=\"./search_form.pl\" NAME=\"Results\" METHOD=\"POST\">\n";
  &searchForm();
  if ($objParams->getSEARCH ne "MIS") {
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
      print "  <INPUT TYPE=\"hidden\" NAME=\"QUE\" VALUE=\"";
      print common_subs::htmlEscape($objResFile->getRepeatSearchString($queryNum));
      print "\">\n";
    }    
  } else {
    print "  <INPUT TYPE=\"hidden\" NAME=\"QUERYLIST\" VALUE=\"All\">\n";
  }
  print <<'end_of_static_HTML_text_block';

</FORM>
</TD></TR>
</TABLE>

end_of_static_HTML_text_block

  common_subs::printFooter();
  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &searchForm()
# prints out parameters as hidden fields to support repeat searches, etc.
# globals:
# my($fileIn, $reportType, $shipper, $objResFile, $objParams, $objQuantMethod);
###############################################################################

sub searchForm {

  # User parameters are included in $params.
  my $params = MascotResults::repeat_search_parameters($objResFile);

  delete @$params{qw(ACCESSION ERRORTOLERANT)}; # See ErrTolRepeat below.

  print '  <INPUT TYPE="hidden" NAME="FORMVER" VALUE=2>', "\n";
  print '  <INPUT TYPE="hidden" NAME="ACCESSION">', "\n";

  if (defined $sessionID and $sessionID ne '') {
    print '  <INPUT TYPE="hidden" NAME="sessionID" VALUE="'.$sessionID.'">', "\n";
  }

  for my $key (sort keys %$params) {
    next if $$params{$key} eq '';
    printf '  <INPUT TYPE="hidden" NAME="%s" VALUE="%s">'."\n", $key, common_subs::htmlEscape($$params{$key});
  }
  
  my @db_list = map { $objParams->getDB($_)} 1 .. $objParams->getNumberOfDatabases();

  printf '  <INPUT TYPE="hidden" NAME="DB" VALUE="%s">'."\n", common_subs::htmlEscape(join(",", @db_list));
  
  printf '  <INPUT TYPE="hidden" NAME="INTERMEDIATE" VALUE="%s">'."\n", common_subs::htmlEscape($fileIn);
  print '  <INPUT TYPE="hidden" NAME="repeatSearch" VALUE="not again">', "\n";

  if ($reportType eq 'select') {
    print '  <INPUT TYPE="hidden" NAME="QUERYLIST" VALUE="All">', "\n";
  }
  
# display command buttons to select queries and submit repeat search
  if ($reportType eq 'archive' || $reportType eq 'unassigned'){
    print "\n";
  } elsif ($reportType eq 'select'){
    my $r1 = "";
    my $r2 = "";
    my $r3 = "";
    my $r4 = "";
    if ($thisScript->param($urlParams{_querylist})) {
      if ($thisScript->param($urlParams{_querylist}) =~ /unassigned/i) {
        $r2 = "CHECKED";
      } elsif ($thisScript->param($urlParams{_querylist}) =~ /homology/i) {
        $r3 = "CHECKED";
      } elsif ($thisScript->param($urlParams{_querylist}) =~ /identity/i) {
        $r4 = "CHECKED";
      } else {
        $r1 = "CHECKED";
      }
    } else {
      $r1 = "CHECKED";
    }
    print '<P><INPUT TYPE="submit" VALUE="Re-Search">', "\n";
    print qq{&nbsp;&nbsp;&nbsp;<INPUT TYPE="radio" VALUE="all" name="_querylistswitch" onClick="window.document.forms['Re-format']._querylist.value='all'; return true" $r1>&nbsp;All queries}, "\n";
    print qq{&nbsp;&nbsp;&nbsp;<INPUT TYPE="radio" VALUE="unassigned" name="_querylistswitch" onClick="window.document.forms['Re-format']._querylist.value='unassigned'; return true" $r2>&nbsp;Unassigned}, "\n";
    print qq{&nbsp;&nbsp;&nbsp;<INPUT TYPE="radio" VALUE="homology" name="_querylistswitch" onClick="window.document.forms['Re-format']._querylist.value='homology'; return true" $r3>&nbsp;Below homology threshold}, "\n";
    print qq{&nbsp;&nbsp;&nbsp;<INPUT TYPE="radio" VALUE="identity" name="_querylistswitch" onClick="window.document.forms['Re-format']._querylist.value='identity'; return true" $r4>&nbsp;Below identity threshold}, "\n";
  } elsif ((($reportType eq 'protein' || $reportType eq 'concise') && $objResFile->getNumHits($sec_summary) < 1)
    || ($reportType eq 'peptide' && (!$anyPepSumMatch && !$anyLibraryMatch))) {
    print '<P><INPUT TYPE="submit" VALUE="Repeat Search">', "\n";
  } elsif ($reportType eq 'peptide' || ($reportType eq 'protein' && $objParams->getOVERVIEW())){
    print '<P><INPUT TYPE="button" VALUE="Select All"'; 
    print ' onClick="checkQuery(1, ', $objResFile->getNumQueries(), ', this.form)">', "\n";
    print '&nbsp;&nbsp;<INPUT TYPE="button" VALUE="Select None"'; 
    print ' onClick="checkQuery(0, ', $objResFile->getNumQueries(), ', this.form)">', "\n";
    print '&nbsp;&nbsp;<INPUT TYPE="submit" VALUE="Search Selected">', "\n";
    if ($reportType eq 'peptide' && $objResFile->anyMSMS()
      && $anyPepSumMatch && &atLeastVersion(1,7,15) && not($objQuantMethod)) {
      print '&nbsp;&nbsp;<INPUT TYPE="checkbox" NAME="ErrTolRepeat" VALUE=1';
      if ($OldStyleErrTolReport) {
        print ' CHECKED';
      }
      print ' onClick="checkQuery(1, ', $objResFile->getNumQueries(), ', this.form)"><B>Error&nbsp;tolerant</B>&nbsp;&nbsp;&nbsp;', "\n";
    } else {
      print '&nbsp;&nbsp;&nbsp;&nbsp;';
    }
  } else {
    print '<P><INPUT TYPE="submit" VALUE="Re-Search All">', "\n";
    print '&nbsp;&nbsp;<INPUT TYPE="button" VALUE="Search Unmatched" onClick="unMatched(this.form); return true">', "\n";
  }
  if ($reportType eq 'peptide' && $shipper eq 'TRUE') {
    print '<INPUT TYPE="button" VALUE="Archive Report" onClick="makeArchive(this.form); return true">', "\n";
  }

}

###############################################################################
# &outputJavaScript ()
# output the bulk of the javascript
# globals:
# my($reportType, $showOverview, $fileIn, %et_accession_list, $myName,
# @accession_list, @query_title_list);
###############################################################################

sub outputJavaScript {

  print <<'end_of_static_HTML_text_block';

<STYLE TYPE="text/css">
.tooltip { position: absolute; visibility: hidden; left: 0; top: 0; }
span.abbr { cursor: help; border-bottom: 1px dotted black; }
</STYLE>

<SCRIPT LANGUAGE="JavaScript">
<!-- Begin hiding Javascript from old browsers.
 
  var arrayReady=0;
  
  if (document.images){
    redBall = new Image();
    redBall.src = "../images/red_ball_white.gif";  
    cyanBall = new Image();
    cyanBall.src = "../images/cyan_ball_white.gif";  
  }
  
  function cover(hit,query,protSeq){
    document.Results.OV_Title.value = queryTitle[query];
    document.Results.OV_Accession.value = protName[hit];
    document.Results.OV_Sequence.value = protSeq;
    if (hit == 0) {
      self.status = queryTitle[query];
    } else {
      self.status = protName[hit] + ": " + protSeq;
    }
    if (protSeq != "" && document.fred && arrayReady == 1){
      var i = 0;
      var element = 0;
      element = eval(protSeq+"["+i+"]");
      while(element) {
        document.images[element].src = redBall.src;
        i++;
        element = eval(protSeq+"["+i+"]");
      }
    }
  }

  function uncover(hit,query,protSeq){
    document.Results.OV_Title.value = "-Query-";
    document.Results.OV_Accession.value="-Accession-";
    document.Results.OV_Sequence.value="-Sequence-";
    self.status="";
    if (protSeq != "" && document.fred && arrayReady == 1){
      var i = 0;
      var element = 0;
      element = eval(protSeq+"["+i+"]");
      while(element) {
        document.images[element].src = cyanBall.src;
        i++;
        element = eval(protSeq+"["+i+"]");
      }
    }
  }

  function checkQuery(checkState,numChecks,form){
    if (numChecks == 1){
      form.QUE.checked = checkState;
    } else {
      for (var i = 0; i < form.elements.length; i++) {
        if (form.elements[i].name == "QUE"){
          form.elements[i].checked = checkState;
        }
      }
    }
  }

  function SaveClicks(checkBox,form){
    if (form.ErrTolRepeat && form.ErrTolRepeat.checked) {
      return;
    }
    var checkState = checkBox.checked;
    var thisHit = 0;
    for (var i = 0; i < form.elements.length; i++) {
      if (thisHit == 1){
        if (form.elements[i].name == "QUE") {
          form.elements[i].checked = !checkState;
        } else {
          break;
        }
      } else {
        if (form.elements[i] == checkBox){
          thisHit = 1;
        }
      }
    }
  }

  function unMatched(form){
    var numMisses = 0;
    for (var i = 0; i < form.elements.length; i++) {
      if (form.elements[i].name == "MISSES"){
        numMisses++;
      } 
    }
    if (numMisses < 1) {
      alert("No unmatched mass values to be searched");
      return false;
    }
    for (var i = 0; i < form.elements.length; i++) {
      if (form.elements[i].name == "QUE"){
        form.elements[i].name = "DUMMY";
      } 
    }
    for (var i = 0; i < form.elements.length; i++) {
      if (form.elements[i].name == "MISSES"){
        form.elements[i].name = "QUE";
      } 
    }
    form.submit();
    return true;
  }

  function setReFormatAction(form){
    if (form.REPTYPE.options[form.REPTYPE.selectedIndex].value == "export"){
      form.action = "./export_dat_2.pl";
    } else if (form.REPTYPE.options[form.REPTYPE.selectedIndex].value == "family") {
      form.action = "./master_results_2.pl";
    } else {
end_of_static_HTML_text_block

  print "      form.action = \"./$myName\";\n";
  print <<'end_of_static_HTML_text_block';
    }
  }
  
  function chooseQueryList(form) {
    var formatForm = window.document.forms['Re-format'];
    if (formatForm._querylist.value == 'unassigned') {
      form.QUERYLIST.value = queryListUnassigned;
    } else if (formatForm._querylist.value == 'identity') {
      form.QUERYLIST.value = queryListIdentity;
    } else if (formatForm._querylist.value == 'homology') {
      form.QUERYLIST.value = queryListHomology;
    } else {
      form.QUERYLIST.value = "All";
    }
    return true;
  }
    
  function makeAccessionString(form){
    if (!form.ErrTolRepeat || !form.ErrTolRepeat.checked) {
      form.ACCESSION.value = "";
      return true;
    }
    var accessions = [];
    for (var i = 0, ilen = form.elements.length; i != ilen; i++) {
      var elt = form.elements[i];
      if (elt.name == "INCLUDE" && elt.checked){
        var et_accs = et_list[elt.value];
        if (!et_accs) 
          continue;
        for (var j = 0, jlen = et_accs.length; j != jlen; j++) 
          accessions.push(et_accs[j]);
      }
    }
    if (accessions.length == 0) {
      alert("No hits selected for error tolerant search");
      return false;
    }
end_of_static_HTML_text_block

  my $maxETAccessions = $objMascotOptions->getErrTolMaxAccessions + 0;
  my $objMaxETAccessions = $session->getPermittedTasks->getTaskFromID($msparser::ms_security_tasks::SECTASK_MAXETACCESSIONS);
  if ($objMaxETAccessions && $objMaxETAccessions->getID() != -1) {
    $maxETAccessions = $objMaxETAccessions->getLongParams()->[0];
  }
  if ($maxETAccessions) {
    print "    if (accessions.length > $maxETAccessions) {\n";
    print "      alert(\"Number of protein entries for error tolerant search has been limited to $maxETAccessions\");\n";
    print "      accessions = accessions.slice(0, $maxETAccessions);\n";
    print "    }\n";
  }

  # Assumption: accessions don't contain ".
  print "    form.ACCESSION.value = '\"' + accessions.join('\",\"') + '\"';\n";
  print "    checkQuery(1," . $objResFile->getNumQueries() . ", form);\n";

  print <<'end_of_static_HTML_text_block';
    return true;
  }

  function clearEl() {}

end_of_static_HTML_text_block

  print "  function makeArchive(form){\n";
  my $tempUrl = $thisScript->self_url;
  if ($tempUrl =~ /REPTYPE=/) {
    $tempUrl =~ s/REPTYPE=\w+/REPTYPE=archive/;
  } else {
    $tempUrl .= "&REPTYPE=archive";
  }
  $tempUrl .= "&INTERMEDIATE=" . &common_subs::urlEscape($fileIn);
  print "    var tempUrl = \"$tempUrl\";";
  print "    for (var i = 0; i < form.elements.length; i++) {\n";
  print "      if (form.elements[i].name == \"INCLUDE\" && form.elements[i].checked){\n";
  print "        tempUrl = tempUrl + \"&INCLUDE=\" + form.elements[i].value;\n";
  print "      }\n";
  print "    }\n";
  print "    window.location.href = tempUrl;\n";
  print "  }\n\n";

  if ($showOverview) { 

    print "protName = new Array(1);\n";
    print "protName[0] = \"\";\n";
    my $objProt;
    for (my $hitNum=1; $hitNum <= $#accession_list; $hitNum++){
      print "protName[$hitNum] = \"$accession_list[$hitNum]\";\n";
    }
    print "\n";
    
    print "queryTitle = new Array(1);\n";
    print "queryTitle[0] = \"\";\n";
    for (my $queryNum=1; $queryNum <= $#query_title_list; $queryNum++){
      if ($query_title_list[$queryNum]) {
        print "queryTitle[$queryNum] = unescape(\"";
        # Query titles are already URL-escaped in the results file.
        print $query_title_list[$queryNum];
        print "\");\n";
      } else {
        print "queryTitle[$queryNum] = \"\";\n";
      }
    }
    print "\n";

  }

  if ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    print "\n";
  }

  if (%et_accession_list) {
    print "  et_list = {};\n";
    for my $hitNum (sort { $a <=> $b } keys %et_accession_list) {
      print "  et_list[$hitNum] = [";
      print join(", ", 
        map { 
          my $str = $_; $str =~ s/'/\\'/g; sprintf("'%s'", $str)
        } @{ $et_accession_list{$hitNum} }
      );
      print "];\n";
    }
    print "\n";
  }

  if ($reportType eq 'select') {
    my @subStrings = $objSummary->getQueryList($msparser::ms_peptidesummary::QL_UNASSIGNED) =~ /(.{1,72})/gs;
    print "var queryListUnassigned = \"" . join("\"\n      + \"",@subStrings) . "\";\n";
    @subStrings = $objSummary->getQueryList($msparser::ms_peptidesummary::QL_BELOW_IDENTITY) =~ /(.{1,72})/gs;
    print "var queryListIdentity = \"" . join("\"\n      + \"",@subStrings) . "\";\n";
    @subStrings = $objSummary->getQueryList($msparser::ms_peptidesummary::QL_BELOW_HOMOLOGY) =~ /(.{1,72})/gs;
    print "var queryListHomology = \"" . join("\"\n      + \"",@subStrings) . "\";\n";
  }
  

  print "// End hiding Javascript from old browsers. -->\n</SCRIPT>\n\n";  
  
}

###############################################################################
# &outputBody()
# outputs the report "header block"
# globals:
# my($reportType, @headerText, $objResFile, $saveMixHit, &objParams);
###############################################################################

sub outputBody {

  my(
  $bestScore,
  $hour,
  $i,
  $isdst,
  $mday, 
  $min, 
  $mon, 
  $sec, 
  $temp,
  $wday, 
  $yday, 
  $year
  );

# 'select' report requires additional padding
  my $spacer = ":";
  if ($reportType eq 'select' || $reportType eq 'unassigned') {
    $spacer = "       :";
  }
  print "<BODY BGCOLOR=\"#ffffff\" ALINK=\"#0000ff\" VLINK=\"#0000ff\">\n";
  print "<H1><IMG SRC=\"../images/88x31_logo_white.gif\" WIDTH=88 HEIGHT=31\n";
  print "  ALIGN=\"TOP\" BORDER=0 NATURALSIZEFLAG=3 Name=\"fred\"> Mascot Search Results</H1>\n";
  print "<FONT FACE='Courier New,Courier,monospace'>\n";
  print "<PRE>\n";
  print "<B>User            $spacer ".common_subs::htmlEscape($objParams->getUSERNAME())."</B>\n";
  print "<B>Email           $spacer ".common_subs::htmlEscape($objParams->getUSEREMAIL())."</B>\n";
  print "<B>Search title    $spacer ".common_subs::htmlEscape($objParams->getCOM())."</B>\n";
  if ($objParams->getFILENAME()){
    print "<B>MS data file    $spacer ".common_subs::htmlEscape($objParams->getFILENAME())."</B>\n";
  }  
  if ($objParams->getDATAURL()){
    print "<B>MS data URL     $spacer ".common_subs::htmlEscape($objParams->getDATAURL())."</B>\n";
  }  
  for (my $i = 1; $i <= $objParams->getNumberOfDatabases; $i++){
    if ($objParams->getNumberOfDatabases > 1 && $i > 9) {
      print "<B>Database $i     $spacer ";
    } elsif ($objParams->getNumberOfDatabases > 1) { 
      print "<B>Database $i      $spacer ";
    } else {
      print "<B>Database        $spacer ";
    }
    if ($ShowDecoyReport) {
      print "Decoy database equivalent to "
    }
    print common_subs::htmlEscape($objParams->getDB($i))
      . " "
      . common_subs::htmlEscape(($objResFile->getFastaVer($i) =~ /_(.*?)\.[^.]*$/)[0])
      . " ("
      . $objResFile->getNumSeqs($i)
      . " sequences; "
      . $objResFile->getNumResidues($i)
      . " residues)</B>\n";
    if ($objParams->getTAXONOMY 
      && !($objParams->getTAXONOMY =~ /All entries/)
      && $objResFile->getNumSeqs($i) != $objResFile->getNumSeqsAfterTax($i)) {
      $temp = $objParams->getTAXONOMY();
      $temp =~ s/^[\.\s]*//;
      if ($objParams->getNumberOfDatabases > 1 && $i > 9) {
        print "<B>Taxonomy $i     $spacer ";
      } elsif ($objParams->getNumberOfDatabases > 1) { 
        print "<B>Taxonomy $i      $spacer ";
      } else {
        print "<B>Taxonomy        $spacer ";
      }
      print common_subs::htmlEscape($temp)." ("
        . $objResFile->getNumSeqsAfterTax($i)
        . " sequences)</B>\n";
    }
  }
  if ($objResFile->isErrorTolerant()) {
    if (my $temp = $objParams->getACCESSION()) {
      $i = $temp =~ tr/,//;
      $i++;
      print "<B>                $spacer Error tolerant search of $i entries</B>\n";
    } else {
      if ($objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
        print "<B>                $spacer Error tolerant search of all significant protein hits</B>\n";
      } else {
        print "<B>                $spacer Error tolerant search of all entries</B>\n";
      }
    }
  }
  if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
    print "<B>Quantitation    $spacer " 
      . common_subs::htmlEscape($objParams->getQUANTITATION()) 
      . "&nbsp;&nbsp;&nbsp;<A HREF=\"./client.pl?quant_method&file=".common_subs::urlEscape($fileIn)."\" target=_blank>method details</A></B>\n";
    if ($objQuantMethod && $objQuantMethod->haveDescription()) {
      print "<B>                $spacer " . common_subs::htmlEscape($objQuantMethod->getDescription()) . "</B>\n";
    }
  }
  ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) 
    = gmtime ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'date'));
  printf "<B>Timestamp       $spacer %d %s %d at %02d:%02d:%02d GMT</B>\n", $mday,
  (qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec))[$mon],$year+1900,$hour,$min,$sec;
# Display any warnings from result file header
  if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn0')){
    $i=0;
    while ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn'.$i)){
      print "<B>Warning         $spacer <font color=#ff0000>"
      . common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn'.$i))
      . "</font></B>\n";
      $i++;
    }
  }
# Display a warning if this is a protein summary and there are ions scores
  if (($reportType eq 'protein' || $reportType eq 'concise')
    && $anyPepSumMatch) {
    print "<B>Warning         $spacer <font color=#ff0000>"
      . "A Peptide summary report will usually give a much clearer picture of MS/MS search results."
      . "</font></B>\n";
  }
# Display a warning if this is an orphan error tolerant search
  if ($objResFile->isErrorTolerant() 
    && !$objParams->getErrTolParentFilename()
    && !$objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
    print "<B>Warning         $spacer <font color=#ff0000>"
      . "Scores from standard search not available. Better matches may exist in other proteins."
      . "</font></B>\n";
  }
# Display any warnings from Parser
  for (my $i = 1; $i <= $objResFile->getNumberOfErrors(); $i++) {
    print "<B>Warning         $spacer <font color=#ff0000>" . common_subs::htmlEscape($objResFile->getErrorString($i)) . "</font></B>\n";
  }
# Display any warnings from quantitation
  if (@quantWarnings) {
    foreach (@quantWarnings) {
      print "<B>Warning         $spacer <font color=#ff0000>" . common_subs::htmlEscape($_) . "</font></B>\n";
    }
  }

# @headerText must be created 'ready to print' at top level
  if (@headerText) {
    for (my $i=0; $i <= $#headerText; $i++) {
      print $headerText[$i];
    }
  }

# search parameter summary
  if ($reportType eq 'select' || $reportType eq 'unassigned') {
    outputParams();  
  }

# for a peptide or archive or select report, significant hit list is created here
# if there is quantitation information, also include this for select
  unless ($reportType eq 'protein' || $reportType eq 'concise' || $reportType eq 'unassigned') {
    my $spacer = "&nbsp;" x 4;
    if ($reportType eq 'select') {
      $spacer = "&nbsp;" x 11;
    }
    if (@headerAccession) {
      print "<TABLE BGCOLOR=#FFFF99 BORDER=0 CELLPADDING=0 CELLSPACING=0>\n";
      if ($objQuantMethod) {
        print "<TR>";
        print  "<TD BGCOLOR=#FFFFFF NOWRAP><TT><B>Protein hits$spacer:</B></TT></TD>";
        for (@reportRatioNames) {
          print "<TD ALIGN=CENTER NOWRAP><TT><B>&nbsp;";
          print common_subs::htmlEscape($_). "&nbsp;";
          print "</B></TT></TD>";
        }
        print "<TD BGCOLOR=#FFFFFF NOWRAP>&nbsp;</TD><TD BGCOLOR=#FFFFFF NOWRAP>&nbsp;</TD>";
        print "</TR>\n";
      }
      for (my $j = 0; $j <= $#headerAccession; $j++) {
        print "<TR>";
        if ($objQuantMethod || $j > 0) {
          print "<TD BGCOLOR=#FFFFFF NOWRAP>&nbsp;</TD>";
        } else {
          print  "<TD BGCOLOR=#FFFFFF NOWRAP><TT><B>Protein hits$spacer:</B></TT></TD>";
        }
        if ($objQuantMethod) {
          for (my $i = 0; $i < @reportRatioNames; $i++) {
            print "<TD ALIGN=CENTER NOWRAP><TT>" .  ${ $headerQuant[$j + 1] }[$i] . "</TT></TD>";
          }
        }
        print "<TD BGCOLOR=#FFFFFF NOWRAP><TT><B>&nbsp;<A HREF=\"#Hit" 
          . ($j + 1)
          . "\">" 
          . common_subs::htmlEscape($headerAccession[$j]) 
          . "</A></B></TT></TD>";
        print "<TD BGCOLOR=#FFFFFF NOWRAP><TT>&nbsp;"
          . common_subs::htmlEscape($headerDescription[$j])
          . "</TT></TD>";
        print "</TR>\n";
      }
      print "</TABLE>\n";
    }
  }

  print "</PRE>\n</FONT>\n";
  
}

###############################################################################
# &outputDecoy()
# outputs decoy database summary stats
# globals:
# my($reportType, @headerText, $objResFile, $saveMixHit, &objParams);
###############################################################################

sub outputDecoy {

  my $DB = $objParams->getDB(1);
  if ($objParams->getNumberOfDatabases > 1) {
    $DB = "Target";
  }
  my $IF = $objSummary->getNumHitsAboveIdentity(1 / $sigThreshold);
  my $ID = $objSummary->getNumDecoyHitsAboveIdentity(1 / $sigThreshold);
  my $HF = $objSummary->getNumHitsAboveHomology(1 / $sigThreshold);
  my $HD = $objSummary->getNumDecoyHitsAboveHomology(1 / $sigThreshold);
  my $IR="&nbsp;";
  if ($IF) {
    $IR = sprintf("%.2f", $ID * 100 / $IF) . " %";
  }
  my $HR="&nbsp;";
  if ($HF) {
    $HR = sprintf("%.2f", $HD * 100 / $HF) . " %";
  }
  my $URL = $thisScript->self_url . "&_show_decoy_report=1";
  
  if (($reportType eq 'protein' || $reportType eq 'concise') && !$anyPepSumMatch && $objResFile->getNumHits($sec_summary) > 0) {

    my $TF = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_SUMMARY, "h1");
    $TF = (split(/,/, $TF))[1];
    $TF = sprintf("%.0f", $TF);

    my $TD = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_DECOYSUMMARY, "h1");
    $TD = (split(/,/, $TD))[1];
    $TD = sprintf("%.0f", $TD);

    print <<"end_of_static_HTML_text_block";
<TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
  <tr>
    <th bgcolor=#EEEEFF>&nbsp;</th>
    <th bgcolor=#EEEEFF>$DB</th>
    <th bgcolor=#EEEEFF><a href="$URL">Decoy</a></th>
  </tr>
  <tr>
    <td bgcolor=#EEEEFF>Protein hits above identity threshold</td>
    <td bgcolor=#EEEEFF align=center>$IF</td>
    <td bgcolor=#EEEEFF align=center>$ID</td>
  </tr>
  <tr>
    <td bgcolor=#EEEEFF>Highest scoring protein hit</td>
    <td bgcolor=#EEEEFF align=center>$TF</td>
    <td bgcolor=#EEEEFF align=center>$TD</td>
  </tr>
</table>
end_of_static_HTML_text_block
    
  } elsif (($reportType eq 'peptide' || $reportType eq 'select') && $anyPepSumMatch) {
      
    print <<"end_of_static_HTML_text_block";
<TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
  <tr>
    <th bgcolor=#EEEEFF>&nbsp;</th>
    <th bgcolor=#EEEEFF>$DB</th>
    <th bgcolor=#EEEEFF><a href="$URL">Decoy</a></th>
    <th bgcolor=#EEEEFF>False discovery rate</th>
  </tr>
  <tr>
    <td bgcolor=#EEEEFF>Peptide matches above identity threshold</td>
    <td bgcolor=#EEEEFF align=center>$IF</td>
    <td bgcolor=#EEEEFF align=center>$ID</td>
    <td bgcolor=#EEEEFF align=center>$IR</td>
  </tr>
end_of_static_HTML_text_block

    unless ($percolate == 1) {
      
      print <<"end_of_static_HTML_text_block";
  <tr>
    <td bgcolor=#EEEEFF>Peptide matches above homology or identity threshold</td>
    <td bgcolor=#EEEEFF align=center>$HF</td>
    <td bgcolor=#EEEEFF align=center>$HD</td>
    <td bgcolor=#EEEEFF align=center>$HR</td>
  </tr>
end_of_static_HTML_text_block
  
    }

    print "</table>\n";
  }
}

###############################################################################
# &overView()
# output an overview table 
# globals:
# my($objResFile, $objSummary, $reportType, @includeHit, $massDP, %pepRankByHQ,
# $fileIn, %pepSeqByHQ, $thisScript, %urlParams);
###############################################################################

sub overView {
  
  print <<'end_of_static_HTML_text_block';

<H3>Overview Table</H3>
Click on column header to jump to entry in results list.<BR>
Move mouse over any indicator to highlight identical peptides.<BR>  
Click on an indicator to see details of individual match.<BR>
Use check boxes to select sub-set of queries for new search.

<P><TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0 WIDTH=600>
<TR><TD NOWRAP><B>Mouse over:&nbsp;</B></TD>
<TD><INPUT TYPE="text" NAME="OV_Title" VALUE="-Query-" SIZE=62></TD>
</TR>
<TR><TD>&nbsp;</TD><TD NOWRAP>
<INPUT TYPE="text" NAME="OV_Accession" VALUE="-Accession-" SIZE=20>
&nbsp;<INPUT TYPE="text" NAME="OV_Sequence" VALUE="-Sequence-" SIZE=40>
</TD></TR></TABLE>

<P><TABLE BORDER=1 CELLSPACING=0 CELLPADDING=1>
  <TR>
    <TH BGCOLOR="#CCCCCC" ALIGN="RIGHT">Hit:</TH>
end_of_static_HTML_text_block

  my $hitNum = 1;
  while ($hitNum <= $objSummary->getNumberOfHits()){
    unless ($reportType eq 'archive' && !$includeHit[$hitNum]) {
      print "    <TH BGCOLOR=\"#CCCCCC\" ALIGN=\"CENTER\" WIDTH=16>";
      print "<A HREF=\"#Hit$hitNum\"";
      print " onMouseOver=  \"cover($hitNum,0,''); return true\"";
      print " onMouseOut = \"uncover($hitNum,0,''); return true\">";
      print "$hitNum</A></TH>\n";     
    }
    $hitNum++;
  }
  print "  </TR>\n";

  my %tableHash;
  my $queryNum = 1;
  my $imageNum = 2;  #NB image[0] is logo, image[1] is score gif
  while ($queryNum <= $objResFile->getNumQueries()){
    print "  <TR>\n";
    print "    <TD BGCOLOR=#CCCCCC NOWRAP>";
    print "<B><INPUT TYPE=\"checkbox\" NAME=\"QUE\" VALUE=\""
      . common_subs::htmlEscape($objResFile->getRepeatSearchString($queryNum))
      . "\" CHECKED> ";
    print "<A HREF='#Query" 
      . $queryNum
      . "' onMouseOver= \"cover(0,$queryNum,''); return true\""
      . " onMouseOut = \"uncover(0,$queryNum,''); return true\">";
    my $fieldCode = '%.' . $massDP . 'f</A> ';
    print sprintf($fieldCode, $objResFile->getObservedMass($queryNum));
    my $local_charge_string;
    if ($objResFile->getObservedCharge($queryNum) < -99) {
      $local_charge_string = "Unknown";
    } elsif ($objResFile->getObservedCharge($queryNum) == 0) {
      $local_charge_string = "Mr";
    } elsif ($objResFile->getObservedCharge($queryNum) < 0) {
      $local_charge_string = "" . $objResFile->getObservedCharge($queryNum) . "-";
    } else {
      $local_charge_string = "" . $objResFile->getObservedCharge($queryNum) . "+";
    }
    print "($local_charge_string)";
    print " </B></TD>\n";    
    my $hitNum = 1;
    while ($hitNum <= $objSummary->getNumberOfHits()){
      unless ($reportType eq 'archive' && !$includeHit[$hitNum]) {
        if ($pepSeqByHQ{"h$hitNum\_q$queryNum"}) {        
          my $sequence = $pepSeqByHQ{"h$hitNum\_q$queryNum"};
          $sequence =~ s/@/_/g;
          if ($tableHash{$sequence}){
            $tableHash{$sequence}.= ",$imageNum";
          }else{
            $tableHash{$sequence}= "$imageNum";
          }
          $imageNum++;
          print "    <TD ALIGN=\"CENTER\">";
          if ($objResFile->isPMF()){
          # can't just use hit number because could be mixture
            if ($objSummary->getHit($hitNum)->isPMFMixture() || $ShowDecoyReport) {
              print "<A HREF=\"javascript:void(0)\" ";
            } else {
              print "<A HREF=\"$protViewLink?file=".common_subs::urlEscape($fileIn)."&hit="
                . $objSummary->getHit($hitNum)->getProteinSummaryHit()
                . "&db_idx=" . $objSummary->getHit($hitNum)->getDB()
                . ($sessionID ? '&sessionID='.common_subs::urlEscape($sessionID) : ())
                . "\" TARGET=\"_blank\" ";        
            }
          } else {
            print "<A HREF=\"$pepViewLink?file=".common_subs::urlEscape($fileIn)."&query=$queryNum&hit=";
            if ($reportType eq 'peptide' || $reportType eq 'archive') {
              print $objSummary->getSrcRank($queryNum, $pepRankByHQ{"h$hitNum\_q$queryNum"}) . "&index=";
              unless ($UniGeneFile) {
                print common_subs::urlEscape($objSummary->getHit($hitNum)->getAccession());
                print "&db_idx=", $objSummary->getHit($hitNum)->getDB();
              }
              print "&px=1&section=";
              print $objSummary->getSrcSection($queryNum, $pepRankByHQ{"h$hitNum\_q$queryNum"});
              print "&ave_thresh=";
              print $averageThreshold;
              if ($ignoreIonsScoreBelowDefined) {
                print "&_ignoreionsscorebelow=";
                print $ignoreIonsScoreBelow;
              }
              print "&report=";
              print $numHits;
              print "&_sigthreshold=";
              print $sigThreshold;
              print "&_msresflags=";
              print $msresFlags;
              print "&_msresflags2=";
              print $msresFlags2;
              print "&percolate=";
              print $percolate;
              print "&percolate_rt=";
              print $percolate_rt;
              print "&_minpeplen=";
              print $minPepLen;
            } else {
              print "$hitNum";
              if ($ShowDecoyReport) {
                print "&section=" . $msparser::ms_mascotresfile::SEC_DECOYSUMMARY;
              }
            }
            if ($sessionID) {
              print '&sessionID='.common_subs::urlEscape($sessionID);
            }
            print "\" TARGET=\"_blank\"";
          }
          print " onMouseOver=  \"cover($hitNum,$queryNum,'$sequence'); return true\"";
          print " onMouseOut = \"uncover($hitNum,$queryNum,'$sequence'); return true\">";
          print "<IMG SRC=\"../images/cyan_ball_white.gif\" BORDER=0"; 
          print " HEIGHT=16 WIDTH=16></A></TD>\n"; 
        }else{
          print "    <TD>&nbsp;</TD>\n";
        }
      }
      $hitNum++;      
    }
    print "  </TR>\n";
    $queryNum++;
  }
    
  print "  </TABLE>\n";

  print "<SCRIPT LANGUAGE=\"JavaScript\">\n";
  print "<!-- \n\n";

  while (my($first,$last) = each(%tableHash)){
    print "$first = new Array(1);\n";
    my @arrayVals = (split(",",$last));
    for (my $i = 0; $i <= $#arrayVals; $i++){
      print "$first\[$i\] = $arrayVals[$i];\n";
    } 
  }
  print "arrayReady = 1;\n";
  print "\n";
  
  print "// -->\n";
  print "</SCRIPT>\n";

}

###############################################################################
# &outputParams()
# outputs search parameter summary
# globals:
# my($objResFile, $reportType, $massDP, @query_title_list, $objParams, 
# $objSummary);
###############################################################################

sub outputParams {

  my(
  $anchor,
  $fieldCode,
  $queryNum,
  $temp
  );

  unless ($reportType eq 'select' || $reportType eq 'unassigned') {
    print "<H3>Search Parameters</H3>\n";
    print "<FONT FACE='Courier New,Courier,monospace'>\n";
    print "<PRE>\n"; 
    print "<B>Type of search         : ";
    if ($objResFile->isMSMS()){
      print "MS/MS Ion Search</B>\n";
    } elsif ($objResFile->isPMF()){
      print "Peptide Mass Fingerprint</B>\n";
    } elsif ($objResFile->isSQ()){
      print "Sequence Query</B>\n";
    } else {
      print "Unknown</B>\n";
    }
    if ($objResFile->isErrorTolerant()){
      if ($objParams->getACCESSION()) {
        print "<B>Error tolerant search  : " . common_subs::htmlEscape($objParams->getACCESSION()) . "</B>\n";
      } else {
        if ($objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
          print "<B>Error tolerant search  : All significant protein hits</B>\n";
        } else {
          print "<B>Error tolerant search  : All entries</B>\n";
        }
      }
    }
  }
  do {
    my $target_fdr = $objParams->getTARGET_FDR_PERCENT;
    if ($target_fdr > 0) {
      printf "<B>Target FDR             : %g%%</B>\n", $target_fdr;
    }
  };
  print "<B>Enzyme                 : " . common_subs::htmlEscape($objParams->getCLE()) . "</B>\n";
  my $local_mods;
  if ($objParams->getFixedModsName(1)) {
    my $i = 1;
    my @tmpFixedMods;
    while ($objParams->getFixedModsName($i)) {
      push @tmpFixedMods, $objParams->getFixedModsName($i);
      $i++;
    }
    $local_mods = join(",", @tmpFixedMods);
  } else {
    $local_mods = $objParams->getMODS();
  }
  if ($local_mods){
    my @tempMods = split(/,/, $local_mods);
    for (my $i = 0; $i <= $#tempMods; $i++) {
      $tempMods[$i] = "<A HREF=\"./client.pl?modification&mod_name=" 
        . common_subs::urlEscape($tempMods[$i])
        . "&file=" 
        . common_subs::urlEscape($fileIn)
        . "\"target=_blank>" 
        . common_subs::htmlEscape($tempMods[$i])
        . "</A>";
    }
    $local_mods = join(",", @tempMods);
    print "<B>Fixed modifications    : $local_mods</B>\n";
  }
  my $local_it_mods;
  if ($objParams->getVarModsName(1)) {
    my $i = 1;
    my @tmpVarMods;
    while ($objParams->getVarModsName($i)) {
      push @tmpVarMods, $objParams->getVarModsName($i);
      $i++;
    }
    $local_it_mods = join(",", @tmpVarMods);
  } else {
    $local_it_mods = $objParams->getIT_MODS();
  }
  if ($objParams->getICAT()) {
    print "<B>ICAT                   : Cysteine filter and ICAT modifications applied</B>\n";
    my $ICATLight = $objMascotOptions->getICATLight;
    $ICATLight = "ICAT_light" unless $ICATLight;
    my $ICATHeavy = $objMascotOptions->getICATHeavy;
    $ICATHeavy = "ICAT_heavy" unless $ICATHeavy;
    my @tempMods = split(/,/, $local_it_mods);
    for (my $i = $#tempMods; $i >= 0; $i--) {
      if (($tempMods[$i] eq $ICATLight) || ($tempMods[$i] eq $ICATHeavy)) {
        splice @tempMods, $i, 1;
      }
    }
    if (@tempMods) {
      $local_it_mods = join(',', @tempMods);
    } else {
      $local_it_mods = "";
    }
  }
  if ($local_it_mods){
    my @tempMods = split(/,/, $local_it_mods);
    for (my $i = 0; $i <= $#tempMods; $i++) {
      $tempMods[$i] = "<A HREF=\"./client.pl?modification&mod_name=" 
        . common_subs::urlEscape($tempMods[$i])
        . "&file=" 
        . common_subs::urlEscape($fileIn)
        . "\"target=_blank>" 
        . common_subs::htmlEscape($tempMods[$i])
        . "</A>";
    }
    $local_it_mods = join(",", @tempMods);
    print "<B>Variable modifications : $local_it_mods</B>\n";
  }
  print "<B>Mass values            : ". $objParams->getMASS(). "</B>\n";
  if ($objParams->getSEG() <= 0){
    print "<B>Protein Mass           : Unrestricted</B>\n"; 
  } elsif ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    print "<B>Protein Mass           : Restriction does not apply in peptide summary</B>\n"; 
  } else {
    print "<B>Protein Mass           : ". $objParams->getSEG() ." kDa</B>\n";
  }
  print "<B>Peptide Mass Tolerance : &#177; "
    . $objParams->getTOL() . " " . $objParams->getTOLU();
  if ($objParams->getPEP_ISOTOPE_ERROR() > 0 && !$objResFile->isPMF()) {  
    print " (#&nbsp;<sup>13</sup>C = " . $objParams->getPEP_ISOTOPE_ERROR() . ")";
  }
  print "</B>\n";
  if ($objResFile->isPMF()){
    print "<B>Peptide Charge State   : " . $objParams->getCHARGE() . "</B>\n";
  } else {
    print "<B>Fragment Mass Tolerance: &#177; "
    . $objParams->getITOL() . " " . $objParams->getITOLU() . "</B>\n";
  }
  if ($objParams->getFORMVER() > 1.00) {
    print "<B>Max Missed Cleavages   : " . $objParams->getPFA() . "</B>\n";
  } else {
    print "<B>Partials               : " . $objParams->getPFA() . "</B>\n";
  }
  if (!$objResFile->isPMF() && $objResFile->anyMSMS() && !&atLeastVersion(1,1,9)) {
    print "<B>Ion Series Weightings  : a=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'iatol'));
    print ", a*=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'iastol'));
    print ", a++=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'ia2tol'));
    print "</B>\n";
    print "<B>Ion Series Weightings  : b=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'ibtol'));
    print ", b*=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'ibstol'));
    print ", b++=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'ib2tol'));
    print "</B>\n";
    print "<B>Ion Series Weightings  : y=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'iytol'));
    print ", y*=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'iystol'));
    print ", y++=";
    print common_subs::htmlEscape($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, 'iy2tol'));
    print "</B>\n";
  }
  if (!$objResFile->isPMF() && $objParams->getINSTRUMENT()) {
    print "<B>Instrument type        : " . $objParams->getINSTRUMENT() . "</B>\n";
  }
#  if ($objParams->getFILENAME()){
#    print "<B>Data File Name         : " . &common_subs::noTag($objParams->getFILENAME()) . "</B>\n";
#  }  
  if (!$objResFile->isPMF() && ($reportType eq 'protein' || $reportType eq 'concise')){
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
      $fieldCode = '%.'.$massDP.'f';
      my $local_charge_string;
      if ($objResFile->getObservedCharge($queryNum) < -99) {
        $local_charge_string = "Unknown";
      } elsif ($objResFile->getObservedCharge($queryNum) == 0) {
        $local_charge_string = "Mr";
      } elsif ($objResFile->getObservedCharge($queryNum) < 0) {
        $local_charge_string = "" . abs($objResFile->getObservedCharge($queryNum)) . "-";
      } else {
        $local_charge_string = "" . $objResFile->getObservedCharge($queryNum) . "+";
      }
      $temp = "Query" . $queryNum . " (" 
        . sprintf("$fieldCode,%s", $objResFile->getObservedMass($queryNum), $local_charge_string) . ")";
      $anchor = "\"Query".$queryNum."\"";
      print sprintf("<A NAME=%s><B>%-s</A>: ", $anchor, common_subs::htmlEscape(sprintf("%23s", $temp)));
      if ($query_title_list[$queryNum]){
        my $local_query_title = $query_title_list[$queryNum];
        print common_subs::htmlEscape(uri_unescape($local_query_title));
      } else {
        print "&lt;no title&gt;";
      }
      print "</B>\n";
    # create a query object
      my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
      if ($objQuery->getPepTolString()) {                                  
        print "<B>                       : " . common_subs::htmlEscape($objQuery->getPepTolString()) . "</B>\n";
      }
      my $local_seq_string = "";
      for (my $i = 1; $i <= 20; $i++) {
        if ($objQuery->getSeq($i)) {
          $local_seq_string .= "seq(" . $objQuery->getSeq($i) . ") "; 
        } else {
          last;
        }
      }
      if ($local_seq_string) {
        print "<B>                       : " . common_subs::htmlEscape($local_seq_string) . "</B>\n";
      }
      my $local_comp_string = "";
      for (my $i = 1; $i <= 20; $i++) {
        if ($objQuery->getComp($i)) {
          $local_comp_string .= "comp(" . $objQuery->getComp($i) . ") "; 
        } else {
          last;
        }
      }
      if ($local_comp_string) {
        unless ($objParams->getICAT() && $local_comp_string =~ /^\s*comp\(\*\[C\]\)\s*$/i) {
          print "<B>                       : ", common_subs::htmlEscape($local_comp_string), "</B>\n";
        }
      }
      my $local_tag_string = "";
      for (my $i = 1; $i <= 20; $i++) {
        if ($local_tag_string = $objQuery->getTag($i)) {
          $local_tag_string =~ /([et]),(.*)/i;
          if ($1 eq "e") {
            print "<B>                       : etag($2)</B>\n";
          } else {
            print "<B>                       : tag($2)</B>\n";
          }
        } else {
          last;
        }
      }
    }
  } else {
    print "<B>Number of queries      : " . $objResFile->getNumQueries() . "</B>\n";
    if ($objResFile->isPMF()) {
      if ($objSummary->getNumPmfQueriesUsed() 
        && $objSummary->getNumPmfQueriesUsed() != $objResFile->getNumQueries()) {
        print "<B>Selected for scoring   : " . $objSummary->getNumPmfQueriesUsed() . "</B>\n";
      }
    }
  }
  unless ($reportType eq 'select' || $reportType eq 'unassigned') {
    print "</PRE>\n";
    print "</FONT>\n";
  }
}

###############################################################################
# &toolTips()
# JavaScript for displaying the yellow pop-up windows
###############################################################################

sub toolTips {

  print <<'end_of_static_HTML_text_block';

<SCRIPT LANGUAGE="JavaScript1.2">
<!--
var timerID = null;
var active = null;

function redo() {
  window.location.reload();
}

function init() {
  setTimeout("window.onresize = redo", 1000);
}

function displayEl(left, top) {
  document.onmousemove = null;
  var unit = '', whichEl;
  if (typeof browser_NS4 != 'undefined' && browser_NS4) {
    whichEl = document[active];
    whichEl.visibility = "show";
  } else if (typeof document.getElementById != 'undefined') {
    whichEl = document.getElementById(active).style;
    unit = 'px';
    whichEl.visibility = "visible";
  } else {
    whichEl = document.all[active].style;
    unit = 'px';
    whichEl.visibility = "visible";
  }
  whichEl.left = left + unit;
  whichEl.top = top + unit;
}

function clearEl() {
  self.status = "";
  if (typeof browser_NS4 != 'undefined' && browser_NS4) {
    whichEl = document[active];
    whichEl.visibility = "hide";
  } else if (typeof document.getElementById != 'undefined') {
    whichEl = document.getElementById(active).style;
    whichEl.visibility = "hidden";
  } else {
    whichEl = document.all[active].style;
    whichEl.visibility = "hidden";
  }
  active = null;
  if (timerID) clearTimeout(timerID);
  document.onmousemove = null;
}

function activateEl(id, e) {
  active = id;
  document.onmousemove = checkEl;
  checkEl(e);
}

function checkEl(e) {
  if (typeof e == 'undefined')
    e = event;
  if (timerID) clearTimeout(timerID);
  var left, top;
  if (typeof e.clientX != 'undefined') {
    left = e.clientX;
  } else if (typeof e.pageX != 'undefined') {
    left = e.pageX;
  }
  if (typeof e.clientY != 'undefined') {
    top = e.clientY + 20;
  } else if (typeof e.pageY != 'undefined') {
    top = e.pageY + 20;
  }
  var coordElt;
  if (document.documentElement) {
    left += (document.documentElement.scrollLeft || document.body.scrollLeft);
    top += (document.documentElement.scrollTop || document.body.scrollTop);
  } else {
    left += document.body.scrollLeft;
    top += document.body.scrollTop;
  }
  timerID = setTimeout("displayEl(" + left + ", " + top + ")", 300);
}

// -->
</SCRIPT>

end_of_static_HTML_text_block

}

###############################################################################
# &testBrowser()
# return 0 if browser is positively recognised as _not_
# supporting yellow 'top 10' pop-ups, otherwise, return 1
###############################################################################

sub testBrowser {

  my $agent = lc($ENV{'HTTP_USER_AGENT'});
  
  my $is_aol = (index($agent, 'aol') != -1);
  my $is_opera = (index($agent, 'opera') != -1);
  my $is_webtv = (index($agent, 'webtv') != -1);
  my $is_hotjava = (index($agent, 'hotjava') != -1);
  my $is_ie = (index($agent, 'msie') != -1
                           && !$is_aol
                           && !$is_opera
                           && !$is_webtv
                           && !$is_hotjava);
  my $is_nav = (index($agent, 'mozilla') != -1 
                           && !$is_ie
                           && !$is_aol
                           && !$is_opera
                           && !$is_webtv
                           && !$is_hotjava);
  
  if ($is_ie || $is_aol || $is_webtv) {
    $agent =~ /msie (\d*)/;
    if ($1 < 4){ 
      return 0;
    }
  } elsif ($is_nav) {
    $agent =~ /mozilla\/(\d*)/;
    if ($1 < 4){
      return 0;
    }
  } elsif ($is_opera) {
    $agent =~ /opera[ \/](\d*)/;
    if ($1 < 4){ 
      return 0;
    }
  } elsif ($is_hotjava) {
    return 0;
  }
  
  return 1;

}

###############################################################################
# &toolText()
# create the <DIV> blocks for the the yellow pop-up windows
# globals:
# my($objResFile, $objSummary, $sigThreshold, $objParams);
###############################################################################

sub toolText {

  my $delta = "Delta";
  if ($objParams->getTOLU() eq "%") {
    $delta = "  %  ";
  } elsif ($objParams->getTOLU() eq "ppm") {
    $delta = " ppm ";
  } elsif ($objParams->getTOLU() eq "mmu") {
    $delta = " mmu ";
  }      

  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++) {
    if (($reportType eq 'select' || $reportType eq 'unassigned') && !$selectList[$queryNum]) {
    # skip loop for unassigned queries in a 'select' report
    # or assigned queries in an 'unassigned' report 
      next;
    }
    print "<DIV CLASS='tooltip' ID='Q$queryNum'>";
    print "<TABLE BORDER=0 CELLPADDING=2 CELLSPACING=0 BGCOLOR=#0000ff>";
    print "<TR><TD><TABLE BORDER=0 CELLPADDING=5 CELLSPACING=0 BGCOLOR=#ffffcc><TR><TD><TT>";
    print "Top scoring peptide matches to query $queryNum";

  # create a query object
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
    if ($objQuery->getStringTitle(1)) {
      print "<BR>" . common_subs::htmlEscape($objQuery->getStringTitle(1));
    }
    if ($objQuery->getPepTolString()) {                           
      print "<BR>Tolerance qualifier: " . common_subs::htmlEscape($objQuery->getPepTolString());
    }
    if ($objQuery->getINSTRUMENT(1)) {                           
      print "<BR>Instrument qualifier: " . common_subs::htmlEscape($objQuery->getINSTRUMENT(1));
    }
    if ($objQuery->getIT_MODS(1)) {                           
      print "<BR>Variable mods qualifier: " . $objQuery->getIT_MODS(1);
    }
    my $local_seq_string = "";
    for (my $i = 1; $i <= 20; $i++) {
      if ($objQuery->getSeq($i)) {
        $local_seq_string .= "seq(" . $objQuery->getSeq($i) . ") "; 
      } else {
        last;
      }
    }
    if ($local_seq_string) {
      print "<BR>Sequence qualifier(s): ", common_subs::htmlEscape($local_seq_string);
    }
    my $local_comp_string = "";
    for (my $i = 1; $i <= 20; $i++) {
      if ($objQuery->getComp($i)) {
        $local_comp_string .= "comp(" . $objQuery->getComp($i) . ") "; 
      } else {
        last;
      }
    }
    if ($local_comp_string) {
      unless ($objParams->getICAT() && $local_comp_string =~ /^\s*comp\(\*\[C\]\)\s*$/i) {
        print "<BR>Composition qualifier: ", common_subs::htmlEscape($local_comp_string);
      }
    }
    my $local_tag_string = "";
    for (my $i = 1; $i <= 20; $i++) {
      if ($local_tag_string = $objQuery->getTag($i)) {
        $local_tag_string =~ /([et]),(.*)/i;
        if ($1 eq "e") {
          print "<BR>etag($2)";
        } else {
          print "<BR>tag($2)";
        }
      } else {
        last;
      }
    }
   
    if ($objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant and $OmitErrTolThresholds) {
      # Bug 13542: old behaviour, i.e., don't display thresholds.
    } else {
      if ($objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold)) {
        print "<BR>Score greater than " 
          . int($objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold)) 
          . " indicates homology";
      }
      my $dbType = $objSummary->getPeptide($queryNum, 1)->getIsFromLibrary()
          ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
          : $msparser::ms_mascotresults::DM_FASTA
          ;
      if ($objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold, $dbType)
        && $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold, $dbType) > 0) {
        print "<BR>Score greater than "
          . int($objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold, $dbType))
          . " indicates identity";
      }
    }
    print "</TT><PRE>";
    my @columns;
    for (my $p = 1; $p <= $objSummary->getMaxRankValue(); $p++) {
      my $objPepMatch = $objSummary->getPeptide($queryNum, $p);
      if (defined($objPepMatch) && (my $local_pep = $objPepMatch->getPeptideStr(0))) { 
        push @{ $columns[0] }, sprintf("%5.1f", $objPepMatch->getIonsScore());
        my $expect = $objSummary->getPeptideExpectationValue($objPepMatch->getIonsScore(), $queryNum, $p);
        if ($expect > 0) {
          push @{ $columns[1] }, sprintf(" %8.2g", $expect);
        } else {
          push @{ $columns[1] }, "";
        }
        push @{ $columns[2] }, &err2Tolu($objPepMatch->getDelta(), $objPepMatch->getMrCalc());
        my($hit, $protein, $morethan) = 
          $objSummary->getProteinsWithThisPepMatch($queryNum, $p, 1) =~ /^(\d+):"([^"]*)"\s+(.*)/;
        if ($hit) {
          if ($morethan) {
            push @{ $columns[3] }, sprintf("%4d+", $hit);
          } else {
            push @{ $columns[3] }, sprintf("%4d ", $hit);
          }
          push @{ $columns[4] }, " " . $protein;
        } else {
          push @{ $columns[3] }, "     ";
          push @{ $columns[4] }, "";
        }
      # use the variable mods string to underline the peptide string at modified residues
        my $varModsString = $objPepMatch->getVarModsStr();
        my $summedModsString = $objPepMatch->getSummedModsStr();
        $summedModsString = '0' x length($varModsString) if $summedModsString eq '';
        my $localModsString = $objPepMatch->getLocalModsStr();
        $localModsString = '0' x length($varModsString) if $localModsString eq '';
        if ($varModsString =~ /[1-Y]/ or $summedModsString =~ /[1-Y]/ or $localModsString =~ /[1-Y]/) {
          # The strings have the same length by definition.
          for (my $i = length($varModsString) - 2; $i > 0; $i--) {
            if (substr($varModsString, $i, 1) =~ /[1-Y]/
                    or substr($summedModsString, $i, 1) =~ /[1-Y]/
                    or substr($localModsString, $i, 1) =~ /[1-Y]/) {
              substr($local_pep, $i - 1, 1) = "<U>" . substr($local_pep, $i - 1, 1) . "</U>";
            }
          }
        }
        if (&atLeastVersion(2, 0, 99)) {
          my $termlist = $objSummary->getTerminalResiduesString($queryNum, $p);
          if ($termlist) {
            my @terms = split(/:/, $termlist);
            my $termPair = $terms[0];
            for (my $i = 1; $i <= $#terms; $i++) {
              if ($termPair ne $terms[$i]) {
                $termPair = "";
                last;
              }
            }
            if ($termPair) {
              $termPair =~ s/@/-/g;
              (my $N, my $C) = $termPair =~ /(.),(.)/;
              if (substr($varModsString, 0, 1) =~ /[1-Y]/
                      or substr($summedModsString, 0, 1) =~ /[1-Y]/
                      or substr($localModsString, 0, 1) =~ /[1-Y]/) {
                $N = "$N<U>.</U>";
              } else {
                $N = "$N.";
              }
              if (substr($varModsString, length($varModsString) - 1, 1) =~ /[1-Y]/ 
                      or substr($summedModsString, length($summedModsString) - 1, 1) =~ /[1-Y]/
                      or substr($localModsString, length($localModsString) - 1, 1) =~ /[1-Y]/) {
                $C = "<U>.</U>$C";
              } else {
                $C = ".$C";
              }
              $local_pep = "$N$local_pep$C";
            }
          }
        }
        push @{ $columns[5] }, $local_pep;
        if ($OldStyleErrTolReport or $expect <= 0) {
          push @{ $columns[6] }, 1;
        } else {
          push @{ $columns[6] }, 0;
        }
      } else {
        last;
      }
    }
    my $widest = 0;
    foreach (@{ $columns[4] }) {
      if (length($_) > $widest) {
        $widest = length($_);
      }
    }
    if ($OldStyleErrTolReport) {
      print join(" ", 
        "<B>Score",
        (" " x $massDP) . $delta,
        " Hit ",
        "Protein" . (" " x ($widest - 7)),
        "Peptide</B>");
      for (my $i = 0; $i <= $#{ $columns[0] }; $i++) {
        print "<BR>";
        print join(" ", 
          ${ $columns[0] }[$i],
          ${ $columns[2] }[$i],
          ${ $columns[3] }[$i],
          common_subs::htmlEscape(${ $columns[4] }[$i]) . (" " x ($widest - length(${ $columns[4] }[$i]))),
          ${ $columns[5] }[$i]);
      }
    } else {
      print join(" ", 
        "<B>Score",
        "   Expect",
        (" " x $massDP) . $delta,
        " Hit ",
        "Protein" . (" " x ($widest - 7)),
        "Peptide</B>");
      for (my $i = 0; $i <= $#{ $columns[0] }; $i++) {
        print "<BR>";
        if (${ $columns[6] }[$i]) {
          print "<I>";
          ${ $columns[1] }[$i] = " " x length(${ $columns[1] }[$i]);
        }
        print join(" ", 
          ${ $columns[0] }[$i],
          ${ $columns[1] }[$i],
          ${ $columns[2] }[$i],
          ${ $columns[3] }[$i],
          common_subs::htmlEscape(${ $columns[4] }[$i]) . (" " x ($widest - length(${ $columns[4] }[$i]))),
          ${ $columns[5] }[$i]);
        if (${ $columns[6] }[$i]) {
          print "</I>";
        }
      }
    }
    print "</PRE></TD></TR></TABLE></TD></TR></TABLE></DIV>\n";
  }
}

###############################################################################
# &spoolProtMatch()
# output a protein hit
# globals:
# my($objResFile, $objSummary, @spoolText, $fileIn, $massDP,
# $reportType, %et_accession_list, $shipper, $msresFlags, $objParams);
###############################################################################

sub spoolProtMatch {
  my ($objProtein, $hitNum, $length_cache, $sameset_proteins, $sameset_ordering) = @_;

  my $spoolTextBoundary;
  my $hitLineBoundary;
  my %PAI;  # track # observed peptides for emPAI
  
  printHitLine($objProtein, $hitNum);

  my $preferredAcc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  my $score4Main = $objProtein->getScore();

# place holder for emPAI info
  $hitLineBoundary = $#spoolText;

  if ($reportType eq 'peptide'
    && (atLeastVersion(1,7,15) || $shipper eq 'TRUE')) {
    push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n";
    push @spoolText, "<TR><TD ALIGN=RIGHT><INPUT TYPE=\"checkbox\" NAME=\"INCLUDE\" VALUE=\"$hitNum\"";
    push @{ $et_accession_list{$hitNum} }, $objProtein->getAccession();
    if ($objResFile->isErrorTolerant()) {
      my $tempString = $objProtein->getAccession();
      $tempString =~ s/([\\|()[\]{}^\$*+?.])/\\$1/g;
      if ($objParams->getACCESSION() =~ /\"$tempString\"/) {
        push @spoolText, " CHECKED";
      }
    }
    if ($showOverview) {
      push @spoolText, ">";
    } else {
      push @spoolText, " onClick=\"SaveClicks(this,this.form)\">";
    }
    push @spoolText, "</TD><TD NOWRAP><TT>";
    if (atLeastVersion(1,7,15)) {
      push @spoolText, "Check to include this hit in error tolerant search";
      if ($shipper eq 'TRUE') {
        push @spoolText, " or archive report";
      }
    } elsif ($shipper eq 'TRUE') {
      push @spoolText, "Check to include this hit in archive report";
    }
    push @spoolText, "</TT></TD></TR>\n";
    push @spoolText, "<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>";
    push @spoolText, "<TD>&nbsp;</TD></TR>\n</TABLE>\n";
  }

# place holder for protein level quantitation info
  $spoolTextBoundary = $#spoolText;
    
  if ($reportType ne 'concise') {
    tableHeader($objProtein);

  # Enumerate the peptides that matched this protein
    if ($reportType eq 'select') {
    # For Select report, need to preliminary loop to make lists of duplicate matches
      my(%selectFamilyArray, %selectHighScore, %selectHighScorePepNum);
      my @dupe_key = ();
      for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
        if ($objProtein->getPeptideDuplicate($i) != $msparser::ms_protein::DUPE_DuplicateSameQuery) {
          my $queryNum = $objProtein->getPeptideQuery($i);
          $selectList[$queryNum] = 1;
          my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
          my $key = duplicate_group_main(\%dupe_state_cache, $objResFile, $objParams, $objSummary, $objPeptide, $quantProtocolType);
          $dupe_key[$i] = $key;
          push @{ $selectFamilyArray{$key} }, $i;
          unless ($selectHighScore{$key} && $objPeptide->getIonsScore() <= $selectHighScore{$key}) {
            $selectHighScore{$key} = $objPeptide->getIonsScore();
            $selectHighScorePepNum{$key} = $i;
          }
        }
      }
      for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
        if ($objProtein->getPeptideDuplicate($i) != $msparser::ms_protein::DUPE_DuplicateSameQuery) {
          my $queryNum = $objProtein->getPeptideQuery($i);
          my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
          my $key = $dupe_key[$i];
          if ($selectHighScorePepNum{$key} == $i) {
            push @spoolText, "<TR>";
            tableBody($objProtein, $hitNum, $queryNum, $i, $objPeptide, 0);
            foreach my $j (@{ $selectFamilyArray{$key} }) {
              unless ($i == $j) {
                my $queryNum = $objProtein->getPeptideQuery($j);
                my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($j));
                tableBody($objProtein, $hitNum, $queryNum, $j, $objPeptide, 1);
              }
            }
            push @spoolText, "</TD></TR>\n";
          }
        }
      }
      push @spoolText, "</TABLE>\n";
    } else {
      for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
        if ($objProtein->getPeptideDuplicate($i) != $msparser::ms_protein::DUPE_DuplicateSameQuery) {
          my $queryNum = $objProtein->getPeptideQuery($i);
          my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
          push @spoolText, "<TR>";
          tableBody($objProtein, $hitNum, $queryNum, $i, $objPeptide, 0);
          push @spoolText, "</TD></TR>\n";
        }
      }
      push @spoolText, "</TABLE>\n";
    }

  # In protein summary, list unmatched experimental mass values
    if ($reportType eq 'protein' && $objProtein->getUnmatchedMasses($objResFile, $massDP)) {
      push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>";
      push @spoolText, "<TD><TT><B>No match to: </B>";
      push @spoolText, $objProtein->getUnmatchedMasses($objResFile, $massDP);
      push @spoolText, "</TT></TD></TR>\n</TABLE>";
    } 
    push @spoolText, "<BR>\n";
  }

# insert emPAI information
  do {
    my $L = -1;
    if ($objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB)) {
      # no need to provide a length
    } else {
      if ($length_cache) {
        if (exists $$length_cache{$objParams->getDB($objProtein->getDB)}{$objProtein->getAccession()}) {
          $L = $$length_cache{$objParams->getDB($objProtein->getDB)}{$objProtein->getAccession()}{'prot_len'};
          # It's possible the database is offline. In that case, $length_cache
          # will be given, but it won't contain anything (for this database).
          $L = -1 if not defined $L;
        }
      }
    }

    my $empai = $objSummary->getProteinEmPAI($objProtein->getAccession(), $objProtein->getDB(), $L);

    if ($empai > 0) {
      for (my $i = $hitLineBoundary; $i > $hitLineBoundary - 10; $i--) {
        if ($spoolText[$i] =~ /Sequences:/) {
          $spoolText[$i] .= "&nbsp;&nbsp;<B>emPAI:</B>&nbsp;"
            . sprintf("%.2f", $empai);
          last;
        }
      }
    }
  };
  
# insert quantitation summary information for the primary hit
  if ($objQuantMethod) {
    splice @spoolText, $spoolTextBoundary + 1, 0, 
      quant_subs::quant_summary($objMS2Quant, $hitNum, $objProtein->getAccession(), $objProtein->getDB(), \@headerQuant);
  }
  
# Now display list of all proteins that contained subsets or and
# identical list of the matching peptides
  if ($msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS) {
    if ($reportType ne 'concise' and @$sameset_proteins) {
      push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n<TR><TD COLSPAN=2><TT>&nbsp;</TT></TD></TR>\n"
        . "<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>"
        . "<TD><TT><B>Proteins matching the same set of peptides:</B></TT></TD></TR>\n</TABLE>\n"
        ;
    }

    for my $objProtein (@$sameset_proteins) {
      if ($reportType eq 'peptide' && (&atLeastVersion(1,7,15) || $shipper eq 'TRUE')) {
        push @{ $et_accession_list{$hitNum} }, $objProtein->getAccession();
      }
      printHitLine($objProtein);
    }

    if ($msresFlags & $msparser::ms_mascotresults::MSRES_SHOW_SUBSETS) {
      if ($reportType ne 'concise'
              and $objSummary->getNextSubsetProtein($hitNum, 1)) {
        push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n<TR><TD COLSPAN=2><TT>&nbsp;</TT></TD></TR>\n"
          . "<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>"
          . "<TD><TT><B>Proteins matching a subset of these peptides:</B></TT></TD></TR>\n</TABLE>\n"
          ;
      }

      my @subsets = ();
      my $i = 1;
      while (my $objSub = $objSummary->getNextSubsetProtein($hitNum, $i++)) {
        push @subsets, $objSub;
      }

      if ($sameset_ordering) {
        @subsets = common_subs::sort_by_sameset_ordering(\@subsets, $sameset_ordering);
      }

      for my $objSub (@subsets) {
        my @sameset = common_subs::compileSameset($objSummary, $objSub, 1, $sameset_ordering);
        my $objProtein = shift @sameset;
        unless ($reportType ne 'concise' && $objProtein->getScore() < $score4Main * (1 - $ShowSubSets)) {
          push @{ $et_accession_list{$hitNum} }, $objProtein->getAccession();
          printHitLine($objProtein);
        }
      }
    }
  }
 
  if ($reportType ne 'protein') {
    push @spoolText, "\n<HR>\n";  
  }  

}

###############################################################################
# &tableBody()
# output row of peptide match table
# $_[0] $objProtein (undef indicates the unassigned list)
# $_[1] $hitNum (0 indicates the unassigned list)
# $_[2] $queryNum
# $_[3] peptide 'Number'
# $_[4] $objPeptide
# $_[5] $justLink (1 to output just link for select report)
# globals:
# my($objResFile, @spoolText, $massDP, $reportType, $showOverview,
# %pepSeqByHQ, $objProtein, $fileIn, $objSummary
# %pepRankByHQ, @colourKey, $thisScript, %urlParams);
###############################################################################

sub tableBody {
  my ($objProtein, $hitNum, $queryNum, $i, $objPeptide, $justLink) = @_;

  my $fieldCode = $massDP + 7  ;
  $fieldCode = '%' . $fieldCode . '.' . $massDP . 'f';
  
  my $tempMods;
  my $peptideString = $objPeptide->getPeptideStr(1);
  
  if ($hitNum) {
    if ($showOverview 
      && $objPeptide->getAnyMatch 
      && (!$objResFile->isMSMS() || $objPeptide->getIonsScore())) {
      $pepSeqByHQ{"h$hitNum\_q$queryNum"} = $peptideString;
      $pepRankByHQ{"h$hitNum\_q$queryNum"} = $objPeptide->getRank();
    }
  }
  
  my $startColour = "";
  my $endColour = "";
  my $startBold = "";
  my $endBold = "";
  unless ($justLink) {
    if ($reportType eq 'protein') {
    # In a protein summary for a PMF mixture, matches are colour coded back to protein
      if ($objProtein->isPMFMixture()) {
        $startColour = "<FONT COLOR="
          . $colourKey[$objProtein->getPeptideComponentID($i)]
          . ">";
        $endColour = "</FONT>";
      }
    } else {
    # In peptide type summaries, top ranking peptide matches are shown in red
      if ($hitNum) {
        if ($objPeptide->getPrettyRank() == 1) {
          $startColour = "<FONT COLOR=#FF0000>";
          $endColour = "</FONT>";
        }
      } else {
        if ($peptideString) {
        # unassigned list: if there's a peptide match, then must be rank 1
          $startColour = "<FONT COLOR=#FF0000>";
          $endColour = "</FONT>";
        } else {
        # unassigned list: if there isn't a peptide match, then use italics
          $startColour = "<I>";
          $endColour = "</I>";
        }
      }
    }
  # In peptide type summaries, the first appearance of each query is shown in bold
    unless ($reportType eq 'protein') {
      if ($hitNum) {
        if ($objProtein->getPeptideIsBold($i)) {
          $startBold = "<B>";
          $endBold = "</B>";
        }
      } else {
        if ($objSummary->getUnassignedIsBold($i)) {
          $startBold = "<B>";
          $endBold = "</B>";
        }
      }
    }
  # If report is peptide summary and there is no overview table, include a checkbox
  # to allow a query to be selected for a repeat search. The checkbox is shown
  # alongside the first appearance of the rank 1 match, which may be in the unassigned.
  # Always include checkboxes for rank 1 matches in unasssigned list
    my $checkBoxString = "&nbsp;";
    unless ($showOverview && $hitNum) {
      if ((!$hitNum && $reportType ne 'unassigned')
        || ($reportType eq 'peptide' && $objProtein && $objProtein->getPeptideShowCheckbox($i))) {
        $checkBoxString = "<INPUT TYPE=\"checkbox\" NAME=\"QUE\" VALUE=\""
          . common_subs::htmlEscape($objResFile->getRepeatSearchString($queryNum))
          . "\" CHECKED>";    
      }
    }
    push @spoolText, "<TD ALIGN=RIGHT>$checkBoxString</TD>";
  }
  
# collect quantitation information for this match before potential exit if $justLink is true
  my @spoolQuantText; 
  if ($objQuantMethod and $hitNum > 0) {
    my $key = msparser::ms_peptide_quant_key->new($objPeptide->getQuery(), $objPeptide->getRank());

      # XXX
    if (0 == @{ $objMS2Quant->getComponentIntensities($key) }) {
      if ($objQuantMethod->isReportDetail() && !$justLink) {
        for (my $i = 0; $i < @reportRatioNames; $i++) {
          push @spoolQuantText, "<TD ALIGN=CENTER NOWRAP BGCOLOR=#FFFF99><TT>$startColour$startBold" 
            . "---&nbsp;$endBold$endColour</TT></TD>";
        }
      }
    } else {
      if ($objQuantMethod->isReportDetail() && !$justLink) {
        for my $name (@reportRatioNames) {
          push @spoolQuantText, "<TD ALIGN=CENTER NOWRAP BGCOLOR=#FFFF99><TT>$startColour$startBold";
          if ($objMS2Quant->hasPeptideRatio($key, $name)) {
            my $ratio = $objMS2Quant->getPeptideRatio($key, $name);
            if ($ratio->isInfinite()) {
              push @spoolQuantText, "###";
            } else {
              push @spoolQuantText, sprintf("%.3f", $ratio->getValue());
            }
          } else {
            push @spoolQuantText, "---";
          }
          push @spoolQuantText, "&nbsp;$endBold$endColour</TT></TD>";
        }
      }
    }
  }

# For an MS/MS report, the query number is a hyperlink to a peptide view
# report. Mouse over the link causes the top 10 peptide matches list to pop-up.
  if ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    if ($justLink) {
      push @spoolText, "&nbsp;";
    } else {
      push @spoolText, "<TD ALIGN=RIGHT>";
    }
    push @spoolText, "<TT><A HREF=\"$pepViewLink?file=".common_subs::urlEscape($fileIn)."&query=$queryNum&hit=";
    if ($hitNum) {
      push @spoolText, $objSummary->getSrcRank($queryNum, $objPeptide->getRank()) . "&index=";
      unless ($UniGeneFile) {
        push @spoolText, common_subs::urlEscape($objProtein->getAccession()); 
        push @spoolText, "&db_idx=", $objSummary->getHit($hitNum)->getDB();
      }
    } else {
      push @spoolText, "1&index=";
    }
    push @spoolText, "&px=1&section=";
    push @spoolText, $objSummary->getSrcSection($queryNum, $objPeptide->getRank());
    push @spoolText, "&ave_thresh=";
    push @spoolText, $averageThreshold;
    if ($ignoreIonsScoreBelowDefined) {
      push @spoolText, "&_ignoreionsscorebelow=";
      push @spoolText, $ignoreIonsScoreBelow;
    }
    push @spoolText, "&report=";
    push @spoolText, $numHits;
    push @spoolText, "&_sigthreshold=";
    push @spoolText, $sigThreshold;
    push @spoolText, "&_msresflags=";
    push @spoolText, $msresFlags;
    push @spoolText, "&_msresflags2=";
    push @spoolText, $msresFlags2;
    push @spoolText, "&percolate=";
    push @spoolText, $percolate;
    push @spoolText, "&percolate_rt=";
    push @spoolText, $percolate_rt;
    push @spoolText, "&_minpeplen=";
    push @spoolText, $minPepLen;
    if ($sessionID) {
      push @spoolText, '&sessionID='.common_subs::urlEscape($sessionID);
    }
    if (@peptideViewFromTo) {
      push @spoolText, "&from=" . $peptideViewFromTo[0] . "&to=" . $peptideViewFromTo[1];
    }
    push @spoolText, "\" TARGET=\"_blank\"";
    if($showToolTips) {
      push @spoolText, " onMouseOver=\"activateEl('Q$queryNum', event)\"";
      push @spoolText, " onMouseOut=\"clearEl()\"";
    }
    if ($justLink) {
      push @spoolText, ">" . $queryNum . "</A></TT>";
      return 1;
    } else {
      push @spoolText, ">" . $queryNum . "</A>&nbsp;&nbsp;</TT></TD>";
    }
  }

  push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold" 
    . sprintf("$fieldCode", $objPeptide->getObserved()) 
    . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
  my $mrExpt = $objResFile->getObservedMrValue($queryNum);
  if ($objPeptide->getAnyMatch) {
    $mrExpt = $objPeptide->getMrExperimental;    # only works if there's a match
  }
  push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold"
    . sprintf("$fieldCode", $mrExpt)
    . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
  if ($peptideString) {
    push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold" 
      . sprintf("$fieldCode", $objPeptide->getMrCalc()) 
      . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
  } else {
    push @spoolText, "<TD>&nbsp;</TD>";
  }
  if ($peptideString) {
    push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT>$startColour$startBold"
      . &err2Tolu($objPeptide->getDelta(), $objPeptide->getMrCalc()) 
      . "&nbsp;$endBold$endColour</TT></TD>";
  } else {
    push @spoolText, "<TD>&nbsp;</TD>";
  }
  if ($reportType eq 'protein') {
    push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold" . sprintf("%5d", $objProtein->getPeptideStart($i)) . "&nbsp;$endBold$endColour</TT></TD>";
    push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT>$startColour$startBold-&nbsp;$endBold$endColour</TT></TD>";
    push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold" . sprintf("%5d", $objProtein->getPeptideEnd($i)) . "&nbsp;$endBold$endColour</TT></TD>";
  }
  if ($peptideString) {
    push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold"
      . sprintf("%1d", $objPeptide->getMissedCleavages()) 
      . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
  } else {
    push @spoolText, "<TD>&nbsp;</TD>";
  }
  if ($objResFile->isPMF() && $reportType eq 'protein' && $objProtein->isPMFMixture()) {
    push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold"
      . sprintf("%1d", $objProtein->getPeptideComponentID($i)) 
      . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
  } elsif (!$objResFile->isPMF()) {
    if ($peptideString) {
      if ($objPeptide->getIonsScore()){
        push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT>$startColour$startBold"; 
        if ($hitNum && $objProtein->getPeptideDuplicate($i) == $msparser::ms_protein::DUPE_Duplicate) {
          push @spoolText, sprintf("(%.0f)&nbsp;", $objPeptide->getIonsScore()); 
        } else {
          push @spoolText, sprintf("%.0f&nbsp;&nbsp;", $objPeptide->getIonsScore()); 
        }
        push @spoolText, "$endBold$endColour</TT></TD>";
      } else {
        push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT>$startColour$startBold---&nbsp;&nbsp;$endBold$endColour</TT></TD>";
      }
    } else {
      push @spoolText, "<TD>&nbsp;</TD>";
    }
    unless ($reportType eq 'protein' || $reportType eq 'concise') {
      if ($objParams->getFORMVER() < 1.01
        || !$peptideString
        || !$objPeptide->getIonsScore()
        || $OldStyleErrTolReport) {
        push @spoolText, "<TD>&nbsp;</TD>";
      } else {
        my $expect = $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $queryNum, $objPeptide->getRank());
        if ($expect > 0) {
          push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT>$startColour$startBold" 
          . sprintf("%.2g", $expect) 
          . "&nbsp;$endBold$endColour</TT></TD>";
        } else {
          push @spoolText, "<TD>&nbsp;</TD>";
        }
      }
      if ($peptideString) {
        push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold" 
          . sprintf("%1d", $objPeptide->getPrettyRank())  
          . "&nbsp;&nbsp;$endBold$endColour</TT></TD>";
        push @spoolText, "<TD ALIGN=RIGHT><TT>$startColour$startBold"; 
        if ($hitNum && $objSummary->isPeptideUnique($queryNum, $objPeptide->getRank)) {
          push @spoolText, "U";
        } else {
          push @spoolText, "&nbsp;";
        }
        push @spoolText, "&nbsp;&nbsp;&nbsp;$endBold$endColour</TT></TD>";
      } else {
        push @spoolText, "<TD>&nbsp;</TD>";
        push @spoolText, "<TD>&nbsp;</TD>";
      }
    }
  }
# output any quantitation text created earlier
  if (@spoolQuantText) {
    push @spoolText, @spoolQuantText;
  }
  if ($peptideString) {
    my $local_pep = $peptideString;
    unless ($reportType eq 'protein' || $reportType eq 'concise') {
    # use the variable mods string to underline the peptide string at modified residues
      my $varModsString = $objPeptide->getVarModsStr();
      my $summedModsString = $objPeptide->getSummedModsStr();
      $summedModsString = '0' x length($varModsString) if $summedModsString eq '';
      my $localModsString = $objPeptide->getLocalModsStr();
      $localModsString = '0' x length($varModsString) if $localModsString eq '';
      if ($varModsString =~ /[1-Y]/ or $summedModsString =~ /[1-Y]/ or $localModsString =~ /[1-Y]/) {
        for (my $i = length($varModsString) - 2; $i > 0; $i--) {
          if (substr($varModsString, $i, 1) =~ /[1-Y]/
                  or substr($summedModsString, $i, 1) =~ /[1-Y]/
                  or substr($localModsString, $i, 1) =~ /[1-Y]/) {
            substr($local_pep, $i - 1, 1) = "<U>" . substr($local_pep, $i - 1, 1) . "</U>";
          }
        }
      }
    }
    if (atLeastVersion(2, 0, 99) && $hitNum) {
      my $varModsString = $objPeptide->getVarModsStr();
      my $summedModsString = $objPeptide->getSummedModsStr();
      my $localModsString = $objPeptide->getLocalModsStr();
      my $aa_before = $objProtein->getPeptideResidueBefore($i);
      my $aa_after = $objProtein->getPeptideResidueAfter($i);
      $aa_before =~ s/@/-/;
      $aa_after =~ s/@/-/;
      if (substr($varModsString, 0, 1) =~ /[1-Y]/
              or substr($summedModsString, 0, 1) =~ /[1-Y]/
              or substr($localModsString, 0, 1) =~ /[1-Y]/) {
        $aa_before = "$aa_before<U>.</U>";
      } else {
        $aa_before = "$aa_before.";
      }
      if (substr($varModsString, length($varModsString) - 1, 1) =~ /[1-Y]/
              or substr($summedModsString, length($summedModsString) - 1, 1) =~ /[1-Y]/
              or substr($localModsString, length($localModsString) - 1, 1) =~ /[1-Y]/) {
        $aa_after = "<U>.</U>$aa_after";
      } else {
        $aa_after = ".$aa_after";
      }
      $local_pep = "$aa_before$local_pep$aa_after";
    }
    push @spoolText, "<TD NOWRAP><TT>$startColour$startBold&nbsp;$local_pep";

    if ($objPeptide->getComponentStr()) {
      push @spoolText, " [" . $objPeptide->getComponentStr() . "]";
    } elsif ($quantProtocolType eq 'PRECURSOR') {
      my $comp = msparser::ms_quant_component->new();
      my $status = $objSummary->getQuantitationComponentForPeptide($objPeptide, $comp, $objQuantMethod);
      if ($status == $msparser::ms_mascotresults::QCS_OK) {
        push @spoolText, " [" . $comp->getName() . "]";
      }
    }

    $tempMods = $objSummary->getReadableVarMods($queryNum, $objPeptide->getRank(), $massDP);
    if ($reportType ne 'select' && $tempMods) {
      if ($objResFile->isErrorTolerant()) {
        push @spoolText, _possible_etmods(
          $tempMods, $objPeptide, $hitNum, $queryNum
        );
      } else {
        push @spoolText, " + " . $tempMods;
      }
    }
    push @spoolText, "$endBold$endColour</TT>";
  }

}

sub _possible_etmods {
  my ($tempMods, $objPeptide, $hitNum, $queryNum) = @_;

# mod found by error tolerant search, if any, is always last in list
  my @tempList = split(/;/, $tempMods);
  my ($modName, $modDelta) = $tempList[-1] =~ /([^[]*)\[([-+][\d.]+)\]/;

  if (not defined $modName or not defined $modDelta) {
    return " + " . $tempMods;
  }

  $modName =~ s/^\s*//;
  $modName =~ s/\s*$//;

  my %base = (
    'NA_INSERTION'    => '[Base insertion]',
    'NA_DELETION'     => '[Base deletion]',
    'NA_SUBSTITUTION' => '[Base substitution]',
  );

  if (my $text = $base{$modName}) {
    $tempList[-1] = $text;
    return " + " . join(';', @tempList);
  }

  $tempList[-1] = "";
  my @modpositions = etmodified_termini_and_residues($objPeptide);
  if (@modpositions > 1) {
    $tempList[-1] .= ' '.scalar(@modpositions).' ';
  }
  # First one in list could have non-empty $terminus.
  my (undef, $terminus, $residue) = @{ $modpositions[0] };
  if ($showToolTips) {
    findDeltas($modName, $terminus, $residue, $modDelta, "ET$hitNum$queryNum", $objPeptide);
    $tempList[-1] .= " [<A HREF=\"javascript:void(0)\" "
      . " onMouseOver=\"activateEl('ET$hitNum$queryNum', event)\" onMouseOut=\"clearEl(); return true\">"
      . "$modDelta</A> at ";
  } else {
    $tempList[-1] = " [<A HREF=\"javascript:void(0)\">$modDelta</A> at ";
  }
  if ($terminus) {
    $tempList[-1] .= "$terminus-term $residue]";
  } else {
    my @residue_list = map { sprintf("%s%d", $_->[2], $_->[0]) } @modpositions;
    $tempList[-1] .= join(",", @residue_list) . "]";
  }

  return " + " . join(';', @tempList);
}

###############################################################################
# &tableHeader()
# output column headers for peptide match table
# $_[0] $objProtein (undef for the unassigned list)
# globals:
# my($objResFile, @spoolText, $reportType, $objProtein, $objParams);
###############################################################################

sub tableHeader {
  my ($objProtein) = @_;

  push @spoolText, "<TABLE BORDER=0 CELLPADDING=0 CELLSPACING=0>\n";
  push @spoolText, "<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>";
  if ($reportType eq 'peptide' || $reportType eq 'archive' 
    || $reportType eq 'select' || $reportType eq 'unassigned') {
    push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
    push @spoolText, "Query&nbsp;&nbsp;";
    push @spoolText, "</B></TT></TD>";
  }
  push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
  push @spoolText, "Observed&nbsp;&nbsp;";
  push @spoolText, "</B></TT></TD>";
  push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT><B>";
  push @spoolText, "Mr(expt)&nbsp;&nbsp;";
  push @spoolText, "</B></TT></TD>";
  push @spoolText, "<TD ALIGN=RIGHT NOWRAP><TT><B>";
  push @spoolText, "Mr(calc)&nbsp;&nbsp;";
  push @spoolText, "</B></TT></TD>";
  push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
  my $delta = "Delta";
  if ($objParams->getTOLU() eq "%") {
    $delta = "%&nbsp;&nbsp;";
  } elsif ($objParams->getTOLU() eq "ppm") {
    $delta = "ppm&nbsp;";
  } elsif ($objParams->getTOLU() eq "mmu") {
    $delta = "mmu&nbsp;";
  }      
  push @spoolText, "&nbsp;" . $delta . "&nbsp;";
  push @spoolText, "</B></TT></TD>";
  if ($reportType eq 'protein') {
    push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
    push @spoolText, "Start&nbsp;";
    push @spoolText, "</B></TT></TD>";
    push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
    push @spoolText, "&nbsp;</B></TT></TD>";
    push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
    push @spoolText, "End&nbsp;";
    push @spoolText, "</B></TT></TD>";
  }
  push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
  push @spoolText, "Miss&nbsp;";
  push @spoolText, "</B></TT></TD>";
  if ($objResFile->isPMF() && $reportType eq 'protein' && $objProtein->isPMFMixture()) {
    push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
    push @spoolText, "Comp&nbsp;";
    push @spoolText, "</B></TT></TD>";
  } elsif (!$objResFile->isPMF()) {
    if ($reportType eq 'protein' || $reportType eq 'concise') {   
      push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
      push @spoolText, "Ions&nbsp;";
      push @spoolText, "</B></TT></TD>";
    } else {
      push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
      push @spoolText, "Score&nbsp;";
      push @spoolText, "</B></TT></TD>";
      push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
      push @spoolText, "Expect&nbsp;";
      push @spoolText, "</B></TT></TD>";
      push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
      push @spoolText, "Rank&nbsp;";
      push @spoolText, "</B></TT></TD>";
      push @spoolText, "<TD ALIGN=RIGHT><TT><B>";
      push @spoolText, "Unique&nbsp;";
      push @spoolText, "</B></TT></TD>";
    }
  }
  if ($objQuantMethod
    && $objQuantMethod->isReportDetail()
    && $objProtein) {
    for my $name (@reportRatioNames) {
      push @spoolText, "<TD ALIGN=RIGHT BGCOLOR=#FFFF99><TT><B>";
      push @spoolText, common_subs::htmlEscape($name) . "&nbsp;";
      push @spoolText, "</B></TT></TD>";
    }
  }
  push @spoolText, "<TD><TT><B>";
  push @spoolText, "&nbsp;Peptide</B></TT></TD></TR>\n";

}

###############################################################################
# &printHitLine()
# output protein information for individual protein hit
# $_[0] is a protein object
# $_[1] is hit number, or 0 if no hit number required
# globals:
# my(@spoolText, $fileIn, @numMatches, $objSummary, @colourKey, 
# $sigThreshold, $reportType, $thisScript, %urlParams, $objResFile, $objParams);
# spools summary line(s) for protein match onto @spoolText
###############################################################################

sub printHitLine {
  my ($objProtein, $hitNum) = @_;

  push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n";
  push @spoolText, "<TR><TD><TT><B>";
  if ($hitNum) {
    push @spoolText, "<A NAME=\"Hit$hitNum\">$hitNum.</A>";
    push @spoolText, "&nbsp;" x (5 - length($hitNum));
  } else {
    push @spoolText, "&nbsp;" x 6;
  }
  push @spoolText, "</B></TT></TD>";
  push @spoolText, "<TD NOWRAP><TT>";
  if ($objProtein->isPMFMixture || $ShowDecoyReport) {
    push @spoolText, "<B>" . common_subs::htmlEscape($objProtein->getAccession) . "</B>";
  } else {
    push @spoolText, "<A HREF=\"$protViewLink?file=".common_subs::urlEscape($fileIn)."&hit=";
    if ($reportType eq 'protein' || $reportType eq 'concise') {
      push @spoolText, $objProtein->getProteinSummaryHit();
      push @spoolText, "&db_idx=", $objProtein->getDB();
    } else {
      push @spoolText, common_subs::urlEscape($objProtein->getAccession);
      push @spoolText, "&db_idx=";
      push @spoolText, $objProtein->getDB;
      push @spoolText, "&px=1";
      push @spoolText, "&ave_thresh=";
      push @spoolText, $averageThreshold;
      if ($ignoreIonsScoreBelowDefined) {
        push @spoolText, "&_ignoreionsscorebelow=";
        push @spoolText, $ignoreIonsScoreBelow;
      }
      push @spoolText, "&report=";
      push @spoolText, $numHits;
      push @spoolText, "&_sigthreshold=";
      push @spoolText, $sigThreshold;
      push @spoolText, "&_msresflags=";
      push @spoolText, $msresFlags;
      push @spoolText, "&_msresflags2=";
      push @spoolText, $msresFlags2;
      push @spoolText, "&percolate=";
      push @spoolText, $percolate;
      push @spoolText, "&percolate_rt=";
      push @spoolText, $percolate_rt;
      push @spoolText, "&_minpeplen=";
      push @spoolText, $minPepLen;
      if ($UniGeneFile and $objProtein->getDB == 0) {
        push @spoolText, "&UNIGENE=" . common_subs::urlEscape(scalar $thisScript->param($urlParams{'unigene'}));
      }
      if ($sessionID) {
        push @spoolText, "&sessionID=" . common_subs::urlEscape($sessionID);
      }
    }
    push @spoolText, "\" TARGET=\"_blank\">";
    push @spoolText, common_subs::htmlEscape(accession_plus_prefix($objProtein));
    push @spoolText, "</A>";
  }
  push @spoolText, "&nbsp;&nbsp;";
  if ($objProtein->isPMFMixture()) {
    push @spoolText, "&nbsp;&nbsp;<B>Total score:</B>&nbsp;";
  } else {
    if ((my $m = $objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB)) > 0) {
      my $tempString = sprintf ("%.0f", $m);
      push @spoolText, "&nbsp;&nbsp;<B>Mass:</B>&nbsp;$tempString";
      push @spoolText, "&nbsp;" x (7 - length($tempString));
    }
    push @spoolText, "&nbsp;&nbsp;<B>Score:</B>&nbsp;"; 
  }
  if ($objResFile->isPMF()
    && $objProtein->getScore() >= $objSummary->getProteinThreshold(1 / $sigThreshold)) {
    push @spoolText, "<FONT COLOR=#FF0000><B>";
  }
  my $tempString = sprintf ("%.0f", $objProtein->getScore());
  if ($objParams->getFORMVER() < 1.01) {
    $tempString = sprintf ("%9.2e", $objProtein->getScore());
  }
  push @spoolText, $tempString;
  push @spoolText, "&nbsp;" x (5 - length($tempString));
  if ($objResFile->isPMF()
    && $objProtein->getScore() >= $objSummary->getProteinThreshold(1 / $sigThreshold)) {
    push @spoolText, "</B></FONT>";
  }
  if ($reportType eq 'protein' || $reportType eq 'concise') {
    unless ($objParams->getFORMVER() < 1.01) {
      push @spoolText, "&nbsp;&nbsp;<B>Expect:</B>&nbsp;";
      push @spoolText, sprintf ("%.2g", 1 / $objSummary->getProbOfProteinBeingRandomMatch($objProtein->getScore()));
    }
  }
  if ($reportType eq 'protein' || $reportType eq 'concise') {
    push @spoolText, "&nbsp;&nbsp;<B>Matches:</B>&nbsp;" 
      . $objProtein->getNumDisplayPeptides(0);
  } else {
    push @spoolText, "&nbsp;&nbsp;<B>Matches:</B>&nbsp;" 
      . $objProtein->getNumDisplayPeptides(0)
      . "("
      . $objProtein->getNumDisplayPeptides(1)
      . ")";
    push @spoolText, "&nbsp;&nbsp;<B>Sequences:</B>&nbsp;" 
      . $objProtein->getNumDistinctPeptides(0, $msparser::ms_protein::DPF_SEQUENCE)
      . "("
      . $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE)
      . ")";
  }
  if ($objProtein->getFrame() > 0){
    push @spoolText, "&nbsp;&nbsp;<B>Frame:</B>&nbsp;" . $objProtein->getFrame();
  }
  push @spoolText, "</TT></TD></TR>\n";

  my $description;
  if ($objProtein->isPMFMixture() and $reportType eq 'concise') {
    $description = "Components (only one family member shown for each component):";
  } elsif ($objProtein->isPMFMixture()) {
    $description = "";
  } else {
    $description = $protData->mustGetProteinDescription($protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession()), $objProtein->getDB, scalar $thisScript->param($urlParams{_alwaysgettitle}));
  }

  if ($description) {
    push @spoolText, "<TR><TD>&nbsp;</TD><TD NOWRAP><TT>";
    push @spoolText, common_subs::htmlEscape($description);
    push @spoolText, "</TT></TD></TR>\n";
  }

  push @spoolText, "</TABLE>\n";
  
# for a PMF mixture, recurse to print header lines for each component
  if ($objProtein->isPMFMixture()) {
    if ($reportType eq 'concise') {
      for (my $i = 1; $i <= $objProtein->getNumComponents(); $i++) {
        printHitLine($objProtein->getComponent($i));
      }
    } else {
      push @spoolText, "<TABLE BORDER=0 CELLSPACING=0>\n";
      for (my $i = 1; $i <= $objProtein->getNumComponents(); $i++) {
        my $c = $objProtein->getComponent($i);
        push @spoolText, "<TR><TD><TT>" . "&nbsp;" x 6 . "</TT></TD>\n";
        push @spoolText, "<TD NOWRAP><TT>";
        if ($i == 1) {
          push @spoolText, "<B>Components:</B>";
        } else {
          push @spoolText, "&nbsp;" x 11;
        }
        if ($ShowDecoyReport) {
          push @spoolText, " $i. <A HREF=\"javascript:void(0)\">";
        } else {
          push @spoolText, " $i. <A HREF=\"$protViewLink?file=".common_subs::urlEscape($fileIn)."&hit="
            . $c->getProteinSummaryHit()
            . "&db_idx=" . $objProtein->getDB()
            . ($sessionID ? '&sessionID='.common_subs::urlEscape($sessionID) : ())
            . "\" TARGET=\"_blank\">";
        }
        my $prefAcc = $protData->preferredAccession($c->getDB(), $c->getAccession);
        push @spoolText, common_subs::htmlEscape($prefAcc)
          . "</A> <FONT COLOR=" 
          . $colourKey[$i] 
          . ">"; 
	push @spoolText, common_subs::htmlEscape($protData->mustGetProteinDescription($prefAcc, $c->getDB, $thisScript->param($urlParams{_alwaysgettitle})))
	  . "</FONT>\n";
        push @spoolText, "</TT></TD></TR>\n";
      }
      push @spoolText, "</TABLE>\n";
    }
  }

}

###############################################################################
# &formatControls()
# Print controls to re-format report
# globals:
# my($myName, $objResFile, $objParams, $numHits, $sigThreshold, $fileIn,
# $ignoreIonsScoreBelow, $ShowSubSets, $reportType, $sortUnassigned, $thisScript,
# $RequireBoldRed, %urlParams, $mudpitSwitch, $showToolTips);
###############################################################################

sub formatControls {
  my ($preferredTaxonomyAvailable) = @_;

  print "<FORM ACTION=\"./$myName\" NAME=\"Re-format\" METHOD=\"GET\"";
  print " onSubmit=\"return setReFormatAction(this)\">\n";
  print "<INPUT TYPE=\"hidden\" NAME=\"file\" VALUE=\"".common_subs::htmlEscape($fileIn)."\">\n";
  if ($ShowDecoyReport) {
    print "<INPUT TYPE=\"hidden\" NAME=\"_show_decoy_report\" VALUE=\"1\">\n";
  }
  if ($reportType eq 'select'){
    if ($thisScript->param($urlParams{_querylist})) {
      print "<INPUT TYPE=\"hidden\" NAME=\"_querylist\" VALUE=\"" . common_subs::htmlEscape($thisScript->param($urlParams{_querylist})) . "\">\n";
    } else {
      print "<INPUT TYPE=\"hidden\" NAME=\"_querylist\" VALUE=\"all\">\n";
    }
  }
  # Only needed for export_dat_2.pl:
  print "<INPUT TYPE=\"hidden\" NAME=\"group_family\" VALUE=\"0\">\n";
  
  my $titleString;
  my $selectString = "<SELECT NAME=\"REPTYPE\">\n";
  if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_SUMMARY, "h1")
    && !$objResFile->isErrorTolerant()) {
    $selectString .= "  <OPTION VALUE=\"protein\"";
    if ($reportType =~ /protein/i) {
      $titleString = "<H3>Protein Summary Report";
      $selectString .= " SELECTED";
    }
    $selectString .= ">Protein Summary";
    if ($anyPepSumMatch) {
      $selectString .= " (deprecated)";
    }
    $selectString .= "</OPTION>\n";
    unless ($anyPepSumMatch) {
      $selectString .= "  <OPTION VALUE=\"concise\"";
      if ($reportType =~ /concise/i) {
        $titleString = "<H3>Concise Protein Summary Report";
        $selectString .= " SELECTED";
      }
      $selectString .= ">Concise Protein Summary</OPTION>\n";
    }
  }
  if ($anyPepSumMatch or $anyLibraryMatch) {
    $selectString .= "  <OPTION VALUE=\"family\">Protein Family Summary</OPTION>\n";
    $selectString .= "  <OPTION VALUE=\"peptide\"";
    if ($reportType =~ /peptide/i){
      $titleString = "<H3>Peptide Summary Report";
      $selectString .= " SELECTED";
    }
    $selectString .= ">Peptide Summary</OPTION>\n";
    $selectString .= "  <OPTION VALUE=\"select\"";
    if ($reportType =~ /select/i){
      $titleString = "<H3>Select Summary Report";
      $selectString .= " SELECTED";
    }
    $selectString .= ">Select Summary (protein hits)</OPTION>\n";
    $selectString .= "  <OPTION VALUE=\"unassigned\"";
    if ($reportType =~ /unassigned/i){
      $titleString = "<H3>Select Summary Report (unassigned)";
      $selectString .= " SELECTED";
    }
    $selectString .= ">Select Summary (unassigned)</OPTION>\n";
  }
  $selectString .= "  <OPTION VALUE=\"export\">Export Search Results</OPTION>\n";
  $selectString .= "</SELECT>\n";

  
  print $titleString;
  if ($percolate == 1) {
    print " (scores adjusted by Percolator)";
  }
  print "</H3>\n";
  
  print "<TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>\n";
  print "  <TR>\n";
  print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
  print "      <INPUT TYPE=\"submit\" VALUE=\"Format As\">\n";
  print "    </TD>\n";

# REPTYPE
  print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
  print $selectString;
  print "    </TD>\n";

# UNIGENE
  if ($objParams->getNumberOfDatabases() == 1 and ($anyPepSumMatch or $anyLibraryMatch)) {
    my $tempString;
    for (my $i = 1; $i <= $objParams->getNumberOfDatabases(); $i++) {
      if (my $dbs = $objUnigeneOptions->getPropValStringByName($objParams->getDB($i))) {
        unless ($dbs =~ /\//){
          $tempString .= $dbs . " ";
        }
      }
    }
    if ($tempString) {
      my @tempList = split(/\s+/, $tempString);
      print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
      print "UniGene index ";
      print "<SELECT NAME=\"UNIGENE\">\n";
      unshift @tempList, "None";
      my $species = "None";
      if ($thisScript->param($urlParams{'unigene'})) {
        $species = $thisScript->param($urlParams{'unigene'});
      }
      for (my $i = 0; $i < scalar(@tempList); $i++) {
        print "  <OPTION";
        if ($species eq $tempList[$i]) {
          print " SELECTED";
        }
        print ">".common_subs::htmlEscape($tempList[$i])."</OPTION>\n";
      }
      print "</SELECT>\n";
      print "    </TD>\n";
    } else {
      print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
    }
  } else {
  # Help link for PMF 
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "      <A HREF=\"../help/pmf_summaries_help.html\" TARGET=_blank>Help</A>\n";
    print "    </TD>\n";
  }

  if ($anyPepSumMatch or $anyLibraryMatch) {
  # Help link for MS/MS
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "      <A HREF=\"../help/msms_summaries_help.html\" TARGET=_blank>Help</A>\n";
    print "    </TD>\n";
  }
  print "  </TR>\n";
  
# _sigthreshold
  do {
    print "  <TR>\n";
    print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "Significance threshold p&lt; "; 
    print "<INPUT NAME=\"_sigthreshold\" TYPE=text SIZE=8 VALUE=$sigThreshold>\n";
    print "    </TD>\n";
  };

# REPORT
  print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
  print "Max. number of hits ";
  if ($numHits) {
    print "<INPUT NAME=\"REPORT\" TYPE=text SIZE=5 VALUE=$numHits>\n";
  } else {
    print "<INPUT NAME=\"REPORT\" TYPE=text SIZE=5 VALUE=AUTO>\n";
  }
  #if ($libraryMode) {
  #  print "<INPUT TYPE=\"hidden\" NAME=\"library\" VALUE=\"$libraryMode\">\n";
  #}
  print "    </TD>\n";
  if ($anyPepSumMatch or $anyLibraryMatch) {
  # _noerrortolerant
    if ($objResFile->isErrorTolerant()
      && $objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      print "    <TD BGCOLOR=#EEEEFF>";
      print "Hide error tolerant matches&nbsp;";
      print "<INPUT TYPE=\"checkbox\" NAME=\"etchkbox\" onClick=\"";
      print "if (form.etchkbox.checked) { form._noerrortolerant.value = 1; } else { form._noerrortolerant.value = 0; } return true;\"";
      if ($NoErrorTolerant) {
        print " CHECKED";
      }
      print ">\n";
      if ($NoErrorTolerant) {
        print "<INPUT TYPE=\"hidden\" NAME=\"_noerrortolerant\" VALUE=1>";
      } else {
        print "<INPUT TYPE=\"hidden\" NAME=\"_noerrortolerant\" VALUE=0>";
      }
      print "</TD>\n";
    } else {
      if ($percolate != -1) {
      # percolate
        print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "Show Percolator scores ";
        print "<INPUT TYPE=\"checkbox\" NAME=\"percchkbox\" onClick=\"";
        print "if (form.percchkbox.checked) { form.percolate.value = 1; } else { form.percolate.value = 0; } return true;\"";
        if ($percolate == 1) {
          print " CHECKED";
        }
        print ">\n";
        if ($percolate) {
          print "<INPUT TYPE=\"hidden\" NAME=\"percolate\" VALUE=1>\n";
        } else {
          print "<INPUT TYPE=\"hidden\" NAME=\"percolate\" VALUE=0>\n";
        }
        print "    </TD>\n";
      } else {
        print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
      }
    }
  }
  print "  </TR>\n";
  
# Next two rows only for peptide and select
  if (($anyPepSumMatch or $anyLibraryMatch) && ($reportType =~ /peptide/i || $reportType =~ /select/i || $reportType =~ /unassigned/i)) {
    print "  <TR>\n";
    print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";

  # _minpeplen
    print "<INPUT TYPE=\"hidden\" NAME=\"_minpeplen\" VALUE=$minPepLen>\n";
    
  # _server_mudpit_switch
    print "Standard scoring&nbsp;<INPUT TYPE=\"radio\" VALUE=99999999 NAME=\"_server_mudpit_switch\"";
    unless ($objResFile->getNumSeqsAfterTax > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
      print " CHECKED";
    }
    print ">&nbsp;MudPIT scoring&nbsp;<INPUT TYPE=\"radio\" VALUE=0.000000001 NAME=\"_server_mudpit_switch\"";
    if ($objResFile->getNumSeqsAfterTax > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
      print " CHECKED";
    }
    print ">\n";
    print "    </TD>\n";
    
  # _ignoreionsscorebelow
  # If the setting is 2, always show the edit field.
  # If the setting is 0 or 1, show either the edit field (if value is
  # defined and > 0) or a checkbox.
    my $displayNonSigMatches = $objMascotOptions->getDisplayNonSignificantMatches;
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    if ($displayNonSigMatches == 2
            or ($ignoreIonsScoreBelowDefined and $ignoreIonsScoreBelow > 0)) {
      print "Ions score or expect cut-off ";
      print "<INPUT NAME=\"_ignoreionsscorebelow\" TYPE=text SIZE=5 VALUE=$ignoreIonsScoreBelow>\n";
    } else {
      print "Display non-significant matches ";
      print "<INPUT NAME=\"_ignoreionsscorebelow_test\" TYPE=\"checkbox\" onClick=\"";
      print "if (form._ignoreionsscorebelow_test.checked) { form._ignoreionsscorebelow.value = '0'; } else { form._ignoreionsscorebelow.value = '-1'; } return true;";
      print "\"";
      my $default = -1;
      if ($ignoreIonsScoreBelowDefined) {
        if ($ignoreIonsScoreBelow == 0) {
          print " CHECKED";
          $default = 0;
        }
      } else {
        if ($displayNonSigMatches == 1) {
          print " CHECKED";
          $default = 0;
        }
      }
      print ">\n";
      print "<INPUT NAME=\"_ignoreionsscorebelow\" TYPE=\"hidden\" VALUE=\"$default\">\n";
    }
    print "    </TD>\n";
    
  # _showsubsets
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "Show sub-sets ";
    print "<INPUT NAME=\"_showsubsets\" TYPE=text SIZE=2 VALUE=$ShowSubSets>\n";
    print "    </TD>\n";
    print "  </TR>\n";
    
  # _showpopups
    print "  <TR>\n";
    print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "Show pop-ups&nbsp;<INPUT TYPE=\"radio\" VALUE=\"TRUE\" NAME=\"_showpopups\"";
    if ($showToolTips){
      print " CHECKED";
    }
    print ">&nbsp;Suppress pop-ups&nbsp;<INPUT TYPE=\"radio\" VALUE=\"FALSE\" NAME=\"_showpopups\"";
    unless ($showToolTips){
      print " CHECKED";
    }
    print ">\n";
    print "    </TD>\n";
    
  # _sortunassigned
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    if ($reportType =~ /select/i){
      print "&nbsp;<INPUT TYPE=\"hidden\" NAME=\"_sortunassigned\" VALUE=\"$sortUnassigned\">";
    } else {
      print "Sort unassigned ";
      print "<SELECT NAME=\"_sortunassigned\">\n";
      print "  <OPTION VALUE=\"scoredown\"";
      if ($sortUnassigned =~ /scoredown/i){
        print " SELECTED";
      }
      print ">Decreasing Score</OPTION>\n";
      print "  <OPTION VALUE=\"queryup\"";
      if ($sortUnassigned =~ /queryup/i){
        print " SELECTED";
      }
      print ">Increasing query / Mr</OPTION>\n";
      print "  <OPTION VALUE=\"intdown\"";
      if ($sortUnassigned =~ /intdown/i){
        print " SELECTED";
      }
      print ">Decreasing Intensity</OPTION>\n";
      print "</SELECT>\n";
    }
    print "    </TD>\n";
    
  # _requireboldred
    print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    print "Require bold red ";
    print "<INPUT TYPE=\"checkbox\" NAME=\"rbrchkbox\" onClick=\"";
    print "if (form.rbrchkbox.checked) { form._requireboldred.value = 1; } else { form._requireboldred.value = 0; } return true;\"";
    if ($RequireBoldRed) {
      print " CHECKED";
    }
    print ">\n";
    if ($RequireBoldRed) {
      print "<INPUT TYPE=\"hidden\" NAME=\"_requireboldred\" VALUE=1>\n";
    } else {
      print "<INPUT TYPE=\"hidden\" NAME=\"_requireboldred\" VALUE=0>\n";
    }
    print "    </TD>\n";
    print "  </TR>\n";
  }
  
  if ($preferredTaxonomyAvailable) {
    print "  <TR>\n";
    print "    <TD BGCOLOR=#EEEEFF>&nbsp;</TD>\n";
    print "    <TD BGCOLOR=#EEEEFF NOWRAP COLSPAN=2>\n";

    print "Preferred taxonomy&nbsp;";
    
    print DOM->dropdown({
      values => [taxonomy_choice($objResFile)],
      name => '_prefertaxonomy',
      default => $preferTaxonomy,
    })->as_string;
    
    print "    </TD>\n";

    if ($anyPepSumMatch or $anyLibraryMatch) {
      print "    <TD BGCOLOR=#EEEEFF NOWRAP>\n";
    }

    print "  </TR>\n";
  }

  if ($objQuantMethod) {
    print @quantFormatControlText;
  }
  
  print "</TABLE>\n";

  if (defined $sessionID and $sessionID ne '') {
    print '<P><INPUT TYPE="hidden" NAME="sessionID" VALUE="'.$sessionID.'"></P>', "\n";
  }
  
  print "</FORM>\n";

}

###############################################################################
# &atLeastVersion()
# returns 1 if result file is this version or later
# $_[0] major version number for earliest acceptable version
# $_[1] minor version number for earliest acceptable version
# $_[2] minimus version number for earliest acceptable version
# globals:
# my($objResFile);
###############################################################################

sub atLeastVersion {
 
  my ($major, $minor, $minimus) = @_;
  my $requiredVersion = 0;
  if ($major) { $requiredVersion += $major }
  $requiredVersion *= 1000;
  if ($minor) { $requiredVersion += $minor }
  $requiredVersion *= 1000;
  if ($minimus) { $requiredVersion += $minimus }
  
  ($major, $minor, $minimus) = split(/\./, $objResFile->getMascotVer());
  my $actualVersion = 0;
  if ($major) { $actualVersion += $major }
  $actualVersion *= 1000;
  if ($minor) { $actualVersion += $minor }
  $actualVersion *= 1000;
  if ($minimus) { $actualVersion += $minimus }
  
  if ($actualVersion < $requiredVersion) {
    return 0;
  } else {
    return 1;
  } 

}

###############################################################################
# &findDeltas()
# $_[0] modification title reported in result file
# $_[1] terminus "N-term", "C-term", or ""
# $_[2] residue
# $_[3] mass error reported in result file
# $_[4] label for DIV element
# $_[5] peptide object
# globals:
# my($massDP, @errorTolDiv);
# create DIV of possible assignments for mass deltas found
# in error tolerant search. Spool to @errorTolDiv
###############################################################################

sub findDeltas {
  my ($modName, $terminus, $residue, $modDelta, $label, $objPeptide) = @_;

  my $fieldCode = '%+.'.$massDP.'f';

  my $spoolText = "<DIV CLASS='tooltip' ID='".common_subs::htmlEscape($label)."'>"
    . "<TABLE BORDER=0 CELLPADDING=2 CELLSPACING=0 BGCOLOR=#0000ff>"
    . "<TR><TD><TABLE BORDER=0 CELLPADDING=5 CELLSPACING=0 BGCOLOR=#ffffcc><TR><TD><PRE>"
    . "<B>Possible Assignments:</B>\n";

  my @mods = possible_etmod_assignments(
    $objResFile, $objSummary, $objPeptide, $massDP, $sessionID
  );
  for (@mods) {
    $spoolText .= "\n" . common_subs::htmlEscape($_->{'title'})
      . " [" . sprintf($fieldCode, $_->{'delta'}) . "]";
  }

  $spoolText .= "</PRE></TD></TR></TABLE></TD></TR></TABLE></DIV>\n";
  push @errorTolDiv, $spoolText;
  
  return;

}

###############################################################################
# &err2Tolu()
# $_[0] error value to be converted in Da
# $_[1] scale value in Da
# globals:
# my($objParams, $massDP);
# returns formatted string for value in TOLU units
###############################################################################

  sub err2Tolu {

    my $value = shift;
    my $scale = shift;
    my $fieldCode;
    my $fieldWidth = $massDP + 5;

    if ($objParams->getTOLU() eq "%"){
      $fieldCode = '%' . $fieldWidth . '.' . $massDP . 'f';
      return sprintf("$fieldCode", $value * 100 / abs($scale));
    } elsif ($objParams->getTOLU() eq "Da"){
      $fieldCode = '%' . $fieldWidth . '.' . $massDP . 'f';
      return sprintf("$fieldCode", $value);
    } elsif ($objParams->getTOLU() eq "ppm"){
      $value = $value * 1000000 / abs($scale);
      if ($value >= 100) {
        $fieldCode = '%' . $fieldWidth . '.0f';
      } elsif ($value >= 10) {
        $fieldCode = '%' . $fieldWidth . '.1f';
      } else {
        $fieldCode = '%' . $fieldWidth . '.2f';
      }
      return sprintf("$fieldCode", $value);
    } elsif ($objParams->getTOLU() eq "mmu"){
      $value = $value * 1000;
      if ($value >= 100) {
        $fieldCode = '%' . $fieldWidth . '.0f';
      } elsif ($value >= 10) {
        $fieldCode = '%' . $fieldWidth . '.1f';
      } elsif ($value >= 1) {
        $fieldCode = '%' . $fieldWidth . '.2f';
      } else {
        $fieldCode = '%' . $fieldWidth . '.3f';
      }
      $fieldCode = '%' . $fieldWidth . '.1f';
      return sprintf("$fieldCode", $value);
    } else {
      common_subs::fatal("Unrecognised peptide mass tolerance unit", __LINE__, $myName, "");
    }

  }

###############################################################################
# &accession_plus_prefix()
# $_[0] protein object
# globals:
# my($objParams);
# returns accession with database index prefix if multi-database search
# otherwise, returns just accession
###############################################################################

  sub accession_plus_prefix {

    my $objProtein = shift;
    my $acc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());
    my $retVal;

    if ($objParams->getNumberOfDatabases > 1) {
      $retVal = $objProtein->getDB . "::" . $acc;
    } else {
      $retVal = $acc;
    }
    
    return $retVal;
    
  }

sub parse_PMF_mixture_accessions {
  my ($objSummary, $objProtein) = @_;

  my @acc = $objSummary->getProteinDescription($objProtein->getAccession(), $objProtein->getDB) =~ /\"(.+?)\"/g;

  return join(" + ", map { $protData->preferredAccession($objProtein->getDB(), $_) } @acc);
}
  
###############################################################################
# &_percolate()
# run Percolator, if required
###############################################################################

  sub _percolate {
     my ($resfile, $keepalive, $use_proteins, $use_rt) = @_;

     my $bin_createpip = join('/', MascotDat->bindir, MascotDat->bin_createpip);
     my $bin_percolator = join('/', MascotDat->bindir, MascotDat->bin_percolator);

     if (not -x $bin_createpip or not -x $bin_percolator) {
       common_subs::fatal("Unable to find Percolator executables", __LINE__, $myName, "");
     }

     my %filenames = MascotResults::percolator_filenames($resfile);

     _run_createpip(
        $bin_createpip, $resfile->getFileName,
        $filenames{'pipname'}, $keepalive, 
        $use_proteins, $use_rt
     );

     _run_percolator(
        $bin_percolator, MascotResults::percolator_flags($resfile, $use_rt),
        $keepalive
     );
  }

###############################################################################
# &_run_createpip()
# called by &_percolate()
###############################################################################

  sub _run_createpip {
      my ($bin_createpip, $inputname, $pipname, $keepalive, $use_proteins, $use_rt) = @_;

      my $sessionID = $thisScript->param($urlParams{sessionid});

      my @a = ();
      my @r = ();

      if ($use_proteins) {
         push @a, 'numUniqPeps';
      } else {
         push @r, 'numUniqPeps';
      }

      if ($use_rt) {
         # XXX: does createpip need parameters for this?
      }

      my @args = (
         -i => $inputname,
         -o => $pipname,

         -p => $keepalive ? 1 : 0,

         (@a ? (map {+ '-a', $_ } @a) : ()),
         (@r ? (map {+ '-r', $_ } @r) : ()),

         $sessionID ? ('--sessionID' => $sessionID) : (),
      );

      my $status = _run(
          MascotDat->bin_createpip, # for error reporting and pretty-printing
          $keepalive, 
          $bin_createpip, 
          @args,
      );

      return $status;
  }

###############################################################################
# &_run_percolator()
# called by &_percolate()
###############################################################################

  sub _run_percolator {
       my ($bin_percolator, $flags, $keepalive) = @_;

       $flags =~ s/-v \d+//;
       my @args = split /\s+/, $flags;
       push @args, '-v', 2;

       my $status = _run(
          MascotDat->bin_percolator,
          $keepalive,
          $bin_percolator,
          @args,
       );

       return $status;
  }

###############################################################################
# &_run()
# execute with keepalive comments
###############################################################################

  sub _run {
      my ($cmd_str, $keepalive, $cmd, @args) = @_;

      if (not -x $cmd) {
        &common_subs::fatal("Cannot execute ".common_subs::htmlEscape($cmd), __LINE__, $myName, "");
      }

      my $print_comments = sub {};

      if ($keepalive) {
         $print_comments = sub {
             for (@_) {
                chomp;
                print '<!-- ', $_, " -->\n";
             }
          };
      }

      my $status = eval {
          pipefork($print_comments, $print_comments, $cmd, @args)
      };

      if ($@) {
        &common_subs::fatal("Could not fork external process", __LINE__, $myName, "");
      }

      if ($status != 0) {
        &common_subs::fatal("Non-zero exit code from ".common_subs::htmlEscape($cmd), __LINE__, $myName, "");
      }

      return $status;
  }
  
