#!/usr/local/mascot/perl64/bin/perl -w
##############################################################################
# Mascot search progress updates                                             #
##############################################################################
#     $Author: villek@matrixscience.com $ 
#       $Date: 2022-06-24 10:57:52 +0100 $ 
#   $Revision: cae025d431bd23ebef15a79dfbf663d50a0d79ba | MASCOT_REL_2_8_1-3-gcae025d431 $ 
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 2008-2022 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################

package SearchStatus::Parameters;

use strict;


use WebApp::Validators;

use base qw(WebApp::Parameters);

our @BANNED_VARS;

BEGIN {
    @BANNED_VARS = qw(fakemimetype mimetype streamtoken);
}

our %VALIDATE;

BEGIN {
    %VALIDATE = (
        __PACKAGE__->_inheritvar('%VALIDATE'),

        task_id => WebApp::Validators->positive_integer('task_id'),

        _debugjs => WebApp::Validators->boolean('_debugjs'),

        mimetype => WebApp::Validators->mimetype('mimetype'),
        fakemimetype => WebApp::Validators->mimetype('fakemimetype'),

        # If cookies are not supported or enabled...
        'streamtoken' => WebApp::Validators->string_nonempty(
            'streamtoken', '[0-9a-zA-Z;]',
        ),
    );
}

1;

package SearchStatus;

use strict;

use HTML::Entities qw();

use WebApp;
use WebApp::Redirect;

use Request;

use WebApp::Defaults;
use WebApp::Variants;

use AnyML::Node;
use DOM;

use Util::JSON qw(json_script_receive);
use Util::MascotExtBatch qw(parse_ms_status_xml_output);
use Util::Text qw(shell_quote);
use Util::Time qw(iso8601_to_timestamp human_readable_seconds);
use Util::URI qw(make_versioned_resource_link);

our %Variants = (
    WebApp::Variants->html,
    WebApp::Variants->msl_html,
);

use MascotDat;

use Util::Hash qw(query_string);

use msparser;

use Debugging;

sub new {
    my $class = shift;
    bless {}, $class;
}

sub etag { 
    my $self = shift;
    my ($method, $content_type) = @_;

    return unless $method eq 'GET';

    my $raw = join('--',
        Request->complete_uri,
        time,
    );

    my $etag = msparser::ms_fileutilities::getMD5Sum($raw);
    return $etag;
}

sub last_modified { time }

sub expires {
    return time + shift->max_age;
}

sub max_age {
    return 1; # seconds
}

sub filename { 'search_status' }

sub POST {
    my $self = shift;
    my ($content_type) = @_;

    my $uri = URI->new(Request->complete_uri);

    return WebApp::Redirect->new(
        $uri,
        { 
            task_id => SearchStatus::Parameters->param('task_id'), 
            sessionID => SearchStatus::Parameters->param('sessionID'),
            mimetype => undef 
        },
        '303 See Other',
        undef,
        'meta'
    )->GET($content_type);
}

BEGIN {
    my %ERRORS = (
        -2 => 'Job has been killed',
        -1 => 'Job has crashed',
        0 => 'No error',
        1 => 'Failed to open ../data/task_id',
        2 => 'Failed to create ../data/task_id',
        3 => 'Failed to read ../data/task_id',
        4 => 'Failed to write ../data/task_id',
        5 => 'Failed to close ../data/task_id',
        6 => 'The record in the file ../data/task_id has changed -- another task has deleted it?',
        7 => 'mascot.dat is invalid or missing',
        8 => 'Results file cannot be found',
        9 => 'Results file name is longer than 256 characters',
        10 => 'Session timed out',
        11 => 'Permission denied',
    );

    my $sc = '../bin/ms-searchcontrol.exe';
    my $ms_status = '../x-cgi/ms-status.exe';
    my $ms_geterror = '../cgi/ms-geterror.exe';

    sub _run_searchcontrol {
        my ($task_id, $cmd) = @_;

        if (not defined($cmd) or $cmd eq '') {
            die 'must have $cmd';
        }

        if (not -x $sc) {
            return (undef, { error => 'Error: ms-searchcontrol.exe is not executable' });
        }

        # We know $task_id is an integer, as that's validated by
        # SearchStatus::Parameters, so we don't need to quote it.

        my $sessionID = SearchStatus::Parameters->param('sessionID');
        my $ssid = $sessionID ? '--sessionID '.shell_quote($sessionID) : '';

        my $status = qx{$sc --$cmd --taskID $task_id $ssid};

        if (not $status) {
            return (undef, {
                error => 'Failed to run ms-searchcontrol.exe: '.$@
            });
        }

        chomp($status);

        return $status;
    }

    sub _run_ms_geterror {
        my ($error_number) = @_;

        if (not defined($error_number) or $error_number eq '') {
            die 'must have $error_number';
        }

        if (not -x $ms_geterror) {
            return (undef, { error => 'Error: ms-geterror.exe is not executable' });
        }

        # We know $task_id is an integer, as it comes from 
        # ms-searchcontrol.exe. The local CGI environment must be cleared,
        # or else ms-geterror.exe thinks it's running as a CGI script.
        # See documentation in Util::MascotExtBatch.

        delete local @ENV{qw(
            AUTH_TYPE
            CONTENT_LENGTH
            CONTENT_TYPE
            GATEWAY_INTERFACE
            PATH_INFO
            PATH_TRANSLATED
            QUERY_STRING
            REMOTE_ADDR
            REMOTE_HOST
            REMOTE_IDENT
            REMOTE_USER
            REQUEST_METHOD
            SCRIPT_NAME
            SERVER_NAME
            SERVER_PORT
            SERVER_PROTOCOL
            SERVER_SOFTWARE
        )};

        my $status = qx{$ms_geterror HELP $error_number};

        if (not $status) {
            return (undef, {
                error => 'Failed to run ms-geterror.exe: '.$@
            });
        }

        chomp($status);

        return $status;
    }

    sub _find_search_status {
        my ($task_id) = @_;

        my ($status, $error) = _run_searchcontrol($task_id, 'status');
        return $error if $error;

        if ($status =~ /complete/) {
            ($status, $error) = _run_searchcontrol($task_id, 'result_file_name');
            return $error if $error;

            if ($status =~ /filename=(.+)/) {
                return { status => 'complete', filename => $1 };
            }
        }

        if ($status =~ /id_assigned|queued|job_killed|job_not_running/) {
            return { status => $status };
        }

        if ($status =~ /unknown_id/) {
            return { error => 'Error: Unknown task ID', show_output => 1 };
        }

        if ($status =~ /running=(\d+)%/) {
            my $percentage = $1;

            # Bug 11864: Find out if the process is actually running
            # the ExecAfterSearch step(s).

            my ($extra_status, $error) = _run_searchcontrol($task_id, 'status_current_stage');

            if ($error) {
                # Suppress error. We got this far, so return
                # percentage progress.
                return { status => 'running', percentage => $1 };
            }

            # Unlike above commands, status_current_stage has two-line
            # output.
            my ($stage) = $extra_status =~ /^stage=(.+)/mi;
            my ($step) = $extra_status =~ /^step=(\d+)/mi;

            return { status => 'running', percentage => $percentage, stage => $stage, step => $step };
        }

        # This includes 'searchcontrol-error=nnn'.
        if ($status =~ /error=(-?\d+)/) {
            my $num = $1;

            if (not exists $ERRORS{$num}) {
                # Internal error numbers typically indicate that something
                # went wrong while nph-mascot.exe was receiving the uploaded
                # data or parsing input parameters. In this case, the job
                # has probably been erased, so we can't use ms-status.exe
                # to find the intermediate filename or any other details.
                # If the intermediate filename has been assigned, though,
                # we can still get it from ms-searchcontrol.exe, even if
                # nph-mascot.exe didn't create the file or write anything
                # in it.

                my ($progress_log, $advert) = _quick_progress_log($task_id);
                my ($help_text, $help_error) = _run_ms_geterror($num);

                # Ignore $help_error.

                return {
                    error => 'Mascot terminated with error '.$num,
                    progress_log => $progress_log, # could be undef
                    advert => $advert,
                    help_text => $help_text, # could be undef
                    show_output => 1
                };
            }

            # "Job killed" is still a valid status for a job. We can show
            # some details about job ID and task ID to help the user navigate.
            if ($num == -1 or $num == -2) {
                return { status => 'job_killed', show_output => 1 };
            }
            
            return { error => 'Error: '.$ERRORS{$num}, show_output => 1 };
        }

        return { error => 'Error: ms-searchcontrol.exe returned unrecognised status' };
    }

    sub _quick_progress_log {
        my ($task_id) = @_;

        my ($status, $error) = _run_searchcontrol($task_id, 'result_file_name');

        if ($error) {
            # Ignore error silently.
        } else {
            if ($status =~ /filename=(.+)/) {
                return  _read_progress_log($1);
            }
        }

        return;
    }

    sub _find_search_job_id {
        my ($task_id) = @_;

        my ($status, $error) = _run_searchcontrol($task_id, 'mascot_job_number');
        return $error if $error;

        if ($status =~ /mascotjobnumber=(\d+)/) {
            return { job_id => $1 };
        }

        return;
    }

    sub _find_search_job_data {
        my ($job_id) = @_;

        if (not -x $ms_status) {
            return { error => 'Error: ms-status.exe is not executable' };
        }

        my $sessionID = SearchStatus::Parameters->param('sessionID');
        my $session = msparser::ms_session->new($sessionID || '');

        if ($session->isSecurityEnabled()) {
            if (not $session->isPermitted($msparser::ms_security_tasks::SECTASK_USEMSSTATUSEXE)) {
                return { error => 'Error: Not allowed to run ms-status.exe' };
            }
        }

        my $ssid = $sessionID ? 'sessionID='.shell_quote($sessionID) : '';
        my $output;

        do {
            open(my $fh, sprintf("%s Show=JOBDETAILS_XML Job=%d %s |", $ms_status, $job_id, $ssid))
                or return { error => "Cannot run ms-status.exe: $! ($^E)" };

            local $/ = undef;
            $output = readline($fh);

            if (not defined $output) {
                return { error => "Cannot read temporary file: $! ($^E)" };
            }
        };

        my $parse_result = parse_ms_status_xml_output($output);

        if (not $parse_result) {
            return { error => 'Error: No data from ms-status.exe' };
        }

        if ($$parse_result{'error'}) {
            return { error => $$parse_result{'error'} };
        }

        my ($root, $nsprefix) = @$parse_result{qw(root nsprefix)};
        my ($start_time, $end_time, $search_title, $results_filepath);

        for my $entry (@{ $$root{$nsprefix.'JobDetails'} }) {
            # ms-status.exe returns gmtime() in StartTime and EndTime.
            $start_time = $$entry{$nsprefix.'StartTime'}[0]{'_'};
            $end_time = $$entry{$nsprefix.'EndTime'}[0]{'_'};
            $search_title = $$entry{$nsprefix.'SearchTitle'}[0]{'_'};
            $results_filepath = $$entry{$nsprefix.'IntermediateFile'}[0]{'_'};

            last; # should only be one job details element
        }

        my ($progress_log, $advert) = _read_progress_log($results_filepath);

        return {
            search_title => $search_title, # could be undef
            start_time => $start_time, # could be undef
            end_time => $end_time, # will be undef unless the search has crashed or finished
            progress_log => $progress_log,
            advert => $advert,
        };
    }
}

sub _read_progress_log {
    my ($results_filepath) = @_;

    return unless defined $results_filepath and $results_filepath ne '';

    my $log_filepath = MascotDat->decrypt($results_filepath) . '.log';
    my ($progress_log, $advert);

    if (-e $log_filepath and -r _) {
        if (open(my $fh, '<', $log_filepath)) {
            my @lines = <$fh>;
            my $advert_place = -1;

            for my $i (0 .. $#lines) {
                if ($lines[$i] =~ m#\[M00381\]#) {
                    $advert_place = $i;
                    last;
                }
            }

            if ($advert_place >= 0) {
                # Slurp all lines from here backwards until an empty line is
                # encountered. The advert could span multiple lines.
                my $j = $advert_place;

                while ($j >= 0) {
                    if ($lines[$j] =~ m#^\s*$#) {
                        last;
                    }
                    --$j;
                }

                $progress_log = join('', @lines[0 .. $j-1], @lines[$advert_place+1 .. $#lines]);
                $advert = join('', @lines[$j+1 .. $advert_place]);
                $advert =~ s#\[M00381\]##s;
            } else {
                $progress_log = join('', @lines);
            }
        }
    }

    # Bug 12075: nph-mascot.exe encodes HTML entities in error
    # messages, including adverts. Decode here so they're displayed
    # correctly, i.e. passed through as raw HTML.

    return map {
        # Bug 14086: Peculiarity of decode_entities() is, it warns
        # if the argument is undef but correctly returns undef.
        # Could disable warnings, but it's safer to map undef to undef.
        defined($_) ? HTML::Entities::decode_entities($_) : $_
    } $progress_log, $advert;
}


sub GET {
    my $self = shift;
    my ($content_type) = @_;

    my $task_id = SearchStatus::Parameters->param('task_id');
    my $status_data;

    if ($task_id) {
        $status_data = _find_search_status($task_id);
    } else {
        $status_data = { task_id => 'missing' };
    }
    
    if ($$status_data{'status'} and $$status_data{'status'} =~ /complete/) {
        my $opts = MascotDat->MascotOptions;
        my $uri = URI->new(join('/', Request->cgidir, $opts->getResultsPerlScript_2()));

        return WebApp::Redirect->new(
            $uri, 
            { 
                file => $$status_data{'filename'}, 
                mimetype => undef 
            }, 
            '303 See Other', 
            undef, 
            'meta'
        )->GET($content_type);
    }

    if ($content_type->category ne 'msl_html') {
        return $self->_make_html($content_type, $status_data, $task_id);
    }

    if ($$status_data{'show_output'}) {
        my $uri = URI->new(Request->complete_uri);

        return WebApp::Redirect->new(
            $uri,
            { 
                task_id => SearchStatus::Parameters->param('task_id'), 
                sessionID => SearchStatus::Parameters->param('sessionID'),
                mimetype => undef 
            },
            '303 See Other',
            undef,
            'meta'
        )->GET($content_type);
    }

    my $body = DOM->dtd($content_type->type).<<EOF;
<body onload="Request.finish()">
EOF

    my $json = $self->_make_json($content_type, $status_data, $task_id);
    return ({}, $body.$json.'</body></html>');
}

BEGIN {
    my %STATUS_MESSAGE = (
        job_killed => 'Job crashed or has been killed',
        job_not_running => 'Job is not running',
        id_assigned => 'Preparing search...',
        queued => 'Search is currently queued, waiting for it to start...',
        running => 'Search running...',
    );

    sub _status_message {
        my ($ms_mascotoptions, $status, $stage, $step) = @_;

        if (not exists $STATUS_MESSAGE{$status}) {
            return 'Cannot determine search status';
        }

        if ($status eq 'running') {
            if (defined $stage and $stage ne '') {
                if ($stage eq 'ExecAfterSearch') {
                    $step ||= 1; # default to first task

                    # Find the task title. Rest of the arguments are
                    # ignored.
                    my $commands = msparser::VectorString->new;
                    my $titles = msparser::VectorString->new;
                    my $waitfor = msparser::vectori->new;
                    my $logging = msparser::vectori->new;
                    my $additionalFlags = msparser::VectorString->new;
                    $ms_mascotoptions->getExecAfterSearch('unused', $commands, $titles, $waitfor, $logging, $additionalFlags);

                    if ($step >= 1 and $step <= $titles->size) {
                        return sprintf('Search is finished. ExecAfterSearch %d: %s', $step, $titles->get($step-1));
                    }

                    return 'Search is finished. Running ExecAfterSearch tasks';
                }
            }
        }

        return $STATUS_MESSAGE{$status};
    }
}

sub _make_html {
    my $self = shift;
    my ($content_type, $status_data, $task_id) = @_;

    my %headers = ();

    if (Request->user_agent and Request->user_agent =~ /MSIE/) {
        $headers{'X-UA-Compatible'} = 'IE=edge';
    }

    my %namespaces = (
        xmlns => 'http://www.w3.org/1999/xhtml',
        'xml:lang' => 'en',
        'xml:base' => Request->current_cgidir.'/',
    );

    my $head = $self->dom_head($content_type);
    my $elements = $self->dom($content_type, $status_data, $task_id);

    my $DTD = DOM->dtd($content_type->type) || '';
    my $root = AnyML::Node->new('html', \%namespaces,
        [DOM->head($head), DOM->body($elements)]
    );
    my $body = AnyML::Node->serialize('HTML', {}, $root);

    return (\%headers, $DTD.$body);
}

sub _make_json {
    my $self = shift;
    my ($content_type, $status_data, $task_id) = @_;

    if ($$status_data{'task_id'} and $$status_data{'task_id'} eq 'missing') {
        return json_script_receive({ 
                'tc:metadata' => {
                    'tc:error' => 'Task ID is missing' 
                },
            }
        );
    }

    if ($$status_data{'error'}) {
        return json_script_receive({ 
                'tc:metadata' => {
                    'tc:error' => $$status_data{'error'} 
                },
            }
        );
    }

    my $status = _status_message(MascotDat->MascotOptions, $$status_data{'status'}, $$status_data{'stage'}, $$status_data{'step'});
    my $job_id = undef;
    
    if ($task_id) {
        $job_id = _find_search_job_id($task_id);
    }

    my ($search_title, $start_time, $elapsed_time, $end_time, $details_uri, $progress_log, $advert);

    if ($job_id and $$job_id{'job_id'}) {
        $details_uri = URI->new(MascotDat->uri_ms_status);
        $details_uri->query(query_string({ 
            Show => 'JOBDETAILS', 
            JobId => $$job_id{'job_id'}, 
            Autorefresh => 'true' 
        }, '+'));

        my $job_data = _find_search_job_data($$job_id{'job_id'});

        if ($$job_data{'error'}) {
            # Fail silently.
            debug($$job_data{'error'});

            ($progress_log, $advert) = _quick_progress_log($task_id);
        } else {
            $search_title = $$job_data{'search_title'};

            my $now_time = time;
            $start_time = iso8601_to_timestamp($$job_data{'start_time'});
            $end_time = iso8601_to_timestamp($$job_data{'end_time'});

            if (defined($end_time) and $end_time > 0) {
                $elapsed_time = $end_time - $start_time;
            } elsif (defined($start_time) and $start_time > 0) {
                $elapsed_time = $now_time - $start_time;
            }

            if (defined $start_time) {
                $start_time = scalar(localtime($start_time));
            }

            if (defined $end_time) {
                $end_time = scalar(localtime($end_time));
            }

            $progress_log = $$job_data{'progress_log'}; # raw HTML
            $advert = $$job_data{'advert'};
        }
    }

    return json_script_receive({
            'tc:metadata' => {
                'tc:status-message' => $status,
                'tc:percentage' => $$status_data{'percentage'},
                'tc:now-time' => scalar(localtime), # date string
                defined($search_title)
                    ? ('tc:search-title' => $search_title)
                    : ()
                    ,
                defined($start_time)
                    ? ('tc:start-time' => $start_time) # date string
                    : ()
                    ,
                defined($elapsed_time)
                    ? ('tc:elapsed-time' => human_readable_seconds($elapsed_time)) # time string
                    : ()
                    ,
                defined($end_time)
                    ? ('tc:end-time' => $end_time) # date string
                    : ()
                    ,
                defined($details_uri) 
                    ? ('tc:details-uri' => $details_uri->as_string)
                    : ()
                    ,
                defined($progress_log)
                    ? ('tc:progress-log' => $progress_log)
                    : ()
                    ,
                # Bug 12075: search_status.js inserts advert and
                # progress log as raw HTML, so text in $advert can
                # have arbitrary HTML.
                defined($advert)
                    ? ('tc:advert' => $advert)
                    : ()
                    ,
            },
        });
}

sub dom_head {
    my $self = shift;
    my ($content_type) = @_;

    return [
        DOM->meta_content_type($content_type),
        DOM->base({ href => Request->script_uri }),
        DOM->title('Mascot Search'),
        DOM->link({ rel => 'shortcut icon', href => '../favicon.ico' }),
        DOM->link({
            rel => 'stylesheet',
            type => 'text/css',
            href => make_versioned_resource_link(WebApp::Defaults->generic_style_path_rel, 'search_status.css'),
        }),
    ];
}

sub dom {
    my $self = shift;
    my ($content_type, $status_data, $task_id) = @_;

    my $session = msparser::ms_session->new(SearchStatus::Parameters->param('sessionID') || '');

    my @page_title = DOM->h1(
        DOM->img({ 
                src => '../images/88x31_logo_white.gif',
                width => 88,
                height => 31,
            }),
        'Mascot Search'
    );

    my $MascotMessage = MascotDat->MascotOptions->getMascotMessage();

    if ($MascotMessage ne '') {
        # Prevent AnyML from quoting HTML tags by passing the string
        # by reference.
        push @page_title, DOM->div({ class => 'infoBanner' }, \$MascotMessage);
    }

    if ($$status_data{'task_id'} and $$status_data{'task_id'} eq 'missing') {
        return [
            @page_title,

            DOM->p({ class => 'error' },
                'Error: Task ID is missing'
            ),
        ];
    }

    if ($$status_data{'error'}) {
        my @progress_log = ();

        if ($$status_data{'show_output'} and defined $$status_data{'progress_log'}) {
            # Taking a reference forces raw HTML output in AnyML::Node.
            @progress_log = (
                DOM->pre({ class => 'progress-log' }, \$$status_data{'progress_log'}),
            );
        }

        my @help_text = ();

        if (defined $$status_data{'help_text'}) {
            @help_text = DOM->p({ class => 'warning' }, $$status_data{'help_text'});
        }

        return [
            @page_title,
            DOM->p({ class => 'error' }, 
                $$status_data{'error'}
            ),

            @progress_log,

            @help_text,
        ];
    }

    my $can_use_ms_status = 0;

    if ($session->isSecurityEnabled()) {
        if ($session->isPermitted($msparser::ms_security_tasks::SECTASK_USEMSSTATUSEXE)) {
            $can_use_ms_status = 1;
        }
    } else {
        $can_use_ms_status = 1;
    }

    my $job_id = undef;
    
    if ($task_id) {
        $job_id = _find_search_job_id($task_id);
    }

    my $job_data = undef;

    if ($can_use_ms_status) {
        if ($job_id and $$job_id{'job_id'}) {
            $job_data = _find_search_job_data($$job_id{'job_id'});
        }
    }

    my @scripts = ();
    my @body = ();

    if ($$status_data{'status'} =~ /job_not_running/) {
        @body = DOM->p(_status_message(Mascotdat->MascotOptions, $$status_data{'status'}, $$status_data{'stage'}, $$status_data{'step'}));
    } else {
        my $percentage = 0;
        my $show_ajax_loader = 1;

        if ($$status_data{'status'} =~ /running/) {
            $percentage = $$status_data{'percentage'};
        }

        my $ajax_loader_src = WebApp::Defaults->img_path_rel.'/ajax-loader.gif';

        my $details_uri = URI->new(MascotDat->uri_ms_status);

        my $now_time = time;
        my ($start_time, $elapsed_time, $end_time, $search_title, @progress_log, @advert);

        if ($job_id and $$job_id{'job_id'}) {
            $details_uri->query(query_string({ 
                Show => 'JOBDETAILS', 
                JobId => $$job_id{'job_id'}, 
                Autorefresh => 'true' 
            }, '+'));

            if (not $job_data or $$job_data{'error'}) {
                # Fail silently.
                debug($$job_data{'error'});

                my ($progress_log, $advert) = _quick_progress_log($task_id);

                if (defined $progress_log) {
                    @progress_log = \$progress_log;
                }

                if (defined $advert) {
                    @advert = \$advert;
                }
            } else {
                $start_time = iso8601_to_timestamp($$job_data{'start_time'});
                $end_time = iso8601_to_timestamp($$job_data{'end_time'});

                if (defined($end_time) and $end_time > 0) {
                    $elapsed_time = $end_time - $start_time;
                } elsif (defined($start_time) and $start_time > 0) {
                    $elapsed_time = $now_time - $start_time;
                }

                if (defined $$job_data{'search_title'} 
                        and $$job_data{'search_title'} ne '') {
                    $search_title = $$job_data{'search_title'};
                }

                if (defined $$job_data{'progress_log'}) {
                    # Taking a reference forces raw HTML output in AnyML::Node.
                    @progress_log = \$$job_data{'progress_log'};
                }

                if (defined $$job_data{'advert'}) {
                    @advert = \$$job_data{'advert'};
                }
            }
        } 

        if ($$status_data{'status'} =~ /job_killed/) {
            $show_ajax_loader = 0;

            if (not @progress_log) {
                # Either no error output from Mascot or user is not allowed
                # to run ms-status.exe.

                if ($task_id) {
                    @progress_log = 'Mascot search ended with a fatal error (task ID '.$task_id.'). Please contact the system administrator.';
                } else {
                    @progress_log = 'Mascot search ended with a fatal error. Please contact the system administrator.';
                }
            }
        }

        if ($show_ajax_loader) {
            push @scripts, (
                DOM->script({ 
                    type => 'text/javascript', 
                    src => make_versioned_resource_link(WebApp::Defaults->generic_js_path_rel, 'search_status-vendor.js'),
                }),
                DOM->ie_comment($content_type, 'lte IE 9',
                    DOM->script({ 
                            type => 'text/javascript', 
                            src => make_versioned_resource_link(WebApp::Defaults->generic_js_path_rel, 'search_status-vendor-ie.js'),
                        }),
                ),
                DOM->script({ 
                    type => 'text/javascript', 
                    src => make_versioned_resource_link(WebApp::Defaults->generic_js_path_rel, '/search_status.js'),
                }),
            );
        }

        my @progressbar = ();

        if ($show_ajax_loader) {
            @progressbar = DOM->p({ 
                    class => 'progressbar', 
                    style => sprintf("width: %.0f%%;", $percentage)
                }, sprintf("%.0f%%", $percentage),
            );
        }

        my @search_title_tr = ();

        if ($can_use_ms_status) {
            if (defined $search_title) {
                @search_title_tr = DOM->tr(
                    DOM->th('Title'),
                    DOM->td({ class => 'search-title' }, DOM->span($search_title)),
                );
            } else {
                @search_title_tr = DOM->tr(
                    DOM->th('Title'),
                    DOM->td({ class => 'search-title' }, DOM->span('')),
                );
            }
        }

        my @start_time_tr = ();

        if ($can_use_ms_status) {
            my @start_time = ();

            if ($start_time) {
                @start_time = DOM->span(scalar(localtime($start_time)));
            } else {
                @start_time = DOM->span('');
            }

            @start_time_tr = DOM->tr(
                DOM->th('Started'),
                DOM->td({ class => 'start-time' }, @start_time),
            );
        }

        my @elapsed_time_tr = ();

        if ($can_use_ms_status) {
            my @elapsed_time = ();
            
            if (defined $elapsed_time) {
                @elapsed_time = DOM->span(human_readable_seconds($elapsed_time));
            } else {
                @elapsed_time = DOM->span('');
            }

            @elapsed_time_tr = DOM->tr(
                DOM->th('Elapsed'),
                DOM->td({ class => 'elapsed-time' }, @elapsed_time),
            );
        }

        my @time_now_tr = ();

        if ($end_time) {
            @time_now_tr = DOM->tr(
                DOM->th('Ended'),
                DOM->td({ class => 'end-time' }, 
                    DOM->span(scalar(localtime($end_time)))
                ),
            );
        } elsif ($show_ajax_loader) {
            @time_now_tr = DOM->tr(
                DOM->th('Time now'),
                DOM->td({ class => 'now-time' }, 
                    DOM->span(scalar(localtime($now_time)))
                ),
            );
        }

        my @job_status_link = ();

        if ($can_use_ms_status and not -e './www.pl') {
            @job_status_link = DOM->p(
                'The ',
                DOM->a({ href => $details_uri->as_string, target => '_blank', class => 'job-details-uri' },
                    'job status page'
                ),
                ' (new window) allows controlling search priorities and shows more detailed process information.',
            );
        }

        my @refresh_form = ();

        if ($show_ajax_loader) {
            @refresh_form = DOM->form({ 
                    action => Request->script_uri,
                    method => 'get'
                },
                DOM->p(
                    'Your Mascot search is running.',
                    ' This page will refresh at the end of the search to show the results report.',

                    DOM->hidden({ name => 'task_id', value => SearchStatus::Parameters->param('task_id') }),
                    DOM->hidden({ name => 'sessionID', value => SearchStatus::Parameters->param('sessionID') }),
                    DOM->button_submit({ value => 'Refresh' }),

                    DOM->br,

                    ' If you close your browser window before the search ends, the search will finish in the background.',
                ),
            );
        }

        # Session ID must be explicitly given. Otherwise page updates will 
        # fail in IE 6-8, which use an IFRAME hack to load the data. For some
        # reason, the IFRAME request doesn't carry over the session cookie.
        
        push @body, (
            DOM->p({ class => 'advert' }, @advert),
        );

        push @body, DOM->div({ class => 'messagebox', 'data-update-uri' => SearchStatus::Parameters->repeat_uri_abs(sessionID => $session->getID()) },
            DOM->h2({ class => 'status-message' }, 
                DOM->span(_status_message(MascotDat->MascotOptions, $$status_data{'status'}, $$status_data{'stage'}, $$status_data{'step'})),
                DOM->nbsp,
                $show_ajax_loader
                    ? DOM->img({ src => $ajax_loader_src, alt => '', class => 'ajax-loader' })
                    : ()
                    ,
            ),

            @progressbar,
            DOM->table({ class => 'status-times' },
                DOM->tbody(
                    @search_title_tr,
                    @start_time_tr,
                    @time_now_tr,
                    @elapsed_time_tr,
                ),
            ),

            @refresh_form,

            @job_status_link,
        );

        push @body, (
            DOM->pre({ class => 'progress-log' }, @progress_log),
        );
    }

    my @debug_logger = ();

    if (SearchStatus::Parameters->param('_debugjs')) {
        push @debug_logger, DOM->ul({ id => 'MSL-logger' });
    }

    return [
        @page_title,
        @body,

        @debug_logger,
        @scripts,
    ];
}

1;

package main;

use strict;


use URI;
use Carp qw(confess);

use msparser;

use WebApp;
use WebApp::Redirect;

use Request;

use MascotDat;

if (not caller) {
    $SIG{'SEGV'} = sub { confess 'Segmentation fault (Parser problem?)'; };

    $|++;

    my $mimetype = SearchStatus::Parameters->param('mimetype');
    my $fakemimetype = SearchStatus::Parameters->param('fakemimetype');
    my $sid = SearchStatus::Parameters->param('sessionID') || '';

    my $session = new msparser::ms_session($sid || '');

    if (not $session->isValid) {
        my $uri = URI->new_abs(MascotDat->uri_login, Request->cgidir.'/');
        my $referer = URI->new_abs(Request->script_name, Request->current_cgidir.'/');

        $referer->query(
            SearchStatus::Parameters->paramstring(
                'mimetype' => undef,
                'fakemimetype' => undef,
            )
        );

        my $params = {
            referer => $referer,
            error => $session->getLastError,
            errorstring => $session->getLastErrorString,
        };

        my $res = WebApp::Redirect->new(
            $uri, $params, '303 See Other', undef, 'meta'
        );

        exit WebApp->show($res, Request->method, $mimetype, $fakemimetype);
    }

    my $res = eval { SearchStatus->new };

    if ($@) {
        WebApp->died($@, 0, $mimetype, $fakemimetype);
    }

    WebApp->show($res, Request->method, $mimetype, $fakemimetype);
}

