##############################################################################
# Class hierarchy to generate MzIdentML files
##############################################################################
#     $Author: robertog@matrixscience.com $
#       $Date: 2022-05-16 14:27:02 +0100 $
#   $Revision: 2248db7aed94d66543e99125f32fa6d43c1e70b4 | MASCOT_REL_2_8_1-3-gcae025d431 $
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 2010-2021 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################

#!/usr/local/mascot/perl64/bin/perl
 use v5.14;

package ExportDat::MzIdentML {

use Moose::Role;

 # required Perl packages
  use msparser;
  use MascotDat;

  use PeptideSummary::Util qw(key_acc key_acc_str);
  use PeptideSummary::Backend qw(duplicate_group_main sort_duplicate_group);

  use Util::Peptides qw(lookup_query_level_mods);

  use Util::Text qw(round_evenly notags begins_with);

  use ExportDat::Util;

  use constant {
    # Debug only: print ms_peptidesummary parameters in stderr.
      WARN_ABOUT_PEPSUM_PARAMS => 0,
  };

require "./common_subs.pl";

has 'fh' =>(is => 'rw', isa => 'FileHandle', default => sub { \*STDOUT }, writer=>'_set_fh');
has 'objResFile' =>(is => 'ro', isa => 'msparser::ms_mascotresfile');
has 'objModFile' =>(is => 'ro', isa => 'msparser::ms_modfile');
has 'objXlModFile' =>(is => 'ro', isa => 'Maybe[msparser::ms_modfile]');
has 'objQuantMethod' =>(is => 'ro', isa => 'Maybe[msparser::ms_quant_method]');
has 'objUmodSection' =>(is => 'ro', isa => 'msparser::ms_umod_configfile');
has 'objUmodXlSection' =>(is => 'ro', isa => 'Maybe[msparser::ms_umod_configfile]');

has 'objMassesFile' =>(is => 'ro', isa => 'msparser::ms_masses');
has 'urlParams' =>(is => 'ro', isa => 'HashRef[Str]');
has 'libraryMode' =>(is => 'ro', isa => 'Str');
has 'thisScript' =>(is => 'ro', isa => 'CGI');

has 'objUniGene' =>(is => 'ro', isa => 'Maybe[msparser::ms_unigene]');
has 'ShowSameSets' =>(is => 'ro', isa => 'Bool');
has 'ShowSubSets' =>(is => 'ro', isa => 'Bool');
has 'objFullUmodFile' =>(is => 'ro', isa => 'msparser::ms_umod_configfile');

has 'vmMass' =>(is => 'ro', isa => 'HashRef[Str]');
has 'vmString' =>(is => 'ro', isa => 'HashRef[Str]');
has 'objParams' =>(is => 'ro', isa => 'msparser::ms_searchparams');
has 'objSummary' =>(is => 'ro', isa => 'msparser::ms_mascotresults'); #msparser::ms_peptidesummary or msparser::ms_proteinsummary

has 'numHits' =>(is => 'ro', isa => 'Int');
has 'ignoreIonsScoreBelow' =>(is => 'ro', isa => 'Int');
has 'use_homology' =>(is => 'ro', isa => 'Bool');
has 'group_family' =>(is => 'ro', isa => 'Int');

has 'minNumSigUniqueSeqs' =>(is => 'ro', isa => 'Int');
has 'mudpitSwitch' =>(is => 'ro', isa => 'Num');
has 'UniGeneFile' =>(is => 'ro', isa => 'Str');
has 'RequireBoldRed' =>(is => 'ro', isa => 'Bool');

has 'objSummaryDecoy' =>(is => 'ro', isa => 'Maybe[msparser::ms_peptidesummary]');
has 'NoErrorTolerant' =>(is => 'ro', isa => 'Bool');
has 'ShowDecoyReport' =>(is => 'ro', isa => 'Bool');
has 'preferTaxonomy' =>(is => 'ro', isa => 'Int');

has 'percolate' =>(is => 'ro', isa => 'Int');
has 'percolate_rt' =>(is => 'ro', isa => 'Int');
has 'showFDRs' =>(is => 'ro', isa => 'Bool');
has 'taxNames' =>(is => 'ro', isa => 'HashRef[Str]');

has 'fileIn' =>(is => 'ro', isa => 'ArrayRef[Str]');
has 'sessionID' =>(is => 'ro', isa => 'Str');
has 'msresFlags2' =>(is => 'ro', isa => 'Int');
has 'elementalMasses' =>(is => 'ro', isa => 'HashRef[Str]');

has 'massDP' =>(is => 'ro', isa => 'Int');
has 'reportType' =>(is => 'ro', isa => 'Str');
has 'shipper' =>(is => 'ro', isa => 'Str');
has 'mass_type' =>(is => 'ro', isa => 'Str');

has 'anyLibraryMatch' =>(is => 'ro', isa => 'Bool');
has 'objMascotOptions' =>(is => 'ro', isa => 'msparser::ms_mascotoptions');

has 'proteinData' =>(is => 'ro', isa => 'ExportDat::ProteinData');
has 'query_mod_and_dupe_cache' =>(is => 'rw', isa => 'HashRef', default => sub { {} });

has 'analyseSiteData' =>(is => 'rw', isa => 'Bool');

has 'modTypeComp' =>(is => 'rw', isa => 'HashRef', default => sub { {} });
has 'modDeltaComp' =>(is => 'rw', isa => 'HashRef', default => sub { {} });
has 'modIndicesComp' =>(is => 'rw', isa => 'HashRef', default => sub { {} });

# CV terms used in ProteinAmbiguityGroup
has 'cv_anchor_protein' =>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001591', name => 'anchor protein'} });
has 'cv_family_member'  =>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001592', name => 'family member protein'} });

## In Mascot <= 2.2, grouping was based on sequence.
has 'cv_sequence_sameset'   =>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001594', name => 'sequence same-set protein'} });
has 'cv_sequence_subset'=>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001596', name => 'sequence sub-set protein'} });
has 'cv_sequence_intersect'=>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001598', name => 'sequence subsumable protein'} });

## In Mascot >= 2.3, family grouping is based on spectrum.
has 'cv_spectrum_sameset'=>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001595', name => 'spectrum same-set protein'} });
has 'cv_spectrum_subset'=>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001597', name => 'spectrum sub-set protein' } });
has 'cv_spectrum_intersect'=>(is => 'ro', isa => 'HashRef', default => sub { {acc => 'MS:1001599', name => 'spectrum subsumable protein'} });

sub BUILD {
  my $self = shift;

  my $analyseSiteData = 0;
  if ($self->thisScript->param($self->urlParams->{'export_site_data'})) {
    $analyseSiteData = 1;
  }
  $self->analyseSiteData($analyseSiteData);
}

sub _mzid_bool { my $self = shift; $_[0] ? "1" : "0" }

sub _mzid_DBSeq_id { my $self = shift; sprintf("DBSeq_%d_%s", $_[0], notags($_[1])) }
sub _mzid_Enzyme_id { my $self = shift; sprintf("ENZ_%d", $_[0]) }
sub _mzid_MassTable_id { my $self = shift; sprintf("MT_%s", notags($_[0])) }
sub _mzid_Measure_id { my $self = shift; sprintf("m_%s", notags($_[0])) }
sub _mzid_PeptideEvidence_id {
  my $self = shift;
  my $psmComp = ($_[2] != $msparser::ms_peptide::PSM_COMPLETE) ? "_$_[2]" : '';
  sprintf("PE_%d_%d%s_%d_%s_%d_%d_%d", $_[0], $_[1], $psmComp, $_[3], notags($_[4]), $_[5], $_[6], $_[7])
}
sub _mzid_Peptide_id {
  my $self = shift;
  my $psmComp = ($_[2] != $msparser::ms_peptide::PSM_COMPLETE) ? "_p$_[2]" : '';
  sprintf("peptide_%d_%d%s", $_[0], $_[1], $psmComp)
}
sub _mzid_ProteinAmbiguityGroup_id { my $self = shift; sprintf("PAG_hit_%d", $_[0]) }
sub _mzid_ProteinDetectionHypothesis_id { my $self = shift; sprintf("PDH_%s_%d", notags($_[0]), $_[1]) }
sub _mzid_ProteinDetectionList_id { my $self = shift; sprintf("PDL_%d", $_[0]) }
sub _mzid_ProteinDetectionProtocol_id { my $self = shift; sprintf("PDP_%s_%d", notags($_[0]), $_[1]) }
sub _mzid_ProteinDetection_id { my $self = shift; sprintf("PD_%d", $_[0]) }
sub _mzid_Sample_id { my $self = shift; sprintf("Sample_%s", notags($_[0])) }
sub _mzid_SearchDatabase_id { my $self = shift; sprintf("SDB_%s", notags($_[0])) }
sub _mzid_SourceFile_id { my $self = shift; sprintf("SF_%d", $_[0]) }
sub _mzid_SpectraData_id { my $self = shift; sprintf("SD_%d", $_[0]) }
sub _mzid_SpectrumIdentificationItem_id {
  my $self = shift;
  my $psmComp = ($_[2] != $msparser::ms_peptide::PSM_COMPLETE) ? "_p$_[2]" : '';
  sprintf("SII_%d_%d%s", $_[0], $_[1], $psmComp)
}
sub _mzid_SpectrumIdentificationList_id { my $self = shift; sprintf("SIL_%d", $_[0]) }
sub _mzid_SpectrumIdentificationProtocol_id { my $self = shift; sprintf("SIP") }
sub _mzid_SpectrumIdentificationResult_id { my $self = shift; sprintf("SIR_%d", $_[0]) }
sub _mzid_SpectrumIdentification_id { my $self = shift; sprintf("SI") }
sub _mzid_TranslationTable_id { my $self = shift; sprintf("TT_%s", notags($_[0])) }

sub _psmComponents {
  my $self = shift;
  my $objPeptide = $_[0];

  if ($objPeptide->getNumberOfLinkedPeptides() == 2) {
    return ($msparser::ms_peptide::PSM_CROSSLINK_ALPHA, $msparser::ms_peptide::PSM_CROSSLINK_BETA);
  }
  return ($msparser::ms_peptide::PSM_COMPLETE);
}

sub markExportableQueries {
  my ($queryAll, $isPMF, $sameset, $exported_queryList) = @_;

  unless ($queryAll or $isPMF) {
    for my $i (1 .. $sameset->getNumPeptides()) {
      $$exported_queryList[$sameset->getPeptideQuery($i)] = 1;
    }
  }
}


###############################################################################
# &print()
# output PSI mzIdentML elements
# my($fh);
###############################################################################
sub print {
  my $self = shift;
  my ($fh) = @_;

  $self->_set_fh($fh);

  my $isPMF = $self->objResFile->isPMF;
  my $isSQ = $self->objResFile->isSQ;
  my $vecFixed = ExportDat::Util::make_mod_vector($self->objResFile, $self->objModFile, $self->objXlModFile, 'FixedMod');
  my $vecVariable = ExportDat::Util::make_mod_vector($self->objResFile, $self->objModFile, $self->objXlModFile, 'delta');

# make hash for fast retrieval of fixed mods
  my %fixedMods = ExportDat::Util::make_fixed_mods_map($vecFixed);

  my (@componentNames, $numcomponents, %aaHelpers);

  if ($self->objQuantMethod) {
    $numcomponents = $self->objQuantMethod->getNumberOfComponents;

    for my $c (0 .. $numcomponents-1) {
      my $component = $self->objQuantMethod->getComponentByNumber($c);
      $componentNames[$c] = $component->getName;
      my $objLocalMassesFile = new msparser::ms_masses($self->objMassesFile);
      if ($component->getNumberOfIsotopes > 0) {
        $objLocalMassesFile->applyIsotopes($self->objUmodSection, $component);
      }
      $aaHelpers{$component->getName} = new msparser::ms_aahelper($self->objResFile, "../config/enzymes");
      $aaHelpers{$component->getName}->setMasses($objLocalMassesFile);
      $aaHelpers{$component->getName}->setAvailableModifications($vecFixed, $vecVariable);
    # success?
      if (!$aaHelpers{$component->getName}->isValid()) {
        die "Unable to create aahelper object: " . common_subs::htmlEscape($aaHelpers{$component->getName}->getLastErrorString());
      }
    }
  } else {
    $numcomponents = 1;
    $componentNames[0] = "1";
    my $objLocalMassesFile = new msparser::ms_masses($self->objMassesFile);
    $aaHelpers{"1"} = new msparser::ms_aahelper($self->objResFile, "../config/enzymes");
    $aaHelpers{"1"}->setMasses($objLocalMassesFile);
    $aaHelpers{"1"}->setAvailableModifications($vecFixed, $vecVariable);
  # success?
    if (!$aaHelpers{"1"}->isValid()) {
      die "Unable to create aahelper object: " . common_subs::htmlEscape($aaHelpers{"1"}->getLastErrorString());
    }
  }

  my @resfile_params = ();
  my @resfile_children = ();

  for my $i (1 .. $self->objResFile->getNumberOfResfiles()) {
    my $resfile = $self->objResFile->getResfile($i);
    $resfile_children[$i] = $resfile;
    $resfile_params[$i] = $resfile->params;
  }

  $self->printCVList();

  $self->printAnalysisSoftwareList();

  $self->printProvider();

  $self->printAuditCollection(\@resfile_params);

  $self->printAnalysisSampleCollection();

  my $sameset_ordering = $self->proteinData->sameset_ordering;
  my $queryAll = $self->thisScript->param($self->urlParams->{'query_all'});

  my $protLen = $self->thisScript->param($self->urlParams->{'prot_len'});
  my $protSeq = $self->thisScript->param($self->urlParams->{'prot_seq'});
  my $protDesc = $self->thisScript->param($self->urlParams->{'prot_desc'});
  my $protTaxStr = $self->thisScript->param($self->urlParams->{'prot_tax_str'});
  my $protTaxId = $self->thisScript->param($self->urlParams->{'prot_tax_id'});
  my $filter_duplicates = 1;
  if ($self->thisScript->param($self->urlParams->{'show_pep_dupes'})) {
    $filter_duplicates = 0;
  }

  my ($exported_sequences, $exported_queryList, $peptide_evidence_list) = $self->printSequenceCollection(
    $queryAll, $protLen, $protSeq, $protDesc,
    $protTaxStr, $protTaxId,
      $sameset_ordering, \%fixedMods, $filter_duplicates,
  );

  $self->printAnalysisCollection(_tolerance_cv_param_map());

  $self->printAnalysisProtocolCollection(
    \@resfile_params,
    $protTaxStr,
      _tolerance_cv_param_map(),
  );

  my $queryRaw = $self->thisScript->param($self->urlParams->{'query_raw'});

  $self->printDataCollection(
    \@resfile_params, \@resfile_children,
    $exported_queryList, $exported_sequences,
    $peptide_evidence_list, \%aaHelpers,
    _tolerance_cv_param_map(), $protTaxStr,
    $protTaxId, $queryAll, $protLen, $queryRaw,
      $sameset_ordering, $numcomponents, $componentNames[0],
  );

  $self->printBibliographicReference();
}

sub printCVList {
  # Implemented in subclasses
}

sub printAnalysisSoftwareList {
  my $self = shift;

  # (When results are merged: we pick Mascot version from the first results
  # file.)
  my $mascotver = notags($self->objResFile->getMascotVer());

  my $parserver = notags($self->objResFile->getMSParserVersion());

  my $args = ExportDat::Util::cmdline_arguments($self->thisScript, $self->urlParams);

  my $pretty_args = notags(join(' ', map {
    $$args{$_} =~ /\s/
      ? sprintf("%s='%s'", $_, $$args{$_})
      : sprintf("%s=%s", $_, $$args{$_})
  } sort keys %$args));

  print {$self->{fh}} <<EOF;
  <AnalysisSoftwareList>
EOF
  # Bug 12187: In integrated mode, we list MSPepSearch first, because Mascot
  # performs further processing on the library search results. This makes the
  # pipeline appear in the correct logical order.
  if ($self->libraryMode eq 'onlylibrary' or $self->libraryMode eq 'integrated') {
    # MSPepSearch version is omitted, since it's not recorded in the results
    # file.
    print {$self->{fh}} <<EOF;
    <AnalysisSoftware id="AS_nist_mspepsearch" name="NIST MSPepSearch" uri="http://chemdata.nist.gov/dokuwiki/doku.php?id=peptidew:mspepsearch">
      <ContactRole contact_ref="ORG_NIST">
        <Role>
          <cvParam accession="MS:1001267" name="software vendor" cvRef="PSI-MS" />
        </Role>
      </ContactRole>
      <SoftwareName>
        <cvParam accession="MS:1002750" name="NIST MSPepSearch" cvRef="PSI-MS" />
      </SoftwareName>
      <Customizations>
        No customisations
      </Customizations>
    </AnalysisSoftware>
EOF
  }
  # Even in SL-only mode, Mascot processes the library search results, so it's
  # always listed.
  do {
    print {$self->{fh}} <<EOF;
    <AnalysisSoftware id="AS_mascot_server" name="Mascot Server" version="$mascotver" uri="http://www.matrixscience.com/search_form_select.html">
      <ContactRole contact_ref="ORG_MSL">
        <Role>
          <cvParam accession="MS:1001267" name="software vendor" cvRef="PSI-MS" />
        </Role>
      </ContactRole>
      <SoftwareName>
        <cvParam accession="MS:1001207" name="Mascot" cvRef="PSI-MS" />
      </SoftwareName>
      <Customizations>
        No customisations
      </Customizations>
    </AnalysisSoftware>
EOF
  };
  print {$self->{fh}} <<EOF;
    <AnalysisSoftware id="AS_mascot_parser" name="Mascot Parser" version="$parserver" uri="http://www.matrixscience.com/msparser.html">
      <ContactRole contact_ref="ORG_MSL">
        <Role>
          <cvParam accession="MS:1001267" name="software vendor" cvRef="PSI-MS" />
        </Role>
      </ContactRole>
      <SoftwareName>
        <cvParam accession="MS:1001478" name="Mascot Parser" cvRef="PSI-MS" />
      </SoftwareName>
      <Customizations>
        No customisations
      </Customizations>
    </AnalysisSoftware>
  </AnalysisSoftwareList>
EOF
}

sub printProvider {
  my $self = shift;
  print {$self->{fh}} <<EOF;
  <Provider id="PROVIDER">
    <ContactRole contact_ref="PERSON_DOC_OWNER">
      <Role>
        <cvParam accession="MS:1001271" name="researcher" cvRef="PSI-MS" />
      </Role>
    </ContactRole>
  </Provider>
EOF
}

sub printAuditCollection {
  my $self = shift;
  my ($resfile_params) = @_;

  my @users = ();
  my %users = ();

  # (When results are merged: the contact person is the user in the first
  # results file.)

  for my $i (1 .. $#$resfile_params) {
    my $params = $$resfile_params[$i];
    my $username = notags($params->getUSERNAME);
    my $email = notags($params->getUSEREMAIL);

    if (not $users{$username}) {
        push @users, $username;
    }

    $users{$username}{$email} = 1;
  }

  print {$self->{fh}} <<EOF;
  <AuditCollection>
    <Person id="PERSON_MSL">
      <Affiliation organization_ref="ORG_MSL" />
    </Person>
EOF

  for my $i (0 .. $#users) {
    my $id = $i == 0 ? 'PERSON_DOC_OWNER' : sprintf('PERSON_DOC_OWNER_%d', $i);
    my $username = $users[$i];
    print {$self->{fh}} <<EOF;
    <Person id="$id" name="$username">
EOF

    for my $email (sort keys %{ $users{$username} }) {
      next if $email eq '';
      print {$self->{fh}} <<EOF;
      <cvParam accession="MS:1000589" name="contact email" value="$email" cvRef="PSI-MS" />
EOF
    }

    print {$self->{fh}} <<EOF;
      <Affiliation organization_ref="ORG_DOC_OWNER" />
    </Person>
EOF
  }

  if ($self->libraryMode eq 'onlylibrary' or $self->libraryMode eq 'integrated') {
    print {$self->{fh}} <<EOF;
    <Organization id="ORG_NIST" name="National Institute of Standards and Technology">
      <cvParam accession="MS:1000588" name="contact URL" value="https://www.nist.gov" cvRef="PSI-MS" />
    </Organization>
EOF
  }

  print {$self->{fh}} <<EOF;
    <Organization id="ORG_MSL" name="Matrix Science Limited">
      <cvParam accession="MS:1000589" name="contact email" value="support\@matrixscience.com" cvRef="PSI-MS" />
      <cvParam accession="MS:1000587" name="contact address" value="64 Baker Street, London W1U 7GB, UK" cvRef="PSI-MS" />
      <cvParam accession="MS:1001755" name="contact phone number" value="+44 (0)20 7486 1050" cvRef="PSI-MS" />
      <cvParam accession="MS:1001756" name="contact fax number" value="+44 (0)20 7224 1344" cvRef="PSI-MS" />
      <cvParam accession="MS:1000588" name="contact URL" value="http://www.matrixscience.com" cvRef="PSI-MS" />
    </Organization>
    <Organization id="ORG_DOC_OWNER" />
  </AuditCollection>
EOF
}

sub printAnalysisSampleCollection {
  my $self = shift;

  return if not $self->objQuantMethod;

  my $numcomponents = $self->objQuantMethod->getNumberOfComponents;

  return if $numcomponents <= 1;

  $self->fh->print("  <AnalysisSampleCollection>\n");

  for my $c (0 .. $numcomponents-1) {
    my $component = $self->objQuantMethod->getComponentByNumber($c);
    $self->fh->print('    <Sample id="' . $self->_mzid_Sample_id($component->getName) . "\">\n");
    $self->fh->print("      <userParam name=\"" . notags($component->getName) . "\" />\n");
    if ($component->getNumberOfIsotopes > 0) {
      for my $i (0 .. $component->getNumberOfIsotopes()-1) {
        my $isotope = $component->getIsotope($i);
        if ($isotope->haveOld && $isotope->haveNew) {
          $self->fh->print("      <userParam name=\"Isotope: " . $isotope->getOld . " -> " .  $isotope->getNew . "\" />\n");
        } else {
          $self->fh->print("      <userParam name=\"Isotope: None\" />\n");
        }
      }
    }
    $self->fh->print("    </Sample>\n");
  }

  $self->fh->print('    <Sample id="' . $self->_mzid_Sample_id('Combined') . "\">\n");
  for my $c (0 .. $numcomponents-1) {
    my $component = $self->objQuantMethod->getComponentByNumber($c);
    $self->fh->print('      <SubSample sample_ref="' . $self->_mzid_Sample_id($component->getName) . "\" />\n");
  }
  $self->fh->print("    </Sample>\n");

  $self->fh->print("  </AnalysisSampleCollection>\n");
}

sub printSequenceCollectionPeptides {
  my $self = shift;
  my ($exported_sequences, $exported_queryList, $queryAll,
      $fixedMods, $protTaxId, $protTaxStr, $filter_duplicates) = @_;

  my $isPMF = $self->objResFile->isPMF;
  my $max_p = $self->objSummary->getMaxRankValue();

  # We need to make two loops over the peptides, because PeptideEvidence
  # elements must follow all Peptide elements. Additionally, we need to
  # collect peptide evidence in the first loop and skip peptides for which
  # there is no evidence; otherwise we'll end up outputting Peptide elements
  # with no corresponding PeptideEvidence. See bug #10759.
  my %peptide_evidence_list = ();

  # We need to keep some further details about peptide evidence (residue
  # before, residue after, etc.) for the second loop.
  my %peptide_evidence_details = ();
  my %umod_mod_cache = ();
  my %dupe_key_cache = ();

  for my $query (1 .. $self->objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      my $objPeptide = $self->objSummary->getPeptide($query, $rank);
      next if $objPeptide->getPeptideStr(0) eq '';

      $self->collectPeptideEvidence(
          $query, $rank, $exported_sequences,
          \%peptide_evidence_list, \%peptide_evidence_details,
      );
      if ($filter_duplicates) {
        $dupe_key_cache{$query}{$rank} = {
          obj => $objPeptide,
          key => duplicate_group_main($self->query_mod_and_dupe_cache, $self->objResFile, $self->objParams, $self->objSummary, $objPeptide),
        };
      }
    }
  }

  if ($filter_duplicates) {
    # Bug #10777: If duplicates should be filtered out, we have some extra
    # work to do. The duplicate status of a peptide match depends on which
    # protein hit we are considering. Thus, for each protein hit, we need
    # to find its "duplicate groups" of peptide matches. We then remove
    # the evidence of a peptide match from protein hits where the peptide match
    # is not the highest scoring match in its duplicate group within the hit.


    # Step 1: reconstruct the list of matches for each protein hit, with
    # each peptide match grouped in its duplicate group.
    my %protein_data = ();

    for my $query (keys %peptide_evidence_details) {
      for my $rank (keys %{ $peptide_evidence_details{$query} }) {
	my $key = $dupe_key_cache{$query}{$rank}{'key'};
	my $objPeptide = $dupe_key_cache{$query}{$rank}{'obj'};
	for my $psmComp (keys %{ $peptide_evidence_details{$query}{$rank} }) {
	  my $details = $peptide_evidence_details{$query}{$rank}{$psmComp};
	  for my $d (@$details) {
	    push @{ $protein_data{$$d{'db'}}{$$d{'acc'}}{$key} }, $objPeptide;
	  }
	}
      }
    }

    # Step 2: find the master peptide in each duplicate group.
    my %surviving_peptide = ();

    for my $db (keys %protein_data) {
      for my $acc (keys %{ $protein_data{$db} }) {
        for my $key (keys %{ $protein_data{$db}{$acc} }) {
          my $sorted = sort_duplicate_group($protein_data{$db}{$acc}{$key});
          my $master_peptide = $$sorted[0];
	  for my $psmComp($self->_psmComponents($master_peptide)) {
	    $surviving_peptide{$master_peptide->getQuery()}{$master_peptide->getRank()}{$psmComp}{$db}{$acc} = 1;
	  }
        }
      }
    }

    # Step 3: remove peptide match details from protein hits where the peptide
    # match is not the master peptide in the protein hit.
    for my $query (keys %peptide_evidence_details) {
      for my $rank (keys %{ $peptide_evidence_details{$query} }) {
	for my $psmComp (keys %{ $peptide_evidence_details{$query}{$rank} }) {
	  my $details = $peptide_evidence_details{$query}{$rank}{$psmComp};
	  my @filtered = ();

	  for my $d (@$details) {
	    if ($surviving_peptide{$query}{$rank}{$psmComp}{$$d{'db'}}{$$d{'acc'}}) {
	      push @filtered, $d;
	    }
	  }

	  if (@filtered) {
	    $peptide_evidence_details{$query}{$rank}{$psmComp} = \@filtered;
	    $peptide_evidence_list{$query}{$rank}{$psmComp} = {};
	    for my $d (@filtered) {
	      $peptide_evidence_list{$query}{$rank}{$psmComp}{$$d{'id'}} = 1;
	    }
	  } else {
	    delete $peptide_evidence_details{$query}{$rank}{$psmComp};
	    if (not %{ $peptide_evidence_details{$query}{$rank} }) {
	      delete $peptide_evidence_details{$query}{$rank};
	    }
	    if (not %{ $peptide_evidence_details{$query} }) {
	      delete $peptide_evidence_details{$query};
	    }

	    delete $peptide_evidence_list{$query}{$rank}{$psmComp};
	    if (not %{ $peptide_evidence_list{$query}{$rank} }) {
	      delete $peptide_evidence_list{$query}{$rank};
	    }
	    if (not %{ $peptide_evidence_list{$query} }) {
	      delete $peptide_evidence_list{$query};
	    }
	  }
	}#psmComp
      }
    }
  }

  for my $query (1 .. $self->objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      next unless exists $peptide_evidence_list{$query}{$rank};

      my $objPeptide = $self->objSummary->getPeptide($query, $rank);
      for my $psmComp ($self->_psmComponents($objPeptide)){
	next unless exists $peptide_evidence_list{$query}{$rank}{$psmComp};

	$self->printSequenceCollectionPeptide(
	  $objPeptide,
	  $query, $rank, $psmComp,
	  $fixedMods, \%umod_mod_cache
	  );
      }
    }
  }

  for my $query (1 .. $self->objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      my $objPeptide = $self->objSummary->getPeptide($query, $rank);
      next if $objPeptide->getPeptideStr(0) eq '';

      next if not $peptide_evidence_details{$query}{$rank};

      for my $psmComp ($self->_psmComponents($objPeptide)) {
	$self->printSequenceCollectionPeptideEvidence(
	  $objPeptide, $query, $rank, $psmComp,
	  $protTaxId, $protTaxStr, \%peptide_evidence_details
	);
      }
    }
  }

  return \%peptide_evidence_list;
}

sub _find_unigene_ids {
  my ($objUniGene, $accession) = @_;

  my $i = 0;
  my @ids = ();

  while (1) {
    my $str = $objUniGene->getUnigeneForAccession($accession, ++$i);
    last if $str eq '';

    push @ids, $str;
  }

  return \@ids;
}

sub collectPeptideEvidence {
  my $self = shift;
  my ($query, $rank, $exported_sequences,
      $peptide_evidence_list, $peptide_evidence_details) = @_;

  my $start = new msparser::vectori;
  my $end   = new msparser::vectori;
  my $pre   = new msparser::VectorString;
  my $post  = new msparser::VectorString;
  my $frame = new msparser::vectori;
  my $multiplicity = new msparser::vectori;
  my $db    = new msparser::vectori;
  my $psmComponent = new msparser::vectori;

  my $accessions = $self->objSummary->getAllProteinsWithThisPepMatch($query, $rank, $start, $end, $pre, $post, $frame, $multiplicity, $db, $psmComponent);
  # $self->fh->print("Errors: " . $resfile->getLastErrorString() . "\n");

  my @unigene_ids = ();

  if ($self->objUniGene) {
    for my $i (0 .. $#$accessions) {
      $unigene_ids[$i] = _find_unigene_ids($self->objUniGene, $$accessions[$i]);
    }
  }

  for my $i (0 .. $#$accessions) {
    my $preStr  = $pre->get($i);
    my $postStr = $post->get($i);
    my $psmComp = $psmComponent->get($i);

    if ($preStr eq '@') {
      $preStr = '-';
    }
    if ($postStr eq '@') {
      $postStr = '-';
    }

    if (my @ids = @{ $unigene_ids[$i] || [] }) {
      for my $unigene_id (@ids) {
        next unless $$exported_sequences{ key_acc_str(0, $unigene_id) };

        my $id = $self->_mzid_PeptideEvidence_id(
          $query, $rank, $psmComp, 0, $unigene_id, $frame->get($i),
          $start->get($i), $end->get($i)
        );

        # Several accessions can (and usually do) map to the same UniGene ID.
        # We only want one peptide evidence item in this case.
        next if $$peptide_evidence_list{$query}{$rank}{$psmComp}{$id};
        $$peptide_evidence_list{$query}{$rank}{$id} = 1;

        push @{ $$peptide_evidence_details{$query}{$rank}{$psmComp} }, {
          id => $id,
          acc => $unigene_id,
          db => 0,
          preferred_acc => $unigene_id,
          pre => $preStr,
          post => $postStr,
          start => $start->get($i),
          end => $end->get($i),
          frame => $frame->get($i),
        };
      }

      next;
    }

    next unless $$exported_sequences{ key_acc_str($db->get($i), $$accessions[$i]) };

    my $preferred_acc = $self->proteinData->preferredAccession($db->get($i), $$accessions[$i]);

    my $id = $self->_mzid_PeptideEvidence_id(
      $query, $rank, $psmComp, $db->get($i), $preferred_acc, $frame->get($i),
      $start->get($i), $end->get($i)
    );

    $$peptide_evidence_list{$query}{$rank}{$psmComp}{$id} = 1;

    push @{ $$peptide_evidence_details{$query}{$rank}{$psmComp} }, {
      id => $id,
      acc => $$accessions[$i],
      db => $db->get($i),
      preferred_acc => $preferred_acc,
      pre => $preStr,
      post => $postStr,
      start => $start->get($i),
      end => $end->get($i),
      frame => $frame->get($i),
    };
  }
}

sub printSequenceCollectionPeptide {
    my $self = shift;
     my ($objPeptide,
	 $query, $rank, $psmComp,
	 $fixedMods, $umod_mod_cache) = @_;

     $self->fh->print('    <Peptide id="' . $self->_mzid_Peptide_id($query, $rank, $psmComp) . "\">\n");
     $self->printSequenceCollectionPeptideBody($objPeptide, $query, $rank, $psmComp, $fixedMods, $umod_mod_cache);
     $self->fh->print("    </Peptide>\n");
}
sub printSequenceCollectionPeptideBody {
  my $self = shift;
  my ($objPeptide,
      $query, $rank, $psmComponent,
      $fixedMods, $umod_mod_cache,
    ) = @_;

  my %vmMass = %{$self->vmMass};
  my %vmString = %{$self->vmString};

  my $pepstr = $objPeptide->getPeptideStr(0, $psmComponent);

  $self->fh->print("      <PeptideSequence>" . $pepstr . "</PeptideSequence>\n");

  # Any modifications?
  my $varModsStr = $objPeptide->getVarModsStr($psmComponent);
  my $summedModsStr = $objPeptide->getSummedModsStr($psmComponent);
  $summedModsStr = '0' x length($varModsStr) if $summedModsStr eq '';

  my $local_mods = lookup_query_level_mods($self->query_mod_and_dupe_cache, $self->objResFile, $objPeptide, $self->objUmodSection);

  my $nlStr = $objPeptide->getPrimaryNlStr($psmComponent);
  $nlStr = '0' x length($varModsStr) if $nlStr eq '';
  my $localNlStr = $objPeptide->getLocalModsNlStr($psmComponent);
  $localNlStr = '0' x length($varModsStr) if $localNlStr eq '';

  my %varmod_index_map = do {
    my @indices = (0..9, 'A'..'W');
    map { $indices[$_] => $_ } 0 .. $#indices
  };

  my @pepRes = split(//, lc($objPeptide->getPeptideStr(1, $psmComponent)));

  for my $p (0 .. length($varModsStr)-1) {
    my $has_var_mod = 0;

    if (my $varmod_id = substr($varModsStr, $p, 1)) {
      # $varmod_id is not '' and not '0' (both evaluate to false).
      $has_var_mod = 1;
      my ($delta, $umod_title, $modfile_name);
      if ($varmod_id eq "X") {
        my $srcRank = $self->objSummary->getSrcRank($query, $rank);
        my $tempString =
            $self->objResFile->getSectionValueStr($self->objSummary->getSrcSection($query, $rank), "q$query\_p$srcRank\_et_mods");
        if ($tempString) {
          ($delta, undef, $modfile_name) = split(/,/, $tempString, 3);
        }
        if ($modfile_name =~ /^NA_/) {
          my %pretty_value = (
              NA_SUBSTITUTION => 'substitution',
              NA_INSERTION => 'insertion',
              NA_DELETION => 'deletion',
          );

          my @cv_params;
          push @cv_params, "        <cvParam accession=\"MS:1002028\" name=\"nucleic acid base modification\" cvRef=\"PSI-MS\" value=\"" . $pretty_value{uc $modfile_name} . "\" />\n";

          my $tempString =
              $self->objResFile->getSectionValueStr($self->objSummary->getSrcSection($query, $rank), "q$query\_p$srcRank\_na_diff");
          if ($tempString) {
            my ($old, $new) = split(/,/, $tempString, 2);
            push @cv_params, "        <cvParam accession=\"MS:1002029\" name=\"original nucleic acid sequence\" cvRef=\"PSI-MS\" value=\"" . notags($old) . "\" />\n";
            push @cv_params, "        <cvParam accession=\"MS:1002030\" name=\"modified nucleic acid sequence\" cvRef=\"PSI-MS\" value=\"" . notags($new) . "\" />\n";
          }
          $self->printModificationElementNA($p, $pepstr, \@cv_params);
          next;
        }
        # Otherwise treat as "regular" mod.
        $umod_title = _modfile_name_to_Unimod_title($modfile_name);
      } elsif ($varmod_id eq 'Y') {
        my $lib_mods_str = $self->objSummary->getLibraryModString($objPeptide->getQuery, $objPeptide->getRank);
        for (split /,/, $lib_mods_str) {
          my ($pos, $mod_id) = split /:/;
          next unless $pos == $p;
          # Bug 12187:
          # Libraries don't record the specificity used when the search or
          # searches were run, so we will pretend position = Anywhere. The
          # specificity ultimately only has an effect for neutral loss lookup.
          # Since library entries don't specify neutral losses, constructing
          # the modfile-style name is only relevant for consistency.
          $delta = $self->objParams->getLibraryModDelta($mod_id);
          $umod_title = $self->objParams->getLibraryModName($mod_id);
          my $spec_site = substr($pepstr, $p-1, 1);
          $modfile_name = "$umod_title ($spec_site)";
          last;
        }
      } else {
        $delta = $vmMass{$varmod_id};
        $modfile_name = $vmString{$varmod_id};
        $umod_title = _modfile_name_to_Unimod_title($modfile_name);
      }
      my $nl_id = substr($nlStr, $p, 1);
      if (defined $nl_id) {
        $nl_id = $varmod_index_map{$nl_id};
      }
      my $nl_backup_value = undef;
      if (defined($nl_id) and $nl_id > 0) {
        # Old results files without Unimod section store neutral loss
        # information in the parameter section.
        my @arr = @{ $self->objParams->getVarModsNeutralLosses($varmod_index_map{$varmod_id}) };
        if ($nl_id-1 < @arr) {
          $nl_backup_value = $arr[$nl_id-1];
        }
      }

      $self->printModificationElement(
        $psmComponent, $umod_mod_cache, $pepstr, $p, $delta, $umod_title,
        $modfile_name, $nl_id,
        $nl_backup_value,
      );
    }

    if (my $varmod_id = substr($summedModsStr, $p, 1)) {
      my $umod_title = _modfile_name_to_Unimod_title($vmString{$varmod_id});
      $self->printModificationElement(
        $psmComponent, $umod_mod_cache, $pepstr, $p, $vmMass{$varmod_id},
        $umod_title, $vmString{$varmod_id},
      );
      $has_var_mod = 1;
    }

    if ($local_mods and $$local_mods{$psmComponent}{$p}) {
      my $nl_id = substr($localNlStr, $p, 1);
      if (defined $nl_id) {
        $nl_id = $varmod_index_map{$nl_id};
      }
      my $umod_title = _modfile_name_to_Unimod_title($$local_mods{$psmComponent}{$p}{'name'});
      $self->printModificationElement(
        $psmComponent, $umod_mod_cache, $pepstr, $p, $$local_mods{$psmComponent}{$p}{'delta'},
        $umod_title, $$local_mods{$psmComponent}{$p}{'name'},
        $nl_id,
      );
      $has_var_mod = 1;
    }

    if (not $has_var_mod) {
      my $modfile_name;
    # check for fixed mod
      if ($p == 0) {
      # n-term
        $modfile_name = $$fixedMods{'n-term'}{'name'};
      } elsif ($p == (length($varModsStr) - 1)) {
      # c-term
        $modfile_name = $$fixedMods{'c-term'}{'name'};
      } else {
      # residue $pepRes[$p-1];
        $modfile_name = $$fixedMods{$pepRes[$p-1]}{'name'};
      }
      if ($modfile_name) {
        my $umod_title = _modfile_name_to_Unimod_title($modfile_name);
        my $objMod = $self->objModFile->getModificationByName($modfile_name);
        $self->printModificationElement(
          $psmComponent, $umod_mod_cache, $pepstr, $p,
          $objMod->getDelta($self->objParams->getMassType()), $umod_title,
          $modfile_name, 1,
        );
      }
    }
  }
  # Any substitutions?
  my $ambiguity = $objPeptide->getAmbiguityString($psmComponent);
  if ($ambiguity) {
    # Format: 4,X,A
    my ($location, $before, $after) = split(/,/, $ambiguity);
    $self->fh->print("      <SubstitutionModification location=\"". $location . "\" originalResidue=\"" . $before . "\" replacementResidue=\"" . $after . "\" />\n");
  }

}

sub _modfile_name_to_Unimod_title { ($_[0] =~ m{(.+?) \s+ \( }x)[0] // $_[0] }

sub printModificationElementNA {
  my $self = shift;
  my ($p, $pepstr, $cv_params) = @_;
  $self->fh->print("      <Modification location=\"" . $p . "\" ");
  if ($p > 0 and $p <= length($pepstr)) {
    $self->fh->print("residues=\"" . substr($pepstr, $p-1, 1) . "\" ");
  }
  $self->fh->print(">\n");
  $self->fh->print(@$cv_params);
  $self->fh->print("      </Modification>\n");
}

sub printModificationElement {
  my $self = shift;
  my ($psmComponent, $mod_cache, $pepstr, $p, $delta, $umod_title, $modfile_name,
      $nl_id, $nl_backup_value) = @_;

  $self->fh->print("      <Modification location=\"" . $p . "\" ");
  if ($p > 0 and $p <= length($pepstr)) {
    $self->fh->print("residues=\"" . substr($pepstr, $p-1, 1) . "\" ");
  }

  my $modTypes = $self->modTypeComp->{$psmComponent};
  my $isLink = (defined $modTypes) && ($$modTypes[$p] eq 'I');
  my $isMonolink = (defined $modTypes) && ($$modTypes[$p] eq 'M');
  my $isLooplink = (defined $modTypes) && begins_with($$modTypes[$p], 'L');
  if ($isMonolink or $isLooplink) {
    # If monolink, delta is linker delta - monolink nl
    # If looplink, delta is the delta of the end of the looplink that is represented as a monolink.
    $delta = $self->modDeltaComp->{$psmComponent}->[$p];
  }

  # It's an intact link in the receiving peptide
  if ( ($psmComponent == $msparser::ms_peptide::PSM_CROSSLINK_BETA) && $isLink ) {
    $self->fh->print("monoisotopicMassDelta=\"0.0\"");
  }
  elsif ($delta) {
    # Delta could be undefined in a spectral library search if the library
    # modification isn't in Unimod. Luckily the delta attribute is optional.
    if ($self->objParams->getMassType == $msparser::MASS_TYPE_MONO) {
      $self->fh->print("monoisotopicMassDelta=\"" . $delta . "\"");
    } else {
      $self->fh->print("avgMassDelta=\"" . $delta . "\"");
    }
  }

  $self->fh->print(">\n");
  my ($cv_acc, $cv_name, $cv_ref, $nl_arr);
  my $mod2;
  # Title could be undefined for a library mod.
  if (defined $umod_title) {
    $mod2 = $self->objUmodSection->getModificationByName($umod_title);
    if (not $mod2 or not $mod2->haveRecordID()) {
      # There are two cases where the modification isn't found:
      # 1) If the Unimod section is too old or doesn't contain record IDs.
      # 2) If this is a library mod.
      #
      # Library mods are not necessarily mentioned in the Unimod section of
      # the results file, for example if the user didn't select them in the
      # search form.
      #
      # In these cases, use the full unimod.xml file.
      if ($self->objUmodSection != $self->objFullUmodFile) {
        $mod2 = $self->objFullUmodFile->getModificationByName($umod_title);
      }
    }
  }
  if ($mod2) {
    $cv_acc = 'UNIMOD:' . notags($mod2->getRecordID);
    $cv_name = notags($umod_title);
    $cv_ref = 'UNIMOD';
    if (exists $$mod_cache{$modfile_name}) {
      if ($$mod_cache{$modfile_name}) {
        $nl_arr = $$mod_cache{$modfile_name};
      }
    } else {
      my $modfile_mod = msparser::ms_modification->new;
      # Here we only need to look up in the Unimod section. Library mods don't
      # have neutral losses, so no need to use the full unimod.xml file
      # -- except if the file is too old to have a Unimod section.
      if ($modfile_mod->getFromUnimod($modfile_name, $self->objUmodSection)) {
        $nl_arr = $modfile_mod->getNeutralLoss($self->objParams->getMassType());
      } elsif ($self->objUmodSection != $self->objFullUmodFile
              and $modfile_mod->getFromUnimod($modfile_name, $self->objFullUmodFile)) {
        $nl_arr = $modfile_mod->getNeutralLoss($self->objParams->getMassType());
      }
      $$mod_cache{$modfile_name} = $nl_arr;
    }
  } else {
    $cv_acc = "MS:1001460";
    $cv_name = 'unknown modification';
    $cv_ref = 'PSI-MS';
  }

  my $nl_value = $nl_backup_value; # could be undef
  if (defined $nl_id) {
    if ($nl_arr and @$nl_arr and $nl_id-1 < @$nl_arr and $$nl_arr[$nl_id-1] > 0) {
      $nl_value = $$nl_arr[$nl_id-1];
    }
  }

  $self->printModificationCvTerms($umod_title, $mod2, $cv_acc, $cv_name, $cv_ref, $nl_value, $psmComponent, $p);
  $self->fh->print("      </Modification>\n");
}

sub printModificationCvTerms {
  my $self = shift;
  my ($umod_title, $mod2, $cv_acc, $cv_name, $cv_ref, $nl_value, $psmComponent, $p) = @_;

  $self->fh->print("        <cvParam accession=\"$cv_acc\" name=\"$cv_name\" cvRef=\"$cv_ref\" />\n");
  if (defined $nl_value) {
    $self->fh->print("        <cvParam accession=\"MS:1001524\" name=\"fragment neutral loss\" cvRef=\"PSI-MS\" value=\"" . $nl_value . "\" unitAccession=\"UO:0000221\" unitName=\"dalton\" unitCvRef=\"UO\" />\n");
  }
}

sub printSequenceCollectionPeptideEvidence {
  my $self = shift;
  my ($objPeptide, $query, $rank, $psmComp,
      $protTaxId, $protTaxStr, $peptide_evidence_details) = @_;

  # The evidence needs to be printed in sorted order to keep the output
  # constant across different runs with the same parameters.
  for my $evidence (sort { $$a{'id'} cmp $$b{'id'} } @{ $$peptide_evidence_details{$query}{$rank}{$psmComp} }) {
    my ($id, $start, $end, $preStr, $postStr) = @$evidence{qw(id start end pre post)};

    $self->fh->print('    <PeptideEvidence id="' . $id
      . '" start="' . $start
      . '" end="'   . $end
      . '" pre="'   . $preStr
      . '" post="'  . $postStr
      . '" peptide_ref="' . $self->_mzid_Peptide_id($query, $rank, $psmComp)
      );

    my ($db, $preferred_acc) = @$evidence{qw(db preferred_acc)};

    if (ExportDat::Util::db_is_NA($self->objResFile, $self->objSummary, $db)) {
      my $ttNum = 1;  # if unspecified, Mascot uses standard genetic code
      my ($acc, $framenum) = @$evidence{qw(acc frame)};
      if ($protTaxStr || $protTaxId) {
        $ttNum = common_subs::getProteinTranslationTableID($self->objParams, $acc, $db);
      }
      $self->fh->print('" translationTable_ref="' . $self->_mzid_TranslationTable_id($ttNum));
      if ($framenum >= 1 && $framenum <=3) {
        $self->fh->print('" frame="' . $framenum);
      } else {
        $self->fh->print('" frame="' . ($framenum -7));  # 6 -> -1, 5 -> -2, 4 -> -3
      }
    }
    $self->fh->print("\" isDecoy=\"" . "false"
      . "\" dBSequence_ref=\"" . $self->_mzid_DBSeq_id($db, $preferred_acc) . "\" />\n");
  }
}

sub printSequenceCollection {
  my $self = shift;
  my ($queryAll, $protLen, $protSeq, $protDesc,
      $protTaxStr, $protTaxId,
      $sameset_ordering, $fixedMods,
      $filter_duplicates) = @_;

  $self->fh->print("  <SequenceCollection>\n");

  my ($exported_sequences, $exported_queryList) = $self->printDBSequences(
    $queryAll, $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
    $sameset_ordering,
  );

  my $peptide_evidence_list = $self->printSequenceCollectionPeptides(
    $exported_sequences, $exported_queryList, $queryAll,
    $fixedMods, $protTaxId, $protTaxStr, $filter_duplicates
  );

  $self->fh->print("  </SequenceCollection>\n");

  return ($exported_sequences, $exported_queryList, $peptide_evidence_list);
}

sub printDBSequences {
  my $self = shift;
  my ($queryAll, $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
      $sameset_ordering) = @_;

  my $isPMF = $self->objResFile->isPMF;

  my %exported_sequences = ();
  my @queryList = ();

  for my $h (1 .. $self->objSummary->getNumberOfHits()) {

    my $objProtein = $self->objSummary->getHit($h);
    my @sameset = common_subs::compileSameset(
      $self->objSummary, $objProtein, $self->ShowSameSets, $$sameset_ordering{$h}
    );

    # Note that markExportableQueries() must match exactly the calls done later
    # for printProteinDetectionHypothesis(). Otherwise some PeptideEvidence
    # elements can be missing from the exported file. See bug #11013.

    unless ($exported_sequences{ key_acc($sameset[0]) }) {
      $self->printDBSequence(
        $sameset[0]->getDB, $sameset[0]->getAccession,
        $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
      );
      $exported_sequences{ key_acc($sameset[0]) } = 1;
      markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
    }

    if ($self->group_family) {

    # Other proteins that match the same set of peptides
      if ($self->ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          unless ($exported_sequences{ key_acc($_) }) {
            $self->printDBSequence(
              $_->getDB, $_->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
            );
            $exported_sequences{ key_acc($_) } = 1;
            markExportableQueries($queryAll, $isPMF, $_, \@queryList);
          }
        }
      }

    # Other family members
      my $k = 1;
      while (my $objProtein = $self->objSummary->getNextFamilyProtein($h, $k++)) {
        my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, $self->ShowSameSets,
          $$sameset_ordering{$h}
        );

        unless ($exported_sequences{ key_acc($sameset[0]) }) {
          $self->printDBSequence(
            $sameset[0]->getDB, $sameset[0]->getAccession,
            $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
          );
          $exported_sequences{ key_acc($sameset[0]) } = 1;
          markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
        }

      # And their same-sets
        if ($self->ShowSameSets) {
          for (@sameset[1 .. $#sameset]) {
            unless ($exported_sequences{ key_acc($_) }) {
              $self->printDBSequence(
                $_->getDB, $_->getAccession,
                $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
              );
              $exported_sequences{ key_acc($_) } = 1;
              markExportableQueries($queryAll, $isPMF, $_, \@queryList);
            }
          }
        }
      }


    # sub-sets and intersections for the entire family
      if ($self->ShowSubSets > 0) {
        my $k = 1;
        while (my $objProtein = $self->objSummary->getNextSubsetProtein($h, $k++, 1)) {
          my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, $self->ShowSameSets,
            $$sameset_ordering{$h}
          );

          unless ($exported_sequences{ key_acc($sameset[0]) }) {
            $self->printDBSequence(
              $sameset[0]->getDB, $sameset[0]->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
            );
            $exported_sequences{ key_acc($sameset[0]) } = 1;
            markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
          }

        # And their same-sets
          if ($self->ShowSameSets) {
            for (@sameset[1 .. $#sameset]) {
              unless ($exported_sequences{ key_acc($_) }) {
                $self->printDBSequence(
                  $_->getDB, $_->getAccession,
                  $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
                );
                $exported_sequences{ key_acc($_) } = 1;
                markExportableQueries($queryAll, $isPMF, $_, \@queryList);
              }
            }
          }
        }

      }

    } else {

      if ($self->ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          unless ($exported_sequences{ key_acc($_) }) {
            $self->printDBSequence(
              $_->getDB, $_->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
            );
            $exported_sequences{ key_acc($_) } = 1;
            markExportableQueries($queryAll, $isPMF, $_, \@queryList);
          }
        }
      }
      my $j = 1;
      while (my $objProtein = $self->objSummary->getNextSubsetProtein($h, $j++)) {
        my $key_acc = key_acc($objProtein);
        unless ($exported_sequences{ $key_acc }) {
          $self->printDBSequence(
            $objProtein->getDB, $objProtein->getAccession,
            $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
          );
          $exported_sequences{ $key_acc } = 1;
          markExportableQueries($queryAll, $isPMF, $objProtein, \@queryList);
        }
      }

    }

  }

  # if query_all, then want to output all possible DBSequence values so that we
  # can output <PeptideEvidence> for everything.
  if ($queryAll) {
    my @ranks = (1 .. $self->objSummary->getMaxRankValue());
    my ($query, $rank);
    my $db    = new msparser::vectori;
    my $tmp1  = new msparser::vectori;
    my $tmp2  = new msparser::vectori;
    my $tmp3  = new msparser::VectorString;
    my $tmp4  = new msparser::VectorString;
    my $tmp5  = new msparser::vectori;
    my $tmp6  = new msparser::vectori;
    for my $query (1 .. $self->objResFile->getNumQueries()) {
      for my $rank (@ranks) {
        $db->clear;
        $_->clear for ($tmp1, $tmp2, $tmp3, $tmp4, $tmp5, $tmp6);
        my $accessions = $self->objSummary->getAllProteinsWithThisPepMatch($query, $rank,
                                                                     $tmp1, # discard (start)
                                                                     $tmp2, # discard (end)
                                                                     $tmp3, # discard (pre)
                                                                     $tmp4, # discard (post)
                                                                     $tmp5,      # discard (frame)
                                                                     $tmp6, # discard (multiplicity)
                                                                     $db);

        for my $i (0 .. $#$accessions) {
          if ($self->objUniGene) {
            my $unigene_ids = _find_unigene_ids($self->objUniGene, $$accessions[$i]);

            if (@$unigene_ids) {
              for my $unigene_id (@$unigene_ids) {
                my $key_acc = key_acc_str(0, $unigene_id);

                unless ($exported_sequences{ $key_acc }) {
                  $self->printDBSequence(
                    0, $unigene_id,
                    $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
                  );
                  $exported_sequences{ $key_acc } = 1;
                }
              }

              next;
            }
          }

          my $accession = $$accessions[$i];
          my $key_acc = key_acc_str($db->get($i), $accession);

          unless ($exported_sequences{ $key_acc }) {
            $self->printDBSequence(
              $db->get($i), $accession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId
            );
            $exported_sequences{ $key_acc } = 1;
          }
        }
      }
    }
  }

  return (\%exported_sequences, \@queryList);
}

sub printDBSequence {
  my $self = shift;
  my ($dbIdx, $accession, $protLen, $protSeq,
      $protDesc, $protTaxStr, $protTaxId) = @_;

  my $preferred_acc = $self->proteinData->preferredAccession($dbIdx, $accession);

  $self->fh->print('    <DBSequence id="' . $self->_mzid_DBSeq_id($dbIdx, $preferred_acc)."\" ");

  if ($protLen) {
    my $length = $self->proteinData->getProteinLen($accession, 0, $dbIdx); # best to use NA length? $objProtein->getFrame()
    if ($length) {
      $self->fh->print("length=\"" . $length . "\" ");
    }
  }

  if ($dbIdx == 0) {
    # UniGene
    $self->fh->print('searchDatabase_ref="' . $self->_mzid_SearchDatabase_id('UniGene') . '" accession="' . $preferred_acc . '" ');
  } else {
    $self->fh->print('searchDatabase_ref="' . $self->_mzid_SearchDatabase_id($self->objParams->getDB($dbIdx)) . '" accession="' . $preferred_acc . '" ');
  }

  $self->fh->print(">\n");

  if ($protSeq) {
    # Have to use NA because PI doesn't support anything other than [A-Z] and
    # frame may be mixed.
    my ($seq, $err) = $self->proteinData->mustGetProteinSeq($accession, 0, $dbIdx, $self->shipper);
    if ($seq) {
        $self->fh->print("      <Seq>" . $seq . "</Seq>\n");
    }
  }

  if ($protDesc) {
    my $description = $self->proteinData->mustGetProteinDescription($preferred_acc, $dbIdx);
    if ($description) {
        $self->fh->print("      <cvParam accession=\"MS:1001088\" name=\"protein description\" cvRef=\"PSI-MS\" value=\"" . notags($description) . "\" />\n");
    }
  }

  if ($protTaxStr or $protTaxId) {
    my ($taxID, $species) = $self->proteinData->getProteinTaxonomy($preferred_acc, $dbIdx);

    if ($protTaxStr and $species) {
      $self->fh->print("      <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . notags($species) . "\" />\n");
    }

    if ($protTaxId and $taxID) {
      $self->fh->print("      <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"" . $taxID . "\" />\n");
    }
  }

  $self->fh->print("    </DBSequence>\n");
}

sub printAnalysisCollection {
  my $self = shift;
  my ($tolCVparam) = @_;

  # (When results are merged: it's desirable to have a single master list of
  # spectrum identifications, which is why we do not export one per input
  # file. For this reason activityDate is the date from the first results
  # file.)

  $self->fh->print("  <AnalysisCollection>\n");
  $self->fh->print('    <SpectrumIdentification id="' . $self->_mzid_SpectrumIdentification_id() . "\" spectrumIdentificationProtocol_ref=\"" . $self->_mzid_SpectrumIdentificationProtocol_id() . "\"  spectrumIdentificationList_ref=\"" . $self->_mzid_SpectrumIdentificationList_id(1) . "\" ");
  $self->fh->printf("activityDate=\"%s\">\n", common_subs::_fmtdate_iso8601($self->objResFile->getDate()));
  for my $i (1 .. $self->objResFile->getNumberOfResfiles()) {
    $self->fh->print('      <InputSpectra spectraData_ref="' . $self->_mzid_SpectraData_id($i) . "\" />\n");
  }
  for my $i (1 .. $self->objParams->getNumberOfDatabases()) {
    $self->fh->print('      <SearchDatabaseRef searchDatabase_ref="' . $self->_mzid_SearchDatabase_id($self->objParams->getDB($i)) . "\" />\n");
  }
  $self->fh->print("    </SpectrumIdentification>\n");

  # Protein determination is performed 'now'
  $self->fh->print('    <ProteinDetection id="' . $self->_mzid_ProteinDetection_id(1) . "\" proteinDetectionProtocol_ref=\"" . $self->_mzid_ProteinDetectionProtocol_id('MascotParser', '1') . "\" proteinDetectionList_ref=\"" . $self->_mzid_ProteinDetectionList_id(1) . "\" ");
  $self->fh->printf("activityDate=\"%s\">\n", common_subs::_fmtdate_iso8601(time));
  $self->fh->print('      <InputSpectrumIdentifications spectrumIdentificationList_ref="' . $self->_mzid_SpectrumIdentificationList_id(1) . "\" />\n");
  $self->fh->print("    </ProteinDetection>\n");
  $self->fh->print("  </AnalysisCollection>\n");
}

sub printAnalysisProtocolCollection {
  my $self = shift;
  my ($resfile_params,
      $protTaxStr,
      $tolCVparam) = @_;

  # (When results are merged: nearly all parameters must be the same between
  # the input files, so we simply use $self->objParams. The exception is the search
  # title (COM).)

  # Bug 12187: If there are only SL databases in the search, omit parameters
  # MSPepSearch doesn't use.
  my $SL_only = do {
    my %num_db_types = ExportDat::Util::count_db_types($self->objResFile, $self->objParams);
    $num_db_types{$msparser::SL} && 1 == keys %num_db_types
  };

  $self->fh->print("  <AnalysisProtocolCollection>\n");

  $self->printSpectrumIdentificationProtocol($resfile_params, $protTaxStr, $tolCVparam, $SL_only);

  my $scoringType;
  if ($self->anyLibraryMatch
          or ($self->objResFile->getNumSeqsAfterTax() > 0 and $self->objResFile->getNumQueries() / $self->objResFile->getNumSeqsAfterTax() > $self->mudpitSwitch)) {
    $scoringType = "MudPIT";
  } else {
    $scoringType = "Standard";
  }
  my $maxProteinHits = $self->numHits;
  if ($maxProteinHits == 0) {
    $maxProteinHits = "Auto";
  }
  my $useUnigene = "true";
  if ($self->UniGeneFile eq "") {
    $useUnigene = "false";
  }
  $self->fh->print('    <ProteinDetectionProtocol id="' . $self->_mzid_ProteinDetectionProtocol_id('MascotParser', 1) . "\"  analysisSoftware_ref=\"AS_mascot_parser\">\n");
  $self->fh->print("      <AnalysisParams>\n");
  do {
    my $value = $self->objSummary->getProbabilityThreshold;
    if ($SL_only) {
      $value = msparser::ms_peptidesummary::getSLThresholdFromMinProbability($value);
    }
    $self->fh->print("        <cvParam accession=\"MS:1001316\" name=\"Mascot:SigThreshold\"                               cvRef=\"PSI-MS\" value=\"" . $value     . "\" />\n");
  };
  $self->fh->print("        <cvParam accession=\"MS:1001317\" name=\"Mascot:MaxProteinHits\"                             cvRef=\"PSI-MS\" value=\"" . $maxProteinHits       . "\" />\n");
  unless ($self->objResFile->isPMF) {
    $self->fh->print("        <cvParam accession=\"MS:1001318\" name=\"Mascot:ProteinScoringMethod\"                       cvRef=\"PSI-MS\" value=\"" . $scoringType          . "\" />\n");
    # Bug 12007: Print the value regardless of whether it's defined in the URL.
    # It's what we used to open the results file.
    $self->fh->print("        <cvParam accession=\"MS:1001319\" name=\"Mascot:MinMSMSThreshold\"                           cvRef=\"PSI-MS\" value=\"" . $self->ignoreIonsScoreBelow . "\" />\n");
    if ($self->use_homology) {
      $self->fh->print("        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\"                           cvRef=\"PSI-MS\" value=\"homology\" />\n");
    } else {
      $self->fh->print("        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\"                           cvRef=\"PSI-MS\" value=\"identity\" />\n");
    }
  }
  if ($self->group_family) {
    $self->fh->print("        <cvParam accession=\"MS:1001759\" name=\"Mascot:ProteinGrouping\"                            cvRef=\"PSI-MS\" value=\"family clustering\" />\n");
    $self->fh->print("        <cvParam accession=\"MS:1003030\" name=\"Mascot:MinNumSigUniqueSeqs\"                        cvRef=\"PSI-MS\" value=\"" . $self->minNumSigUniqueSeqs . "\" />\n");
  } else {
    $self->fh->print("        <cvParam accession=\"MS:1001759\" name=\"Mascot:ProteinGrouping\"                            cvRef=\"PSI-MS\" value=\"Occam's razor\" />\n");
  }
  $self->fh->print("        <cvParam accession=\"MS:1001320\" name=\"Mascot:ShowHomologousProteinsWithSamePeptides\"     cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool($self->ShowSameSets) . "\" />\n");
  $self->fh->print("        <cvParam accession=\"MS:1001321\" name=\"Mascot:ShowHomologousProteinsWithSubsetOfPeptides\" cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool($self->ShowSubSets) . "\" />\n");
  unless ($self->objResFile->isPMF) {
    $self->fh->print("        <cvParam accession=\"MS:1001322\" name=\"Mascot:RequireBoldRed\"                             cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool($self->RequireBoldRed) . "\" />\n");
    $self->fh->print("        <cvParam accession=\"MS:1001323\" name=\"Mascot:UseUnigeneClustering\"                       cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool($useUnigene) . "\" />\n");
    $self->fh->print("        <cvParam accession=\"MS:1001324\" name=\"Mascot:IncludeErrorTolerantMatches\"                cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool(!$self->NoErrorTolerant) . "\" />\n");
  }
  $self->fh->print("        <cvParam accession=\"MS:1001325\" name=\"Mascot:ShowDecoyMatches\"                           cvRef=\"PSI-MS\" value=\"" . $self->_mzid_bool($self->ShowDecoyReport) . "\" />\n");
  if ($self->preferTaxonomy > 0) {
    my $taxonomyfile = MascotDat->taxonomy_file();
    my $taxonomy_entry;

    if ($self->preferTaxonomy < $taxonomyfile->getNumberOfEntries()) {
      $taxonomy_entry = $taxonomyfile->getEntryByNumber($self->preferTaxonomy);
    }

    my @tax_ids = ();

    if ($taxonomy_entry) {
      for my $i (0 .. $taxonomy_entry->getNumberOfIncludeTaxonomies()-1) {
        push @tax_ids, $taxonomy_entry->getIncludeTaxonomy($i);
      }
    }

    if (@tax_ids) {
      $self->fh->print("        <cvParam accession=\"MS:1001793\" name=\"Mascot:PreferredTaxonomy\"                           cvRef=\"PSI-MS\" value=\"" . join(' ', @tax_ids). "\" />\n");
    }
  }

  if ($self->percolate == 1) {
    my @features = split /\s*,\s*/, $self->objMascotOptions->getPercolatorFeatures();
    if ($self->percolate_rt) {
      push @features, 'retentionTime' unless 0 < grep { $_ eq 'retentionTime' } @features;
    }
    $self->fh->print("        <cvParam accession=\"MS:1001760\" name=\"Percolator:features\" cvRef=\"PSI-MS\" value=\"".join(' ', map { notags($_) } @features)."\" />\n");
  }

  if ($self->anyLibraryMatch) {
    my $y = ($self->libraryMode eq 'integrated') ? 1 : 0;
    $self->fh->print("        <cvParam accession=\"MS:1002749\" name=\"Mascot:IntegratedSpectralLibrarySearch\" cvRef=\"PSI-MS\" value=\"$y\" />\n");
  }

  $self->fh->print("      </AnalysisParams>\n");

  $self->fh->print("      <Threshold>\n");

  if ($self->showFDRs and not $self->objResFile->isPMF) {
    my ($targetCount, $decoyCount, $proteinFDR) = ExportDat::Util::calculateProteinFDR($self->objSummary, $self->objSummaryDecoy, $self->objResFile);
    if (defined $proteinFDR and $proteinFDR > 0) {
      $self->fh->print("        <cvParam accession=\"MS:1001214\" name=\"protein-level global FDR\" cvRef=\"PSI-MS\" value=\"@{[ round_evenly($proteinFDR, 4) ]}\" />\n");
    } else {
      $self->fh->print("        <cvParam accession=\"MS:1001494\" name=\"no threshold\" cvRef=\"PSI-MS\" />\n");
    }
  } else {
    $self->fh->print("        <cvParam accession=\"MS:1001494\" name=\"no threshold\" cvRef=\"PSI-MS\" />\n");
  }

  print {$self->{fh}} <<EOF;
      </Threshold>
    </ProteinDetectionProtocol>
  </AnalysisProtocolCollection>
EOF
}

sub printSpectrumIdentificationProtocol {
    my $self = shift;
    my ($resfile_params,
	$protTaxStr,
	$tolCVparam,
	$SL_only) = @_;

    $self->fh->print('    <SpectrumIdentificationProtocol id="' . $self->_mzid_SpectrumIdentificationProtocol_id() . "\" analysisSoftware_ref=\"AS_mascot_server\">\n");
    $self->printSpectrumIdentificationProtocolBody($resfile_params, $protTaxStr, $tolCVparam, $SL_only);
    $self->fh->print("    </SpectrumIdentificationProtocol>\n");

}

sub printSpectrumIdentificationProtocolBody {
    my $self = shift;
    my ($resfile_params,
	$protTaxStr,
	$tolCVparam,
	$SL_only) = @_;

    $self->printSearchType();

  $self->printAdditionalSearchParams($resfile_params,	$protTaxStr, $tolCVparam, $SL_only);

  my $fixed_mods = $self->objParams->getMODS();

  if ($self->objQuantMethod) {
    # add in any fixed mods specified in the quant method
    my $qMods = quant_subs::getQuantFixedModsAsString($self->objQuantMethod);
    if ($qMods && $fixed_mods) {
      $fixed_mods .= "," . $qMods;
    } elsif ($qMods) {
      $fixed_mods = $qMods;
    }
  }

  my $variable_mods = $self->objParams->getIT_MODS();

  if ($self->objQuantMethod) {
    # add in any variable or exclusive mods specified in the quant method
    my $qMods = quant_subs::getQuantVariableModsAsString($self->objQuantMethod);
    if ($qMods && $variable_mods) {
      $variable_mods .= "," . $qMods;
    } elsif ($qMods) {
      $variable_mods = $qMods;
    }
  }

  if (not $SL_only and ($fixed_mods || $variable_mods)) {
    $self->fh->print("      <ModificationParams>\n");
    if ($fixed_mods) {
      my @mod_list = split(/,/, $fixed_mods);
      foreach (@mod_list) {
        $self->printSearchModification($_, "true");
      }
    }
    if ($variable_mods) {
      my @mod_list = split(/,/, $variable_mods);
      foreach (@mod_list) {
        $self->printSearchModification($_, "false");
      }
    }
    $self->fh->print("      </ModificationParams>\n");
  }

# create an enzymes file object
  my $objEnzymesFile = new msparser::ms_enzymefile;

  if (not $SL_only and $self->objResFile->getEnzyme($objEnzymesFile, "../config/enzymes")) {
    my $enzyme = $objEnzymesFile->getEnzymeByNumber(0);
    if ($enzyme->isValid) {
      $self->fh->print("      <Enzymes");
      if ($enzyme->getNumberOfCutters > 1) {
        $self->fh->print(" independent=\"");
        if ($enzyme->isIndependent) {
          $self->fh->print("1\"");
        } else {
          $self->fh->print("0\"");
        }
      }
      $self->fh->print(">\n");

      my %cvParam = (
        '::none' => { id => "MS:1001955", name => "no cleavage" },
        'kr:p:trypsin' => { id => "MS:1001251", name => "Trypsin" },
        'r:p:arg-c' => { id => "MS:1001303", name => "Arg-C" },
        'bd::asp-n' => { id => "MS:1001304", name => "Asp-N" },
        'de::asp-n_ambic' => { id => "MS:1001305", name => "Asp-N_ambic" },
        'flwy:p:chymotrypsin' => { id => "MS:1001306", name => "Chymotrypsin" },
        'm::cnbr' => { id => "MS:1001307", name => "CNBr" },
        'd::formic_acid' => { id => "MS:1001308", name => "Formic_acid" },
        'k:p:lys-c' => { id => "MS:1001309", name => "Lys-C" },
        'k::lys-c/p' => { id => "MS:1001310", name => "Lys-C/P" },
        'fl::pepsina' => { id => "MS:1001311", name => "PepsinA" },
        'fklrwy:p:trypchymo' => { id => "MS:1001312", name => "TrypChymo" },
        'kr::trypsin/p' => { id => "MS:1001313", name => "Trypsin/P" },
        'bdez:p:v8-de' => { id => "MS:1001314", name => "V8-DE" },
        'ez:p:v8-e' => { id => "MS:1001315", name => "V8-E" },
      );

      for my $i (0 .. $enzyme->getNumberOfCutters()-1) {
        my $cTermGain = $self->objParams->getOxygenMass + $self->objParams->getHydrogenMass;
        my $nTermGain = $self->objParams->getHydrogenMass;
        $self->fh->print('        <Enzyme id="' . $self->_mzid_Enzyme_id($i)
            . "\" cTermGain=\"" . "OH" . "\" nTermGain=\"" . "H"
            . "\" missedCleavages=\"" . $self->objParams->getPFA . "\" semiSpecific=\"");
        if ($enzyme->isSemiSpecific) {
          $self->fh->print("1");
        } else {
          $self->fh->print("0");
        }
        $self->fh->print("\">\n");
        $self->fh->print("          <SiteRegexp>");
        $self->fh->print("<![CDATA[");
        if (length($enzyme->getCleave($i)) == 1) {
          $self->fh->print("(?<=" . $enzyme->getCleave($i) . ")");
        } elsif (length($enzyme->getCleave($i)) > 1) {
          $self->fh->print("(?<=[" . $enzyme->getCleave($i) . "])");
        }
        if (length($enzyme->getRestrict($i)) == 1) {
          $self->fh->print("(?!" . $enzyme->getRestrict($i) . ")");
        } elsif (length($enzyme->getRestrict($i)) > 1) {
          $self->fh->print("(?![" . $enzyme->getRestrict($i) . "])");
        }
        $self->fh->print("]]>");

      # sort multi-residue strings to simplify comparison
        my($cleaver, $restrictor, @tmpRes);
        if (length($enzyme->getCleave($i)) > 1) {
          @tmpRes = split(//, lc($enzyme->getCleave($i)));
          @tmpRes = sort { $a cmp $b } @tmpRes;
          $cleaver = join("", @tmpRes)
        } else {
          $cleaver = lc($enzyme->getCleave($i));
        }
        if (length($enzyme->getRestrict($i)) > 1) {
          @tmpRes = split(//, lc($enzyme->getRestrict($i)));
          @tmpRes = sort { $a cmp $b } @tmpRes;
          $restrictor = join("", @tmpRes)
        } else {
          $restrictor = lc($enzyme->getRestrict($i));
        }

        $self->fh->print("</SiteRegexp>\n");
        $self->fh->print("          <EnzymeName>\n");

        my $cv = $cvParam{
          join(':', $cleaver, $restrictor ? $restrictor : '', lc($enzyme->getTitle))
        };

        if ($cv) {
          my $name = notags($$cv{'name'});
          $self->fh->print("            <cvParam accession=\"$$cv{'id'}\" name=\"$name\" cvRef=\"PSI-MS\" />\n");
        } else {
          # Alternative is to use M:1001956 "unspecified cleavage".
          $self->fh->print("            <userParam name=\"Enzyme\" value=\"" . notags($enzyme->getTitle) . "\" />\n");
        }
        $self->fh->print("          </EnzymeName>\n");
        $self->fh->print("        </Enzyme>\n");
      }
      $self->fh->print("      </Enzymes>\n");
    }
  }

  my $multiMassTables = 1;

  if ($self->objQuantMethod and $self->objQuantMethod->getNumberOfComponents > 1) {
    for my $c (0 .. $self->objQuantMethod->getNumberOfComponents()-1) {
      my $component = $self->objQuantMethod->getComponentByNumber($c);
      if ($component->getNumberOfIsotopes > 0) {
        $multiMassTables = $self->objQuantMethod->getNumberOfComponents;
        last;
      }
    }
  }

  for my $c (0 .. $multiMassTables-1) {
    my $objLocalMassesFile = new msparser::ms_masses($self->objMassesFile);
    if ($multiMassTables > 1) {
      $objLocalMassesFile->applyIsotopes($self->objUmodSection, $self->objQuantMethod->getComponentByNumber($c));
      $self->fh->print('      <MassTable id="' . $self->_mzid_MassTable_id($self->objQuantMethod->getComponentByNumber($c)->getName) . "\"");
    } else {
      $self->fh->print('      <MassTable id="' . $self->_mzid_MassTable_id(1) . "\"");
    }
    #my $err = new msparser::ms_errs;
  #  my $dummy = $objLocalMassesFile->applyFixedMods($vecFixed, $err);  # decided MassTable should be completely unmodified
    #unless ($err->isValid) {
    #  my $errorList = "Mascot Parser error(s):<BR>\n";
    #  for (my $i = 1; $i <= $err->getNumberOfErrors(); $i++) {
    #    $errorList .= $err->getErrorNumber($i)
    #      . ": " . $err->getErrorString($i) . "<BR>\n";
    #  }
    #  die($errorList, " ", $idForFatalError);
    #}
    $self->fh->print(" msLevel=\"1 2\">\n");
    for my $residue ('A' .. 'Z') {
      my $mass = $objLocalMassesFile->getResidueMass($self->mass_type, $residue);
      if ($residue ne 'B' and $residue ne 'X' and $residue ne 'Z' and $mass != 0) {
        $self->fh->print("        <Residue code=\"" . $residue . "\" mass=\"" . $mass . "\" />\n");
      }
    }
    print {$self->{fh}} <<EOF;
        <AmbiguousResidue code="B">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="D N" />
        </AmbiguousResidue>
        <AmbiguousResidue code="Z">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="E Q" />
        </AmbiguousResidue>
        <AmbiguousResidue code="X">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="A C D E F G H I K L M N O P Q R S T U V W Y" />
        </AmbiguousResidue>
      </MassTable>
EOF
  }

  if (!$self->objResFile->isPMF) {
    $self->fh->print("      <FragmentTolerance>\n");

    if (my $cv = $$tolCVparam{$self->objParams->getITOLU}) {
      my $val = $$cv{'fmt'}->($self->objParams->getITOL);
      $self->fh->print("        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
      $self->fh->print("        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
    }

    $self->fh->print("      </FragmentTolerance>\n");
  }

  $self->fh->print("      <ParentTolerance>\n");

  if (my $cv = $$tolCVparam{$self->objParams->getTOLU}) {
    my $val = $$cv{'fmt'}->($self->objParams->getTOL);
    $self->fh->print("        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
    $self->fh->print("        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
  }

  $self->fh->print("      </ParentTolerance>\n");

  $self->fh->print("      <Threshold>\n");
  my $value = $self->objSummary->getProbabilityThreshold;
  if ($SL_only) {
    $value = msparser::ms_peptidesummary::getSLThresholdFromMinProbability($value);
  }
  $self->fh->print("        <cvParam accession=\"MS:1001316\" name=\"Mascot:SigThreshold\" cvRef=\"PSI-MS\" value=\"" . $value . "\" />\n");
  if ($self->use_homology) {
    $self->fh->print("        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"homology\" />\n");
  } else {
    $self->fh->print("        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"identity\" />\n");
  }
  if ($self->showFDRs and not $self->objResFile->isPMF) {
    for my $count_type ($msparser::ms_mascotresults::DS_COUNT_SEQUENCE, $msparser::ms_mascotresults::DS_COUNT_PSM) {
      my $FDR;
      if ($self->use_homology) {
	my $HF = $self->objSummary->getNumHitsAboveHomology(1 / $self->objSummary->getProbabilityThreshold, $count_type);
	my $HD = $self->objSummary->getNumDecoyHitsAboveHomology(1 / $self->objSummary->getProbabilityThreshold, $count_type);
	if ($HF > 0) {
	  $FDR = round_evenly($HD / $HF, 4);
	}
      } else {
	my $IF = $self->objSummary->getNumHitsAboveIdentity(1 / $self->objSummary->getProbabilityThreshold, $count_type);
	my $ID = $self->objSummary->getNumDecoyHitsAboveIdentity(1 / $self->objSummary->getProbabilityThreshold, $count_type);
	if ($IF > 0) {
	  $FDR = round_evenly($ID / $IF, 4);
	}
      }
      # NB: Exported as an absolute value between 0 and 1, not as a percentage.
      if (defined $FDR) {
	if ($count_type == $msparser::ms_mascotresults::DS_COUNT_SEQUENCE) {
	  $self->fh->print("        <cvParam accession=\"MS:1001364\" name=\"peptide sequence-level global FDR\" cvRef=\"PSI-MS\" value=\"$FDR\" />\n");
	} else {
	  $self->fh->print("        <cvParam accession=\"MS:1002350\" name=\"PSM-level global FDR\" cvRef=\"PSI-MS\" value=\"$FDR\" />\n");
	}
      }
    }
  }
  $self->fh->print("      </Threshold>\n");

  print {$self->{fh}} <<EOF;
      <DatabaseFilters>
        <Filter>
          <FilterType>
            <cvParam accession="MS:1001020" name="DB filter taxonomy" cvRef="PSI-MS" />
          </FilterType>
EOF

  my $taxonomy = new msparser::ms_taxonomyfile;
  if (not $SL_only and $self->objResFile->getTaxonomy($taxonomy)) {
    if ($taxonomy->getNumberOfEntries == 0 ) {
      print {$self->{fh}} <<EOF;
          <Include>
            <cvParam accession="MS:1001467" name="taxonomy: NCBI TaxID" cvRef="PSI-MS" value="1" />
          </Include>
EOF
    } else {
      my $choice = $taxonomy->getEntryByNumber(0);
      if ($choice->getNumberOfIncludeTaxonomies > 0) {
        $self->fh->print("          <Include>\n");
        for my $i (0 .. $choice->getNumberOfIncludeTaxonomies()-1) {
          $self->fh->print("            <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"" . $choice->getIncludeTaxonomy($i) . "\" />\n");
          if ($protTaxStr) {
            my $tax_title = ExportDat::TaxonomyData::getTaxonomyNameFromId($choice->getIncludeTaxonomy($i), $self->taxNames);
            if ($tax_title) {
              $self->fh->print("            <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . notags($tax_title) . "\" />\n");
            }
          }
        }
        $self->fh->print("          </Include>\n");
      }

      if ($choice->getNumberOfExcludeTaxonomies > 0) {
        $self->fh->print("          <Exclude>\n");
        for my $i (0 .. $choice->getNumberOfExcludeTaxonomies()-1) {
          $self->fh->print("            <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"". $choice->getExcludeTaxonomy($i) . "\" />\n");
          if ($protTaxStr) {
            my $tax_title = ExportDat::TaxonomyData::getTaxonomyNameFromId($choice->getExcludeTaxonomy($i), $self->taxNames);
            if ($tax_title) {
              $self->fh->print("            <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . notags($tax_title) . "\" />\n");
            }
          }
        }
        $self->fh->print("          </Exclude>\n");
      }
    }
  }
  print {$self->{fh}} <<EOF;
        </Filter>
      </DatabaseFilters>
EOF

  my $any_NA;

  # Bug 12416: In Mascot 2.6 and later files, we have a shortcut to database
  # types. In earlier files, the search results are either all AA or all NA.
  if ($self->objResFile->isDatabaseTypeAvailable) {
    for my $dbIdx (1..$self->objParams->getNumberOfDatabases) {
      if ($self->objResFile->getDatabaseType($dbIdx) == $msparser::NA) {
        $any_NA = 1;
      }
    }
  } else {
    $any_NA = $self->objSummary->isNA;
  }

  if ($any_NA) {
    open(my $fh, '<', "../taxonomy/gencode.dmp")
      or die sprintf("Cannot open gencode.dmp (%s:%d): %s (%s)", __FILE__, __LINE__, $!, $^E);

    $self->fh->print("      <DatabaseTranslation frames=\"1 2 3 -1 -2 -3\">\n");
    while (defined(my $gencode = <$fh>)) {
      my ($id, $dummy, $name, $table, $startcodons) = split(/\s*\|\s*/, $gencode);
      if ($table && $startcodons) {
        $self->fh->print('        <TranslationTable id="' . $self->_mzid_TranslationTable_id($id) . "\" name=\"" . $name . "\">\n");
        $self->fh->print("          <cvParam accession=\"MS:1001025\" name=\"translation table\" cvRef=\"PSI-MS\" value=\"" . $table . "\" />\n");
        $self->fh->print("          <cvParam accession=\"MS:1001410\" name=\"translation start codons\" cvRef=\"PSI-MS\" value=\"" . $startcodons . "\" />\n");
        $self->fh->print("          <cvParam accession=\"MS:1001423\" name=\"translation table description\" cvRef=\"PSI-MS\" value=\"https://www.ncbi.nlm.nih.gov/Taxonomy/taxonomyhome.html/index.cgi?chapter=cgencodes#SG" . $id . "\" />\n");
        $self->fh->print("        </TranslationTable>\n");
      }
    }
    $self->fh->print("      </DatabaseTranslation>\n");

    close($fh)
      or die sprintf("Cannot close gencode.dmp (%s:%d): %s (%s)", __FILE__, __LINE__, $!, $^E);
  }
}

sub printSearchType {
    my $self = shift;

    $self->fh->print("      <SearchType>\n");
    $self->printSearchTypeBody();
    $self->fh->print("      </SearchType>\n");
}
sub printSearchTypeBody {
    my $self = shift;

  if ($self->objResFile->anyMSMS) {
    if ($self->objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_PEPTIDES)
            and $self->objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES)) {
      $self->fh->print("        <cvParam accession=\"MS:1002755\" name=\"combined ms-ms + spectral library search\" cvRef=\"PSI-MS\" value=\"\" />\n");
    } elsif ($self->objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES)) {
      $self->fh->print("        <cvParam accession=\"MS:1001031\" name=\"spectral library search\" cvRef=\"PSI-MS\" value=\"\" />\n");
    } else {
      $self->fh->print("        <cvParam accession=\"MS:1001083\" name=\"ms-ms search\" cvRef=\"PSI-MS\" value=\"\" />\n");
    }
  } elsif ($self->objResFile->anySQ || $self->objResFile->anyTag) {
    $self->fh->print("        <cvParam accession=\"MS:1001082\" name=\"tag search\" cvRef=\"PSI-MS\" value=\"\" />\n");
  } else {
    $self->fh->print("        <cvParam accession=\"MS:1001081\" name=\"pmf search\" cvRef=\"PSI-MS\" value=\"\" />\n");
  }

}

sub printAdditionalSearchParams {
    my $self = shift;
        my ($resfile_params,
	$protTaxStr,
	$tolCVparam,
	$SL_only) = @_;

	$self->fh->print("      <AdditionalSearchParams>\n");
	$self->printAdditionalSearchParamsBody($resfile_params, $protTaxStr, $tolCVparam, $SL_only);
	$self->fh->print("      </AdditionalSearchParams>\n");
}
sub printAdditionalSearchParamsBody {
    my $self = shift;
    my ($resfile_params,
	$protTaxStr,
	$tolCVparam,
	$SL_only) = @_;

  my %printed_COM = ();
  for my $i (1 .. $#$resfile_params) {
    my $params = $$resfile_params[$i];
    if (my $com = $params->getCOM) {
      next if $printed_COM{$com};
      $printed_COM{$com} = 1;
      $self->fh->print("        <userParam name=\"Mascot User Comment\" value=\"" . notags($com) . "\" />\n");
    }
  }
  if ($self->objParams->getINSTRUMENT and not $SL_only) {
    $self->fh->print("        <userParam name=\"Mascot Instrument Name\" value=\"" . notags($self->objParams->getINSTRUMENT)  . "\" />\n");
  }

  if ($self->objParams->getACCESSION and not $SL_only) {
    $self->fh->print("        <cvParam accession=\"MS:1001021\" name=\"DB filter on accession numbers\" cvRef=\"PSI-MS\" value=\"" . notags($self->objParams->getACCESSION) . "\" />\n");
  }
  if (lc($self->objParams->getMASS) eq "monoisotopic") {
    $self->fh->print("        <cvParam accession=\"MS:1001211\" name=\"parent mass type mono\"    cvRef=\"PSI-MS\" />\n");
  } else {
    $self->fh->print("        <cvParam accession=\"MS:1001212\" name=\"parent mass type average\" cvRef=\"PSI-MS\" />\n");
  }
  if (not $SL_only and ($self->objResFile->anyMSMS || $self->objResFile->anySQ || $self->objResFile->anyTag)) {
    my %cvParam = (
      # 1 => { name => "singly charged" },
      # 2 => { name => "doubly charged if precursor 2+ or higher (not internal or immonium)" },
      # 3 => { name => "doubly charged if precursor 3+ or higher (not internal or immonium)" },
      4 =>  { id => "MS:1001259", name => "param: immonium ion" },
      5 =>  { id => "MS:1001108", name => "param: a ion" },
      6 =>  { id => "MS:1001146", name => "param: a ion-NH3" },
      7 =>  { id => "MS:1001148", name => "param: a ion-H2O" },
      8 =>  { id => "MS:1001118", name => "param: b ion" },
      9 =>  { id => "MS:1001149", name => "param: b ion-NH3" },
      10 => { id => "MS:1001150", name => "param: b ion-H2O" },
      11 => { id => "MS:1001119", name => "param: c ion" },
      12 => { id => "MS:1001261", name => "param: x ion" },
      13 => { id => "MS:1001262", name => "param: y ion" },
      14 => { id => "MS:1001151", name => "param: y ion-NH3" },
      15 => { id => "MS:1001152", name => "param: y ion-H2O" },
      16 => { id => "MS:1001263", name => "param: z ion" },
      17 => { id => "MS:1001406", name => "param: internal yb ion" },
      18 => { id => "MS:1001407", name => "param: internal ya ion" },
      # 19 => { name => "", id => "" },
      # 20 => { name => "", id => "" },
      21 => { id => "MS:1001408", name => "param: z+1 ion" },
      22 => { id => "MS:1001258", name => "param: d ion" },
      23 => { id => "MS:1001257", name => "param: v ion" },
      24 => { id => "MS:1001260", name => "param: w ion" },
      25 => { id => "MS:1001409", name => "param: z+2 ion" },
    );
    my @rules_list = split(/,/, $self->objParams->getRULES);

    foreach my $ruleNum (@rules_list) {
      my $cv = $cvParam{$ruleNum};
      next unless $cv;

      $self->fh->print("        <cvParam accession=\"" . $$cv{'id'}. "\" name=\"" . notags($$cv{'name'}) . "\" cvRef=\"PSI-MS\" />\n");
    }
  }

}

sub printSearchModification {
  my $self = shift;
  my ($mod_file_name, $fixed) = @_;

  my ($mod_name, $specificity) = $mod_file_name =~ /(.*)\s+\((.*)\)/;

  my ($specificityAccession, $specificityName, $residues);

  if ($specificity =~ /^N-term/i) {
    $specificityAccession = "MS:1001189";
    $specificityName      = "modification specificity N-term";
    ($residues) = $specificity =~ /^N-term\s+(.*)/i;
  } elsif ($specificity =~ /^C-term/i) {
    $specificityAccession = "MS:1001190";
    $specificityName      = "modification specificity C-term";
    ($residues) = $specificity =~ /^C-term\s+(.*)/i;
  } elsif( $specificity =~ /^Protein N-term/i) {
    $specificityAccession = "MS:1001189";
    $specificityName      = "modification specificity N-term";
    ($residues) = $specificity =~ /^Protein N-term\s+(.*)/i;
  } elsif ($specificity =~ /^Protein C-term/i) {
    $specificityAccession = "MS:1001190";
    $specificityName      = "modification specificity C-term";
    ($residues) = $specificity =~ /^Protein C-term\s+(.*)/i;
  } else {
    $specificityAccession = "";
    $specificityName      = "";
    $residues             = $specificity;
  }

  $residues = '' unless defined($residues);

  my $cv_param;
  my $mod2 = $self->objUmodSection->getModificationByName($mod_name);

  if ($mod2) {
    $cv_param = "accession=\"UNIMOD:" . $mod2->getRecordID . "\" name=\"" . notags($mod_name) . "\" cvRef=\"UNIMOD\"";
  } else {
    $cv_param = "accession=\"MS:1001460\" name=\"unknown modification\" cvRef=\"PSI-MS\"";
  }
  my $delta = $self->objModFile->getModificationByName($mod_file_name)->getDelta($self->mass_type);

  if (length($residues) > 1) {
    $residues = join(" ", split(//, $residues));
  }

  $self->fh->print("        <SearchModification fixedMod=\"" . $fixed . "\" massDelta=\"" . $delta . "\" residues=\"" . $residues . "\">\n");
  if ($specificityAccession) {
    $self->fh->print("          <SpecificityRules>\n");
    $self->fh->print("            <cvParam accession=\"" . $specificityAccession . "\" cvRef=\"PSI-MS\" name=\"" . $specificityName . "\" />\n");
    $self->fh->print("          </SpecificityRules>\n");
  }
  $self->fh->print("          <cvParam $cv_param />\n");
  $self->fh->print("        </SearchModification>\n");
}

sub printDataCollection {
  my $self = shift;
  my ($resfile_params, $resfile_children,
      $exported_queryList, $exported_sequences,
      $peptide_evidence_list, $aaHelpers, $tolCVparam,
      $protTaxStr, $protTaxId, $queryAll,
      $protLen, $queryRaw, $sameset_ordering, $numcomponents,
      $firstComponentName,) = @_;

  $self->fh->print("  <DataCollection>\n");
  my ($indexIsQueryNumber) = $self->printDataCollectionInputs(
    $resfile_params, $resfile_children
  );

  $self->fh->print("    <AnalysisData>\n");

  $self->printSpectrumIdentificationList(
    $resfile_params,
    $exported_queryList, $exported_sequences,
    $peptide_evidence_list, $protTaxStr, $protTaxId, $queryAll, $aaHelpers,
    $tolCVparam, $queryRaw, $indexIsQueryNumber,
    $numcomponents, $firstComponentName,
  );

  $self->printProteinDetectionList(
    $peptide_evidence_list,
    $protLen, $sameset_ordering
  );

  $self->fh->print("    </AnalysisData>\n");
  $self->fh->print("  </DataCollection>\n");
}

sub printDataCollectionInputs {
  my $self = shift;
  my ($resfile_params, $resfile_children) = @_;

  $self->fh->print("    <Inputs>\n");

  for my $i (1 .. @{$self->fileIn}) {
    $self->fh->print("      <SourceFile location=\"file:///" . notags($self->fileIn->[$i-1]) . "\" id=\"" . $self->_mzid_SourceFile_id($i) . "\" >\n");
    $self->fh->print("        <FileFormat>\n");
    $self->fh->print("          <cvParam accession=\"MS:1001199\" name=\"Mascot DAT format\" cvRef=\"PSI-MS\" />\n");
    $self->fh->print("        </FileFormat>\n");
    $self->fh->print("      </SourceFile>\n");
  }

  if ($self->UniGeneFile) {
    $self->fh->print("      <SearchDatabase location=\"file:///" . notags($self->UniGeneFile)
                 . "\" id=\"" . $self->_mzid_SearchDatabase_id('UniGene')
                 . "\" name=\"" . 'UniGene'
                 . "\">\n");
    $self->fh->print("        <ExternalFormatDocumentation>https://www.ncbi.nlm.nih.gov/unigene/</ExternalFormatDocumentation>\n");
    $self->fh->print("        <DatabaseName>\n");
    $self->fh->print("          <userParam name=\"" . notags($self->UniGeneFile) . "\" />\n");
    $self->fh->print("        </DatabaseName>\n");
    $self->fh->print("      </SearchDatabase>\n");
  }

  # (When results are merged: merged results have the same databases, so we
  # don't need to merge database information manually.)

  for my $i (1 .. $self->objParams->getNumberOfDatabases()) {
    # releaseDate is not easy to figure out. We could, in principle, check
    # what the last modified time of the local FASTA file is (if it's
    # available), but that's not reliable. It's best to omit.

    # Mascot 2.3 added multi-database search support. In files since then,
    # getNumSeqs(1) and getNumResidues(1) correctly return the number of
    # sequences and residues in database 1. In files older than that, we
    # should use getNumSeqs(0) to be on the safe side. No version test is
    # needed; just use index 0 if there's only one database in the search.
    my $ii = $self->objParams->getNumberOfDatabases() == 1 ? 0 : $i;

    $self->fh->print("      <SearchDatabase location=\"file:///" . notags($self->objResFile->getFastaPath($i))
                 . "\" id=\"" . $self->_mzid_SearchDatabase_id($self->objParams->getDB($i))
                 . "\" name=\"" . notags($self->objParams->getDB($i))
                 . "\" numDatabaseSequences=\"" . $self->objResFile->getNumSeqs($ii)    # wait for parser function ### debug ###
                 . "\" numResidues=\""          . $self->objResFile->getNumResidues($ii)
                 . "\" version=\""              . $self->objResFile->getFastaVer($i)
                 . "\">\n");
    $self->fh->print("        <FileFormat>\n");
    my $db_type = $self->objResFile->getDatabaseType($i);
    if ($db_type == $msparser::SL) {
      $self->fh->print("          <cvParam accession=\"MS:1002751\" name=\"NIST MSP format\" cvRef=\"PSI-MS\" />\n");
    } else {
      $self->fh->print("          <cvParam accession=\"MS:1001348\" name=\"FASTA format\" cvRef=\"PSI-MS\" />\n");
    }
    $self->fh->print("        </FileFormat>\n");
    $self->fh->print("        <DatabaseName>\n");
    $self->fh->print("          <userParam name=\"" . $self->objResFile->getFastaVer($i) . "\" />\n");
    $self->fh->print("        </DatabaseName>\n");

    # We need different tests for database type, because type information is
    # only recorded explicitly from Mascot 2.6 onwards. In earlier files, NA
    # must be deduced. On the other hand, type can only be SL from 2.6 onwards
    # when database type is recorded in the results file, so no special case
    # is needed in that case.
    if (ExportDat::Util::db_is_NA($self->objResFile, $self->objSummary, $i)) {
      $self->fh->print("        <cvParam accession=\"MS:1001079\" name=\"database type nucleotide\" cvRef=\"PSI-MS\" />\n");
    } elsif ($db_type == $msparser::SL) {
      $self->fh->print("        <cvParam accession=\"MS:1002752\" name=\"database type spectral library\" cvRef=\"PSI-MS\" />\n");
    } else {
      $self->fh->print("        <cvParam accession=\"MS:1001073\" name=\"database type amino acid\" cvRef=\"PSI-MS\" />\n");
    }

    $self->fh->print("      </SearchDatabase>\n");
  }

  my @indexIsQueryNumber = ();

  for my $i (1 .. $#$resfile_children) {
    my $resfile = $$resfile_children[$i];
    my $params = $$resfile_params[$i];

    my ($location) = $self->fileIn->[0] =~ /([^\/\\]*)\.dat$/;
    #BUG#13297: Use the the name of the exported .mgf file in location attribute.
    $location = 'file:///' . notags("$location.mgf");
    my $name = $params->getDATAURL();
    if ($name eq '') {
      $name = $params->getFILENAME();
    }
    $name = notags($name);

    my $indexIsQueryNumber = 1;
    my $objQueryTmp = new msparser::ms_inputquery($resfile, 1);
    if ($objQueryTmp->getIndex > -1) {
      $indexIsQueryNumber = 0;
    }

    $self->fh->print("      <SpectraData location=\"$location\" id=\"" . $self->_mzid_SpectraData_id($i) . "\" name=\"$name\">\n");
    $self->fh->print("        <FileFormat>\n");
    if ($params->getFORMAT eq "Mascot generic") {
      $self->fh->print("          <cvParam accession=\"MS:1001062\" name=\"Mascot MGF format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "Sequest (.DTA)") {
      $self->fh->print("          <cvParam accession=\"MS:1000613\" name=\"DTA format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "Finnigan (.ASC)") {
      $self->fh->print("          <cvParam accession=\"PI:99999\" name=\"TODO: Finnigan List file\" cvRef=\"MASCOT:PSI:PI\" />\n");   # not worth worrying about
    } elsif ($params->getFORMAT eq "Micromass (.PKL)") {
      $self->fh->print("          <cvParam accession=\"MS:1000565\" name=\"Micromass PKL format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "PerSeptive (.PKS)") {
      $self->fh->print("          <cvParam accession=\"MS:1001245\" name=\"PerSeptive PKS format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "Sciex API III") {
      $self->fh->print("          <cvParam accession=\"MS:1001246\" name=\"Sciex API III format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "Bruker (.XML)") {
      $self->fh->print("          <cvParam accession=\"MS:1001247\" name=\"Bruker XML format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "mzData (.XML)") {
      $self->fh->print("          <cvParam accession=\"MS:1000564\" name=\"PSI mzData format\" cvRef=\"PSI-MS\" />\n");
      $indexIsQueryNumber = 0;
    } elsif ($params->getFORMAT eq "mzML (.mzML)") {
      $self->fh->print("          <cvParam accession=\"MS:1000584\" name=\"mzML format\" cvRef=\"PSI-MS\" />\n");
      $indexIsQueryNumber = 0;
    } elsif ($resfile->isPMF || $resfile->isSQ) {
      $self->fh->print("          <cvParam accession=\"MS:1001369\" name=\"text format\" cvRef=\"PSI-MS\" />\n");
    }
    $self->fh->print("        </FileFormat>\n");
    $self->fh->print("        <SpectrumIDFormat>\n");
    if ($resfile->isPMF || $resfile->isSQ) {
      $self->fh->print("          <cvParam accession=\"MS:1000775\" name=\"single peak list nativeID format\" cvRef=\"PSI-MS\" />\n");
    } elsif ($indexIsQueryNumber) {
      $self->fh->print("          <cvParam accession=\"MS:1001528\" name=\"Mascot query number\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "mzML (.mzML)") {
      $self->fh->print("          <cvParam accession=\"MS:1001530\" name=\"mzML unique identifier\" cvRef=\"PSI-MS\" />\n");
    } elsif ($params->getFORMAT eq "mzData (.XML)") {
      $self->fh->print("          <cvParam accession=\"MS:1000777\" name=\"spectrum identifier nativeID format\" cvRef=\"PSI-MS\" />\n");
    } else {
      $self->fh->print("          <cvParam accession=\"MS:1000774\" name=\"multiple peak list nativeID format\" cvRef=\"PSI-MS\" />\n");
    }
    $self->fh->print("        </SpectrumIDFormat>\n");
    $self->fh->print("      </SpectraData>\n");

    $indexIsQueryNumber[$i] = $indexIsQueryNumber;
  }

  $self->fh->print("    </Inputs>\n");

  return (\@indexIsQueryNumber);
}


sub printSpectrumIdentificationList {
  my $self = shift;
  my ($resfile_params,
      $exported_queryList, $exported_sequences,
      $peptide_evidence_list, $protTaxStr, $protTaxId, $queryAll, $aaHelpers,
      $tolCVparam, $queryRaw, $indexIsQueryNumber,
      $numcomponents, $firstComponentName,) = @_;

  my $isPMF = $self->objResFile->isPMF;

  my ($modsByDelta, $full_unimod_file, $query_is_sig);

  if (not $isPMF) {
    # Site analysis initialisation.
    $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
      $self->objUmodSection, $self->objParams,
    );
    # $self->objUmodSection could be the reduced file embedded in the results file.
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($self->sessionID);
    $full_unimod_file = new msparser::ms_umod_configfile(
      "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
      $cs
    );

    $query_is_sig = $self->msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
      ? sub { Util::Peptides::above_homology($self->objSummary, $self->objSummary->getPeptide($_[0], 1)) }
      : sub { Util::Peptides::above_identity($self->objSummary, $self->objSummary->getPeptide($_[0], 1)) }
      ;
  }

  $self->fh->print('      <SpectrumIdentificationList id="' . $self->_mzid_SpectrumIdentificationList_id(1) . '"'
               . " numSequencesSearched=\"" . $self->objResFile->getNumSeqsAfterTax
               . "\">\n"); # . $componentNames[$c] . "\">\n";

  # Now output all peptide scores matches
  # For PMF, there's just one spectrum, so just one SpectrumIdentificationResult
  if ($isPMF) {
    my ($location) = $self->fileIn->[0] =~ /([^\/\\]*)\.dat$/;
    $location = 'file:///' . notags("$location.mgf");
    my $name = $self->objParams->getDATAURL();
    if ($name eq '') {
        $name = $self->objParams->getFILENAME();
    }
    $name = notags($name);

    $self->fh->print('        <SpectrumIdentificationResult id="' . $self->_mzid_SpectrumIdentificationResult_id(1) . "\" spectrumID=\"file=$location\" spectraData_ref=\"" . $self->_mzid_SpectraData_id(1) . "\" name=\"$name\">\n");
  } else {
    $self->fh->print("        <FragmentationTable>\n");
    $self->fh->print('          <Measure id="' . $self->_mzid_Measure_id('mz') . "\">\n");
    $self->fh->print("            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001225\" name=\"product ion m/z\" unitCvRef=\"PSI-MS\" unitAccession=\"MS:1000040\" unitName=\"m/z\" />\n");
    $self->fh->print("          </Measure>\n");
    $self->fh->print('          <Measure id="' . $self->_mzid_Measure_id('intensity') . "\">\n");
    $self->fh->print("            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001226\" name=\"product ion intensity\" unitCvRef=\"PSI-MS\" unitAccession=\"MS:1000131\" unitName=\"number of counts\" />\n");
    $self->fh->print("          </Measure>\n");
    $self->fh->print('          <Measure id="' . $self->_mzid_Measure_id('error') . "\">\n");
    $self->fh->print("            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001227\" name=\"product ion m/z error\" unitAccession=\"MS:1000040\" unitName=\"m/z\" unitCvRef=\"PSI-MS\" />\n");
    $self->fh->print("          </Measure>\n");
    $self->fh->print("        </FragmentationTable>\n");
  }

  for my $query (1 .. $self->objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    my $objQuery   = new msparser::ms_inputquery($self->objResFile, $query);
    my (undef, $src_q, $src_file_id) = $self->objResFile->getSrcQueryAndFileIdForMultiFile($query);

    # Try scanID first - for an mzML file, this will be the unique id of the spectrum.
    my $scanID = ExportDat::Util::determine_scan_id($$indexIsQueryNumber[$src_file_id], $$resfile_params[$src_file_id], $objQuery, $src_q);

    my $haveOutputSIRelement = 0;
    my $any_ET_or_SL_in_query = 0;
    my $siteAnalysisData;

    if (not $isPMF) {
      if ($self->analyseSiteData and &$query_is_sig($query)) {
        # (When results are merged: $self->objParams is the right thing to pass
        # to site analysis, because it must contain the same tolerances
        # and mods in all input files.)
        $siteAnalysisData = ExportDat::Util::analyse_sites_in_query(
          $self->query_mod_and_dupe_cache, $self->objResFile, $self->objSummary, $self->objParams,
          $query, $modsByDelta, $full_unimod_file,
          $self->objSummary->getMaxRankValue(),
        );
      }

      for my $rank (1 .. $self->objSummary->getMaxRankValue()) {
        my $objPeptide = $self->objSummary->getPeptide($query, $rank);
        next if $objPeptide->getPeptideStr(0) eq '';

        if ($objPeptide->getIsFromErrorTolerant() or $objPeptide->getIsFromLibrary()) {
          $any_ET_or_SL_in_query = 1;
          last;
        }
      }
    }

    my @frag_rules = ();

    if (not $isPMF and $queryRaw) {
      @frag_rules = split(/,/, $self->objResFile->params->getRULES);
      @frag_rules = sort { $a <=> $b } @frag_rules; # ensure sorted ascending
    }

    for my $rank (1 .. $self->objSummary->getMaxRankValue()) {
      my $objPeptide = $self->objSummary->getPeptide($query, $rank);
      my $pepstr     = $objPeptide->getPeptideStr;
      next if $pepstr eq '';

      for my $psmComp ($self->_psmComponents($objPeptide)) {
	my @peptide_evidence = ();

	# The evidence needs to be printed in sorted order to keep the output
	# constant across different runs with the same parameters.
	for my $id (sort keys %{ $$peptide_evidence_list{$query}{$rank}{$psmComp} || {} }) {
	  push @peptide_evidence, "            <PeptideEvidenceRef peptideEvidence_ref=\"$id\" />\n";
	}

	if (not @peptide_evidence) {
	  # It used to be legal in mzIdentML 1.0 to not have peptide evidence
	  # for a SpectrumIdentificationItem. In mzIdentML 1.1, we must omit
	  # the SpectrumIdentificationItem entirely. See bug #10759.
	  next;
	}

	if (!$isPMF && !$haveOutputSIRelement) {
	  $self->fh->print('        <SpectrumIdentificationResult id="' . $self->_mzid_SpectrumIdentificationResult_id($query)
			   . "\" spectrumID=\"" . $scanID . "\" spectraData_ref=\"" . $self->_mzid_SpectraData_id($src_file_id) . "\">\n");
	  $haveOutputSIRelement = 1;
	}

	$self->printSpectrumIdentificationItem($objPeptide, $query, $rank, $psmComp, $isPMF,
					       $any_ET_or_SL_in_query, \@peptide_evidence, $queryRaw, $numcomponents,
					       $aaHelpers, \@frag_rules, $firstComponentName, $siteAnalysisData);
	}#psmComp
      }

    if (!$isPMF && $haveOutputSIRelement) {
      my $qmatch = $self->objSummary->getQmatch($query);

      if ($any_ET_or_SL_in_query) {
        # Individual matches have their own CV parameters for thresholds.
      } else {
        # No ET or SL matches, so the identity and homology thresholds of
        # all the matches in the query are the same.
        my $identityThreshold = $self->objSummary->getPeptideIdentityThreshold($query, 1 / $self->objSummary->getProbabilityThreshold, $msparser::ms_mascotresults::DM_FASTA);
        my $homologyThreshold = $self->objSummary->getHomologyThreshold($query, 1 / $self->objSummary->getProbabilityThreshold, $msparser::ms_mascotresults::HOMTHR_FASTA);
        $self->fh->print("          <cvParam accession=\"MS:1001371\" name=\"Mascot:identity threshold\"  cvRef=\"PSI-MS\" value=\"" . $identityThreshold . "\" />\n");
        if ($homologyThreshold > 0) {
          $self->fh->print("          <cvParam accession=\"MS:1001370\" name=\"Mascot:homology threshold\"  cvRef=\"PSI-MS\" value=\"" . $homologyThreshold . "\" />\n");
        }
      }

      $self->fh->print("          <cvParam accession=\"MS:1001030\" name=\"number of peptide seqs compared to each spectrum\"  cvRef=\"PSI-MS\" value=\"" . $qmatch . "\" />\n");

      if (my @values = ExportDat::Util::collect_indexed_MGF_values($objQuery, 'getScanNumbers')) {
        $self->fh->print("          <cvParam accession=\"MS:1000797\" name=\"peak list scans\"  cvRef=\"PSI-MS\" value=\"" . join(',', @values) . "\" />\n");
      }

      for my $str (ExportDat::Util::collect_indexed_MGF_values($objQuery, 'getRetentionTimes')) {
        # Retention times are a list of numbers or number ranges, e.g.
        #   56.5
        #   100.3-100.5
        #
        # Note: it may be incorrect to print out multiple lower and upper
        # offsets; the PSI-MS working group offers no guidance on this
        # matter.
        if ($str =~ /([^\s-]+)-([^\s-]+)/) {
            my ($start, $finish) = ($1, $2);
            $self->fh->print("          <cvParam accession=\"MS:1000016\" name=\"scan start time\"  cvRef=\"PSI-MS\" value=\"" . notags($start) . "\" unitAccession=\"UO:0000010\" unitName=\"second\" unitCvRef=\"UO\" />\n");
            # It won't hurt to print the range parameters too.
            $self->fh->print("          <cvParam accession=\"MS:1000916\" name=\"retention time window lower offset\"  cvRef=\"PSI-MS\" value=\"" . notags($start) . "\" unitAccession=\"UO:0000010\" unitName=\"second\" unitCvRef=\"UO\" />\n");
            $self->fh->print("          <cvParam accession=\"MS:1000917\" name=\"retention time window upper offset\"  cvRef=\"PSI-MS\" value=\"" . notags($finish) . "\" unitAccession=\"UO:0000010\" unitName=\"second\" unitCvRef=\"UO\" />\n");
        } else {
            my ($start) = $str =~ /([^\s]+)/; # first token only
            $self->fh->print("          <cvParam accession=\"MS:1000016\" name=\"scan start time\"  cvRef=\"PSI-MS\" value=\"" . notags($start) . "\" unitAccession=\"UO:0000010\" unitName=\"second\" unitCvRef=\"UO\" />\n");
        }
      }

      if ($objQuery->getStringTitle(1)) {
        $self->fh->print("          <cvParam accession=\"MS:1000796\" name=\"spectrum title\"  cvRef=\"PSI-MS\" value=\"" . notags($objQuery->getStringTitle(1)) . "\" />\n");
      }

      if ($objQuery->getPepTol && $objQuery->getPepTolUnits) {
        my $cv = $$tolCVparam{$objQuery->getPepTolUnits};
        if ($cv) {
          my $val = $$cv{'fmt'}->($objQuery->getPepTol);
          $self->fh->print("        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
          $self->fh->print("        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $objQuery->getPepTol . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n");
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getSeq($j)) {
          $self->fh->print("            <userParam name=\"SEQ_$j\" value=\"" . $objQuery->getSeq($j) . "\" />\n");
        } else {
          last;
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getComp($j)) {
          $self->fh->print("            <userParam name=\"COMP_$j\" value=\"" . $objQuery->getComp($j) . "\" />\n");
        } else {
          last;
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getTag($j)) {
          $objQuery->getTag($j) =~ /([et]),(.*)/i;
          if (lc($1) eq "e") {
            $self->fh->print("            <userParam name=\"ETAG_$j\" value=\"" . $2 . "\" />\n");
          } else {
            $self->fh->print("            <userParam name=\"TAG_$j\" value=\"" . $2 . "\" />\n");
          }
        } else {
          last;
        }
      }

      if ($objQuery->getIT_MODS(1)) {
      # Need reference to query level mods, otherwise cannot fully understand results.
      # Structured information for mods in matched peptides will be found in
      # /mzIdentML/SequenceCollection/Peptide/Modification
        $self->fh->print("            <userParam name=\"IT_MODS\" value=\"" . notags($objQuery->getIT_MODS(1)) . "\" />\n");
      }
      if ($objQuery->getINSTRUMENT(1)) {
        $self->fh->print("            <userParam name=\"INSTRUMENT\" value=\"" . notags($objQuery->getINSTRUMENT(1)) . "\" />\n");
      }
      if ($objQuery->getRULES()) {
        $self->fh->print("            <userParam name=\"RULES\" value=\"" . $objQuery->getRULES() . "\" />\n");
      }
      $self->fh->print("        </SpectrumIdentificationResult>\n");
    }
  }
  if ($isPMF) {
    $self->fh->print("        </SpectrumIdentificationResult>\n");
  }
  $self->fh->print("      </SpectrumIdentificationList>\n");
}

sub printSpectrumIdentificationItem {
  my $self = shift;

  my ($objPeptide, $query, $rank, $psmComp,
      $isPMF, $any_ET_or_SL_in_query, $peptide_evidence, $queryRaw,
      $numcomponents, $aaHelpers, $frag_rules, $firstComponentName,
      $siteAnalysisData) = @_;

  my ($pepCharge, $charge);

  if ($objPeptide->getCharge > 0) {
    $charge =  $self->elementalMasses->{'hydrogen'} - $self->elementalMasses->{'electron'};
    $pepCharge = $objPeptide->getCharge();
  } elsif ($objPeptide->getCharge == 0) {
    $charge = 0;
    $pepCharge = 1;
  } else {
    $charge = - $self->elementalMasses->{'hydrogen'} + $self->elementalMasses->{'electron'};
    $pepCharge = $objPeptide->getCharge();
  }

  my $calcMz = ($objPeptide->getMrCalc / abs($pepCharge)) + $charge;

  $self->fh->print('          <SpectrumIdentificationItem id="' . $self->_mzid_SpectrumIdentificationItem_id($query, $rank, $psmComp)
		   . '"  calculatedMassToCharge="' . $calcMz
		   . '" chargeState="' . $objPeptide->getCharge
		   . '" experimentalMassToCharge="' . $self->objResFile->getObservedMass($query)
		   . '" peptide_ref="' . $self->_mzid_Peptide_id($query, $rank, $psmComp)
    );

  my $thisThreshold;
  my @threshold_and_expect_cv = ();

  if (!$isPMF) {
    my $dbType = $objPeptide->getIsFromLibrary()
      ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
      : $msparser::ms_mascotresults::DM_FASTA
      ;
    my $identityThreshold = $self->objSummary->getPeptideIdentityThreshold($query, 1 / $self->objSummary->getProbabilityThreshold, $dbType);
    my $homologyThreshold = $self->objSummary->getHomologyThreshold($query, 1 / $self->objSummary->getProbabilityThreshold, $rank);
    if ($self->use_homology and $homologyThreshold > 0) {
      $thisThreshold = $homologyThreshold;
    } else {
      $thisThreshold = $identityThreshold;
    }

    my $expect = $self->objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $query, $objPeptide->getRank());

    if ($expect > 0) {
      @threshold_and_expect_cv = (
	{ id => "MS:1001172", name => "Mascot:expectation value",
	  value => round_evenly($expect, 2, 'g')
	},
	);
    }

    if ($any_ET_or_SL_in_query) {
      # Bug 12187: For an SL match, the threshold is actually a library
      # score between 0 and 1000. However, since we compute this
      # threshold ourselves (it doesn't come from MSPepSearch), we'll
      # report it as a Mascot threshold.
      push @threshold_and_expect_cv, (
	{ id => "MS:1001371", name => "Mascot:identity threshold",
	  value => $identityThreshold },
	$homologyThreshold > 0
	? { id => "MS:1001370", name => "Mascot:homology threshold",
	    value => $homologyThreshold }
	: ()
	,
	);
    }
  }

  if ($isPMF) {
    $self->fh->print("\" rank=\"0");
    $self->fh->print("\" passThreshold=\"true");     # according to spec, if no threshold then set true(!)
  } else {
    $self->fh->print("\" rank=\"" . $rank);

    if (defined $thisThreshold) {
      if ($objPeptide->getIonsScore > $thisThreshold) {
	$self->fh->print("\" passThreshold=\"true");
      } else {
	$self->fh->print("\" passThreshold=\"false");
      }
    } else {
      $self->fh->print("\" passThreshold=\"true");
    }
  }

  if ($objPeptide->getComponentStr) {
    $self->fh->print('" massTable_ref="' . $self->_mzid_MassTable_id($objPeptide->getComponentStr));
    $self->fh->print('" sample_ref="' . $self->_mzid_Sample_id($objPeptide->getComponentStr));
  }

  $self->fh->print("\">\n");

  $self->printSpectrumIdentificationItemBody($objPeptide, $query, $rank, $psmComp, $peptide_evidence,
					     $isPMF, $queryRaw, $numcomponents, $aaHelpers,
					     $firstComponentName, $frag_rules, \@threshold_and_expect_cv, $siteAnalysisData);


  $self->fh->print("          </SpectrumIdentificationItem>\n");
}

sub printSpectrumIdentificationItemBody {
  my $self = shift;
  my ($objPeptide, $query, $rank, $psmComp, $peptide_evidence,
      $isPMF, $queryRaw, $numcomponents, $aaHelpers,
      $firstComponentName, $frag_rules, $threshold_and_expect_cv, $siteAnalysisData) = @_;

  $self->fh->print(@$peptide_evidence);

  if ($isPMF) {
    # Nothing required or desired
  } else {
    if ($queryRaw) {
      if ($numcomponents > 1 && $objPeptide->getComponentStr) {
	$self->printFragmentation($query, $rank, $objPeptide, $$aaHelpers{$objPeptide->getComponentStr}, $frag_rules);
      } elsif ($numcomponents > 1) {
	$self->printFragmentation($query, $rank, $objPeptide, $$aaHelpers{$firstComponentName}, $frag_rules);
      } else {
	$self->printFragmentation($query, $rank, $objPeptide, $$aaHelpers{"1"}, $frag_rules);
      }
    }

    if ($objPeptide->getIsFromLibrary()) {
      $self->fh->print("            <cvParam accession=\"MS:1002754\" name=\"MSPepSearch:score\" cvRef=\"PSI-MS\" value=\"");
    } else {
      $self->fh->print("            <cvParam accession=\"MS:1001171\" name=\"Mascot:score\" cvRef=\"PSI-MS\" value=\"");
    }
    printf "%.2f\" />\n", $objPeptide->getIonsScore;

    if ($self->percolate == 1) {
      my ($pep, $q) = $objPeptide->getPercolatorScores();
      $self->fh->print("            <cvParam accession=\"MS:1001491\" name=\"percolator:Q value\" cvRef=\"PSI-MS\" value=\"$q\" />\n");
      $self->fh->print("            <cvParam accession=\"MS:1001493\" name=\"percolator:PEP\" cvRef=\"PSI-MS\" value=\"$pep\" />\n");
    }

    for (@$threshold_and_expect_cv) {
      $self->fh->print("            <cvParam accession=\"$$_{'id'}\" name=\"$$_{'name'}\" cvRef=\"PSI-MS\" value=\"$$_{'value'}\" />\n");
    }

    if ($self->objSummary->isPeptideUnique($query, $rank)) {
      $self->fh->print("            <cvParam accession=\"MS:1001363\" name=\"peptide unique to one protein\" cvRef=\"PSI-MS\" />\n");
    } else {
      $self->fh->print("            <cvParam accession=\"MS:1001175\" name=\"peptide shared in multiple proteins\" cvRef=\"PSI-MS\" />\n");
    }

    if ($siteAnalysisData) {
      if (exists $$siteAnalysisData{$rank}) {
	$self->fh->print("            <cvParam accession=\"MS:1002012\" name=\"Mascot:PTM site assignment confidence\" cvRef=\"PSI-MS\" value=\"" . round_evenly(100 * $$siteAnalysisData{$rank}{'confidence'}, 2) . "\" unitAccession=\"UO:0000187\" unitName=\"percent\" unitCvRef=\"UO\" />\n");
      }
    }
  }

}

BEGIN {
  my %fragCVparam = (
    $msparser::ms_fragmentationrules::FRAG_IMMONIUM => {
      accession => "MS:1001239", name => "frag: immonium ion"
    },
    $msparser::ms_fragmentationrules::FRAG_A_SERIES => {
      accession => "MS:1001229", name => "frag: a ion"
    },
    $msparser::ms_fragmentationrules::FRAG_A_MINUS_NH3 => {
      accession => "MS:1001235", name => "frag: a ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_A_MINUS_H2O => {
      accession => "MS:1001234", name => "frag: a ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_B_SERIES => {
      accession => "MS:1001224", name => "frag: b ion"
    },
    $msparser::ms_fragmentationrules::FRAG_B_MINUS_NH3 => {
      accession => "MS:1001232", name => "frag: b ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_B_MINUS_H2O => {
      accession => "MS:1001222", name => "frag: b ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_C_SERIES => {
      accession => "MS:1001231", name => "frag: c ion"
    },
    $msparser::ms_fragmentationrules::FRAG_X_SERIES => {
      accession => "MS:1001228", name => "frag: x ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_SERIES => {
      accession => "MS:1001220", name => "frag: y ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_MINUS_NH3 => {
      accession => "MS:1001233", name => "frag: y ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_MINUS_H2O => {
      accession => "MS:1001223", name => "frag: y ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_SERIES => {
      accession => "MS:1001230", name => "frag: z ion"
    },
    $msparser::ms_fragmentationrules::FRAG_INTERNAL_YB => {
      accession => "MS:1001365", name => "frag: internal yb ion"
    },
    $msparser::ms_fragmentationrules::FRAG_INTERNAL_YA => {
      accession => "MS:1001366", name => "frag: internal ya ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_PLUS_1 => {
      accession => "MS:1001367", name => "frag: z+1 ion"
    },
    $msparser::ms_fragmentationrules::FRAG_D_SERIES => {
      accession => "MS:1001236", name => "frag: d ion"
    },
    $msparser::ms_fragmentationrules::FRAG_V_SERIES  => {
      accession => "MS:1001237", name => "frag: v ion"
    },
    $msparser::ms_fragmentationrules::FRAG_W_SERIES => {
      accession => "MS:1001238", name => "frag: w ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_PLUS_2  => {
      accession => "MS:1001368", name => "frag: z+2 ion"
    },
  );

  sub printFragmentation {
    my $self = shift;
    my ($query, $rank, $peptide, $objAahelper, $rules) = @_;

    my $any_printed = 0;

    my ($singleCharge, $doubleCharge) = (0, 0);

    foreach my $rule (@$rules) {
      if ($rule == 1) {
        $singleCharge = 1;
        next;
      }

      if ($rule == 2 || $rule == 3) {
        if (($rule == 2 && $peptide->getCharge > 1)
          || ($rule == 3 && $peptide->getCharge > 2)) {
          $doubleCharge = 1;
        }
        next;
      }

      for my $charge (1 .. 2) {
        if ($charge == 1 && $singleCharge ||
            $charge == 2 && $doubleCharge
                         && $rule != $msparser::ms_fragmentationrules::FRAG_IMMONIUM
                         && $rule != $msparser::ms_fragmentationrules::FRAG_INTERNAL_YB
                         && $rule != $msparser::ms_fragmentationrules::FRAG_INTERNAL_YA) {
          my $cv = $fragCVparam{$rule};
          next if not $cv;

          my $fragments  = new msparser::ms_fragmentvector;
          my $err        = new msparser::ms_errs;
          if (!$objAahelper->calcFragmentsEx($peptide,
                                      $rule,
                                      $charge,
                                      0,  # min fragment mass to return
                                      $peptide->getMrCalc, # max fragment mass to return
                                      $self->objResFile->params->getMassType, # monoisotopic / average
                                      $fragments,
                                      $err)) {
            die "Error creating fragments";
          }

          $fragments->addExperimentalData($self->objResFile, $query);

          my (@idx, @exptMasses, @exptIntensities, @exptErrors);

          for my $i (0 .. $fragments->getNumberOfFragments()-1) {
            my $fragment = $fragments->getFragmentByNumber($i);

            if ($fragment->getMatchedIonMass > 0) {
              if ($fragment->isInternal) {
                push @idx, $fragment->getStart, $fragment->getEnd;
              } else {
                push @idx, $fragment->getColumn;
              }
              push @exptMasses, $fragment->getMatchedIonMass();
              push @exptIntensities, $fragment->getMatchedIonIntensity();
              push @exptErrors, round_evenly(($fragment->getMatchedIonMass - $fragment->getMass), $self->massDP);
            }
          }

          if (@idx) {
            if (!$any_printed) {
              $self->fh->print("            <Fragmentation>\n");
              $any_printed = 1;
            }
            $self->fh->print("              <IonType index=\"" . join(' ', @idx) . "\" charge=\"" . $charge . "\">\n");
            $self->fh->print("                <FragmentArray values=\"" . join(' ', @exptMasses)      . '" measure_ref="' . $self->_mzid_Measure_id('mz') . "\" />\n");
            $self->fh->print("                <FragmentArray values=\"" . join(' ', @exptIntensities) . '" measure_ref="' . $self->_mzid_Measure_id('intensity') . "\" />\n");
            $self->fh->print("                <FragmentArray values=\"" . join(' ', @exptErrors)      . '" measure_ref="' . $self->_mzid_Measure_id('error') . "\" />\n");
            $self->fh->print("                <cvParam cvRef=\"PSI-MS\" accession=\"" . $$cv{'accession'} . "\" name=\"" . $$cv{'name'} . "\" />\n");
            $self->fh->print("              </IonType>\n");
          }
        }
      }
    }

    if ($any_printed) {
      $self->fh->print("            </Fragmentation>\n");
    }

    return 1;
  }
}

sub printProteinDetectionList {
    my $self = shift;
    my ($peptide_evidence_list,
      $protLen,
	$sameset_ordering) = @_;

    $self->fh->print('      <ProteinDetectionList id="' . $self->_mzid_ProteinDetectionList_id(1) . "\">\n");
    $self->printProteinDetectionListBody($peptide_evidence_list, $protLen, $sameset_ordering);
    $self->fh->print("      </ProteinDetectionList>\n");
}

sub printProteinDetectionListBody {
    my $self = shift;
    my ($peptide_evidence_list,
      $protLen,
      $sameset_ordering) = @_;

  for my $i (1 .. $self->objSummary->getNumberOfHits()) {
    $self->printProteinAmbiguityGroup($i, $peptide_evidence_list, $protLen, $sameset_ordering);
  }
}

sub printProteinAmbiguityGroup {
    my $self = shift;
       my ($i, $peptide_evidence_list,
	   $protLen,
	   $sameset_ordering) = @_;

       $self->fh->print('        <ProteinAmbiguityGroup id="' . $self->_mzid_ProteinAmbiguityGroup_id($i) . "\" >\n");
       $self->printProteinAmbiguityGroupBody($i, $peptide_evidence_list,
					     $protLen, $sameset_ordering);
       $self->fh->print("        </ProteinAmbiguityGroup>\n");
}
sub printProteinAmbiguityGroupBody {
    my $self = shift;
    my ($i, $peptide_evidence_list,
	$protLen,
	$sameset_ordering) = @_;

    my $isPMF = $self->objResFile->isPMF;
    my $objProtein = $self->objSummary->getHit($i);
    return if $isPMF and $objProtein->isPMFMixture();

    my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, $self->ShowSameSets,
					      $$sameset_ordering{$i}
    );

    my %cv = @sameset > 1 ? %{$self->cv_anchor_protein} : ();
    $self->printProteinDetectionHypothesis(
      1, $peptide_evidence_list,
      $sameset[0], $protLen, { %cv }
    );

    my $cv_anchor = $self->_mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);

    if ($self->group_family) {

    # Other proteins that match the same set of peptides
      if ($self->ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          $self->printProteinDetectionHypothesis(
            0, $peptide_evidence_list,
            $_, $protLen,
            { %{$self->cv_spectrum_sameset}, val => $cv_anchor }
          );
        }
      }

    # Other family members
      my $k = 1;
      while (my $objProtein = $self->objSummary->getNextFamilyProtein($i, $k++)) {
        my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, $self->ShowSameSets,
          $$sameset_ordering{$i},
	  );
	my %cv = @sameset > 1 ? %{$self->cv_anchor_protein} : ();
        $self->printProteinDetectionHypothesis(
            $k+1, $peptide_evidence_list,
            $sameset[0], $protLen,
            (%cv, { %{$self->cv_family_member}, val => $cv_anchor })
        );

      # And their same-sets
        if ($self->ShowSameSets) {
          my $cv_fm = $self->_mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);
          for (@sameset[1 .. $#sameset]) {
            $self->printProteinDetectionHypothesis(
              0, $peptide_evidence_list,
              $_, $protLen,
              { %{$self->cv_spectrum_sameset}, val => $cv_fm }
            );
          }
        }
      }

    # sub-sets and intersections for the entire family
      if ($self->ShowSubSets > 0) {
        my $k = 1;
        while (my $objProtein = $self->objSummary->getNextSubsetProtein($i, $k++, 1)) {
          # We need the full sameset to filter out proteins from
          # getSimilarProteins().
          my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, 1,
            $$sameset_ordering{$i},
          );
          my %is_sameset = map { key_acc($_) => 1 } @sameset;

          my @supersets = ExportDat::Util::get_supersets($sameset[0], \%is_sameset);
          my %cv = @supersets > 1 ? %{$self->cv_spectrum_intersect} : %{$self->cv_spectrum_subset};

          $self->printProteinDetectionHypothesis(
            0, $peptide_evidence_list,
            $sameset[0], $protLen,
            map {
              { %cv, val => $self->_mzid_DBSeq_id($_->{'db'}, $_->{'acc'}) }
            } @supersets
          );

        # And their same-sets
          if ($self->ShowSameSets) {
            my $cv_subset = $self->_mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);
            for (@sameset[1 .. $#sameset]) {
              $self->printProteinDetectionHypothesis(
               0, $peptide_evidence_list,
               $_, $protLen,
               { %{$self->cv_spectrum_sameset}, val => $cv_subset })
            }
          }
        }

      }

    } else {
      if ($self->ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          $self->printProteinDetectionHypothesis(
            0, $peptide_evidence_list,
            $_, $protLen,
            { %{$self->cv_sequence_sameset}, val => $cv_anchor })
        }
      }

      my $j = 1;
      while (my $objProtein = $self->objSummary->getNextSubsetProtein($i, $j++)) {
        # We need the full sameset to filter out proteins from
        # getSimilarProteins().
        my @sameset = common_subs::compileSameset($self->objSummary, $objProtein, 1);
        my %is_sameset = map { key_acc($_) => 1 } @sameset;

        my @supersets = ExportDat::Util::get_supersets($objProtein, \%is_sameset);
        my %cv = @supersets > 1 ? %{$self->cv_sequence_intersect} : %{$self->cv_sequence_subset};

        $self->printProteinDetectionHypothesis(
          0, $peptide_evidence_list,
          $objProtein, $protLen,
          map {
            { %cv, val => $self->_mzid_DBSeq_id($_->{'db'}, $_->{'acc'}) }
          } @supersets
        );
      }

    }


}

sub printProteinDetectionHypothesis {
  my $self = shift;
  my ($famMemberNum, $peptide_evidence_list,
      $objProtein, $protLen, @cv_relationships) = @_;

  my $passThreshold = "true";
  if ($self->objResFile->isPMF && $objProtein->getScore < $self->objSummary->getProteinThreshold(1 / $self->objSummary->getProbabilityThreshold)) {
    $passThreshold = "false"
  }
  my $preferred_acc = $self->proteinData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());

  $self->fh->print('          <ProteinDetectionHypothesis id="' . $self->_mzid_ProteinDetectionHypothesis_id($preferred_acc, $objProtein->getFrame()) .
		   '" dBSequence_ref="' . $self->_mzid_DBSeq_id($objProtein->getDB, $preferred_acc) . "\"  passThreshold=\"" . $passThreshold . "\">\n");
  $self->printProteinDetectionHypothesisBody($famMemberNum, $preferred_acc, $passThreshold, $objProtein, $peptide_evidence_list, $protLen, @cv_relationships);
  $self->fh->print("          </ProteinDetectionHypothesis>\n");
}

sub printProteinDetectionHypothesisBody {
  my $self = shift;
  my ($famMemberNum, $preferred_acc, $passThreshold, $objProtein, $peptide_evidence_list,
       $protLen, @cv_relationships) = @_;

  for my $pepnum (1 .. $objProtein->getNumPeptides()) {
    my $query     = $objProtein->getPeptideQuery($pepnum);
    my $rank      = $objProtein->getPeptideP($pepnum);
    my $startpos  = $objProtein->getPeptideStart($pepnum);
    my $endpos    = $objProtein->getPeptideEnd($pepnum);
    my $resbefore = $objProtein->getPeptideResidueBefore($pepnum);
    my $resafter  = $objProtein->getPeptideResidueAfter($pepnum);
    my $objPeptide = $self->objSummary->getPeptide($query, $rank);
    for my $psmComp($self->_psmComponents($objPeptide)) {
      my $id = $self->_mzid_PeptideEvidence_id($query, $rank, $psmComp, $objProtein->getDB(), $preferred_acc, $objProtein->getPeptideFrame($pepnum), $startpos, $endpos);
      next unless $$peptide_evidence_list{$query}{$rank}{$psmComp}{$id};

      $self->fh->print("            <PeptideHypothesis peptideEvidence_ref=\"" . $id . "\">\n");
      $self->fh->print("              <SpectrumIdentificationItemRef spectrumIdentificationItem_ref=\"" . $self->_mzid_SpectrumIdentificationItem_id($query, $rank, $psmComp) . "\" />\n");
      $self->fh->print("            </PeptideHypothesis>\n");
    }
  }

  $self->fh->print("            <cvParam accession=\"MS:1001171\" name=\"Mascot:score\" cvRef=\"PSI-MS\" value=\"" . $objProtein->getScore() . "\" />\n");

  if ($self->reportType eq 'concise') {
    my $expect = $self->objSummary->getProteinExpectationValue($objProtein->getScore());
    $self->fh->print("            <cvParam accession=\"MS:1001172\" name=\"Mascot:expectation value\" cvRef=\"PSI-MS\" value=\"" .
      round_evenly($expect, 3, 'g') . "\" />\n");
  }

  for (@cv_relationships) {
    my ($name, $acc, $val) = map { notags($_) } @$_{qw(name acc val)};
    next unless defined($name) and defined($acc);
    $val = '' unless defined $val;
    $self->fh->print("            <cvParam accession=\"$acc\" name=\"$name\" cvRef=\"PSI-MS\" value=\"$val\" />\n");
  }

  if ($protLen) {
    my $coverage = 0;
    my ($length) = $self->proteinData->mustGetProteinLen(
        $objProtein,
        1, # always allow fetching length
        1, # always allow fetching mass
    );
    if ($length > 0) {
      $coverage = $objProtein->getCoverage() * 100 / $length;
    }
    $coverage = int $coverage;

    $self->fh->print("            <cvParam accession=\"MS:1001093\" name=\"sequence coverage\" cvRef=\"PSI-MS\" value=\"" . $coverage . "\" />\n");
  }

  if ($self->reportType eq 'concise') {
    my $numUnmatched = $self->objResFile->getNumQueries - $objProtein->getNumPeptides;
    $self->fh->print("            <cvParam accession=\"MS:1001362\" name=\"number of unmatched peaks\" cvRef=\"PSI-MS\" value=\"" . $numUnmatched . "\" />\n");
  } else {
    $self->fh->print("            <cvParam accession=\"MS:1001097\" name=\"distinct peptide sequences\" cvRef=\"PSI-MS\" value=\"" . $objProtein->getNumDistinctPeptides(0, $msparser::ms_protein::DPF_SEQUENCE) . "\" />\n");
  }

}

sub printBibliographicReference {
  my $self = shift;
  print {$self->{fh}} <<EOF;
  <BibliographicReference 
    id="10.1002/(SICI)1522-2683(19991201)20:18&lt;3551::AID-ELPS3551&gt;3.0.CO;2-2"
    authors="David N. Perkins, Darryl J. C. Pappin, David M. Creasy, John S. Cottrell"
    editor="" 
    title="Probability-based protein identification by searching sequence databases using mass spectrometry data"
    name="Probability-based protein identification by searching sequence databases using mass spectrometry data"
    publication="Electrophoresis" 
    volume="20" 
    issue="18" 
    pages="3551-3567" 
    publisher="Wiley VCH"
    year="1999"
    />
EOF
}

sub _tolerance_cv_param_map {
  return {
    Da => {
      unitAccession => "UO:0000221", unitName => "dalton",
      fmt => sub { $_[0] },
    },
    mmu => {
      unitAccession => "UO:0000221", unitName => "dalton",
      fmt => sub { $_[0]/1000 },
    },
    '%' => {
      unitAccession => "UO:0000187", unitName => "percent",
      fmt => sub { $_[0] },
    },
    ppm => {
      unitAccession => "UO:0000169", unitName => "parts per million",
      fmt => sub { $_[0] },
    },
  }
}

}# --MzIdentML

package ExportDat::MzIdentML_1_1 {

    use Moose;
    with 'ExportDat::MzIdentML';

    around 'print' => sub {
      my $orig = shift;
      my $self = shift;
      my $isCGI = shift;
      my $URI = shift;

      $self->fh->print("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
      $self->fh->print("<MzIdentML id=\"report\" version=\"1.1.0\"\n");
      $self->fh->print("             xsi:schemaLocation=\"http://psidev.info/psi/pi/mzIdentML/1.1 ");
      if ($isCGI) {
	$self->fh->print($URI);
      } else {
	$self->fh->print("http://www.matrixscience.com/");
      }
      $self->fh->print("xmlns/schema/mzIdentML/mzIdentML1.1.0.xsd\"\n");
      $self->fh->print("             xmlns=\"http://psidev.info/psi/pi/mzIdentML/1.1\"\n");
      $self->fh->print("             xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
      $self->fh->printf("             creationDate=\"%s\">\n", common_subs::_fmtdate_iso8601(time));

      $self->$orig(@_);

      $self->fh->printf("</MzIdentML>\n");
    };

    around 'printCVList' => sub {
      my $orig = shift;
      my $self = shift;

      $self->fh->print(qq(  <cvList>\n));
      $self->fh->print(qq(    <cv id="PSI-MS" fullName="Proteomics Standards Initiative Mass Spectrometry Vocabularies"   uri="http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo" version="4.1.30"></cv>\n));
      $self->fh->print(qq(    <cv id="UNIMOD" fullName="UNIMOD"        uri="http://www.unimod.org/obo/unimod.obo"></cv>\n));
      $self->fh->print(qq(    <cv id="UO"     fullName="UNIT-ONTOLOGY" uri="http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/phenotype/unit.obo"></cv>\n));
      $self->fh->print(qq(  </cvList>\n));
    };

}# --MzIdentML_1_1


package ExportDat::MzIdentML_1_2 {

  use Moose;
  use Util::Text qw(notags begins_with);
  with 'ExportDat::MzIdentML';

  # A unique identifier for each intact link
  # (See Fig.3 in specification)
  has 'xlIndex' =>(is => 'rw', isa => 'Int', default => 0);

  sub BUILD {
    my $self = shift;
    #Don't export site analysis data in 1.2
    $self->analyseSiteData(0);
  }

  around 'print' => sub {
    my $orig = shift;
    my $self = shift;
    my $isCGI = shift;
    my $URI = shift;

    $self->fh->print("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    $self->fh->print("<MzIdentML id=\"report\" version=\"1.2.0\"\n");
    $self->fh->print("             xsi:schemaLocation=\"http://psidev.info/psi/pi/mzIdentML/1.2 ");
    if ($isCGI) {
      $self->fh->print($URI);
    } else {
      $self->fh->print("http://www.matrixscience.com/");
    }
    $self->fh->print("xmlns/schema/mzIdentML/mzIdentML1.2.0.xsd\"\n");
    $self->fh->print("             xmlns=\"http://psidev.info/psi/pi/mzIdentML/1.2\"\n");
    $self->fh->print("             xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
    $self->fh->printf("             creationDate=\"%s\">\n", common_subs::_fmtdate_iso8601(time));

    $self->$orig(@_);

    $self->fh->printf("</MzIdentML>\n");
  };

  around 'printCVList' => sub {
    my $orig = shift;
    my $self = shift;

    $self->fh->print(qq(  <cvList>\n));
    $self->fh->print(qq(    <cv id="PSI-MS" fullName="Proteomics Standards Initiative Mass Spectrometry Vocabularies"   uri="http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo" version="4.1.69"></cv>\n));
    $self->fh->print(qq(    <cv id="UNIMOD" fullName="UNIMOD"        uri="http://www.unimod.org/obo/unimod.obo"></cv>\n));
    $self->fh->printf(qq(    <cv id="UO"     fullName="UNIT-ONTOLOGY" uri="%s"></cv>\n),common_subs::htmlEscape("http://bioportal.bioontology.org/ontologies/UO?p=classes&conceptid=root"));
    $self->fh->print(qq(  </cvList>\n));
  };

  around 'printSpectrumIdentificationProtocolBody' => sub {
    my $orig = shift;
    my $self = shift;

    $self->$orig(@_);
  };

  # around 'printSearchTypeBody' => sub {
  #   my $orig = shift;
  #   my $self = shift;

  #   $self->$orig(@_);
  #   # Table 1 and Section 5.2.2
  #   $self->fh->print("        <cvParam accession=\"MS:1001010\" name=\"de novo search\" cvRef=\"PSI-MS\"/>\n");
  # };

  around 'printAdditionalSearchParamsBody' => sub {
    my $orig = shift;
    my $self = shift;

    my $nTerms = 0;

    $self->$orig(@_);
    # Table 1 and Section 5.2.7
    #$self->fh->print("        <cvParam accession=\"MS:1002490\" name=\"peptide-level scoring\" cvRef=\"PSI-MS\"/>\n");
    #Table 1 and Section 5.2.8 N/A
    #$self->fh->print("        <cvParam accession=\"MS:1002491\" name=\"modification localization scoring\" cvRef=\"PSI-MS\"/>\n");

    # Table 1 and Section 5.2.4 N/A? http://www.ontobee.org/ontology/MS?iri=http://purl.obolibrary.org/obo/MS_1002492
    #$self->fh->print("        <cvParam accession=\"MS:1002492\" name=\"consensus scoring\" cvRef=\"PSI-MS\"/>\n");
    # Table 1 and Section 5.2.5 N/A?
    #$self->fh->print("        <cvParam accession=\"MS:1002493\" name=\"sample pre-fractionation\" cvRef=\"PSI-MS\"/>\n");

    my $anyXlPepSumMatch = $self->objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_CROSSLINK_PEPTIDES);
    if($anyXlPepSumMatch) {
      # Table 1 and Section 5.2.9
      $self->fh->print("        <cvParam accession=\"MS:1002494\" name=\"cross-linking search\" cvRef=\"PSI-MS\"/>\n");
      $nTerms++;
    }

    # To print MS:1002635, CV terms in figure 5 in the specification are needed
    # for my $dbIdx (1..$self->objParams->getNumberOfDatabases) {
    #   if ($self->objResFile->getDatabaseType($dbIdx) == $msparser::NA) {
    # 	# Table 1 and Section 5.2.10
    # 	$self->fh->print("        <cvParam accession=\"MS:1002635\" name=\"proteogenomics search\" cvRef=\"PSI-MS\"/>\n");
    # 	last;
    #   }
    # }

    if ($nTerms == 0) {
      # Table 1. N/A
      $self->fh->print("        <cvParam accession=\"MS:1002495\" name=\"no special processing\" cvRef=\"PSI-MS\"/>\n");
    }
  };

  sub printSequenceCollectionPeptide {
    my $self = shift;
     my ($objPeptide,
	 $query, $rank, $psmComp,
	 $fixedMods, $umod_mod_cache) = @_;

    #Initialise the mod types for each site in current peptide. This is used later in printModificationCvTerms
    my ($modTypeComp, $modDeltaComp, undef, $modIndicesComp) = ExportDat::Util::make_mods_data($objPeptide, $self->objResFile, $self->objModFile, $self->mass_type);
    $self->modTypeComp($modTypeComp);
    $self->modDeltaComp($modDeltaComp);
    $self->modIndicesComp($modIndicesComp);

    if ($psmComp == $msparser::ms_peptide::PSM_CROSSLINK_ALPHA) {
      $self->xlIndex($self->xlIndex+1);
    }
    $self->fh->print('    <Peptide id="' . $self->_mzid_Peptide_id($query, $rank, $psmComp) . "\">\n");
    $self->printSequenceCollectionPeptideBody($objPeptide, $query, $rank, $psmComp, $fixedMods, $umod_mod_cache);
    $self->fh->print("    </Peptide>\n");
  }

  around 'printModificationCvTerms' => sub {
    my $orig = shift;
    my $self = shift;

    my ($umod_title, $mod2, $cv_acc, $cv_name, $cv_ref, $nl_value, $psmComponent, $p) = @_;

    # If it's a varmod or fixed mod.
    if ($mod2) {
      $self->$orig(@_);
      return;
    }
    elsif (defined($self->objUmodXlSection)) {
      $mod2 = $self->objUmodXlSection->getModificationByName($umod_title);
      if($mod2) {
	if ($mod2->haveRecordID()) {
	  $cv_acc = 'UNIMOD:' . notags($mod2->getRecordID);
	  $cv_name = notags($umod_title);
	  $cv_ref = 'UNIMOD';
	}
	else {
	  # It's a locally defined mod
	  $cv_acc = 'MS:1001460';
	  $cv_name = 'unknown modification';
	  $cv_ref = 'PSI-MS';
	}

	my $modType = $self->modTypeComp->{$psmComponent}[$p];
	if ($psmComponent != $msparser::ms_peptide::PSM_COMPLETE and $modType eq 'I') { # It's an intact link
	  my $idx = $self->xlIndex;
	  # In Mascot, alpha peptide is always the donor
	  # Rules to choose the donor peptide (Figure 3 in specification) are already implemented in isPeptideAlpha() in seq_iter_array.cpp
	  if ($psmComponent == $msparser::ms_peptide::PSM_CROSSLINK_ALPHA) {
	    if (not $mod2->haveRecordID()) {
	      #TODO: userParams are not allowed here by the schema
	      #my $value = notags($umod_title);
	      #$self->fh->print(qq(        <userParam name="custom linker" value="$value" />\n));
	    }
	    $self->fh->print(qq(        <cvParam accession="$cv_acc" name="$cv_name" cvRef="$cv_ref" />\n));
	    $self->fh->print(qq(        <cvParam accession="MS:1002509" cvRef="PSI-MS" name="cross-link donor" value="$idx"/>\n));
	  }
	  else {
	    $self->fh->print(qq(        <cvParam accession="MS:1002510" cvRef="PSI-MS" name="cross-link receiver" value="$idx"/>\n));
	  }
	  return;
	}
	elsif ($modType eq 'M') { #It's a monolink
	  my $modNum = $self->modIndicesComp->{$psmComponent}[$p][0];
	  my $num = $self->modIndicesComp->{$psmComponent}[$p][1];
	  my $monoLinkMod = $self->objResFile->getMonoLinkModification($modNum, $num);
	  if ($monoLinkMod) {
	    #TODO: userParams are not allowed here by the schema
	    #my $value = $monoLinkMod->getCode();
	    #$self->fh->print(qq(        <userParam name="monolink" value="$value" />\n));
	  }
	}
	elsif (begins_with($modType, 'L')) { # It's a looplink
	  # TODO: Schema doesn't allow userParams here.
	  # my $otherEndIdx = substr($modType, 1);
	  # $self->fh->print(qq(        <userParam name="looplink other end" value="$otherEndIdx"/>\n));
	}
      }
    }
    $self->$orig($umod_title, $mod2, $cv_acc, $cv_name, $cv_ref, $nl_value, $psmComponent, $p);
  };

  around 'printSequenceCollectionPeptideBody' => sub {
    my $orig = shift;
    my $self = shift;
    my $objPeptide = $_[0];
    my $query = $_[1];
    my $rank = $_[2];

    $self->$orig(@_);
  };

  around 'printProteinDetectionListBody' => sub {
    my $orig = shift;
    my $self = shift;
    my $i = $_[0];

    $self->$orig(@_);

    #TODO: PAGs passing threshold ?
    my $numHits = $self->objSummary->getNumberOfHits();
    # Section 5.2.1 (4)
    $self->fh->print("            <cvParam accession=\"MS:1002404\" name=\"count of identified proteins\" cvRef=\"PSI-MS\" value=\"" . $numHits . "\" />\n");
  };

  around 'printProteinAmbiguityGroupBody' => sub{
    my $orig = shift;
    my $self = shift;
    my $i = $_[0];

    $self->$orig(@_);

    my $passThreshold = "true";
    my $objProtein = $self->objSummary->getHit($i);

    if ($self->objResFile->isPMF && $objProtein->getScore < $self->objSummary->getProteinThreshold(1 / $self->objSummary->getProbabilityThreshold)) {
      $passThreshold = "false"
    }
    #Section 5.2.1 (3)
    $self->fh->print("          <cvParam accession=\"MS:1002415\" name=\"protein group passes threshold\" cvRef=\"PSI-MS\" value=\"" . $passThreshold . "\" />\n");
    #Section 5.2.1 (5)
    $self->fh->print("          <cvParam accession=\"MS:1002407\" name=\"cluster identifier\" cvRef=\"PSI-MS\" value=\"" . $objProtein->getHitNumber() . "\" />\n");
  };

  around 'printProteinDetectionHypothesisBody' => sub{
    my $orig = shift;
    my $self = shift;
    my $famMemberNum = $_[0];
    my $preferred_acc = $_[1];
    my $passThreshold = $_[2];
    my $objProtein = $_[3];

    $self->$orig(@_);

    my $isLeading = 0;
    if ($self->group_family) {
      if ($famMemberNum >= 1) {
	$isLeading = 1;
      }
    }
    else {
      if (($objProtein->anyBoldRedPeptides($self->objSummary))) {
	$isLeading = 1;
      }
    }
    if ($isLeading) {
      # Section 5.2.1 (6)
      $self->fh->print("            <cvParam accession=\"MS:1002401\" name=\"leading protein\" cvRef=\"PSI-MS\" />\n");
      if ($famMemberNum == 1) {
	# Section 5.2.1 (6a)
	$self->fh->print("            <cvParam accession=\"MS:1002403\" name=\"group representative\" cvRef=\"PSI-MS\" />\n");
      }
    }
    else {
      $self->fh->print("            <cvParam accession=\"MS:1002402\" name=\"non-leading protein\" cvRef=\"PSI-MS\" />\n");
    }
  };

  around 'printSpectrumIdentificationItemBody' => sub{
    my $orig = shift;
    my $self = shift;

    my $objPeptide = $_[0];
    my $query = $_[1];
    my $rank = $_[2];
    my $psmComp = $_[3];

    $self->$orig(@_);

    if ($objPeptide->getNumberOfLinkedPeptides() == 2) {
      # Section 5.2.9 and Figure 3.
      $self->fh->print(qq(            <cvParam accession="MS:1002511" cvRef="PSI-MS" value="$rank" name="cross-link spectrum identification item"/>\n));
    }

    # Section 5.2.7 N/A
    #$self->fh->print("            <cvParam accession=\"MS:1002520\" cvRef=\"PSI-MS\" value=\"" . $self->_mzid_Peptide_id($query, $rank, $psmComp) . "\" name=\"peptide group ID\" />\n");

    # Section 5.2.7
    my $threshold = $self->objSummary->getPeptideThreshold($query, 1 / $self->objSummary->getProbabilityThreshold);
    my $pass = $self->objSummary->getPeptide($query, $rank)->getIonsScore() < $threshold ? "false" : "true";
    $self->fh->print("            <cvParam accession=\"MS:1002500\" cvRef=\"PSI-MS\" value=\"".$pass."\" name=\"peptide passes threshold\"/>\n");

    #Not in specification but in MzIdentML validator
    #$self->fh->print("            <cvParam accession=\"MS:1002484\" cvRef=\"PSI-MS\" value=\"".$threshold."\" name=\"peptide-level statistical threshold\"/>\n");
  };

}# --MzIdentML_1_2

package ExportDat::MzIdentMLFactory {

  use Moose;

  has 'version'  =>(is => 'ro', isa => 'Str');
  has 'producers' =>(is => 'ro',
		     isa => 'HashRef',
		     default => sub {
		       {
			 "1.1" => sub{ ExportDat::MzIdentML_1_1->new(@_) },
			 "1.2" => sub{ ExportDat::MzIdentML_1_2->new(@_) },
		       }
		     }
    );

  sub create {
    my $self = shift;

    die "Not a valid MzIdentML version: $self->{version}"
      unless my $rc = $self->producers->{$self->version};
    return $rc->(@_);
  };

}# --MzIdentMLFactory

1;
