##############################################################################
# Class to query the protein data retreived from ms-getseq.exe
##############################################################################
#     $Author: robertog@matrixscience.com $ 
#       $Date: 2022-05-16 14:27:02 +0100 $ 
#   $Revision: 2248db7aed94d66543e99125f32fa6d43c1e70b4 | MASCOT_REL_2_8_1-3-gcae025d431 $ 
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 2010-2021 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################

#!/usr/local/mascot/perl64/bin/perl
use v5.14;


package ExportDat::ProteinData {

use Moose;

 # required Perl packages
use msparser;
use ExportDat::TaxonomyData;

  use constant {
    # Debug only: print ms_peptidesummary parameters in stderr.
      WARN_ABOUT_PEPSUM_PARAMS => 0,
  };

require "./common_subs.pl";


has 'objSummary' =>(is => 'ro', isa => 'msparser::ms_mascotresults'); #msparser::ms_peptidesummary or msparser::ms_proteinsummary
has 'objResFile' =>(is => 'ro', isa => 'msparser::ms_mascotresfile');
has 'objDatFile' =>(is => 'ro', isa => 'msparser::ms_datfile');
has 'objUniGene' =>(is => 'ro', isa => 'Maybe[msparser::ms_unigene]');
has 'sessionID' =>(is => 'ro', isa => 'Str');
has 'ShowSameSets' =>(is => 'ro', isa => 'Bool');
has 'ShowSubSets' =>(is => 'ro', isa => 'Bool');
has 'keepalive' =>(is => 'rw', isa => 'Bool');
has 'include_every_protein' =>(is => 'ro', isa => 'Bool');

has 'MIME_boundary'  =>(is => 'ro', isa => 'Maybe[Str]');
has 'shipper' =>(is => 'ro', isa => 'Str');

# sequence length cache (both for emPAI and mustGetProteinMass)
has 'fastaLen' =>(is => 'rw', isa => 'HashRef', default => sub { {} });

has 'objParams' =>(is => 'ro', isa => 'msparser::ms_searchparams');

has 'getseq_cache' =>(is => 'rw', isa => 'HashRef', default => sub { {} });
has 'fastaTitles' =>(is => 'rw', isa => 'Maybe[HashRef]');
has 'populated' =>(is => 'rw', isa => 'Bool');

has 'taxData' =>(is => 'rw', isa => 'ExportDat::TaxonomyData');

has 'sameset_ordering' =>(is => 'rw', isa => 'HashRef', default => sub { {} });

sub BUILD {
    my $self = shift;
    my $args = shift;

    my $taxData = ExportDat::TaxonomyData->new(objSummary => $self->objSummary,
					       objResFile => $self->objResFile,
					       objParams => $self->objParams,
					       objDatFile => $self->objDatFile,
					       objUniGene => $self->objUniGene,
					       sessionID => $self->sessionID,
					       ShowSameSets => $self->ShowSameSets,
					       ShowSubSets => $self->ShowSubSets,
					       keepalive => $self->keepalive,
					       include_every_protein => $self->include_every_protein,
					       MIME_boundary => $self->MIME_boundary,

					       preferTaxonomy => $args->{preferTaxonomy},
					       prot_tax_id => $args->{prot_tax_id},
					       prot_tax_str => $args->{prot_tax_str},
      );
    $self->taxData($taxData);
    $self->sameset_ordering($taxData->sameset_ordering);
    
    if ($args->{do_fetch}) {
      # All proteins in a single NCBI nr entry have the same sequence, so no
      # need to pass \&preferredAccession here.
      my $getseqBatchMass = $self->taxData->queueGetSeqProteins(
        sub { $_[0]->getProteinMass($_[1], $_[2]) }, # only proteins without mass
        1,
      );
      my @data_keys = qw(prot_len prot_sequence);
      # This will be used by both getProteinEmPAI and mustGetProteinMass.
      my $data = $self->batchGetSeqQuery(
        $getseqBatchMass,
        ['showlen', 'showsequence'], \@data_keys
      );
      if ($self->objUniGene) {
        # UniGene entries don't have protein sequences in any meaningful
        # sense.
        $self->taxData->mapUnparsedToUniGeneEntries(
          $data, { prot_len => -1, prot_sequence => -1 }
        );
      }

      $self->fastaLen($self->parseGetSeqBatch(
        $data, $getseqBatchMass, $self->objResFile->params, \@data_keys,
        prot_sequence => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_len => sub { $_[0] eq '' ? -1 : $_[0] },
			));
    }

    $self->keepalive( defined $self->MIME_boundary ? 1 : 0 );
    if($args->{needGetSeqCache}) { #if ($prot_len or $prot_pi or $prot_seq) {
	$self->populateGetSeqCache(
	    {
		prot_len => $args->{prot_len},
		prot_pi => $args->{prot_pi},
		prot_seq => $args->{prot_seq},
	    },
	    $args->{includeFrame}
	    );
    }
    if($args->{needProteinDescriptionCache}) {
      $self->populateProteinDescriptionCache(sub{$self->taxData->preferredAccession(@_)});
    }
}

###############################################################################
# &parseGetSeqBatch()
# Utility function to reformat the data from batchGetSeqQuery().
# $_[0] is the return value from batchGetSeqQuery()
# $_[1] is the original query
# $_[2] is an ms_searchparams object
# $_[3] is the set of data keys to extract (e.g. prot_pi, prot_len)
# $_[4], ... are optional subroutines to massage the cached data
###############################################################################
sub parseGetSeqBatch {
  my $self = shift;
  my ($data, $original_query, $objParams, $data_keys, %massagers) = @_;

  my %ret = ();

  if ($data) {
    while (my ($DB_str, $proteins) = each %$original_query) {
      for my $protein (@$proteins) {
        my ($dbIdx, $acc, $frame, $source_DB_str) = @$protein;
        $frame //= 0;

        if (exists $$data{$source_DB_str}{$acc}) {
          my $d = $$data{$source_DB_str}{$acc};
          my $f = defined($$d{'frame'})
            ? $$d{'frame'} < 0 ? 1 : $$d{'frame'}
            : 0
            ;
          my @data = map {
            exists $massagers{$_}
              ? $massagers{$_}->($$d{$_}, $DB_str)
              : $$d{$_}
          } @$data_keys;
          @{
            $ret{ common_subs::makeProteinCacheKey($dbIdx, $acc, $f) }
          }{ @$data_keys } = @data;

          next;
        }

        my $key = common_subs::makeProteinCacheKey($dbIdx, $acc, $frame);
        $ret{$key} = {
          frame => $frame,
          map { $_ => -1 } @$data_keys
        };
      }
    }
  }

  return \%ret;
}

###############################################################################
# &populateGetSeqCache()
# $_[0] is a hash reference that defines what is to be cached
# $_[1] include_frame argument to queueGetSeqProteins
###############################################################################
sub populateGetSeqCache {
    my $self = shift;
    my ($what, $includeFrame) = @_;

    my $getseqBatchAll = $self->taxData->queueGetSeqProteins(
      undef, $includeFrame,
    );

    my %queue_key = qw(
      prot_pi     showpi
      prot_len    showlen
      prot_seq    showsequence
	);

    my %data_key = qw(
      prot_pi     prot_pi
      prot_len    prot_len
      prot_seq    prot_sequence
	);

    my (@queue_keys, @data_keys);

    for (keys %$what) {
      next unless $queue_key{$_};
      push @queue_keys, $queue_key{$_};
      push @data_keys, $data_key{$_};
    }

    my $data = $self->batchGetSeqQuery(
      $getseqBatchAll,
      \@queue_keys, \@data_keys
	);

    $self->getseq_cache(
      $self->parseGetSeqBatch(
        $data, $getseqBatchAll, $self->objResFile->params, \@data_keys,
        prot_sequence => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_len => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_pi => sub { $_[0] eq '' ? -1 : $_[0] },
      )
    );

    if ($self->objUniGene) {
      $self->taxData->mapToUniGeneEntries(
        \%{$self->getseq_cache}, { prot_pi => -1, prot_len => -1, prot_sequence => -1 }
      );
    }
  }

###############################################################################
# &mustGetProteinMass()
# tries very hard to get a protein mass
# $_[0] accession string
# $_[1] frame number, 0 for protein, -1 for mixed
# $_[2] database number
# globals:
# my($shipper, $fastaLen);
###############################################################################

sub mustGetProteinMass {
  my $self = shift;
  my ($accession, $frame, $dbIdx) = @_;

  my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
  if (defined(my $mass = $self->fastaLen->{$key}{'prot_mass'})) {
    return $mass == -1 ? '' : $mass;
  }

  if (my $mass = $self->objSummary->getProteinMass($accession, $dbIdx)) {
    $self->fastaLen->{$key}{'prot_mass'} = $mass;
    return $mass;
  }

# if this is the public web site, give up
  if ($self->shipper eq 'FALSE') {
    return '';
  }

  if (defined(my $seqString = $self->fastaLen->{$key}{'prot_sequence'})) {
    my $mass = -1;
    if ($seqString ne -1) {
       $mass = $self->objSummary->getSequenceMass($seqString);
    }
    $self->fastaLen->{$key}{'prot_mass'} = $mass;
    return $mass == -1 ? '' : $mass;
  }

  if ($frame == -1) {$frame = 1}

  my ($seqString, $err) = $self->mustGetProteinSeq($accession, $frame, $dbIdx, $self->shipper);
  if ($seqString) {
    my $mass = $self->objSummary->getSequenceMass($seqString);
    $self->fastaLen->{$key}{'prot_mass'} = $mass;
    return $mass;
  }

 # Cannot retrieve sequence
  $self->fastaLen->{$key}{'prot_mass'} = -1;
  return '';
}


###############################################################################
# &mustGetProteinSeq()
# tries very hard to get a description string
# $_[0] is an accession
# $_[1] frame number, 0 for protein, (-1 for mixed is not allowed)
# $_[2] database number
# $_[3] whether the sequence is available to the user or not
###############################################################################
sub mustGetProteinSeq {
    my $self = shift;
    my ($accession, $frame, $dbIdx, $seqAvailable) = @_;

# if this is the public web site, give up
    if ($seqAvailable eq 'FALSE') {
      return ('', 'Protein sequence not available in public website');
    }

# cannot get sequence for mixed frame match
    if ($frame < 0 || $frame > 6) {
      return ('', 'Cannot get sequence for mixed frame match');
    }

# first try the cache
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);

    # Could be undefined if ms-getseq.exe failed. #TODO ROBERTO 12515
    my $seq = $self->getseq_cache->{$key}{'prot_sequence'};
    if (defined($seq)) {
      return ('','Protein sequence not available') if $seq eq -1;
      return ($seq, undef); #Successfully retrieved from cache
    }

    my $source_DB = common_subs::sourceDbName($self->objResFile, $self->objParams, $dbIdx);
    my @retVal = common_subs::getExtInfo($accession, $frame, $source_DB, "sequence");
    my $err = undef;
    if ($retVal[0]) {
      ($seq, $err) = common_subs::parseSequence($retVal[1], $retVal[2]);
    }
    else {
        $err = $retVal[1];
    }
    if ($seq) {
        $self->getseq_cache->{$key}{'prot_sequence'} = $seq;
        return ($seq, $err);
    }
    $self->getseq_cache->{$key}{'prot_sequence'} = -1;
    return ('', $err);
  }

###############################################################################
# &getProteinpI()
# Return the protein pI (from cache if possible)
# $_[0] accession string
# $_[1] frame number, 0 for protein, -1 for mixed
# $_[2] database number
###############################################################################
sub getProteinpI {
    my $self = shift;
    my ($accession, $frame, $dbIdx) = @_;

# if protein contains matches in mixed frames, then return empty handed
    if ($frame == -1) {return ''}

# first try the cache
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
    # Could be undefined if ms-getseq.exe failed.
    my $pI = $self->getseq_cache->{$key}{'prot_pi'};
    if (defined($pI)) {
      return '' if $pI eq -1;
      return $pI;
    }
    # calculate pI from ms-getseq.exe
    my $source_DB = common_subs::sourceDbName($self->objResFile, $self->objParams, $dbIdx);
    my @retVal = common_subs::getExtInfo($accession, $frame, $source_DB, "pI");
    if ($retVal[0]) {
      if ($retVal[1] =~ /([\d\.]+)\s*$/) {
        return $self->getseq_cache->{$key}{'prot_pi'} = $1;
      }
    }
    $self->getseq_cache->{$key}{'prot_pi'} = -1;
    return '';
  }

###############################################################################
# &getProteinLen()
# Return the protein sequence length (from cache if possible)
# $_[0] is an ms_mascotresfile object
# $_[1] is an ms_searchparams object
# $_[2] accession string
# $_[3] frame number, 0 for protein, -1 for mixed
# $_[4] database index for access
###############################################################################
sub getProteinLen {
    my $self = shift;
    my ($accession, $frame, $dbIdx) = @_;

    # if protein contains matches in mixed frames, then choose frame 1 to get approximate length
    if ($frame == -1) {$frame = 1}

    # first try the cache (populated after the header has been output)
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
    my $len = $self->getseq_cache->{$key}{'prot_len'};
    if (defined($len)) {
      return '' if $len eq -1;
      return $len;
    }

    if (defined(my $len = $self->fastaLen->{$key}{'prot_len'})) {
      $self->getseq_cache->{$key}{'prot_len'} = $len;
      return $len == -1 ? '' : $len;
    }

    my $source_DB = common_subs::sourceDbName($self->objResFile, $self->objParams, $dbIdx);
    my @retVal = common_subs::getExtInfo($accession, $frame, $source_DB, "length");
    if ($retVal[0]) {
      if ($retVal[1] =~ /([\d\.]+)\s*$/) {
        return $self->getseq_cache->{$key}{'prot_len'} = $1;
      }
    }
    $self->getseq_cache->{$key}{'prot_len'} = -1;
    return '';
  }

###############################################################################
# my ($length, $mass) = mustGetProteinLen(
#   $objResFile, $objParams, $objSummary, $objProtein,
#   $canFetchLen, $canFetchMass
# );
#
# Fetch protein length if allowed. If not, or if it fails, estimate it
# from mass. If that also fails, return -1.
#
# - If $canFetchLen is false, length is estimated from mass. In this case,
#   $canFetchMass makes a difference:
#     * If $canFetchMass is false, mass is looked up from the results file.
#     * If $canFetchMass is true, mass is looked up from ms-getseq.exe.
# - If $canFetchLen is true, length is always looked up from ms-getseq.exe.
#   If this fails, length is estimated as above.
#
# $length and $mass are looked up values, or undef if no lookup was
# performed. If a lookup was tried and it failed, $length or $mass is
# the empty string.
###############################################################################
sub mustGetProteinLen {
    my $self = shift;
    my ($objProtein, $canFetchLen, $canFetchMass) = @_;

    my $length;

    if ($canFetchLen) {
      $length = $self->getProteinLen($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    }

    my $L = -1;
    my $mass;

    if ($length) {
      $L = $length;
    } else {
      $mass = $self->objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB());
      if ($mass) {
        # OK -- $objSummary will estimate from mass in results file.
      } elsif ($canFetchMass) {
        $mass = mustGetProteinMass($objProtein->getAccession(), $self->objParams, $objProtein->getFrame(), $objProtein->getDB);
        if ($mass) {
          # OK -- estimate length and pass it on.
          $L = int(0.5 + $mass / 110.0);
        } else {
          # Nothing we can do.
        }
      } else {
        # Nothing we can do.
      }
    }

    return ($L, $mass);
  }

###############################################################################
# &populateProteinDescriptionCache()
#
# $_[0] if given, reference to a subroutine that returns a preferred accession
###############################################################################
sub populateProteinDescriptionCache {
    my $self = shift;
    my ($sub_preferred_acc) = @_;

    return if $self->populated;

    # Family members only.
    my $batch = $self->taxData->queueGetSeqProteins(
      sub { $_[0]->getProteinDescription($_[1], $_[2]) },
      1,
      $sub_preferred_acc,
	);

    my $data = $self->batchGetSeqQuery(
      $batch,
      ['showtitle'], ['prot_title']
    );

    my $objParams = $self->objResFile->params;

    my %parse_rules;

    $self->fastaTitles(
      $self->parseGetSeqBatch(
        $data, $batch, $objParams, ['prot_title'],
        prot_title => sub {
          if (not exists $parse_rules{$_[1]}) {
            $parse_rules{$_[1]} = common_subs::findParseRuleForTitle($_[1]);
          }
          my $rule = $parse_rules{$_[1]};
          return unless $rule;
          my ($desc) = $_[0] =~ m#$rule#s;
          return -1 if $desc eq '';
          return $desc;
        }
	));

    if ($self->objUniGene) {
      # -1 is correct: mustGetProteinDescription() will always try the
      # ms_mascotresults object first, and only then try the cache.
      $self->taxData->mapToUniGeneEntries(
        \%{$self->fastaTitles}, { prot_title => -1 }
      );
    }

    $self->populated(1);
  }

###############################################################################
# &mustGetProteinDescription()
# tries very hard to get a description string
# $_[0] is an accession
# $_[1] database number
# $_[2] is _alwaysgettitle
###############################################################################
sub mustGetProteinDescription {
    my $self = shift;
    my ($accession, $dbIdx, $alwaysGetTitle) = @_;

    # first try the Mascot Parser summary object
    my $description = $self->objSummary->getProteinDescription($accession, $dbIdx);
    if ($description) {
      return $description;
    }
    # next try the cache
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession);
    my $desc = $self->fastaTitles->{$key}{'prot_title'};
    if (defined($desc)) {
      # TODO: Assignment to -1 was made on purpose (See comment in populateProteinDescriptionCache),
      # Avoiding this warning requires several design changes
      no warnings 'numeric';
      return '' if $desc == -1;
      return $desc;
    }
    # if this is the public web site, give up
    if ($self->shipper eq 'FALSE') {
      return '';
    }
    if (not $alwaysGetTitle) {
      return '';
    }
    # finally, try ms-getseq.exe
    my $source_DB = common_subs::sourceDbName($self->objResFile, $self->objParams, $dbIdx);
    my @retVal = getExtInfo($accession, 0, $source_DB, "title");
    if ($retVal[0]) {
      # success, save in cache
      return $self->fastaTitles->{$key}{'prot_title'} = parseTitle($retVal[1], $source_DB);
    }
    # failure, set cache entry to -1
    $self->fastaTitles->{$key}{'prot_title'} = -1;
    return '';
}

###############################################################################
# &batchGetSeqQuery()
# Perform a batch query to ms-getseq.exe.
# $_[0] is a batch definition from queueGetSeqProteins()
# $_[1] is an array of types of data to query (e.g. showtitle, showlen, showpi)
# $_[2] is an array of types of data to parse from the query (e.g. prot_title,
#   prot_len, prot_pi)
###############################################################################
sub batchGetSeqQuery {
  my $self = shift;
  my ($batch, $query_keys, $data_keys) = @_;
  return unless %$batch;

  my $proposed_tmpdir = $self->objResFile->getCacheDirectory();

  my $create_message = sub {
    my ($sessionID, $db, $accessions) = @_;
    return Util::MascotExtBatch::create_message($sessionID, $db, $query_keys, 'accession' => $accessions);
  };

  my $parse_message = sub {
    return Util::MascotExtBatch::parse_getseq_message($data_keys, @_);
  };

  my $formatted_query = ExportDat::TaxonomyData::_make_batch_getseq_query($batch);

  my $ret = eval {
    Util::MascotExtBatch::fetch_data(
      $proposed_tmpdir,
      'ms-getseq.exe',
      $create_message,
      $parse_message,
      $formatted_query,
      $self->sessionID,
      progress_message => "<!-- fetching <@$query_keys> from ms-getseq.exe... -->\n",
      keepalive => $self->keepalive ? "<!-- fetching <@$query_keys> from ms-getseq.exe -->\n" : 0,
    );
  };

  if ($@) {
    common_subs::fatal($@, __LINE__, __FILE__, "", "", $self->MIME_boundary);
  }

  return $ret;
}

sub getProteinTaxonomy {
  my $self = shift;
  my ($accession, $dbIdx) = @_;
  return $self->taxData->getProteinTaxonomy($accession, $dbIdx);
}

sub getProteinTranslationTableID {
  my $self = shift;
  my ($accession, $dbIdx) = @_;
  return $self->taxData->getProteinTranslationTableID($accession, $dbIdx);
}

sub preferredAccession {
  my $self = shift;
  my ($db, $acc) = @_;
  return $self->taxData->preferredAccession($db, $acc);
}

}
1;
