##############################################################################
# Class to query the taxonomy data
##############################################################################
#     $Author: robertog@matrixscience.com $ 
#       $Date: 2022-05-16 14:27:02 +0100 $ 
#   $Revision: 2248db7aed94d66543e99125f32fa6d43c1e70b4 | MASCOT_REL_2_8_1-3-gcae025d431 $ 
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 2010-2021 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################

#!/usr/local/mascot/perl64/bin/perl
use v5.14;


package ExportDat::TaxonomyData {

use Moose;

# required Perl packages
use Util::MascotExtBatch; # for ms-getseq.exe batch queries.
use PeptideSummary::Util qw(key_acc key_acc_str make_database_name_map);
use msparser;

  use constant {
    # Debug only: print ms_peptidesummary parameters in stderr.
      WARN_ABOUT_PEPSUM_PARAMS => 0,
  };

require "./common_subs.pl";


has 'objSummary' =>(is => 'ro', isa => 'msparser::ms_mascotresults'); #msparser::ms_peptidesummary or msparser::ms_proteinsummary
has 'objResFile' =>(is => 'ro', isa => 'msparser::ms_mascotresfile');
has 'objParams' =>(is => 'ro', isa => 'msparser::ms_searchparams');
has 'objDatFile' =>(is => 'ro', isa => 'msparser::ms_datfile');
has 'objUniGene' =>(is => 'ro', isa => 'Maybe[msparser::ms_unigene]');
has 'sessionID' =>(is => 'ro', isa => 'Str');
has 'ShowSameSets' =>(is => 'rw', isa => 'Bool');
has 'ShowSubSets' =>(is => 'ro', isa => 'Bool');
has 'keepalive' =>(is => 'rw', isa => 'Bool');
has 'include_every_protein' =>(is => 'ro', isa => 'Bool');
has 'MIME_boundary'  =>(is => 'ro', isa => 'Maybe[Str]');

has 'taxonomy_cache' =>(is => 'rw', isa => 'HashRef', default => sub { {} });
has 'raw_taxonomy_cache' =>(is => 'rw', isa => 'Maybe[HashRef]');
has 'data_key' =>(is => 'ro', isa => 'HashRef', default => sub { { prot_tax_str => 'scientific_name',
								   prot_tax_id => 'taxonomy_id',
								    prot_ttid => 'translation_table_id'} });

has 'sameset_ordering' =>(is => 'rw', isa => 'HashRef', default => sub { {} });
has 'preferred_accession' =>(is => 'rw', isa => 'HashRef', default => sub { {} });

sub BUILD {
  my $self = shift;
  my $args = shift;
  
  if ($args->{preferTaxonomy} > 0 or $args->{prot_tax_id} or $args->{prot_tax_str}) {
    my $prot_ttid = ($args->{prot_tax_id} || $args->{prot_tax_str});
    my $prot_tax_id = $args->{prot_tax_id};
    if ($args->{preferTaxonomy} > 0) {
      $prot_tax_id = 1;
    }
    my $showSameSets = $self->ShowSameSets;
    $self->ShowSameSets($self->{perferTaxonomy} || $self->ShowSameSets);
    $self->populateGetTaxonomyCache(
      {
	prot_ttid => $prot_ttid,
	prot_tax_id => $prot_tax_id,
	prot_tax_str => $args->{prot_tax_str},
      },
      );
    $self->ShowSameSets($showSameSets);
    
  }
  
  if ($args->{preferTaxonomy} > 0) {
    $self->populatePreferredTaxonomyCache(
      $args->{preferTaxonomy},
      );
  }
}

###############################################################################
# &queueGetSeqProteins()
# Create a list of protein accessions, grouped by database, for ms-getseq.exe
# batch calls. pepXML and mzIdentML output (when query_all=1) requires queueing
# even normally inaccessible protein entries.
# $_[0] is a subroutine reference that tests if $_[0] already has the data
# $_[1] should be 1 if protein frames should be part of the query data
# $_[2] if given, should be a hash reference from protein keys (key_acc) to weights in sorting a sameset within a numbered hit
# $_[3] if given, should be reference to a subroutine that returns preferred accession when given DB and accession
#
# UniGene entries are "unpacked" into protein accessions, so that no UniGene
# entry names are passed as query data. Use common_subs::mapToUniGene() to
# add UniGene entry names back after common_subs::batchGetSeqQuery().
#
# Parameter $_[3] will be called with the argument list
#   ($_[0], $accession, $db)
###############################################################################
sub queueGetSeqProteins {
  my $self = shift;
  my ($data_func,
      $include_frame,
      $sub_preferred_acc) = @_;

  my @databases = common_subs::make_database_name_map($self->objParams, $self->objResFile);

  my $objWWWOptions = $self->objDatFile->getWWWOptions();

  for (@databases) {
    my $entry = $objWWWOptions->getSeqEntryByName($$_{'source_DB'});
    if (not $entry) {
      $_ = undef;
      next;
    }
    if ($entry->getHostName() ne 'localhost') {
      $_ = undef;
      next;
    }
    if ($entry->getPath() !~ m#x-cgi/ms-getseq\.exe#) {
      $_ = undef;
      next;
    }
  }

  my $objDatabases = $self->objDatFile->getDatabases();
  my %is_na = ();

  for (@databases) {
    my $i = $objDatabases->findDatabase($$_{'source_DB'});
    if ($i == -1) {
      $_ = undef;
      next;
    }
    my $entry = $objDatabases->getDatabase($i);
    if (not $entry) {
      $_ = undef;
      next;
    }
    $is_na{$$_{'source_DB'}} = $entry->getDBType() == $msparser::NA;
  }

  my %batch = ();

  return \%batch unless grep { defined } @databases;

  my %already_seen = ();

  local *push_protein = sub {
    my ($dbIdx, $acc, $frame) = @_;

    my $source_DB_str = $databases[$dbIdx]{'source_DB'};
    return unless defined $source_DB_str;

    $acc = $sub_preferred_acc->($dbIdx, $acc) if $sub_preferred_acc;

    if ($include_frame) {
      if ($is_na{$source_DB_str}) {
        return if $already_seen{$dbIdx}{$acc}{$frame};
        $already_seen{$dbIdx}{$acc}{$frame} = 1;
      } else {
        return if $already_seen{$dbIdx}{$acc}{''};
        $already_seen{$dbIdx}{$acc}{''} = 1;
      }
    } else {
      return if $already_seen{$dbIdx}{$acc};
      $already_seen{$dbIdx}{$acc} = 1;
    }

    if (defined $data_func) {
      return if $data_func->($self->objSummary, $acc, $dbIdx);
    }

    my $DB_str = $databases[$dbIdx]{'name'};

    if ($include_frame and $is_na{$source_DB_str}) {
      push @{ $batch{$DB_str} }, [
        $dbIdx,
        $acc,
        $frame < 0 ? 1 : $frame,
        $source_DB_str,
      ];
    } else {
      push @{ $batch{$DB_str} }, [
        $dbIdx,
        $acc,
        undef,
        $source_DB_str,
      ];
    }
  };

  if ($self->include_every_protein) {
    my @ranks = (1 .. $self->objSummary->getMaxRankValue());
    # Profiling for bug 11193 revealed that if you have 200,000 protein
    # hits that are pulled from the file (45,000 input spectra), vectori
    # constructors and destructors take an awful lot of time. This is
    # alleviated by keeping the same objects and just clearing them
    # at the start of each loop. (This needs to be done even with values
    # we don't use, such as start and end indices. Otherwise results
    # will accumulate and memory is wasted.)
    my $v_start  = msparser::vectori->new();
    my $v_end    = msparser::vectori->new();
    my $v_pre    = msparser::VectorString->new();
    my $v_post   = msparser::VectorString->new();
    my $v_frames = msparser::vectori->new();
    my $v_multip = msparser::vectori->new();
    my $v_dbs    = msparser::vectori->new();
    for my $q (1 .. $self->objResFile->getNumQueries()) {
      for my $p (@ranks) {
        for ($v_frames, $v_dbs, $v_start, $v_end, $v_pre, $v_post, $v_multip) {
          $_->clear();
        }
        my $accessions = $self->objSummary->getAllProteinsWithThisPepMatch(
          $q, $p, $v_start, $v_end, $v_pre, $v_post, $v_frames, $v_multip, $v_dbs
        );
        for (0 .. $#$accessions) {
          my ($acc, $db, $frame) = ($$accessions[$_], $v_dbs->get($_), $v_frames->get($_));
          # Accessions are always protein accessions, never UniGene entry
          # names. No "unpacking" is needed.
          push_protein($db, $acc, $frame);
        }
      }
    }
    return \%batch;
  }

  my $isPMF = $self->objResFile->isPMF;

  local *add_protein = sub {
    my ($i, $protein) = @_;

    if ($isPMF) {
      if ($protein->isPMFMixture()) {
        for my $i (1 .. $protein->getNumComponents()) {
          my $c = $protein->getComponent($i);
          push_protein($c->getDB(), $c->getAccession(), $c->getFrame());
        }

        return; # Assumption: PMF mixtures don't have samesets.
      }
    }

    my $acc = $protein->getAccession;
    my $db = $protein->getDB;
    my @sameset = ($protein);

    if ($self->ShowSameSets or %{$self->sameset_ordering}) {
      my $i = 1;
      while (my $protein = $self->objSummary->getNextSimilarProteinOf($acc, $db, $i++)) {
        push @sameset, $protein;
      }
    }

    if (%{$self->sameset_ordering}) {
      @sameset = common_subs::sort_by_sameset_ordering(\@sameset, $self->sameset_ordering->{$i});
    }

    my @output_set = $sameset[0];

    if ($self->ShowSameSets) {
      @output_set = @sameset;
    }

    for (@output_set) {
      if ($_->isUnigene()) {
        for my $i (1 .. $_->getNumComponents()) {
          my $c = $_->getComponent($i);
          push_protein($c->getDB, $c->getAccession, $c->getFrame);
        }
      } else {
        push_protein($_->getDB, $_->getAccession, $_->getFrame);
      }
    }
  };

  for my $i (1 .. $self->objSummary->getNumberOfHits) {
    my $hit = $self->objSummary->getHit($i);
    add_protein($i, $hit);

    my $j = 1;
    while (my $protein = $self->objSummary->getNextFamilyProtein($i, $j++)) {
      add_protein($i, $protein);
    }

    if ($self->ShowSubSets) {
      my $j = 1;
      while (my $protein = $self->objSummary->getNextSubsetProtein($i, $j++)) {
        add_protein($i, $protein);
      }
    }
  }

  return \%batch;
}

    
###############################################################################
# &populateGetTaxonomyCache()
# $_[0] is a hash reference that defines what is to be cached
###############################################################################
sub populateGetTaxonomyCache {
  my $self = shift;
  my ($what) = @_;

  # In Mascot 2.4 and later, the results file can contain taxonomy information,
  # but if we would still need to fetch scientific_name and translation_table_id
  # from ms-gettaxonomy.exe, it's easier to just do it all in one batch.
  my $getseqBatchAll = $self->queueGetSeqProteins(
    $$what{'prot_tax_str'} || $$what{'prot_ttid'}
    ? undef
    : sub {
      my $tax_ids = msparser::vectori->new();
      $_[0]->getProteinTaxonomyIDs(
	$_[1], $_[2], $tax_ids, msparser::VectorString->new()
	);
      return $tax_ids->size > 0;
    }
    ,
    0,
    );

  my (@queue_keys, @data_keys);

  for (keys %$what) {
    next unless $self->data_key->{$_};
    push @data_keys, $self->data_key->{$_};
  }

  # All necessary data is included by default, so @queue_keys can be empty.
  
  my $data = $self->batchGetTaxonomyQuery(
    $self->objResFile->getCacheDirectory(), $getseqBatchAll,
    \@queue_keys, \@data_keys,
    );
  
  $self->raw_taxonomy_cache($data || {});
  
  $self->taxonomy_cache(
    $self->parseGetTaxonomyBatch(
      $data, $getseqBatchAll, $self->objResFile->params, \@data_keys,
    ) || {}
    );
  
  if ($self->objUniGene) {
    $self->mapToUniGeneEntries(
      {
	# We can assume that UniGene entries don't include proteins with
	# with different translation table IDs. Otherwise there is something
	# seriously wrong with the mapping. The proteins might come from
	# two different species, which means the UniGene entry species
	# must be undefined.
	scientific_name => -1,
	taxonomy_id => -1,
	translation_table_id => 'merge',
      }
      );
    $self->mapUnparsedToUniGeneEntries(
      {
	scientific_name => -1,
	taxonomy_id => -1,
      }
      );
  }
}

###############################################################################
# &batchGetTaxonomyQuery()
# Perform a batch query to ms-gettaxonomy.exe in order to fetch protein
# taxonomy data.
# $_[0] is the ms_mascotresfile object cache directory
# $_[1] is a batch definition from queueGetSeqProteins()
# $_[2] is an array of types of data to query
# $_[3] is an array of types of data to parse from the query (e.g. taxonomy_id,
#   scientific_name, translation_table_id)
###############################################################################
sub batchGetTaxonomyQuery {
  my $self = shift;
  my ($proposed_tmpdir, $batch, $query_keys,
      $data_keys) = @_;
  return unless %$batch;

  my $create_message = sub {
    my ($sessionID, $db, $accessions) = @_;
    return Util::MascotExtBatch::create_message($sessionID, $db, $query_keys, 'accession' => $accessions);
  };

  my $parse_message = sub {
    return Util::MascotExtBatch::parse_gettaxonomy_message($data_keys, @_);
  };

  my $formatted_query = _make_batch_getseq_query($batch);

  my $ret = eval {
    Util::MascotExtBatch::fetch_data(
      $proposed_tmpdir,
      'ms-gettaxonomy.exe',
      $create_message,
      $parse_message,
      $formatted_query,
      $self->sessionID,
      progress_message => "<!-- fetching data from ms-gettaxonomy.exe... -->\n",
      keepalive => $self->keepalive ? "<!-- fetching data from ms-gettaxonomy.exe -->\n" : 0,
    );
  };

  if ($@) {
    common_subs::fatal($@, __LINE__, __FILE__, "", "", $self->MIME_boundary);
  }

  return $$ret{'acc'};
}

###############################################################################
# &parseGetTaxonomyBatch()
# Utility function to reformat the data from batchGetTaxonomyQuery().
# $_[0] is the return value from batchGetTaxonomyQuery()
# $_[1] is the original query
# $_[2] is an ms_searchparams object
# $_[3] is the set of data keys to extract (e.g. taxonomy_id, scientific_name)
###############################################################################
sub parseGetTaxonomyBatch {
  my $self = shift;
  my ($data, $original_query, $objParams, $data_keys) = @_;

  my %massagers = map { $_ => sub { $_[0] eq '' ? -1 : $_[0] } } @$data_keys;

  my %ret = ();

  while (my ($DB_str, $proteins) = each %$original_query) {
    for my $protein (@$proteins) {
      my ($dbIdx, $acc, undef, $source_DB_str) = @$protein;
      # (We don't use $frame here, because queries in the taxonomy cache don't
      # use it either.)

      if (exists $$data{$source_DB_str}{$acc}) {
        # Databases like NCBInr/NCBIprot contain multiple accessions in each
        # entry. These are collated under the accession used in the query,
        # which is also part of the list.
        my $entry = $$data{$source_DB_str}{$acc};

        for my $i (0 .. $#{ $$entry{'accession'} }) {
          my @data = map {
            exists $massagers{$_}
              ? $massagers{$_}->($$entry{$_}[$i], $DB_str)
              : $$entry{$_}[$i]
          } @$data_keys;
          @{
           $ret{common_subs::makeProteinCacheKey($dbIdx, $$entry{'accession'}[$i])}
          }{ @$data_keys } = @data;
        }

        next;
      }

      $ret{ common_subs::makeProteinCacheKey($dbIdx, $acc) } = {
        map { $_ => -1 } @$data_keys
      };
    }
  }

  return \%ret;
}

###############################################################################
# &mapUnparsedToUniGeneEntries()
# If UniGene is in use, map the batch query results back to UniGene entry
# names. The input data structure will be modified by adding UniGene entry
# names as keys, with field definitions specified by the third parameter:
#   - If the field definition maps to -1, the field value will be set to -1
#   - If the field definition maps to 'merge', the field value will be set
#     to -1 if any data values differ, and otherwise to the common value.
# $_[0] is the data from batchGetSeqQuery()
# $_[1] is the field definition
###############################################################################
sub mapUnparsedToUniGeneEntries {
  my $self = shift;
  my ($data, $field_def) = @_;

  return unless $data;
  for my $db (keys %$data) {
    for my $acc (keys %{ $$data{$db} }) {
      my $j = 0;
      while ((my $str = $self->objUniGene->getUnigeneForAccession($acc, ++$j)) ne '') {
        for my $field (keys %$field_def) {
          $$data{0}{$str}{$field}{ $$data{$db}{$acc}{$field} } = ();
        }
      }
    }
  }

  _merge_unigene_data($$data{0}, $field_def);
}

###############################################################################
# &mapToUniGeneEntries()
# Same as mapUnparsedToUniGeneEntries(), except input data should come
# from batchGetSeqQuery().
# $_[0] is the data from batchGetSeqQuery()
# $_[1] is the field definition
###############################################################################
sub mapToUniGeneEntries {
  my $self = shift;
  my ($data, $field_def) = @_;

  return unless $data;

  my %unigene_data = ();

  for my $key (keys %$data) {
    my ($db_idx, $acc, $frame) = common_subs::parseProteinCacheKey($key);

    my $j = 0;
    while ((my $str = $self->objUniGene->getUnigeneForAccession($acc, ++$j)) ne '') {
      for my $field (keys %$field_def) {
        $unigene_data{$str}{$field}{ $$data{$key}{$field} } = ();
      }
    }
  }

  _merge_unigene_data(\%unigene_data, $field_def);

  for my $str (keys %unigene_data) {
    $$data{ common_subs::makeProteinCacheKey(0, $str) } = $unigene_data{$str};
  }
}

sub _merge_unigene_data {
  my ($data, $field_def) = @_;

  for my $str (keys %$data) {
    for my $field (keys %$field_def) {
      if ($$field_def{$field} eq -1) {
        $$data{$str}{$field} = -1;
      } elsif ($$field_def{$field} eq 'merge') {
        my @values = keys %{ $$data{$str}{$field} };
        if (@values == 1) {
          $$data{$str}{$field} = $values[0];
        } else {
          $$data{$str}{$field} = -1;
        }
      }
    }
  }
}

sub _make_batch_getseq_query {
  my ($batch) = @_;

  # Bugs 12185-12187: The $batch data structure is a mapping from database
  # name to an array of arrays. The innermost elements are
  # [$dbIdx, $acc, $frame, $source_DB], where $dbIdx is the database number in
  # the results file, $acc is the accession, $frame (if defined) is the frame
  # number and $source_DB is the name of the database to query.
  #
  # If $dbIdx is of type AA or NA, the source database name is the same as the
  # $dbIdx database name.
  #
  # If $dbIdx is of type SL, the source database name is the reference
  # database.
  #
  # The query is always targeted at $source_DB, even if the batch
  # specification groups elements by the $dbIdx name. We only need $source_DB
  # and $accframe here; see parseGetseqBatch for the counterpart mapping.
  my %query = ();

  for my $db_name (keys %$batch) {
    for my $protein (@{ $$batch{$db_name} }) {
      my (undef, $acc, $frame, $source_DB) = @$protein;
      my $accframe;
      if (defined $frame) {
        $accframe = sprintf('"%s":%d', $acc, $frame);
      } else {
        $accframe = qq{"$acc"};
      }
      push @{ $query{$source_DB} }, $accframe;
    }
  }

  return \%query;
}

###############################################################################
# &getProteinTaxonomy()
# Return the taxonomy ID and species name of the protein
# $_[0] accession string
# $_[1] database number
###############################################################################
sub getProteinTaxonomy {
  my $self = shift;
  my ($accession, $dbIdx) = @_;
  
  # first try the cache
  my $key = common_subs::makeProteinCacheKey($dbIdx, $accession);
  # Could be undefined if ms-gettaxonomy.exe failed.
  my ($taxID, $species) = @{ $self->taxonomy_cache->{$key} }{qw(taxonomy_id scientific_name)};
  if (defined($taxID)) {
    return ('', '') if $taxID == -1;
    return ($taxID, $species);
  }
  
  my $source_DB = common_subs::sourceDbName($self->objResFile, $self->objParams, $dbIdx);
  my @data;
  if (open(my $fh, sprintf("\"../x-cgi/ms-gettaxonomy.exe\" 1 %s %s | ", shellQuote($source_DB), shellQuote($accession)))) {
    @data = <$fh>;
    close $fh;
  }
  
  if (@data and @data > 2) {
    chomp($data[2]);
    (my $ac, $taxID, $species) = split /\s+/, $data[2], 3;
    if ($ac ne $accession or not defined $taxID) {
      ($taxID, $species) = (-1, -1);
    }
  }

  @{ $self->taxonomy_cache->{$key} }{qw(taxonomy_id scientific_name)} = ($taxID, $species);
  return ('', '') if $taxID == -1;
  return ($taxID, $species);
}

###############################################################################
# &getProteinTranslationTableID()
# Return the translation table ID of the protein, if available
# $_[0] accession string
# $_[1] database number
###############################################################################
sub getProteinTranslationTableID {
  my $self = shift;
  my ($accession, $dbIdx) = @_;
  
  # first try the cache
  my $key = common_subs::makeProteinCacheKey($dbIdx, $accession);
  # Could be undefined if ms-gettaxonomy.exe failed.
  my $ttID = $self->taxonomy_cache->{$key}{'translation_table_id'};
  if (defined($ttID)) {
    return 1 if $ttID == -1;
    return $ttID;
  }
  
  my $db = $self->objParams->getDB($dbIdx);
  my @data;
  if (open(my $fh, sprintf("\"../x-cgi/ms-gettaxonomy.exe\" 9 %s %s | ", shellQuote($db), shellQuote($accession)))) {
    @data = <$fh>;
    close $fh;
  }
  if (@data and @data > 2) {
    ($ttID) = $data[2] =~ /(\d+)/;
    $ttID = -1 unless $ttID;
  }
  
  $self->taxonomy_cache->{$key}{'translation_table_id'} = $ttID;
  return 1 if $ttID == -1;
  return $ttID;
}

###############################################################################
# &populatePreferredTaxonomyCache()
# $_[0] is the preferred taxonomy ID
###############################################################################
sub populatePreferredTaxonomyCache {
  my $self = shift;
  my ($preferTaxonomy) = @_;

  return unless defined $self->raw_taxonomy_cache;
  
  my $taxonomyfile = MascotDat->taxonomy_file();
  my $taxonomy_entry;
  
  if ($preferTaxonomy < $taxonomyfile->getNumberOfEntries()) {
    $taxonomy_entry = $taxonomyfile->getEntryByNumber($preferTaxonomy);
  }
  
  return unless $taxonomy_entry;
  
  my $params = $self->objResFile->params;
  my %databases = map {
    $_ => $params->getDB($_)
  } 1 .. $params->getNumberOfDatabases;
  
  my @sameset_clumps = ();
  my %reverse = ();
  
  # Sameset are included in the initial caching regardless of $ShowSameSets
  # (otherwise you can't determine which sameset protein to prefer).
  local *add_clump = sub {
    my ($ix, $protein) = @_;
    my @sameset = common_subs::compileSameset($self->objSummary, $protein, 1);
    push @sameset_clumps, \@sameset;
    $reverse{\@sameset} = $ix;
    
    # It's possible taxonomy IDs are not actually in $raw_taxonomy_cache
    # if 1) only taxonomy IDs were requested to be fetched and 2) the
    # results file has taxonomy IDs. This is mainly the case for
    # master_results.pl.
    for (@sameset) {
      my $dbName = $databases{$_->getDB()};
      if (not exists $self->raw_taxonomy_cache->{$dbName}{ $_->getAccession }) {
	my $tax_ids = msparser::vectori->new();
	my $acc = msparser::VectorString->new();
	$self->objSummary->getProteinTaxonomyIDs($_->getAccession(), $_->getDB(), $tax_ids, $acc);
	$self->raw_taxonomy_cache->{$dbName}{$_->getAccession} = {
	  taxonomy_id => common_subs::vector_to_array($tax_ids),
	  accession => common_subs::vector_to_array($acc),
	};
      }
    }
  };
  
  for my $ix (1 .. $self->objSummary->getNumberOfHits) {
    my $hit = $self->objSummary->getHit($ix);
    add_clump($ix, $hit);
    
    my $j = 1;
    while (my $protein = $self->objSummary->getNextFamilyProtein($ix, $j++)) {
      add_clump($ix, $protein);
    }
    
    if ($self->ShowSubSets) {
      my $j = 1;
      while (my $protein = $self->objSummary->getNextSubsetProtein($ix, $j++)) {
	add_clump($ix, $protein);
      }
    }
  }
  
  my $data = PeptideSummary::Util::sameset_ordering_by_taxonomy(
    $self->objResFile->params,
    $self->objResFile,
    \@sameset_clumps,
    $self->raw_taxonomy_cache,
    $taxonomy_entry,
    );

  for my $clump (@sameset_clumps) {
    my $sort_order = $$data{$clump}{'sort_order'};
    my $ix = $reverse{$clump};
    
    @{ $self->sameset_ordering->{$ix} }{keys %$sort_order} = values %$sort_order;
    
    my $pref = $$data{$clump}{'preferred_accession'};
    @{$self->preferred_accession}{keys %$pref} = values %$pref;
  }
  
  return 1;
}

sub preferredAccession {
  my $self = shift;
  my ($db, $acc) = @_;
  return $acc unless $self->preferred_accession;
  my $str = $self->preferred_accession->{key_acc_str($db, $acc)};
  return $acc if (not defined($str) or $str eq '');
  return $str;
}

###############################################################################
# &getTaxonomyNameFromId()
# return taxonomy name as string
# $_[0] taxonomy ID
# $_[1] ref to %taxNames (cache)
###############################################################################

sub getTaxonomyNameFromId {

  my ($id, $taxNames_ref) = @_;

# First time through, populate hash with some common tax IDs to reduce number
# of times we have to open names.dmp.
# Also, means we get some information even if names.dmp missing
  unless (scalar(%{ $taxNames_ref })) {
    ${ $taxNames_ref }{'2'} = "Bacteria";
    ${ $taxNames_ref }{'197'} = "Campylobacter jejuni";
    ${ $taxNames_ref }{'358'} = "Agrobacterium tumefaciens";
    ${ $taxNames_ref }{'487'} = "Neisseria meningitidis";
    ${ $taxNames_ref }{'562'} = "Escherichia coli";
    ${ $taxNames_ref }{'590'} = "Salmonella";
    ${ $taxNames_ref }{'1224'} = "Proteobacteria";
    ${ $taxNames_ref }{'1239'} = "Firmicutes";
    ${ $taxNames_ref }{'1313'} = "Streptococcus Pneumoniae";
    ${ $taxNames_ref }{'1423'} = "Bacillus subtilis";
    ${ $taxNames_ref }{'1769'} = "Actinobacteria";
    ${ $taxNames_ref }{'1902'} = "Streptomyces coelicolor";
    ${ $taxNames_ref }{'2093'} = "Mycoplasma";
    ${ $taxNames_ref }{'2157'} = "Archaeobacteria";
    ${ $taxNames_ref }{'2157'} = "Archaeobacteria";
    ${ $taxNames_ref }{'2759'} = "Eukaryota";
    ${ $taxNames_ref }{'3360'} = "Alveolata";
    ${ $taxNames_ref }{'3701'} = "Arabidopsis";
    ${ $taxNames_ref }{'3702'} = "Arabidopsis thaliana";
    ${ $taxNames_ref }{'4530'} = "Oryza sativa";
    ${ $taxNames_ref }{'4754'} = "Pneumocystis carinii";
    ${ $taxNames_ref }{'4896'} = "Schizosaccharomyces pombe";
    ${ $taxNames_ref }{'4932'} = "Saccharomyces Cerevisiae";
    ${ $taxNames_ref }{'5833'} = "Plasmodium falciparum";
    ${ $taxNames_ref }{'6239'} = "Caenorhabditis elegans";
    ${ $taxNames_ref }{'7215'} = "Drosophila";
    ${ $taxNames_ref }{'7711'} = "Chordata";
    ${ $taxNames_ref }{'7898'} = "Actinopterygii";
    ${ $taxNames_ref }{'7955'} = "Danio rerio";
    ${ $taxNames_ref }{'8287'} = "lobe-finned fish and tetrapod clade";
    ${ $taxNames_ref }{'8355'} = "Xenopus laevis";
    ${ $taxNames_ref }{'9443'} = "Primates";
    ${ $taxNames_ref }{'9606'} = "Homo sapiens";
    ${ $taxNames_ref }{'9989'} = "Rodentia";
    ${ $taxNames_ref }{'10088'} = "Mus";
    ${ $taxNames_ref }{'10090'} = "Mus musculus";
    ${ $taxNames_ref }{'10114'} = "Rattus";
    ${ $taxNames_ref }{'10239'} = "Viruses";
    ${ $taxNames_ref }{'11103'} = "Hepatitis C virus";
    ${ $taxNames_ref }{'12908'} = "unclassified";
    ${ $taxNames_ref }{'31033'} = "Takifugu rubripes";
    ${ $taxNames_ref }{'33090'} = "Viridiplantae";
    ${ $taxNames_ref }{'33208'} = "Metazoa";
    ${ $taxNames_ref }{'40674'} = "Mammalia";
    ${ $taxNames_ref }{'45251'} = "Arabidopsis neglecta";
    ${ $taxNames_ref }{'77643'} = "Mycobacterium tuberculosis complex";
    ${ $taxNames_ref }{'117571'} = "bony vertebrates";
  }

# if name not in cache, get from names.dmp and add to cache
  unless (${ $taxNames_ref }{$id}) {
    if (open(TAXFILE, "../taxonomy/names.dmp")) {
      my @matches = grep /^$id\s+/, <TAXFILE>;
      close TAXFILE;
      foreach my $match (@matches) {
        my @fields = split(/\|/, $match);
        if ($fields[3] =~ /\s+scientific name\s+/i) {
          $fields[1] =~ s/^\s+//;
          $fields[1] =~ s/\s+$//;
          ${ $taxNames_ref }{$id} = $fields[1];
          last;
        }
      }
    }
  # if this fails, set cache entry to -1 to prevent further attempts
    unless (${ $taxNames_ref }{$id}) {
      ${ $taxNames_ref }{$id} = "-1";
    }
  }

  if (${ $taxNames_ref }{$id} eq "-1") {
    return "";
  } else {
    return ${ $taxNames_ref }{$id};
  }

}

}
1;
