#!/usr/local/mascot/perl64/bin/perl
##############################################################################
# Export Mascot search results                                               #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 1998-2019 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################
#    $Archive:: /www/cgi/export_dat.pl                                     $ #
#     $Author: robertog@matrixscience.com $ #
#       $Date: 2022-05-16 14:27:02 +0100 $ #
#   $Revision: 2248db7aed94d66543e99125f32fa6d43c1e70b4 | MASCOT_REL_2_8_1-3-gcae025d431 $ #
# $NoKeywords::                                                            $ #
##############################################################################
 use strict;                                                                #
 use warnings;
##############################################################################

# The following (case insensitive) URL arguments are recognised:
#
# file                      - relative path to result file (required)
# sessionid                 - Mascot security sessionID
# do_export                 - 1 to export results, 0 to display form, -1 to display command line arguments
# generate_file             - 1 (and do_export=1) to never redirect if cache files need to be created
# generate_mime             - 1 to export all messages and data in multipart MIME format
# export_format             - "XML" or "CSV" or "mzIdentML" or "mzTab" or "pepXML" or "DTASelect" or MascotDAT or MGF or ...
# _sigthreshold             - probability significance threshold, default 0.05
# report                    - max number of hits to be reported, (0 = AUTO)
# _server_mudpit_switch     - if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
# _ignoreionsscorebelow     - cut off on MS/MS ions scores
# show_same_sets            - 1 to display all proteins that match the same set of peptides, otherwise just one per hit
# _showsubsets              - display protein hits that are missing up to this fraction of the protein score of the main hit
# _requireboldred           - 1 to report protein hits only if they include at least one bold, red peptide match
# group_family              - 1 to group into protein families
# use_homology              - 1 to use homology threshold, 0 to use identity threshold
# unigene                   - UniGene index species to be used to cluster hits
# _showallfromerrortolerant - 1 to display all hits from an error tolerant search, including garbage (no UI)
# _onlyerrortolerant        - 1 to display only error tolerant matches from a new style error tolerant search
# _noerrortolerant          - 1 to display no error tolerant matches from a new style error tolerant search
# _minpeplen                - minimum peptide sequence length to use in protein inference
# min_num_sig_unique_seqs   - minimum number of significant unique sequences in a protein hit
# _show_decoy_report        - 1 to display report for the integrated decoy search (Mascot 2.2. & later)
# percolate                 - 1 to display scores and expect values based on Percolator PEPs
# percolate_rt              - 1 to include retention time feature when running Percolator
# _prefertaxonomy           - taxonomy ID > 0; proteins in samesets with a taxonomy that matches _prefertaxonomy will be preferred over the others
# _target_fdr               - target FDR; if set, this adjusts _sigthreshold (at identity or homology level as specified by use_homology)
# _fdr_count_type           - type of object counted towards FDR: either psm or seq; default comes from FDRCountType in mascot.dat
# library                   - one of integrated, onlyprimary, onlylibrary (only effective in spectral library search results)
# search_master             - Over-ride toggle for show_header, show_decoy, show_mods, show_params, show_format, show_masses, show_quant_header
# show_header               - 1 to include a block containing search level information
# show_decoy                - 1 to include a block containing decoy search information
# show_mods                 - 1 to include a block containing variable modifications information
# show_params               - 1 to include a block containing search parameters
# show_format               - 1 to include a block containing format parameters
# show_masses               - 1 to include a block containing residue and element masses
# show_unassigned           - 1 to include a block containing unassigned matches
# show_header               - 1 to include a block containing quantitation level information
# protein_master            - Over-ride toggle for all protein info
# peptide_master            - Over-ride toggle for all peptide info
# show_pep_dupes            - Over-ride toggle for exporting duplicate peptides
# export_site_data          - Over-ride toggle for exporting site analysis data in mzIdentML
# query_master              - Over-ride toggle for all query info
# mgf_output_order          - 1 if the generated mgf file is sorted by input index
# fasta_output_type         - 1 if only NA sequences are to be exported to the generated FASTA file, 2 if only AA, 0 or undefined if no filtering is needed

# Note: Arguments that allow spaces, such as _quant_pep_threshold_type,
# require special attention when used on the command line and as URL
# parameters.
#  - On the command line, the argument value must be enclosed in quotation
#    marks. For example:
#
#    export_dat_2.pl _quant_pep_threshold_type="at least homology" ...
#
#  - In a URL, the spaces must be URL-escaped:
#
#    export_dat_2.pl?_quant_pep_threshold_type=at%20least%20homology&...
#
# The ellipsis (...) indicates the rest of the argument list. Quoting
# is done correctly when you click on "Show command line arguments"
# in the main form (see sub display_args()).

#
# If the following (case insensitive) URL arguments are set to 1, the corresponding fields
# are included in the report unless over-ridden by "master" checkbox
#
  my @columns;
  push @columns, "prot_hit_num";  # always forced to 1
  push @columns, "prot_family_member";  # slaved off prot_hit_num if group_family true
  push @columns, "prot_acc";      # always forced to 1
  push @columns, "prot_desc";
  push @columns, "prot_score";
  push @columns, "prot_thresh";
  push @columns, "prot_expect";
  push @columns, "prot_mass";
  push @columns, "prot_matches";
  push @columns, "prot_matches_sig";    # slaved off prot_matches
  push @columns, "prot_sequences";      # slaved off prot_matches
  push @columns, "prot_sequences_sig";  # slaved off prot_matches
  push @columns, "prot_cover";
  push @columns, "prot_len";
  push @columns, "prot_pi";
  push @columns, "prot_tax_str";
  push @columns, "prot_tax_id";
  push @columns, "prot_seq";
# prot_empai (no entry in header row)
# prot_quant (no entry in header row)
  push @columns, "pep_query";     # always forced to 1
  push @columns, "pep_index";      # always forced to 1
  push @columns, "pep_rank";      # always forced to 1
  push @columns, "pep_isbold";    # always forced to 1
  push @columns, "pep_isunique";  # always forced to 1
  push @columns, "pep_exp_mz";    # always forced to 1
  push @columns, "pep_exp_mr";
  push @columns, "pep_exp_z";
  push @columns, "pep_calc_mr";
  push @columns, "pep_delta";
  push @columns, "pep_start";
  push @columns, "pep_end";
  push @columns, "pep_miss";
  push @columns, "pep_score";
  push @columns, "pep_homol";
  push @columns, "pep_ident";
  push @columns, "pep_expect";
  push @columns, "pep_res_before";  # slaved off pep_seq
  push @columns, "pep_seq";
  push @columns, "pep_res_after";  # slaved off pep_seq
  push @columns, "pep_frame";
  push @columns, "pep_var_mod";
  push @columns, "pep_var_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_summed_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_local_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_mono_link_mod_pos"; # slaved off pep_var_mod
  push @columns, "pep_var_mod_conf";  # slaved off pep_var_mod
  push @columns, "pep_num_match";
  push @columns, "pep_scan_title";
  push @columns, "pep_source";
# pep_quant (no entry in header row)

# 'Global' variables
  my(
    $anyPepSumMatch,           # $objResFile->anyPeptideSummaryMatches($sec_peptides)
                               # or $objResFile->anyPeptideSummaryMatches(SEC_LIBRARYPEPTIDES)
    $anyLibraryMatch,          # $objResFile->anyPeptideSummaryMatches(SEC_LIBRARYPEPTIDES)
    $anyXlPepSumMatch,         # $objResFile->anyPeptideSummaryMatches(SEC_CROSSLINK_PEPTIDES)
    $hasXlMethod,              # length($objParams->getCROSSLINKING())
    $delimiter,                # delimiter for CSV is "," by default
    $export_format,            # "XML" or "CSV", etc.
    @fileIn,                   # relative paths to result files, passed as required URL argument (file or fileset)
    $idForFatalError,          # $fileIn or results collection set ID for error reporting
    $fileNameRoot,             # root of result file name, e.g. F123456
    $group_family,             # 1 to group into protein families
    $ignoreIonsScoreBelow,     # cut off on MS/MS ions scores
    $ignoreIonsScoreBelowDefined, # true if given in the URL
    $mass_type,                # MONO or AVE
    $massDP,                   # number of decimal places in formatted mass values
    $minPepLen,                # peptides shorter than $minPepLen are ignored when generating Peptide summary
    $minNumSigUniqueSeqs,      # minimum number of significant unique sequences in a protein hit
    $msresFlags,               # flags used to create Mascot Parser summary object
    $msresFlags2,              # flags2 used to create Mascot Parser summary object
    $mudpitSwitch,             # if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
    $mgfOutputOrder,           # 1 if generated mgf file is sorted by input index, 0 otherwise.
    $fastaOutputType,          # undefined if any sequence type, $msparser::NA or $msparser::AA
    $myName,                   # file name of this script for links to self (not complete path)
    $NoErrorTolerant,          # 1 to display only standard matches from a new style error tolerant search
    $numHits,                  # max number of hits to be reported
    $objDatFile,               # Mascot parser mascot.dat file object
    $objMascotOptions,         # Mascot parser configuration options object
    $objMassesFile,            # Mascot parser masses file object
    $objModFile,               # Mascot parser mod_file object
    $objXlModFile,             # Mascot parser mod_file object for croslinks
    $objParams,                # Mascot parser search parameters object
    $objProtein,               # Mascot parser protein object
    $objQuantFile,             # Mascot parser ms_quant_configfile object
    $objQuantMethod,           # Mascot parser quant_method object
    $quantProtocolType,        # REPORTER, MULTIPLEX, PRECURSOR
    $objResFile,               # Mascot parser result file object
    $objSummary,               # Mascot parser result summary object (may be protein or peptide)
    $objSummaryDecoy,          # ms_peptidesummary for decoy section if protein FDR is available
    $objMS2Quant,              # Mascot parser quantitation data object (or undef if file is not Reporter or Multiplex)
    $objUmodSection,           # Unimod section of the results file
    $objUmodXlSection,         # Unimod_xl section of the results file
    $objFullUmodFile,          # The full unimod.xml file
    $objUnigeneOptions,        # Mascot parser configuration unigene object
    $OldStyleErrTolReport,     # 1 to suppress certain parts of the report as per old-style error tolerant search
    $OmitErrTolThresholds,     # 1 to suppress displaying thresholds for ET queries in results files from before 2.8
    $OnlyErrorTolerant,        # 1 to display only error tolerant matches from a new style error tolerant search
    $percolate,                # 1 to display scores and expect values based on Percolator PEPs
    $percolate_rt,             # 1 to include retention time feature when running Percolator
    $preferTaxonomy,           # index > 0 that controls which protein to show at the top of a sameset
    $libraryMode,              # if SEC_LIBRARYPEPTIDES is not empty, one of integrated, onlyprimary, onlylibrary
    $queryColumnCount,         # Used to align peptide match data in CSV output
    $reportType,               # 'peptide' or 'concise'
    $RequireBoldRed,           # 1 to report protein hits only if they include at least one bold, red peptide match
    $sec_peptides,             # SEC_PEPTIDES or SEC_DECOYPEPTIDES
    $sec_summary,              # SEC_SUMMARY or SEC_DECOYSUMMARY
    $session,                  # Mascot security session object
    $sessionID,                # Mascot security sessionID
    $shipper,                  # flag set to 'TRUE' if this is Mascot for intranet, 'FALSE' for public internet site
    $ShowAllFromErrorTolerant, # 1 to display all hits from an error tolerant search, including garbage
    $ShowDecoyReport,          # 1 to display report for the integrated decoy search (Mascot 2.2. & later)
    $ShowSameSets,             # 1 to display same set proteins
    $ShowSubSets,              # display protein hits that are missing up to this fraction of the protein score of the main hit
    $sigThreshold_dbl,         # probability significance threshold (1 / OneInXprobRnd)
    $sigThreshold_str,         # value shown to the user -- could be different in SL-only mode
    $showFDRs,                 # true if there are decoy matches and this is not a decoy report
    $targetFDR,                # target FDR (percentage) or unset; if set, overrides $sigThreshold_dbl
    $targetFDRtype,            # 'psm' or 'seq'
    $thisScript,               # CGI object
    $isCGI,                    # running under CGI?
    $tmpChecked,               # temp variable used to toggle state of CHECKED
    $UniGeneFile,              # path to a UniGene index file
    $objUniGene,               # ms_unigene object if UniGene is in use
    $URI,                      # URL as far as Mascot root directory, if executed as CGI, otherwise "../"
    $use_homology,             # 1 to use homology threshold, 0 to use identity threshold
    $use_resfile_cache,        # if given and 0, don't use resfile cache file
    $use_results_cache,        # if given and 0, don't use results cache file
    $mascotresultsNeedsPercolating, # 1 if Percolator files potentially need to be created
    %elementalMasses,          # mass values for oxygen, carbon, ...
    %groupMasses,              # mass values for residues and terminii, including possible fixed modifications (group means delta + absolute)
    %taxNames,                 # cache for looking up taxonomy name string from tax ID
    %urlParams,                # keys are URL parameter names in lower case, values are names in original case
    %vmMass,                   # variable modification mass
    %vmString,                 # variable modification name
    %db_type_query_restrictions, # shortcuts for determining type of source database
    @quantWarnings,            # contains the HTML text for any warnings returned by quant routines
    @query_title_list,         # list of query titles, HTML escaped, index is query number
    %pepSrcFamilyMembers,      # cache for looking up the family member to which a peptide match is assigned to
    $protData,                 # A ExportDat::ProteinData object
    );

# set current directory to script location for FastTrack on NT
# chdir also handles a Windows drive letter change
  if ($ENV{'WINDIR'} && $0 =~ /(.*)[\\\/]/) {
    chdir($1);
  }

# delimiter for CSV can be changed
  $delimiter = ",";

 # required Perl packages
  use CGI;
  use CGI::Carp qw(fatalsToBrowser);
  use URI;
  use HTTP::Request::Common;
  use LWP::UserAgent;
  use Cwd;
  use File::Spec;
  use Scalar::Util qw(looks_like_number);
  use List::Util qw(sum min max);
  use msparser;
  use MascotDat;
  use MascotResults;
  use Util::IPC qw(pipefork);
  use CachedResults::Build::Util qw(run_ms_createpip run_percolator);
  use PeptideSummary::Util qw(key_acc key_acc_str MAYU_FP_estimate);
  use PeptideSummary::Backend qw(duplicate_group_main sort_duplicate_group peptide_source_family_members);
  use MasterResults::ProteinFamily::ReportFunc qw(taxonomy_choice);
  use Util::Peptides qw(
    lookup_query_level_mods
    determine_peptide_db_type
    varmod_number_to_id
  );
  use Util::SiteAnalysis;
  use Util::Text qw(joinand round_evenly);
  use Util::Hash qw();

  use ExportDat::Util;
  use ExportDat::ProteinData;
  use ExportDat::MzIdentML;

  use constant {
    # Debug only: print ms_peptidesummary parameters in stderr.
      WARN_ABOUT_PEPSUM_PARAMS => 0,
  };

  $| = 1; # set autoflush on STDOUT

# create CGI object
  $thisScript = new CGI;

# pull in some common subroutines
  require "./common_subs.pl";
  require "./quant_subs.pl";


# www.pl contains the code to decrypt result file names
# its existence indicates that this script is running on the public web site
  if (-e "./www.pl") {
    do "./www.pl";
    $shipper = 'FALSE';
  } else {
    $shipper = 'TRUE';
  }

# get the name of this script for creating URL links to self
#  $myName = $thisScript->url(-relative=>1);
  ($myName) = $0 =~ /([^\\\/]*)$/;

# taint mode requires path to be set explicitly
# Unix path may be required to find decompression utilities in &common_subs::decompress()
  if ($ENV{'WINDIR'}){
    $ENV{"PATH"} = "";
  } else {
    $ENV{"PATH"} = "/usr/local/sbin:/usr/sbin:/sbin:/usr/local/bin:/usr/bin:/bin";
  }
# to bypass taint mode check on the path, comment out above lines and uncomment following line
# ($ENV{"PATH"}) = $ENV{"PATH"} =~ /(.*)/;

# create hash for URL arguments, so that we can always use lower case keys
  foreach ($thisScript->param) {
    if ($urlParams{lc($_)} && $urlParams{lc($_)} ne $_) {
      die("URL parameter names must use consistent case");
    }
    $urlParams{lc($_)} = $_;
  }

  $isCGI = 0;

  if (defined($ENV{'REMOTE_ADDRESS'}) or defined($ENV{'GATEWAY_INTERFACE'})
          or defined($ENV{'SERVER_NAME'})) {
    $isCGI = 1;
  }

# some protein options not permitted on public web site
  if ($shipper eq 'FALSE') {
    $thisScript->param(-name=>$urlParams{'prot_cover'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_len'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_pi'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_tax_str'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_tax_id'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_seq'}, -value=>'');
  }

# some options are not optional!
  &set_option('prot_hit_num');
  &set_option('prot_acc');
  &set_option('pep_query');
  &set_option('pep_index');
  &set_option('pep_rank');
  &set_option('pep_isbold');
  &set_option('pep_isunique');
  &set_option('pep_exp_mz');

# save export_format in variable to make script more readable
  $export_format = "XML" unless ($export_format = $thisScript->param($urlParams{'export_format'}));

# verify the result file exists, decompress if necessary
  if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
    @fileIn = common_subs::filesetFilePaths($fileset);
  }

  if (@fileIn == 0) {
    my $fileIn = $thisScript->param($urlParams{'file'})    || do {
      print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
      common_subs::fatal("no file name", __LINE__, $myName, "");
    };
    @fileIn = ($fileIn);
    $idForFatalError = $fileIn;
  } else {
    $idForFatalError = $thisScript->param($urlParams{'fileset'});
  }

# create a datfile file object
  $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objDatFile->isValid) {
    die common_subs::checkErrorHandler($objDatFile);
  }
  $objMascotOptions = $objDatFile->getMascotOptions;
  $objUnigeneOptions = $objDatFile->getUniGeneOptions;

# untaint the file path. This regex may need fine tuning
  for (@fileIn) {
    ($_) = $_ =~ /([\/\w\-\+.$%&()[\]:]+)/;
    if (not $_) {
      print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
      common_subs::fatal('result filename could not be untainted', __LINE__, $myName, $idForFatalError);
    }
    if (not -e common_subs::decompress($_)) {
      # could be master_results_2.pl style filepath
      my $root = $objMascotOptions->getInterFileRelPath;
      my $filepath = join('/', $root, $_);
      if (-e common_subs::decompress($filepath)) {
        $_ = $filepath;
      } else {
        die("result file does not exist ", $_);
      }
    }
  }

  if (defined($thisScript->param($urlParams{'_usecache'}))) {
    $use_resfile_cache = $use_results_cache = $thisScript->param($urlParams{'_usecache'}) + 0;
  } else {
    $use_resfile_cache = $use_results_cache = 1;
    if ($export_format eq 'mzIdentML') {
      # mzIdentML requires lots of query level data, which means the resfile
      # cache creates a large I/O bottleneck. mzIdentML can also require lots
      # of protein level data, which means the results cache can also become
      # a bottleneck. However, we should keep using the cache file until
      # we are actually about to print out the data.
      if ($isCGI) {
        if (($thisScript->param($urlParams{'do_export'}) || 0) == 0) {
          # use caches
        } elsif ($thisScript->param($urlParams{'do_export'}) == -1) {
          # use caches
        } elsif (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
                and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
          # use caches
        } else {
          $use_resfile_cache = 0;
          $use_results_cache = 0;
        }
      } else {
        # Speed up command line.
        $use_resfile_cache = 0;
        $use_results_cache = 0;
      }
    } else {
      # use caches
    }
  }

# see whether to use ms_mascotresfile::RESFILE_USE_CACHE flag
  my $resfileFlags;
  if (common_subs::useResfileCache($objMascotOptions, $myName)) {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_USE_CACHE;
  } else {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_NOFLAG;
  }
  if (not $use_resfile_cache) {
    $resfileFlags &= ~$msparser::ms_mascotresfile::RESFILE_USE_CACHE;
  }
  my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;
  my $resfileWillCache = 0;

  $preferTaxonomy = 0;
  if (defined(my $p = $thisScript->param($urlParams{'_prefertaxonomy'}))) {
    if (looks_like_number($p) and $p > 0) {
      $preferTaxonomy = int($p);
    }
  }

  my $MIME_boundary = undef;

  if ($thisScript->param($urlParams{'generate_mime'})) {
    # We use custom functions to print MIME boundaries. The CGI module
    # provides only "server push" support.

    $MIME_boundary = common_subs::multipart_init(); # sets binmode()
    common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', 'keepalive');
  }

  # In non-multipart mode, redirect to caching here if any of the
  # resfile caches need to be created. In multipart mode, progress
  # messages are taken care of in openResultsCollectionSet().
  if (not defined($MIME_boundary)
      and ($resfileFlags & $msparser::ms_mascotresfile::RESFILE_USE_CACHE)) {
    for (@fileIn) {
      if ((msparser::ms_mascotresfile::willCreateCache(
          common_subs::decompress($_),
          $resfileFlags,
          $resfileCacheDirectory
      ))[0]) {
        maybeRedirectToCaching(1, 'build:resultsfile');
        #warn sprintf("will create resfile cache, flags = 0x%x\n", $resfileFlags);
        # No redirect, so set a flag signalling that caching will be done.
        $resfileWillCache = 1;
      }
    }
  }

# create Mascot Parser result file object
  $objResFile = common_subs::openResultsCollectionSet(
    $objMascotOptions,
    $resfileFlags,
    $resfileCacheDirectory,
    $idForFatalError,
    defined $MIME_boundary ? 1 : 0, # keepalive?
    \@fileIn,
    $MIME_boundary,
  );

# abort on fatal error. Note that isValid() remains true if there are only warnings
  unless ($objResFile->isValid) {
    my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die($errorList, " ", $idForFatalError);
  }

# create a Mascot Parser searchparams object
  $objParams = new msparser::ms_searchparams($objResFile);

# Security
  $sessionID = "" unless ($sessionID = $thisScript->param($urlParams{sessionid}));
  $session = new msparser::ms_session($sessionID);
  if ($session->isValid) {
    unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
      } else {
        print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
      }
      print common_subs::notAuthorised($session->getUserName);
      if (defined $MIME_boundary) {
        common_subs::multipart_final($MIME_boundary);
      }
      exit 1;
    }
    $sessionID = $session->getID();
  } else {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
    }
    &common_subs::printHeader1("Matrix Science - Not logged in");
    &common_subs::printHeader2("", "Not logged in", 1);
    &common_subs::go2login($thisScript->self_url, $session->getLastError, $session->getLastErrorString);
    &common_subs::printFooter;
    print "</BODY>\n";
    print "</HTML>\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 1;
  }

# validate a variety of mascot.dat and URL parameters
  &validateParams();

  if (@fileIn > 1) {
    if ($export_format eq 'CSV' or $export_format eq 'mzIdentML') {
      # OK
    } else {
      if (not defined $MIME_boundary) {
        print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
      }
      common_subs::fatal("export format $export_format is not supported for merged results", __LINE__, $myName, $idForFatalError, $MIME_boundary);
    }
  }

  # Required for redirecting support in maybeRedirectToCaching().
  set_option('report', $numHits);

# fixed settings for DTASelect and pepXML
  if ($export_format eq "DTASelect" or $export_format eq "pepXML") {
    $ShowSubSets = 1;
    $RequireBoldRed = 0;
    $group_family = 0;
    # Required for redirecting support in maybeRedirectToCaching().
    set_option('_showsubsets', $ShowSubSets);
    set_option('_requireboldred', $RequireBoldRed);
    set_option('group_family', $group_family);
  }

  my %params_for_flags;

  if ($group_family) {
    $RequireBoldRed = 0;
    set_option('_requireboldred', $RequireBoldRed);
  }

  if (common_subs::useResultsCache($objMascotOptions, $myName)) {
    $params_for_flags{'_usecache'} = 1;
  } else {
    $params_for_flags{'_usecache'} = 0;
  }

# Choose whether to use identity or homology threshold
  $use_homology = 0;
  if (defined($thisScript->param($urlParams{'use_homology'}))) {
    $use_homology = $thisScript->param($urlParams{'use_homology'}) + 0;
  }

# We need to check if the file has to be cached, in order to write out
# the correct form. So, we construct the flags here as they would be used
# when creating ms_mascotresfile and ms_proteinsummary/ms_peptidesummary.
  if ($reportType eq 'concise') {
    $params_for_flags{':override'}{'grouping'} = 'standard';
    $params_for_flags{':override'}{'scoring'} = 'standard';
    $params_for_flags{':override'}{'showsubsets'} = 1;
    if ($use_homology) {
      $params_for_flags{':override'}{'threshold'} = 'homology';
    }
    if ($objParams->getDECOY() > 0) {
      $params_for_flags{'_show_decoy_report'} = $ShowDecoyReport;
    }
  } else {
    # Note: don't use the ":override" hash in this branch. It's only meant
    # for legacy use.
    #
    # Also be careful about undefined values.
    # make_ms_mascotresults_params() assumes undef means "use default
    # from get_ms_mascotresults_params()", so make sure all parameters
    # set here are defined.
    $params_for_flags{'_showsubsets'} = $ShowSubSets;
    $params_for_flags{'group_family'} = $group_family;
    if (not $group_family) {
      $params_for_flags{'_requireboldred'} = $RequireBoldRed;
    }
    if ($use_homology) {
      $params_for_flags{'_threshold_type'} = 'homology';
    } else {
      $params_for_flags{'_threshold_type'} = 'identity';
    }
    if ($objResFile->isErrorTolerant()) {
      if ($objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      # new-style integrated ET search
        if ($OnlyErrorTolerant) {
          $params_for_flags{'et'} = 'onlyet';
        } elsif ($NoErrorTolerant) {
          $params_for_flags{'et'} = 'onlynormal';
        } else {
          $params_for_flags{'et'} = 'integrated';
        }
      }
      if ($ShowAllFromErrorTolerant) {
        $params_for_flags{'_showallfromerrortolerant'} = 1;
      }
    } elsif ($anyLibraryMatch) {
      $params_for_flags{'library'} = $libraryMode;
    }
    # BUG#13263: This same condition is used to initialise $$params{'_scoring'} in CachedResults::Build::PeptideSummary::new
    if ($anyLibraryMatch
            or ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch)) {
      $params_for_flags{'_scoring'} = 'mudpit';
    } else {
      $params_for_flags{'_scoring'} = 'standard';
    }
    if ($objParams->getDECOY() > 0) {
      $params_for_flags{'_show_decoy_report'} = $ShowDecoyReport;
    }
    if (MascotResults::can_percolate($objResFile)) {
      $params_for_flags{'percolate'} = $percolate;
      if ($percolate == 1) {
	$objMascotOptions->setPercolatorUseRT($percolate_rt);
        $objResFile->setPercolatorFeatures(
            $objMascotOptions,
            '',
        );
        if (MascotResults::no_percolator_files($objResFile)) {
          $mascotresultsNeedsPercolating = 1;
        }
      }
    }
  }

  do {
    # TODO: Use ms_mascotresults_params everywhere.
    my $pepsum_params = PeptideSummary::Util::make_ms_mascotresults_params($objResFile, %params_for_flags);
    $msresFlags = $pepsum_params->getFlags() unless ($msresFlags = $thisScript->param($urlParams{'_msresflags'}));
    $msresFlags2 = $pepsum_params->getFlags2() unless ($msresFlags2 = $thisScript->param($urlParams{'_msresflags2'}));

  };

  if ($export_format eq 'DTASelect') {
    $ShowSameSets = $msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS;
    # Required for redirecting support in maybeRedirectToCaching().
    set_option('show_same_sets', $ShowSameSets);
  }

# branch here if we require the form
  if (($thisScript->param($urlParams{'do_export'}) || 0) == 0) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => "utf-8", -expires => '-1d' );
    }
    &formatControls();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  } elsif ($thisScript->param($urlParams{'do_export'}) == -1) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => "utf-8", -expires => '-1d' );
    }
    &display_args();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "MascotDAT") {
  # Branch off to download raw Mascot result file
  # (we know that @fileIn == 1)
    exit mascotDAT($thisScript, $isCGI, $fileIn[0], $objMascotOptions, $MIME_boundary);
  }

  if ($export_format eq "MGF") {
  # Branch off to download MGF peak list
    exit exportMGF($thisScript, $MIME_boundary);
  }

# pepXML and DTASelect only valid for MS/MS
  if ($export_format eq "pepXML" && $reportType eq 'concise') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print "pepXML format is not valid for peptide mass fingerprint results ", $idForFatalError;
      common_subs::multipart_final($MIME_boundary);
    }
    die("pepXML format is not valid for peptide mass fingerprint results ", $idForFatalError);
  }
  if ($export_format eq "pepXML" and ExportDat::Util::db_type_is_mixed(ExportDat::Util::count_db_types($objResFile, $objParams))) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print "pepXML format is not valid for mixed databases or spectral libraries ", $idForFatalError;
      common_subs::multipart_final($MIME_boundary);
    }
    die("pepXML format is not valid for mixed databases or spectral libraries ", $idForFatalError);
  }
  if ($export_format eq "DTASelect" && $reportType eq 'concise') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print "DTASelect format is not valid for peptide mass fingerprint results ", $idForFatalError;
      common_subs::multipart_final($MIME_boundary);
    }
    die("DTASelect format is not valid for peptide mass fingerprint results ", $idForFatalError);
  }

# $objQuantMethod will encapsulate any quantitation method
# report can only handle multiplex and reporter protocols
# but may need $objQuantMethod for (say) mods and isotope shifts
  if ($objParams->getQUANTITATION()
      and lc($objParams->getQUANTITATION()) ne "none") {
    my $method = msparser::ms_quant_method->new();
    if ($objResFile->getQuantitationMethod($method)) {
      $quantProtocolType = uc($method->getProtocol->getType());
      if ($quantProtocolType eq "MULTIPLEX" or $quantProtocolType eq "REPORTER") {
        $objQuantMethod = $method;
        # default report settings to those in the quantitation method
        quant_subs::quantFormatOverRide($objQuantMethod, $thisScript, \%urlParams, \$mudpitSwitch, \$sigThreshold_dbl, \$ShowSubSets, \$RequireBoldRed);
        $sigThreshold_str = $sigThreshold_dbl;
        # read any URL arguments and create text for format controls
        # text is not used, but need to call the routine to set or over-ride defaults
        quant_subs::quantFormatControls($objQuantMethod, \$thisScript, \%urlParams, []);
        # warn if trying to use unsupported features
        quant_subs::quantWarnings($objQuantMethod, \@quantWarnings);
      }
    }
  }

  $quantProtocolType = '' if not defined $quantProtocolType;
  my $pepsumWillCache = 0;

  my $redirectIsAllowed = 0;

  if (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
          and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
    $redirectIsAllowed = 1;
  }

  do {
    if ($reportType ne 'concise' and $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_CACHE) {
      my $pepsum_params = msparser::ms_mascotresults_params->new(
          $msresFlags,
          $sigThreshold_dbl,
          $numHits,
          $UniGeneFile,
          $ignoreIonsScoreBelow,
          $minPepLen,
          "",
          $msresFlags2,
      );
      common_subs::setTargetFDRandTypeParam($pepsum_params, $targetFDR, $targetFDRtype);
      $pepsum_params->setMinNumSigUniqueSequences($minNumSigUniqueSeqs);
      # Bug 6529: Flag is set to true if target or decoy cache file is
      # missing and we didn't redirect.
      $pepsumWillCache = maybeRedirectIfPeptideSummaryCachingNeeded(
          $redirectIsAllowed, $objResFile, $pepsum_params,
          "regular pepsum cache", __LINE__
      );
      if (WARN_ABOUT_PEPSUM_PARAMS) {
          warn sprintf("will create %s (line %d):\n%s\n",
              "regular pepsum cache", __LINE__, join("\n",
                  PeptideSummary::Util::stringify_params($pepsum_params)
              ));
      }
    }
    if ($reportType ne 'concise' and $mascotresultsNeedsPercolating) {
      maybeRedirectToCaching($redirectIsAllowed, 'build:peptidesummary');
      # If the above function returns, it means we need to percolate right here
      # and now.
      _percolate($objResFile, 0, $percolate_rt, $MIME_boundary);
    }
  };
  if (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
      and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
    if (1) {
      if ($isCGI) {
        print $thisScript->header( -charset => "utf-8", -expires => '-1d' );
        downloadControls();
        exit 0;
      }
    }
  }

# if this is an old-style error tolerant search, may need to decompress the parent
  if ($objResFile->isErrorTolerant()) {
    if ($objParams->getErrTolParentFilename()) {
      &common_subs::decompress($objParams->getErrTolParentFilename());
    }
  # set value for $OldStyleErrTolReport
    if ($ShowAllFromErrorTolerant || $OnlyErrorTolerant) {
      $OldStyleErrTolReport = 1;
    } elsif (!$objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      $OldStyleErrTolReport = 1;
    } else {
      $OldStyleErrTolReport = 0;
    }
  } else {
    $OldStyleErrTolReport = 0;
  }

  # Bug 6529: Always export peptide and protein FDR if there are decoy
  # matches and this isn't a decoy report. In Mascot 2.6 and earlier,
  # peptide FDR was only exported if the user explicitly chose a
  # target FDR.
  if (($objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_DECOYPEPTIDES)
        or $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_ERRTOLDECOYPEPTIDES))
      and not ($msresFlags & $msparser::ms_mascotresults::MSRES_DECOY)) {
    $showFDRs = 1;
  }

# create Mascot Parser summary object of the required type
  if ($reportType eq 'concise') {
    $objSummary = new msparser::ms_proteinsummary(
      $objResFile,
      $msresFlags,
      $sigThreshold_dbl,
      $numHits
    );
  } else {
    # Bug 11436: If $objResFile has been opened with cache and
    # ms_peptidesummary is about to cache, use a non-cached version of
    # $objResFile. This way all the required data is in memory and caching
    # is much faster. This is the same technique used in nph-cache_families.pl.
    if ($resfileWillCache) {
      # We can use the already open object, because it was just cached and
      # holds all data still in memory. (Control flow only comes here in
      # non-CGI mode.)
    } elsif ($pepsumWillCache and ($resfileFlags & $msparser::ms_mascotresfile::RESFILE_USE_CACHE)) {
      my $nocache_flags = $resfileFlags;
      $nocache_flags &= ~$msparser::ms_mascotresfile::RESFILE_USE_CACHE;
      # Reassign $objResFile. We need to keep the correct object in memory,
      # as otherwise internal C++ pointers will break.
      $objResFile = common_subs::openResultsCollectionSet(
        $objMascotOptions,
        $nocache_flags,
        $resfileCacheDirectory,
        $idForFatalError,
        defined $MIME_boundary ? 1 : 0, # keepalive?
        \@fileIn,
        $MIME_boundary,
      );
      # Refresh related things:
      $objParams = new msparser::ms_searchparams($objResFile);
      if (MascotResults::can_percolate($objResFile)) {
        if ($percolate == 1) {
	  $objMascotOptions->setPercolatorUseRT($percolate_rt);
          $objResFile->setPercolatorFeatures(
              $objMascotOptions,
              '',
          );
        }
      }
      # Signal to target FDR branch that a non-cached resfile object is
      # in use.
      $resfileWillCache = 1;
    }
    my $pepsum_params = msparser::ms_mascotresults_params->new(
        $msresFlags,
        $sigThreshold_dbl,
        $numHits,
        $UniGeneFile,
        $ignoreIonsScoreBelow,
        $minPepLen,
        "",
        $msresFlags2,
    );
    common_subs::setTargetFDRandTypeParam($pepsum_params, $targetFDR, $targetFDRtype);
    $pepsum_params->setMinNumSigUniqueSequences($minNumSigUniqueSeqs);
    $objSummary = new msparser::ms_peptidesummary($objResFile, $pepsum_params);
    $objSummary->setSubsetsThreshold($ShowSubSets);

    if (WARN_ABOUT_PEPSUM_PARAMS) {
        warn sprintf("will create %s (line %d):\n%s\n",
            "regular pepsum cache", __LINE__, join("\n",
                PeptideSummary::Util::stringify_params($pepsum_params)
            ));
    }

    if ($showFDRs) {
      # Bug 6529: Decoy summary needed for protein FDR calculation.
      my $decoy_params = msparser::ms_mascotresults_params->new($pepsum_params);
      $decoy_params->setFlag($msparser::ms_mascotresults::MSRES_DECOY, 1);
      # Bug 13372: We must reset keepalive here to display progress
      # messages for the decoy summary, then reset it back to its
      # original value in case the same resfile object is used again
      # with a new ms_peptidesummary. (For example, when determining
      # target FDR.)
      common_subs::resetResultsCollectionSetKeepAlive(
        $objResFile,
        defined $MIME_boundary ? 1 : 0, # keepalive?
        'Protein FDR',
      );
      $objSummaryDecoy = new msparser::ms_peptidesummary($objResFile, $decoy_params);
      common_subs::resetResultsCollectionSetKeepAlive(
        $objResFile,
        defined $MIME_boundary ? 1 : 0, # keepalive?
      );
    }
  }

# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objResFile->isValid) {
    my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die($errorList, " ", $idForFatalError);
  }

  if ($objQuantMethod
      and ($thisScript->param($urlParams{'pep_quant'}) or $thisScript->param($urlParams{'prot_quant'}))) {
    $objMS2Quant = msparser::ms_ms2quantitation->new($objSummary, $objQuantMethod);

    if (not $objMS2Quant->isValid) {
      my $errorList = common_subs::checkErrorHandler($objMS2Quant);
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary);
        print $errorList;
        common_subs::multipart_final($MIME_boundary);
      }
      die $errorList;
    }

    if ($objMS2Quant->hasIntensityNormalisation) {
      $objMS2Quant->normaliseIntensities();
    }
    if ($objMS2Quant->hasPeptideRatioNormalisation) {
      $objMS2Quant->normalisePeptideRatios();
    }

    $objQuantMethod = $objMS2Quant->getQuantitationMethod();
  }

# create an ms_umod_configfile object from the result file unimod section
  do {
    my $m = new msparser::ms_umod_configfile;
    if ($objResFile->getUnimod($m)) {
      $objUmodSection = $m;
    }
  };

  do {
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
    $objFullUmodFile = new msparser::ms_umod_configfile("../config/unimod.xml",
      "../html/xmlns/schema/unimod_2/unimod_2.xsd", $cs);
  };

# If the results file is very old, it won't have a Unimod section. Let's
# set $objUmodSection to a valid Unimod file object and use it for lookup
# where the distinction between the Unimod section and the full Unimod file
# doesn't matter. See library mod handling in mzIdentML for where the
# difference does matter.
  $objUmodSection //= $objFullUmodFile;

# add in any modifications locally defined in any quant method
  if ($objQuantMethod) {
    for (my $i = 0; $i < $objQuantMethod->getNumberOfModificationGroups(); $i++) {
      my $objModGroup = $objQuantMethod->getModificationGroupByNumber($i);
      for (my $j = 0; $j < $objModGroup->getNumberOfLocalDefinitions(); $j++) {
        quant_subs::includeLocalDef($objModGroup->getLocalDefinition($j), $objUmodSection);
      }
    }
    for (my $i = 0; $i < $objQuantMethod->getNumberOfComponents(); $i++) {
      my $objComponent = $objQuantMethod->getComponentByNumber($i);
      for (my $j = 0; $j < $objComponent->getNumberOfModificationGroups(); $j++) {
        my $objModGroup = $objComponent->getModificationGroupByNumber($j);
        for (my $k = 0; $k < $objModGroup->getNumberOfLocalDefinitions(); $k++) {
          quant_subs::includeLocalDef($objModGroup->getLocalDefinition($k), $objUmodSection);
        }
      }
    }
  }

# use this to create an ms_masses object
  $objMassesFile = new msparser::ms_masses($objUmodSection);

# create an ms_modfile object from the ms_umod_configfile object
# this now has all the mods used in the search
  $objModFile = new msparser::ms_modfile($objUmodSection, $msparser::ms_umod_configfile::MODFILE_FLAGS_ALL);
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objModFile->isValid) {
    my $errorList = common_subs::checkErrorHandler($objModFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die $errorList;
  }

# Create a ms_modfile from the unimod_xl section
if ($anyXlPepSumMatch) {
  my $xlM = new msparser::ms_umod_configfile;
  if ($objResFile->getUnimodXL($xlM)) {
    $objUmodXlSection = $xlM;
  }
  $objXlModFile = new msparser::ms_modfile($objUmodXlSection, $msparser::ms_umod_configfile::MODFILE_FLAGS_ALL);
  unless ($objXlModFile->isValid) {
    my $errorList = common_subs::checkErrorHandler($objXlModFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die $errorList;
  }
}

# msparser MASS_TYPE
$mass_type = $msparser::MASS_TYPE_MONO;
if (lc($objParams->getMASS()) eq "average") {
    $mass_type = $msparser::MASS_TYPE_AVE;
}

  if ($UniGeneFile) {
    $objUniGene = msparser::ms_unigene->new($objResFile, $UniGeneFile);

    if (not $objResFile->isValid) {
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary);
        print "Cannot create ms_unigene object: ".$objResFile->getLastErrorString();
        common_subs::multipart_final($MIME_boundary);
      }
      die "Cannot create ms_unigene object: ".$objResFile->getLastErrorString();
    }
  }

# Create hash for masses and variable mods
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_MASSES, $i)) {
    if ($nextKey =~ /^delta(\d+)/i) {
      my $letter = $1;
      if ($letter > 9) {
        $letter = chr($1+55);
      }
      ($vmMass{$letter}, $vmString{$letter}) =
        split(/,/, $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey), 2);
    } elsif ($nextKey =~ /^Ignore(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedMod(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedModResidues(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedModNeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)_master$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)_slave$/i) {
    } elsif ($nextKey =~ /^NeutralLoss\d_/i) {
    } elsif ($nextKey =~ /^ReqPepNeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^PepNeutralLoss(\d+)$/i) {
    } else {
      my $m = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey);
      if (length($nextKey) == 1 or $nextKey =~ /_term/i) {
        $groupMasses{lc($nextKey)} = $m;
      } else {
        $elementalMasses{lc($nextKey)} = $m;
      }
    }
    $i++;
  }
# result files earlier than 1.9 don't list electron mass
  $elementalMasses{electron} = 0.000549 unless $elementalMasses{electron};

# Bug 12185: Type of source database can be determined fast if we know we have
# only AA or only NA databases.
  do {
    my %num_db_types = ExportDat::Util::count_db_types($objResFile, $objParams);
    my $n = sum(values %num_db_types) || 0;
    $db_type_query_restrictions{'only_AA_databases'} = ($num_db_types{$msparser::AA} || 0) == $n;
    $db_type_query_restrictions{'only_NA_databases'} = ($num_db_types{$msparser::NA} || 0) == $n;
  };

  # Bug 13542: If results file is from 2.7 or earlier, or bug 12447
  # isn't fixed, use old behaviour: don't show expect values.
  if ($objResFile->isErrorTolerant and $objResFile->versionGreaterOrEqual(2, 7, 101)
        and $objSummary->isDataCached($msparser::ms_peptidesummary::BUGFIX_12447)) {
    $OmitErrTolThresholds = 0;
  } else {
    $OmitErrTolThresholds = 1;
  }

# parse out result file name root
  ($fileNameRoot) = $fileIn[0] =~ /([^\/\\]*)\.dat$/;
  if (@fileIn > 1) {
    $fileNameRoot .= '_merged';
  }

  my $cache_every_protein = 0;

  if ($export_format eq 'pepXML') {
      $cache_every_protein = 1;
  } elsif ($export_format eq 'mzIdentML') {
      $cache_every_protein = $thisScript->param($urlParams{'query_all'});
  }

  do {

    my $do_fetch = 0;
    if ($export_format eq 'XML' or $export_format eq 'CSV' or $export_format eq 'mzTab') {
      # Only XML, CSV and mzTab support emPAI.
      if ($objSummary->anyEmPAI
           and ($export_format eq 'mzTab' or $thisScript->param($urlParams{'peptide_master'}))
           and $thisScript->param($urlParams{'prot_empai'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_mass'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_cover'})) {
        $do_fetch = 1;
      }
    } else {
      if ($thisScript->param($urlParams{'prot_mass'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_cover'})) {
        $do_fetch = 1;
      }
    }

    my $prot_len = $thisScript->param($urlParams{'prot_len'});
    my $prot_pi = $thisScript->param($urlParams{'prot_pi'});
    my $prot_seq = undef;

    if ($shipper eq 'TRUE') {
      $prot_seq = $thisScript->param($urlParams{'prot_seq'});
    }

    if ($export_format eq "FASTA") {
        $prot_seq = 1;
    }

    my $keepAlive = (defined $MIME_boundary ? 1 : 0);
    my $needProtDescCache = $thisScript->param($urlParams{'prot_desc'});
    my $needGetSeqCache = ($prot_len or $prot_pi or $prot_seq);
    my $includeFrame = ($export_format ne "FASTA");
    my $protTaxId = $thisScript->param($urlParams{'prot_tax_id'});
    my $protTaxStr = $thisScript->param($urlParams{'prot_tax_str'});
    $protData = ExportDat::ProteinData->new(objSummary => $objSummary,
					    objResFile => $objResFile,
					    objDatFile => $objDatFile,
					    objUniGene => $objUniGene,
					    sessionID => $sessionID,
					    ShowSameSets => $ShowSameSets,
					    ShowSubSets => $ShowSubSets,
					    keepalive => $keepAlive,
					    include_every_protein => $cache_every_protein,
					    MIME_boundary => $MIME_boundary,
					    objParams => $objParams,
					    fastaTitles => undef,
					    populated => undef,
					    shipper => $shipper,

					    do_fetch => $do_fetch,
					    prot_len => $prot_len,
					    prot_pi => $prot_pi,
					    prot_seq => $prot_seq,
					    needProteinDescriptionCache => $needProtDescCache,
					    needGetSeqCache => $needGetSeqCache,
					    includeFrame => $includeFrame, # If NA Db, no need of translation frame.

					    preferTaxonomy => $preferTaxonomy,
					    prot_tax_id => $protTaxId,
					    prot_tax_str => $protTaxStr,
      );
  };

  # We need to stop emitting keepalive from here on in order to avoid
  # corrupting the output data.
  $objResFile->resetKeepAlive(0, "");

  if (defined $fastaOutputType) {
      $fileNameRoot .= ($fastaOutputType == $msparser::NA) ? "_NA" : "_AA";
  }

  if (defined $MIME_boundary) {
    common_subs::multipart_end($MIME_boundary);
    if ($export_format eq "XML" || $export_format eq "pepXML") {
      common_subs::multipart_start($MIME_boundary, "text/xml; charset=utf-8", "$fileNameRoot.xml");
    } elsif ($export_format eq "mzIdentML") {
      common_subs::multipart_start($MIME_boundary, "text/xml; charset=utf-8", "$fileNameRoot.mzid");
    } elsif ($export_format eq "mzTab") {
      common_subs::multipart_start($MIME_boundary, "text/plain; charset=utf-8", "$fileNameRoot.mztab");
    } elsif ($export_format eq "CSV") {
      common_subs::multipart_start($MIME_boundary, "application/vnd.ms-excel; charset=utf-8", "$fileNameRoot.csv");
    } elsif ($export_format eq "DTASelect") {
	common_subs::multipart_start($MIME_boundary, "text/plain; charset=utf-8", "DTASelect.txt");
    } elsif ($export_format eq "FASTA") {
	common_subs::multipart_start($MIME_boundary, "text/plain; charset=utf-8", "$fileNameRoot.fasta");
    } elsif ($export_format eq "xiVIEW-CSV") {
	common_subs::multipart_start($MIME_boundary, "application/vnd.ms-excel;; charset=utf-8", "$fileNameRoot.csv");
    } else {
      common_subs::multipart_start($MIME_boundary);
    }
  } elsif ($isCGI) {
    if ($export_format eq "XML" || $export_format eq "pepXML") {
      print $thisScript->header( -type => "text/xml",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.xml");
    } elsif ($export_format eq "mzIdentML") {
      print $thisScript->header( -type => "text/xml",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.mzid");
    } elsif ($export_format eq "mzTab") {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.mztab");
    } elsif ($export_format eq "CSV") {
      print $thisScript->header( -type => "application/vnd.ms-excel",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.csv");
    } elsif ($export_format eq "DTASelect") {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "DTASelect.txt");
    } elsif ($export_format eq "FASTA") {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.fasta");
    } elsif ($export_format eq "xiVIEW-CSV") {
      print $thisScript->header( -type => "application/vnd.ms-excel",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.csv");
    } else {
      print $thisScript->header( -charset => "utf-8" );
    }
  }

  if ($isCGI) {
    $URI = $thisScript->url(-path_info=>1);
    ($URI) = $URI =~ /^(.*\/).+?\/$myName/;
  } else {
    $URI = "../";
  }

# print document opening tags
  if ($export_format eq "XML") {
    print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print "<mascot_search_results xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
    print " xmlns=\"http://www.matrixscience.com/xmlns/schema/mascot_search_results_2\"";
    print " xsi:schemaLocation=\"http://www.matrixscience.com/xmlns/schema/mascot_search_results_2 ";
    if ($isCGI) {
      print $URI;
    } else {
      print "http://www.matrixscience.com/";
    }
    print "xmlns/schema/mascot_search_results_2/mascot_search_results_2.xsd\"";
    print "  majorVersion=\"2\" minorVersion=\"8\">\n";
  } elsif ($export_format eq "CSV") {
    # nothing
  } elsif ($export_format eq "pepXML") {
    print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print "<msms_pipeline_analysis";
    print " xmlns=\"http://regis-web.systemsbiology.net/pepXML\"";
    print " xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
    print " xsi:schemaLocation=\"http://regis-web.systemsbiology.net/pepXML ";
    if ($isCGI) {
      print $URI;
    } else {
      print "http://www.matrixscience.com/";
    }
    print "xmlns/schema/pepXML_v18/pepXML_v18.xsd\"";
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = gmtime(time);
    my $timestamp = sprintf("%04d-%02d-%02dT%02d:%02d:%02dZ",
      $year+1900, $mon+1, $mday, $hour, $min, $sec);
    # optional name
    print " date=\"$timestamp\"";
    print " summary_xml=\"$fileNameRoot.xml\">\n";
  } elsif ($export_format eq "mzIdentML") {
    # nothing here
  } elsif ($export_format eq "DTASelect") {
    # nothing
  } elsif ($export_format eq 'mzTab') {
    # nothing
  } elsif ($export_format eq 'FASTA') {
    # nothing
  } elsif ($export_format eq 'xiVIEW-CSV') {
    # nothing
  } else {
    my $HTMLtitle = "\n\n<HTML>\n<HEAD>\n<LINK REL=\"shortcut icon\" HREF=\"../favicon.ico\">\n<TITLE>$export_format Report (";
    if (@fileIn == 1) {
      if ($objParams->getCOM()) {
        $HTMLtitle .= &noXmlTag($objParams->getCOM());
      } else {
        $HTMLtitle .= $fileIn[0];
      }
    } else {
        $HTMLtitle .= sprintf('results collection set of %d files', scalar(@fileIn));
    }
    $HTMLtitle .= ")</TITLE>\n";
    print $HTMLtitle;
    print "</HEAD>\n\n";
    print "<BODY>\n";
    print "<PRE>\n";
  }

  if ($export_format eq "pepXML") {
  # Branch off for ISB pepXML
    pepXML();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "mzIdentML") {
    # Branch off for PSI mzIdentML
    my $mzIdVer = "1.1";
    if (defined($thisScript->param($urlParams{'mzid_version'}))) {
      $mzIdVer = $thisScript->param($urlParams{'mzid_version'});
    }

    my %params = (
      objResFile => $objResFile,
      objModFile => $objModFile,
      objXlModFile => $objXlModFile,
      objQuantMethod => $objQuantMethod,
      objUmodSection => $objUmodSection,
      objUmodXlSection => $objUmodXlSection,
      objMassesFile => $objMassesFile,
      urlParams => \%urlParams,
      libraryMode => $libraryMode,
      thisScript => $thisScript,
      objUniGene => $objUniGene,
      ShowSameSets => $ShowSameSets,
      ShowSubSets => $ShowSubSets,
      objFullUmodFile => $objFullUmodFile,
      vmMass => \%vmMass,
      vmString => \%vmString,
      objParams => $objParams,
      objSummary => $objSummary,
      numHits => $numHits,
      ignoreIonsScoreBelow => $ignoreIonsScoreBelow,
      use_homology => $use_homology,
      group_family => $group_family,
      minNumSigUniqueSeqs => $minNumSigUniqueSeqs,
      mudpitSwitch => $mudpitSwitch,
      UniGeneFile => $UniGeneFile,
      RequireBoldRed => $RequireBoldRed,
      objSummaryDecoy => $objSummaryDecoy,
      NoErrorTolerant => $NoErrorTolerant,
      ShowDecoyReport => $ShowDecoyReport,
      preferTaxonomy => $preferTaxonomy,
      percolate => $percolate,
      percolate_rt => $percolate_rt,
      showFDRs => $showFDRs,
      taxNames => \%taxNames,
      fileIn => \@fileIn,
      sessionID => $sessionID,
      msresFlags2 => $msresFlags2,
      elementalMasses => \%elementalMasses,
      massDP => $massDP,
      reportType => $reportType,
      shipper => $shipper,
      mass_type => $mass_type,
      anyLibraryMatch => $anyLibraryMatch,
      objMascotOptions => $objMascotOptions,
      proteinData => $protData
      );

    my $obj = ExportDat::MzIdentMLFactory->new(version => $mzIdVer)->create(%params);
    eval { $obj->print($isCGI, $URI, \*STDOUT) };

    my $err;
    if ($@) {
      $err = $@;
    }
    if ($err) {
      common_subs::fatal($err, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "FASTA") {
  # Branch off for FASTA
    eval { fasta() };
    if ($@) {
      common_subs::fatal($@, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "xiVIEW-CSV") {
  # Branch off for xiVIEW
    eval { xiVIEW_CSV() };
    if ($@) {
      common_subs::fatal($@, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "mzTab") {
  # Branch off for PSI mzTab
    eval { mzTab() };
    if ($@) {
      common_subs::fatal($@, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "DTASelect") {
  # Branch off for Scripps DTASelect
    DTASelect();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

# search information
  if ($thisScript->param($urlParams{'search_master'})) {
    if ($thisScript->param($urlParams{'show_header'})) {
      outputHeader();
    }
    if ($thisScript->param($urlParams{'show_decoy'})) {
      outputDecoy();
    }
    if ($thisScript->param($urlParams{'show_mods'})) {
      if ($objParams->getMODS || $objParams->getFixedModsName(1)) {
        outputFixedMods();
      }
      if ($objParams->getIT_MODS || $objParams->getVarModsName(1)) {
	outputVarMods();
	if ($hasXlMethod) {
	  outputLinkers();
	}
      }
    }
    if ($thisScript->param($urlParams{'show_params'})) {
      outputParams();
    }
    if ($thisScript->param($urlParams{'show_format'})) {
      outputFormat();
    }
    if ($thisScript->param($urlParams{'show_masses'})) {
      outputMasses();
    }
    if ($thisScript->param($urlParams{'show_quant_header'})) {
      outputQuantitationHeader();
    }
  }

  if ($thisScript->param($urlParams{'protein_master'})) {
    my $filter_duplicates = 1;
    if ($thisScript->param($urlParams{'show_pep_dupes'})) {
      $filter_duplicates = 0;
    }

    if ($export_format eq "XML") {
      print "<hits>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Protein hits\"$delimiter\"--------------------------------------------------------\"\n\n";
    }

  # print column headers for CSV format
    if ($export_format eq "CSV") {
      my $string = "";
      foreach my $field (@columns) {
        if ($field =~ /^pep/ && !$thisScript->param($urlParams{'peptide_master'})) {
          next;
        }
        if ($thisScript->param($urlParams{lc($field)})) {
	  if ($field eq 'pep_isunique') {
	    $string .= 'pep_isunique' . $delimiter;
	    if ($anyXlPepSumMatch) {
	      $string .= 'prot_acc_alpha' . $delimiter . 'prot_acc_beta' . $delimiter;
	    }
	  }
	  elsif ($field eq 'pep_start') {
	    $string .= 'pep_start' . $delimiter;
	    if ($anyXlPepSumMatch) {
	      $string .= 'pep_start_beta' . $delimiter;
	    }
	  }
	  elsif ($field eq 'pep_end') {
	    $string .= 'pep_end' . $delimiter;
	    if ($anyXlPepSumMatch) {
	      $string .= 'pep_end_beta' . $delimiter;
            }
	  }
	  elsif ($field eq "pep_seq") {
	    $string .= "pep_res_before" . $delimiter . "pep_seq" . $delimiter . "pep_res_after" . $delimiter;
	    if ($anyXlPepSumMatch) {
		$string .= "pep_res_before_beta" . $delimiter . "pep_seq_beta" . $delimiter . "pep_res_after_beta" . $delimiter;
	    }
          } elsif ($field eq "pep_var_mod") {
            $string .= "pep_var_mod" . $delimiter . "pep_var_mod_pos" . $delimiter;
            $string .=  "pep_summed_mod_pos" . $delimiter
              . "pep_local_mod_pos" . $delimiter
              # pep_var_mod_conf is not included here; it's query-level data
		;
	    if ($hasXlMethod) {
	      $string .=  "pep_monolink_mod_pos" . $delimiter;
	    }
	    if ($anyXlPepSumMatch) {
	      $string .= "pep_var_mod_beta" . $delimiter . "pep_var_mod_pos_beta" . $delimiter;
	      $string .=  "pep_summed_mod_pos_beta" . $delimiter
              . "pep_local_mod_pos_beta" . $delimiter
              # pep_var_mod_conf is not included here; it's query-level data
		  ;
	      $string .= "pep_mono_link_mod_pos_beta" . $delimiter;
	    }

	    if ($hasXlMethod) {
	      $string .=  "pep_looplink_pos" . $delimiter;
	    }
	    if ($anyXlPepSumMatch) {
	      $string .= "pep_looplink_pos_beta" . $delimiter;
	      $string .= "pep_linked_sites" . $delimiter;
	    }
          } elsif ($field eq "prot_matches" && $anyPepSumMatch) {
            $string .= $field . $delimiter . "prot_matches_sig" . $delimiter . "prot_sequences" . $delimiter . "prot_sequences_sig" . $delimiter;
          } elsif ($field eq "prot_hit_num" && $group_family && $anyPepSumMatch) {
            $string .= $field . $delimiter . "prot_family_member" . $delimiter;
          } else {
            $string .= $field . $delimiter;
          }
        }
      }
      if ($thisScript->param($urlParams{'pep_quant'})
          || $thisScript->param($urlParams{'prot_quant'})) {
        if ($objQuantMethod) {
        } else {
          # Bug 11992: Output header for component (quant_pep_component).
          $string .= 'quant_pep_component' . $delimiter;
        }
      }
      if ($string) {
        chop $string;
      }
      print $string . "\n";
    }

  sub _mk_full_prot_acc {
    my ($objProtein) = @_;
    my $dbNumPrefix = "";
    if ($objParams->getNumberOfDatabases > 1 && !$objProtein->isPMFMixture) {
      $dbNumPrefix = $objProtein->getDB() . "::";
    }
    my $preferredAcc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession);
    return $dbNumPrefix . $preferredAcc;
  }

  # start of MAIN LOOP to enumerate the protein hits
    my $sameset_ordering = $protData->sameset_ordering;
    my %dupe_state_cache = ();
    my $thisHit = 1;
    $objProtein = $objSummary->getHit($thisHit);
    while (defined($objProtein)) {
    # export the hit
      outputHit($objProtein, $thisHit, $$sameset_ordering{$thisHit}, $filter_duplicates, \%dupe_state_cache);
    # and get the next primary hit
      $thisHit++;
      $objProtein = $objSummary->getHit($thisHit);
    }
  # end of MAIN LOOP to enumerate the protein hits

    if ($export_format eq "XML") {
      print "</hits>\n";
    }

  }

  if ($thisScript->param($urlParams{'show_unassigned'})) {
    outputUnassigned();
  }

  if ($thisScript->param($urlParams{'query_master'})) {
    outputQueries();
  }

# print document closing tags
  if ($export_format eq "XML") {
    print "</mascot_search_results>\n";
  } elsif ($export_format eq "CSV") {
    # none
  } else {
    print "</PRE>\n";
    print "</BODY>\n";
    print "</HTML>\n";
  }

  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }

# Good Night
  exit 0;


###############################################################################
# &validateParams()
# validate a variety of mascot.dat and URL parameters
# globals:
# my($massDP, $minPepLen, $minNumSigUniqueSeqs, $ignoreIonsScoreBelow, $reportType,
# $objParams, $objResFile, $thisScript, %urlParams, $sigThreshold_dbl,
# $sigThreshold_str,
# $targetFDR, $targetFDRtype, $numHits, $ShowSubSets, $RequireBoldRed,
# $mudpitSwitch, $ShowAllFromErrorTolerant, $UniGeneFile, @fileIn,
# $idForFatalError);
###############################################################################

sub validateParams {

  my $tempString;

# $massDP determines precision on peptide mass values
# default is 2 decimal places for peptides
  $massDP = 0 unless ($massDP = $objMascotOptions->getMassDecimalPlaces);
  $massDP = int($massDP);
  if ($massDP < 1 || $massDP > 6) {
    $massDP = 2;
  }

# peptides shorter than $minPepLen are not interesting
# default is to ignore anything shorter than 6 residues
  if (defined($thisScript->param($urlParams{'_minpeplen'}))
    && $thisScript->param($urlParams{'_minpeplen'}) >= 0) {
    $minPepLen = $thisScript->param($urlParams{'_minpeplen'}) + 0;
  } else {
    $minPepLen = 0 unless ($minPepLen = $objMascotOptions->getMinPepLenInPepSummary );
    $minPepLen = int($minPepLen);
    if ($minPepLen < 1 || $minPepLen > 9) {
      $minPepLen = 6;
    }
  }

# Bug 6529: Allow filtering out protein hits with less than specified
# number of significant sequences. It makes no difference to Peptide
# View, but we do need to use the correct constructor parameter.
  if (defined($thisScript->param($urlParams{'min_num_sig_unique_seqs'}))
    && $thisScript->param($urlParams{'min_num_sig_unique_seqs'}) >= 0) {
    $minNumSigUniqueSeqs = $thisScript->param($urlParams{'min_num_sig_unique_seqs'}) + 0;
  } else {
    $minNumSigUniqueSeqs = 1;
  }

# ignoreIonsScoreBelow is a threshold on MS/MS ions scores
# a value of -1 ultimately sets it to _sigthreshold (bug 12007)
# a value of 0 specifies that all peptides are included (default)
# a value greater than 1 indicates an absolute score cutoff
# a value less than 1 is treated as a probability threshold
  $ignoreIonsScoreBelowDefined = 0;
  if (defined($thisScript->param($urlParams{'_ignoreionsscorebelow'}))) {
    $ignoreIonsScoreBelowDefined = 1;
    if ($thisScript->param($urlParams{'_ignoreionsscorebelow'}) < 0) {
      $ignoreIonsScoreBelow = -1;
    } else {
      $ignoreIonsScoreBelow = $thisScript->param($urlParams{'_ignoreionsscorebelow'}) + 0;
    }
  } else {
    my $d = $objMascotOptions->getDisplayNonSignificantMatches;
    if ($d == 2) {
      $ignoreIonsScoreBelow = 0 unless ($ignoreIonsScoreBelow = $objMascotOptions->getIgnoreIonsScoreBelow);
      if ($ignoreIonsScoreBelow < 0) {
        $ignoreIonsScoreBelow = 0;
      }
    } elsif ($d == 1) {
      $ignoreIonsScoreBelow = 0;
    } elsif ($d == 0) {
      $ignoreIonsScoreBelow = -1;
    }
  }

# if queries / entries greater than $mudpitSwitch, default 0.001, switch to Mudpit scoring
  if (defined($thisScript->param($urlParams{'_server_mudpit_switch'}))
    && $thisScript->param($urlParams{'_server_mudpit_switch'}) > 0) {
    $mudpitSwitch = $thisScript->param($urlParams{'_server_mudpit_switch'}) + 0;
  } else {
    $mudpitSwitch = 0.001;
    my $tempString;
    $tempString = 0 unless ($tempString = $objMascotOptions->getMudpitSwitch);
    if ($tempString > 0) {
      $mudpitSwitch = $tempString;
    }
  }

# If generating mgf file, set the output order (input index or query)
  if (defined($thisScript->param($urlParams{'mgf_output_order'}))
    && $thisScript->param($urlParams{'mgf_output_order'}) > 0) {
    $mgfOutputOrder = $thisScript->param($urlParams{'mgf_output_order'}) + 0;
  } else {
      $mgfOutputOrder = 0;
  }

# If generating FASTA
  if (defined($thisScript->param($urlParams{'fasta_output_type'}))) {
      $fastaOutputType = $thisScript->param($urlParams{'fasta_output_type'}) + 0 == 1 ? $msparser::NA : $msparser::AA;
  }

# If $ShowDecoyReport is true, a report for the integrated decoy search is shown
  if (defined($thisScript->param($urlParams{'_show_decoy_report'}))) {
    $ShowDecoyReport = $thisScript->param($urlParams{'_show_decoy_report'}) + 0;
  } else {
    $ShowDecoyReport = 0;
  }

# set $section parameters to point to the correct sections
  $sec_peptides = $msparser::ms_mascotresfile::SEC_PEPTIDES;
  $sec_summary = $msparser::ms_mascotresfile::SEC_SUMMARY;
  if ($objParams->getDECOY() > 0
    && $ShowDecoyReport) {
    $sec_peptides = $msparser::ms_mascotresfile::SEC_DECOYPEPTIDES;
    $sec_summary = $msparser::ms_mascotresfile::SEC_DECOYSUMMARY;
  }
# $objResFile->anyPeptideSummaryMatches($sec_peptides) is used constantly
  $anyPepSumMatch = $objResFile->anyPeptideSummaryMatches($sec_peptides);
  $anyXlPepSumMatch = $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_CROSSLINK_PEPTIDES);
  $hasXlMethod = (length($objParams->getCROSSLINKING()) and ("None" ne $objParams->getCROSSLINKING()));
  # We still  want to use ms_peptidesummary if crosslinked results only,
  $anyPepSumMatch = $anyXlPepSumMatch ? 1 : $anyPepSumMatch;
  $anyLibraryMatch = $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_LIBRARYPEPTIDES);

# set $sigThreshold_dbl to probability significance threshold, default 0.05
  my $SL_only_mode = $anyLibraryMatch && !$anyPepSumMatch;
  $sigThreshold_dbl = $sigThreshold_str = common_subs::validate_sigThreshold(
    scalar $thisScript->param($urlParams{'_sigthreshold'}),
    $objMascotOptions,
    $SL_only_mode,
  );
  if ($SL_only_mode) {
    if ($sigThreshold_dbl >= 1) {
      $sigThreshold_dbl = msparser::ms_peptidesummary::getMinProbabilityForSLScore($sigThreshold_dbl);
    } else {
      # Bug 12834: Display the threshold as library score.
      $sigThreshold_str = msparser::ms_peptidesummary::getSLThresholdFromMinProbability($sigThreshold_dbl);
    }
  } elsif ($sigThreshold_dbl >= 1) {
    # Shouldn't happen, because it's taken care of in the validator.
    $sigThreshold_str = $sigThreshold_dbl = 1 - 1e-9;
  }

# Bug 13750: Target FDR and type parameters
  $targetFDR = 0;
  if (defined($thisScript->param($urlParams{'_target_fdr'}))) {
    my $str = $thisScript->param($urlParams{'_target_fdr'});
    if ($str ne '' and 0 < $str and $str < 100) {
      $targetFDR = 0 + $str;
    }
  }
  $targetFDRtype = 'psm';
  if (defined($thisScript->param($urlParams{'_fdr_count_type'}))) {
    my $str = lc($thisScript->param($urlParams{'_fdr_count_type'}));
    if ($str eq 'seq' or $str eq 'psm') {
      $targetFDRtype = $str;
    }
  }

  $libraryMode = 'onlyprimary';
  if ($anyLibraryMatch) {
    $libraryMode = 'integrated';
    if (defined(my $val = $thisScript->param($urlParams{'library'}))) {
      if ($val eq 'integrated' or $val eq 'onlyprimary' or $val eq 'onlylibrary') {
        $libraryMode = $val;
      }
    }
    # Also force $anyPepSumMatch to true, even if we don't have a peptides
    # section. The flag is really about whether to use ms_peptidesummary
    # than whether there are FASTA matches.
    $anyPepSumMatch = 1;
  }

# Report type is either peptide or concise
  $reportType = 'concise';
  # BUG#13263: This same condition is used to initialise $reportType CachedResults::Build::PeptideSummary::new
  if ($anyPepSumMatch) {
    $reportType = 'peptide';
  }

# set number of hits to display & make sure it is reasonable
# if the value of REPORT is 'AUTO', $numHits is set to 0
  $numHits = 0;
  if (defined($thisScript->param($urlParams{'report'}))) {
    if ($thisScript->param($urlParams{'report'}) =~ /auto/i) {
      $numHits = 0;
    } else {
      $numHits = int($thisScript->param($urlParams{'report'}));
    }
  } else {
    $numHits = int($objParams->getREPORT());
  }
  if ($numHits < 0){
    $numHits = 0;
  }

# Display same-set proteins $ShowSameSets is true
  $ShowSameSets = 0;
  if (defined($thisScript->param($urlParams{'show_same_sets'}))) {
    $ShowSameSets = $thisScript->param($urlParams{'show_same_sets'}) + 0;
  }

# Display a sub-set protein hit if it is missing less than $ShowSubSets fraction of the protein score of the main hit
  if (defined($thisScript->param($urlParams{'_showsubsets'}))) {
    $ShowSubSets = $thisScript->param($urlParams{'_showsubsets'}) + 0;
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $ShowSubSets = 0;
  } else {
    $ShowSubSets = 0 unless ($ShowSubSets = $objMascotOptions->getShowSubsets);
  }

# If RequireBoldRed is true, only display protein hits that include at least one bold, red peptide match
  if (defined($thisScript->param($urlParams{'_requireboldred'}))) {
    $RequireBoldRed = $thisScript->param($urlParams{'_requireboldred'}) + 0;
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $RequireBoldRed = 0;
  } else {
    $RequireBoldRed = 0 unless ($RequireBoldRed = $objMascotOptions->isRequireBoldRed);
  }

# If $ShowAllFromErrorTolerant is true, all hits from an error tolerant search are shown
  if (defined(my $all = $thisScript->param($urlParams{'_showallfromerrortolerant'}))) {
    $ShowAllFromErrorTolerant = looks_like_number($all) ? $all + 0 : 0;
  } else {
    $ShowAllFromErrorTolerant = 0;
  }
# If $OnlyErrorTolerant is true, display only error tolerant matches from a new style error tolerant search
  if (defined(my $only = $thisScript->param($urlParams{'_onlyerrortolerant'}))) {
    $OnlyErrorTolerant = looks_like_number($only) ? $only + 0 : 0;
  } else {
    $OnlyErrorTolerant = 0;
  }
# If $NoErrorTolerant is true, display only standard matches from a new style error tolerant search
  if (defined(my $no = $thisScript->param($urlParams{'_noerrortolerant'}))) {
    $NoErrorTolerant = looks_like_number($no) ? $no + 0 : 0;
  } else {
    $NoErrorTolerant = 0;
  }
# pick up illegal combinations
  if ($NoErrorTolerant) {
    $ShowAllFromErrorTolerant = 0;
  }
  if ($OnlyErrorTolerant && $NoErrorTolerant) {
    die("Meaningless to specify both _onlyerrortolerant and _noerrortolerant", $idForFatalError);
  }

# $UniGeneFile is the path to the Unigene index to be used for clustering EST matches
  if (my $species = $thisScript->param($urlParams{'unigene'})) {
    if ($species eq "None") {
      $UniGeneFile = "";
    } else {
      $UniGeneFile = $objUnigeneOptions->getPropValStringByName($species)
        || die("cannot find UniGene $species in mascot.dat ", $idForFatalError);
      $UniGeneFile =~ s/\s*$//;
    }
  } else {
    $UniGeneFile = "";
  }

# Whether to group protein families
  $group_family = 0;
  if (defined($thisScript->param($urlParams{'group_family'}))) {
    $group_family = $thisScript->param($urlParams{'group_family'}) + 0;
  }

# Whether to display scores and expect values based on Percolator PEPs
  if (defined($percolate = $thisScript->param($urlParams{'percolate'}))) {
    $percolate = $thisScript->param($urlParams{'percolate'});
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $percolate = 0;
  } else {
    $percolate = $objMascotOptions->isPercolator;
  }
  $percolate += 0;
  my($PercolatorMinQueries, $PercolatorMinSequences);
  $PercolatorMinQueries = 100 unless ($PercolatorMinQueries = $objMascotOptions->getPercolatorMinQueries);
  $PercolatorMinSequences = 100 unless ($PercolatorMinSequences = $objMascotOptions->getPercolatorMinSequences);
  if ($reportType eq 'protein' #TODO: Remove this. reportType can only be 'peptide' or 'concise'
    || $reportType eq 'concise'
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_PEPTIDES)
    || $objResFile->getNumQueries < $PercolatorMinQueries
    || $objResFile->getNumSeqsAfterTax < $PercolatorMinSequences
    || $objResFile->isErrorTolerant
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_DECOYPEPTIDES)) {
    $percolate = -1;
  }
  if ($percolate == 1) {
    if (defined($thisScript->param($urlParams{'percolate_rt'}))) {
      $percolate_rt = $thisScript->param($urlParams{'percolate_rt'});
    } elsif ($thisScript->param($urlParams{'do_export'})) {
      $percolate_rt = 0;
    } else {
      $percolate_rt = $objMascotOptions->isPercolatorUseRT;
    }
    $percolate_rt += 0;
  } else {
    $percolate_rt = 0;
  }

}

###############################################################################
# &formatControls()
# Print format form
# globals:
# my($myName, $objResFile, $objParams, $numHits, $sigThreshold_str, @fileIn,
# $idForFatalError,
# $ignoreIonsScoreBelow, $ShowSubSets, $thisScript, $RequireBoldRed,
# %urlParams, $mudpitSwitch, $objMascotOptions, $targetFDR,
# $targetFDRtype);
###############################################################################

sub formatControls {

  &common_subs::printHeader1("Matrix Science - Mascot - Export search results");

  print <<'end_of_static_HTML_text_block';

<SCRIPT LANGUAGE="JavaScript">
<!-- Begin hiding from older browsers

  function check_slaves(checkBox, form){
    if (!(checkBox && form)) {return;}
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
    var updateAll = false;

end_of_static_HTML_text_block

  if ($shipper eq 'TRUE') {
    print "    var shipper = 1;\n";
  } else {
    print "    var shipper = 0;\n";
  }

  print <<'end_of_static_HTML_text_block';
    if (checkName == "search_master") {
      if (form.show_header) {
        form.show_header.disabled = !checkState;
      }
      if (form.show_decoy) {
        form.show_decoy.disabled = !checkState;
      }
      if (form.show_mods) {
        form.show_mods.disabled = !checkState;
      }
      if (form.show_params) {
        form.show_params.disabled = !checkState;
      }
      if (form.show_format) {
        form.show_format.disabled = !checkState;
      }
      if (form.show_masses) {
        form.show_masses.disabled = !checkState;
      }
      if (form.show_quant_header) {
        form.show_quant_header.disabled = !checkState;
      }
    } else if (checkName == "protein_master") {
      updateAll = true;
      if (form.prot_score) {
        form.prot_score.disabled = !checkState;
      }
      if (form.prot_thresh) {
        form.prot_thresh.disabled = !checkState;
      }
      if (form.prot_expect) {
        form.prot_expect.disabled = !checkState;
      }
      if (form.prot_desc) {
        form.prot_desc.disabled = !checkState;
      }
      if (form.prot_mass) {
        form.prot_mass.disabled = !checkState;
      }
      if (form.prot_matches) {
        form.prot_matches.disabled = !checkState;
      }
      if (form.prot_cover) {
        form.prot_cover.disabled = !checkState || !shipper;
      }
      if (form.prot_len) {
        form.prot_len.disabled = !checkState || !shipper;
      }
      if (form.prot_empai) {
        form.prot_empai.disabled = !checkState;
      }
      if (form.prot_pi) {
        form.prot_pi.disabled = !checkState || !shipper;
      }
      if (form.prot_tax_str) {
        form.prot_tax_str.disabled = !checkState || !shipper;
      }
      if (form.prot_tax_id) {
        form.prot_tax_id.disabled = !checkState || !shipper;
      }
      if (form.prot_seq) {
        form.prot_seq.disabled = !checkState || !shipper;
      }
      if (form.prot_quant) {
        form.prot_quant.disabled = !checkState;
      }
end_of_static_HTML_text_block
  if ($export_format eq 'mzTab') {
      print <<'end_of_static_HTML_text_block';
      checkState = form.peptide_master.checked;
end_of_static_HTML_text_block
  }
  else {
    print <<'end_of_static_HTML_text_block';
      if (form.peptide_master) {
        form.peptide_master.disabled = !checkState;
      }
end_of_static_HTML_text_block
  }
  print <<'end_of_static_HTML_text_block';
    } if (checkName == "peptide_master" || updateAll) {
      if (form.pep_exp_mr) {
        form.pep_exp_mr.disabled = !checkState;
      }
      if (form.pep_exp_z) {
        form.pep_exp_z.disabled = !checkState;
      }
      if (form.pep_calc_mr) {
        form.pep_calc_mr.disabled = !checkState;
      }
      if (form.pep_delta) {
        form.pep_delta.disabled = !checkState;
      }
      if (form.pep_start) {
        form.pep_start.disabled = !checkState;
      }
      if (form.pep_end) {
        form.pep_end.disabled = !checkState;
      }
      if (form.pep_miss) {
        form.pep_miss.disabled = !checkState;
      }
      if (form.pep_score) {
        form.pep_score.disabled = !checkState;
      }
      if (form.pep_homol) {
        form.pep_homol.disabled = !checkState;
      }
      if (form.pep_ident) {
        form.pep_ident.disabled = !checkState;
      }
      if (form.pep_expect) {
        form.pep_expect.disabled = !checkState;
      }
      if (form.pep_seq) {
        form.pep_seq.disabled = !checkState;
      }
      if (form.pep_frame) {
        form.pep_frame.disabled = !checkState;
      }
      if (form.pep_var_mod) {
        form.pep_var_mod.disabled = !checkState;
      }
//      if (form.show_unassigned) {
//        form.show_unassigned.disabled = !checkState;
//      }
      if (form.pep_num_match) {
        form.pep_num_match.disabled = !checkState;
      }
      if (form.pep_scan_title) {
        form.pep_scan_title.disabled = !checkState;
      }
      if (form.pep_source) {
        form.pep_source.disabled = !checkState;
      }
      if (form.pep_quant) {
        form.pep_quant.disabled = !checkState;
      }
      if (form.show_pep_dupes) {
        form.show_pep_dupes.disabled = !checkState;
      }
      if (form.export_site_data) {
        form.export_site_data.disabled = !checkState;
      }
      if (form.prot_empai) {
        form.prot_empai.disabled = !(form.peptide_master.checked && form.protein_master.checked);
      }
      if (form.prot_quant) {
        form.prot_quant.disabled = !(form.peptide_master.checked && form.protein_master.checked);
      }

    } else if (checkName == "query_master") {
      if (form.query_title) {
        form.query_title.disabled = !checkState;
      }
      if (form.query_qualifiers) {
        form.query_qualifiers.disabled = !checkState;
      }
      if (form.query_params) {
        form.query_params.disabled = !checkState;
      }
      if (form.query_peaks) {
        form.query_peaks.disabled = !checkState;
      }
      if (form.query_raw) {
        form.query_raw.disabled = !checkState;
      }
    }
  }

  function check_rbr(checkBox, form){
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
    if (form._requireboldred) {
      if (checkState) {
        form._requireboldred.checked = 0;
      }
      form._requireboldred.disabled = checkState;
    }
  }

  function check_target_fdr(checkBox, form){
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
    if (form._target_fdr) {
      form._target_fdr.disabled = checkState;
    }
    if (form._fdr_count_type) {
      form._fdr_count_type.disabled = checkState;
    }
  }

window.onload=function(){
  check_slaves(document.getElementById("protein_master_id"), document.getElementById("export_form_id"));
};


// End hiding Javascript from old browsers. -->
</SCRIPT>

end_of_static_HTML_text_block

  common_subs::printHeader2("", "Export search results", 1, $session);

  print <<"end_of_static_HTML_text_block";
  <FORM ID="export_form_id" ACTION="./$myName" NAME="Re-format" METHOD="POST" ENCTYPE="multipart/form-data">
end_of_static_HTML_text_block

  if ($thisScript->param($urlParams{'fileset'})) {
    my $fileset = common_subs::htmlEscape(scalar $thisScript->param($urlParams{'fileset'}));
    printf('  <INPUT TYPE="hidden" NAME="fileset" VALUE="%s">'."\n", $fileset);
  } else {
    my $fileIn = common_subs::htmlEscape($fileIn[0]);
    printf('  <INPUT TYPE="hidden" NAME="file" VALUE="%s">'."\n", $fileIn);
  }

  print <<"end_of_static_HTML_text_block";
  <INPUT TYPE="hidden" NAME="do_export" VALUE=1>
  <INPUT TYPE="hidden" NAME="prot_hit_num" VALUE=1>
  <INPUT TYPE="hidden" NAME="prot_acc" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_query" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_index" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_rank" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_isbold" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_isunique" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_exp_mz" VALUE=1>
  <INPUT TYPE="hidden" NAME="_showallfromerrortolerant" VALUE=$ShowAllFromErrorTolerant>
  <INPUT TYPE="hidden" NAME="_onlyerrortolerant" VALUE=$OnlyErrorTolerant>
  <INPUT TYPE="hidden" NAME="_noerrortolerant" VALUE=$NoErrorTolerant>
  <INPUT TYPE="hidden" NAME="_show_decoy_report" VALUE=$ShowDecoyReport>
  <INPUT TYPE="hidden" NAME="sessionid" VALUE="$sessionID">
end_of_static_HTML_text_block

# pass through any quantitation parameters
  while (my($key, $value) = each %urlParams) {
    if ($key =~ /^_quant/i || lc($key) eq "_min_precursor_charge") {
      print "  <INPUT TYPE=\"hidden\" NAME=\"$key\" VALUE=\"" . common_subs::htmlEscape(scalar $thisScript->param($urlParams{$key})) . "\">\n";
    }
  }

  print <<"end_of_static_HTML_text_block";
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>

    <TR>
      <TD NOWRAP><H2>Export search results</H2></TD>
      <TD ALIGN=RIGHT VALIGN=baseline><A HREF="../help/export_help.html" target=_blank>Help</A></TD>
    </TR>

    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Export format </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <SELECT NAME=\"export_format\" onChange=\"this.form.do_export.value=0; this.form.submit(); return true\">
end_of_static_HTML_text_block

    if (@fileIn == 1) {
      if ($export_format eq "XML" || $export_format eq "") {
        print "         <OPTION SELECTED>XML</OPTION>\n";
      } else {
        print "         <OPTION>XML</OPTION>\n";
      }
    }
    if ($export_format eq "CSV") {
      print "         <OPTION SELECTED>CSV</OPTION>\n";
    } else {
      print "         <OPTION>CSV</OPTION>\n";
    }
    if ($export_format eq "mzIdentML") {
      print "         <OPTION SELECTED>mzIdentML</OPTION>\n";
    } else {
      print "         <OPTION>mzIdentML</OPTION>\n";
    }

  if(not $anyXlPepSumMatch) { # Bug 13396: Only allow MGF, .dat, xiVIEW, mzIdentML(1.2), CSV and XML for crosslinked searchs
    # Bug 12416: Only allow pepXML if the search results are all AA or all NA.
    if (@fileIn == 1 and not ExportDat::Util::db_type_is_mixed(ExportDat::Util::count_db_types($objResFile, $objParams))) {
      if ($reportType eq 'peptide' && $export_format eq "pepXML") {
        print "         <OPTION SELECTED>pepXML</OPTION>\n";
      } elsif ($reportType eq 'peptide') {
        print "         <OPTION>pepXML</OPTION>\n";
      }
    }
    if(@fileIn == 1) {
      if ($export_format eq "mzTab") {
	print "         <OPTION SELECTED>mzTab</OPTION>\n";
      } else {
	print "         <OPTION>mzTab</OPTION>\n";
      }
    }
    if (@fileIn == 1 and not ExportDat::Util::db_type_is_mixed(ExportDat::Util::count_db_types($objResFile, $objParams))) {
      if ($reportType eq 'peptide' && $export_format eq "DTASelect") {
        print "         <OPTION SELECTED>DTASelect</OPTION>\n";
      } elsif ($reportType eq 'peptide') {
        print "         <OPTION>DTASelect</OPTION>\n";
      }
    }
  }
  # Bug 12515: Only allow FASTA export if a FASTA database was
  # searched. Disallow for SL-only searches.
  if ((@fileIn > 1 ) or ExportDat::Util::db_type_is_SL_only(ExportDat::Util::count_db_types($objResFile, $objParams))) {
    # Omit option.
  } elsif ($export_format eq "FASTA") {
    print "         <OPTION SELECTED>FASTA</OPTION>\n";
  } else {
    print "         <OPTION>FASTA</OPTION>\n";
  }
  if ((@fileIn == 1) and $anyXlPepSumMatch) {
    if ($export_format eq "xiVIEW-CSV") {
      print "         <OPTION SELECTED>xiVIEW-CSV</OPTION>\n";
    }
    else {
      print "         <OPTION>xiVIEW-CSV</OPTION>\n";
    }
  }
  if (@fileIn == 1) {
    if ($export_format eq "MascotDAT") {
      print "         <OPTION value=\"MascotDAT\" SELECTED>Mascot DAT File</OPTION>\n";
    } else {
      print "         <OPTION value=\"MascotDAT\">Mascot DAT File</OPTION>\n";
    }
    if ($export_format eq "MGF") {
	print "       <OPTION value=\"MGF\" SELECTED>MGF Peak List</OPTION>";
    } else {
      print "         <OPTION value=\"MGF\">MGF Peak List</OPTION>\n";
    }
  }
  print "        </SELECT>\n";
  print "      </TD>\n";
  print "    </TR>\n";
  if (@fileIn == 1) {
    if ($export_format eq "MGF") {
	print <<"end_of_static_HTML_text_block";

         <TR>
          <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>
             Output spectra in
          </TD>
          <TD BGCOLOR=#EEEEFF ALIGN=LEFT>
            <table>
              <tr>
                <td>
                  <LABEL FOR=\"mgf_output_order_in\">
                    <INPUT TYPE=\"radio\" VALUE=1 NAME=\"mgf_output_order\" ID=\"mgf_output_order_in\" CHECKED>
	            original order (for use with mzIdentML or mzTab)
	          </LABEL>
                </td>
              </tr>
              <tr>
                <td>
                  <LABEL FOR=\"mgf_output_order_query\">
                    <INPUT TYPE=\"radio\" VALUE=0 NAME=\"mgf_output_order\" ID=\"mgf_output_order_query\">
	            Mascot query order
	          </LABEL>
                </td>
              </tr>
            </table>
          </TD>
        </TR>
end_of_static_HTML_text_block
    }
  }

  my @query_level_information_footnotes;

  unless ($export_format eq "MascotDAT" || $export_format eq "MGF") {

    if ($export_format eq "mzIdentML"){
      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      print "       Version </TD>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
      print "       <SELECT NAME=\"mzid_version\" onChange=\"this.form.do_export.value=0; this.form.submit(); return true\">\n";
      if($anyXlPepSumMatch) {
	print "           <OPTION VALUE=\"1.2\" SELECTED>1.2</OPTION>\n";
      } else {
	my $ver = $thisScript->param($urlParams{'mzid_version'});
	$ver = $ver ? $ver : "1.1";
	if ($ver eq "1.1") {
	  print "           <OPTION VALUE=\"1.1\" SELECTED>1.1</OPTION>\n";
	  print "           <OPTION VALUE=\"1.2\">1.2</OPTION>\n";
	} else {
	  print "           <OPTION VALUE=\"1.1\">1.1</OPTION>\n";
	  print "           <OPTION VALUE=\"1.2\" SELECTED>1.2</OPTION>\n";
	}
      }
      print "       </SELECT>\n";
      print "     </TD>\n";
      print "   </TR>\n";
    }

    print <<"end_of_static_HTML_text_block";
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Significance threshold p&lt; </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
end_of_static_HTML_text_block

    do {
      my $sigThreshold_disabled = '';
      if ($targetFDR and $targetFDR != 0) {
	   $sigThreshold_disabled = ' DISABLED';
      }
      print <<"end_of_static_HTML_text_block";
        <INPUT NAME="_sigthreshold" ID="ctrl_sigthreshold" TYPE=text SIZE=8 VALUE=$sigThreshold_str $sigThreshold_disabled>
end_of_static_HTML_text_block
    };

    if ($anyPepSumMatch) {

      print "      at&nbsp;&nbsp;";

      do {
        print "<INPUT TYPE=\"radio\" VALUE=0 NAME=\"use_homology\" ID=\"use_homology_0\"";
        unless ($use_homology){
          print " CHECKED";
        }
        print "><LABEL FOR=\"use_homology_0\">identity</LABEL>";
        print "&nbsp;&nbsp;&nbsp;&nbsp;<INPUT TYPE=\"radio\" VALUE=1 NAME=\"use_homology\" ID=\"use_homology_1\"";
        if ($use_homology){
          print " CHECKED";
        }
        print "><LABEL FOR=\"use_homology_1\">homology</LABEL>\n";
      };

      print "<INPUT TYPE=\"hidden\" NAME=\"_minpeplen\" VALUE=$minPepLen>\n";
      print "    </TD>\n";
      print "   </TR>\n";

      if ($objParams->getDECOY() > 0) {
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     Target FDR<BR>(overrides significance<BR>threshold if set) </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "     <SELECT NAME=\"_target_fdr\" ONCHANGE=\"var elt = document.getElementById('ctrl_sigthreshold'); if (elt) { elt.disabled = this.options[this.selectedIndex].value != 0; }\" ";
        print ">\n";
        my @options = ();
        my $target_found = 0;
        for my $i (0 .. $objMascotOptions->getNumberOfTargetFDRPercentages()-1) {
          my $t = $objMascotOptions->getTargetFDRPercentage($i);
          my $str = "       <OPTION VALUE=\"$t\"";
          if ($targetFDR and $t == $targetFDR) {
            $str .= " SELECTED";
            $target_found = 1;
          }
          $str .= ">$t%</OPTION>\n";
          push @options, $str;
        }
        if (not $target_found) {
          if ($targetFDR) {
            unshift @options, "       <OPTION VALUE=\"$targetFDR\" SELECTED>$targetFDR%</OPTION>\n";
            $target_found = 1;
          }
        }
        print "       <OPTION VALUE=\"0\"";
        if (not $target_found) {
          print " SELECTED";
        }
        print ">&lt;not set&gt;</OPTION>\n";
        print @options;
        print "     </SELECT>\n";
        print "    </TD>\n";
        print "   </TR>\n";
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     FDR type </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "     <SELECT NAME=\"_fdr_count_type\"";
        print ">\n";
        print "       <OPTION VALUE=\"psm\"";
        if (defined $targetFDRtype and $targetFDRtype eq 'psm') {
          print " SELECTED";
        }
        print ">Distinct PSMs</OPTION>\n";
        print "       <OPTION VALUE=\"seq\"";
        if (defined $targetFDRtype and $targetFDRtype eq 'seq') {
          print " SELECTED";
        }
        print ">Distinct sequences</OPTION>\n";
        print "     </SELECT>\n";
        print "    </TD>\n";
        print "   </TR>\n";
      }

      unless ($export_format eq "pepXML") {
        print "   <TR>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      # If the setting is 2, always show the edit field.
      # If the setting is 0 or 1, show either the edit field (if value is
      # defined and > 0) or a checkbox.
        my $displayNonSigMatches = $objMascotOptions->getDisplayNonSignificantMatches;
        if ($displayNonSigMatches == 2
                or ($ignoreIonsScoreBelowDefined and $ignoreIonsScoreBelow > 0)) {
          print "       Ions score cut-off </TD>\n";
          print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
          print "       <INPUT NAME=\"_ignoreionsscorebelow\" TYPE=text SIZE=8 VALUE=$ignoreIonsScoreBelow>\n";
        } else {
          print "       Display non-significant matches </TD>\n";
          print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
          print "<INPUT NAME=\"_ignoreionsscorebelow_test\" TYPE=\"checkbox\" onClick=\"";
          print "if (form._ignoreionsscorebelow_test.checked) { form._ignoreionsscorebelow.value = '0'; } else { form._ignoreionsscorebelow.value = '-1'; } return true;";
          print "\"";
          my $default = -1;
          if ($ignoreIonsScoreBelowDefined) {
            if ($ignoreIonsScoreBelow == 0) {
              print " CHECKED";
              $default = 0;
            }
          } else {
            if ($displayNonSigMatches == 1) {
              print " CHECKED";
              $default = 0;
            }
          }
          print ">\n";
          print "<INPUT NAME=\"_ignoreionsscorebelow\" TYPE=\"hidden\" VALUE=\"$default\">\n";
        }
        print "     </TD>\n";
        print "   </TR>\n";
      }

      if ($anyLibraryMatch
              and $objResFile->anyPeptideSummaryMatches($sec_peptides)) {
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     Report mode </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "     <SELECT NAME=\"library\"";
        print ">\n";
        print "       <OPTION VALUE=\"integrated\"";
        if ($libraryMode eq 'integrated') {
          print " SELECTED";
        }
        print ">Integrated</OPTION>\n";
        print "       <OPTION VALUE=\"onlyprimary\"";
        if ($libraryMode eq 'onlyprimary') {
          print " SELECTED";
        }
        print ">Only FASTA</OPTION>\n";
        if ($export_format ne "FASTA") { # Bug 12515
            print "       <OPTION VALUE=\"onlylibrary\"";
            if ($libraryMode eq 'onlylibrary') {
                print " SELECTED";
            }
            print ">Only library</OPTION>\n";
        }
        print "     </SELECT>\n";
      }

    } else {
      print "    </TD>\n";
      print "   </TR>\n";
    }

    unless ($export_format eq "pepXML") {
      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      print "       Max. number of hits </TD>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
      if ($numHits) {
        print "       <INPUT NAME=\"report\" TYPE=text SIZE=5 VALUE=$numHits>\n";
      } else {
        print "       <INPUT NAME=\"report\" TYPE=text SIZE=5 VALUE=\"AUTO\">\n";
      }
      print "     </TD>\n";
      print "   </TR>\n";

      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      print "       Min. number of sig. unique sequences </TD>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
      print "       <SELECT NAME=\"min_num_sig_unique_seqs\">\n";
      for my $i (1 .. 5) {
          print "           <OPTION VALUE=\"$i\"";
          if ($i == $minNumSigUniqueSeqs) {
              print " SELECTED";
          }
          print ">$i</OPTION>\n";
      }
      print "       </SELECT>\n";
      print "     </TD>\n";
      print "   </TR>\n";

      if ($anyPepSumMatch) {

        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     Protein scoring </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "      <INPUT TYPE=\"radio\" VALUE=99999999 NAME=\"_server_mudpit_switch\" ID=\"_server_mudpit_switch_99999999\"";
        unless ($anyLibraryMatch
                    or ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch)) {
          print " CHECKED";
        }
        print "><LABEL FOR=\"_server_mudpit_switch_99999999\">Standard</LABEL>";
        print "&nbsp;&nbsp;&nbsp;&nbsp;<INPUT TYPE=\"radio\" VALUE=0.000000001 NAME=\"_server_mudpit_switch\" ID=\"_server_mudpit_switch_0.000000001\"";
        if ($anyLibraryMatch
                or ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch)) {
          print " CHECKED";
        }
        print "><LABEL FOR=\"_server_mudpit_switch_0.000000001\">MudPIT</LABEL>\n";
        print "     </TD>\n";
        print "   </TR>\n";

      }
    }

    unless ($export_format eq "pepXML" || $export_format eq "DTASelect") {
      if ($ShowSameSets) {
        $tmpChecked = "CHECKED";
      } else {
        $tmpChecked = "";
      }
      if ($export_format eq 'xiVIEW-CSV') {
          print <<"end_of_static_HTML_text_block";
          <TR>
          <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
           </TD>
          <TD BGCOLOR=#EEEEFF NOWRAP>
          <INPUT NAME="show_same_sets" TYPE="hidden" VALUE="$ShowSameSets">
          </TD>
          </TR>
end_of_static_HTML_text_block
      } else {
        &checkBox("Include same-set protein hits<BR>(additional proteins that span<BR>the same set of peptides)",
            "show_same_sets", $tmpChecked, "");
      }
      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      if ($export_format eq 'mzTab') {
        print "     </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_showsubsets\" TYPE=hidden VALUE=$ShowSubSets>\n";
      } elsif ($export_format eq 'xiVIEW-CSV') {
        print "     </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_showsubsets\" TYPE=hidden VALUE=$ShowSubSets>\n";
      } else {
        print "       Include sub-set protein hits<BR>(additional proteins that span<BR>a sub-set of peptides) </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_showsubsets\" TYPE=text SIZE=5 VALUE=$ShowSubSets>\n";
      }
      print "     </TD>\n";
      print "   </TR>\n";
      if ($anyPepSumMatch) {

        print "   <TR>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
        print "       Group protein families</TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "     <INPUT TYPE=\"checkbox\" NAME=\"group_family\" VALUE=1 ";
        if ($group_family) {
          print "CHECKED "
        }
        print "onClick=\"check_rbr(this, this.form)\"></TD>\n";
        print "   </TR>\n";
        my $disabled = "";
        if ($group_family) {
          $disabled = "disabled";
        }
        if ($RequireBoldRed) {
          &checkBox("Require bold red", "_requireboldred", "CHECKED", $disabled);
        } else {
          &checkBox("Require bold red", "_requireboldred", "", $disabled);
        }
      }

      # UniGene
      my @db_names = MascotResults::databases($objResFile);
      # Bug 13338: Use the same conditions as in MasterResults::ProteinFamily::TopLevel::form_filters
      # to output UniGene dropdown.
      if (@db_names == 1 and (my @indices = MascotDat->unigene_species_available(@db_names))) {
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
        print "UniGene index </TD>";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "      <SELECT NAME=\"UNIGENE\">\n";

        my $species = "None";
        if ($thisScript->param($urlParams{'unigene'})) {
          $species = $thisScript->param($urlParams{'unigene'});
        }
	foreach (@indices) {
          print "  <OPTION";
          if ($species eq $_) {
            print " SELECTED";
          }
          print ">$_</OPTION>\n";
        }
        print "</SELECT>\n";
        print "    </TD>\n";
        print "  </TR>\n";
      }
    }

    if ($percolate != -1) {
      if ($percolate == 1) {
        $tmpChecked = "CHECKED";
      } else {
        $tmpChecked = "";
      }
      &checkBox("Show Percolator scores", "percolate", $tmpChecked, "", 'check_target_fdr(this, this.form)');
    }

    if($percolate_rt) {
        #BUG12139 fix
        print "    <INPUT NAME=\"percolate_rt\" TYPE=\"HIDDEN\" VALUE=1>\n";
    }

    print <<'end_of_static_HTML_text_block';
    <TR>
end_of_static_HTML_text_block

    if ($export_format eq 'xiVIEW-CSV' or $export_format eq 'FASTA') {

    print <<'end_of_static_HTML_text_block';
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
      </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
end_of_static_HTML_text_block
    print DOM->hidden({
      name => '_prefertaxonomy',
      value => $preferTaxonomy,
    })->as_string;
    print <<'end_of_static_HTML_text_block';
      </TD>
end_of_static_HTML_text_block

    } else {

    print <<'end_of_static_HTML_text_block';
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Preferred Taxonomy<SUP>*</SUP> </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
end_of_static_HTML_text_block

    print DOM->dropdown({
      values => [taxonomy_choice($objResFile)],
      name => '_prefertaxonomy',
      default => $preferTaxonomy,
    })->as_string;
    print <<'end_of_static_HTML_text_block';
      </TD>
end_of_static_HTML_text_block
    }

    print <<'end_of_static_HTML_text_block';
    </TR>
end_of_static_HTML_text_block

    if ($export_format eq "FASTA" and ExportDat::Util::db_type_is_mixed_AA_or_NA(ExportDat::Util::count_db_types($objResFile, $objParams))) {
      print <<"end_of_static_HTML_text_block";
       <TR>
        <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>
           Results from AA and NA databases. Select what to ouput:
        </TD>
        <TD BGCOLOR=#EEEEFF ALIGN=LEFT>
          <table>
            <tr>
              <td>
                <LABEL FOR=\"fasta_output_na_id\">
                  <INPUT TYPE=\"radio\" VALUE=1 NAME=\"fasta_output_type\" ID=\"fasta_output_na_id\" CHECKED>
	            NA sequences
                </LABEL>
              </td>
            </tr>
            <tr>
              <td>
                <LABEL FOR=\"fasta_output_aa_id\">
                  <INPUT TYPE=\"radio\" VALUE=2 NAME=\"fasta_output_type\" ID=\"fasta_output_aa_id\">
	          AA sequences
	        </LABEL>
              </td>
            </tr>
          </table>
        </TD>
      </TR>
end_of_static_HTML_text_block
    }

    print <<'end_of_static_HTML_text_block';
    <TR>
      <TD NOWRAP COLSPAN=2><SUP>* Occasionally requires information to be retrieved from external utilities, which can be slow</TD>
    </TR>
end_of_static_HTML_text_block

    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML" or $export_format eq 'mzTab' or $export_format eq 'FASTA'
        or $export_format eq 'xiVIEW-CSV') {

      print "   <TR>\n";
      print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
      print "   </TR>\n";

      print "   <TR>\n";
      print "     <TD NOWRAP>\n";
      print "     <H3>Search Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"checkbox\" NAME=\"search_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
      print "   </TR>\n";

      &checkBox("Header", "show_header", "CHECKED", "");
      if ($objParams->getDECOY() > 0 && !$ShowDecoyReport) {
        &checkBox("Decoy", "show_decoy", "CHECKED", "");
      }
      &checkBox("Modification deltas", "show_mods", "CHECKED", "");
      &checkBox("Search parameters", "show_params", "CHECKED", "");
      &checkBox("Format parameters", "show_format", "CHECKED", "");
      &checkBox("Residue masses", "show_masses", "", "");
      if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
        &checkBox("Quantitation header", "show_quant_header", "CHECKED", "");
      }
    }

    print "   <TR>\n";
    print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
    print "   </TR>\n";

    print "   <TR>\n";
    print "     <TD NOWRAP>\n";
    if ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
      print "     <H3>Optional Protein Hit Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     &nbsp;</TD>\n";
    } elsif ($export_format ne "FASTA" and $export_format ne 'xiVIEW-CSV') {
      print "     <H3>Protein Hit Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"checkbox\" ID=\"protein_master_id\" NAME=\"protein_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
    }
    print "   </TR>\n";

    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML" or $export_format eq "mzTab" or $export_format eq "FASTA"
            or $export_format eq 'xiVIEW-CSV') {
      &checkBox("Score", "prot_score", "CHECKED", "");
      unless ($anyPepSumMatch) {
        &checkBox("Significance threshold", "prot_thresh", "CHECKED", "");
        &checkBox("Expectation value", "prot_expect", "CHECKED", "");
      }
    }
    unless ($export_format eq "FASTA" or $export_format eq 'xiVIEW-CSV') {
      &checkBox("Description<SUP>*</SUP>", "prot_desc", "CHECKED", "");
      unless ($export_format eq "mzIdentML") {
        &checkBox("Mass (Da)<SUP>*</SUP>", "prot_mass", "CHECKED", "");
      }
      my $disabled = "";
      if ($shipper eq 'FALSE') {
        $disabled = "disabled";
      }
      unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
        &checkBox("Number of queries matched", "prot_matches", "CHECKED", "");
      }
      unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
        &checkBox("Percent coverage<SUP>**</SUP>", "prot_cover", "", $disabled);
      }
      unless ($export_format eq "pepXML" or $export_format eq 'mzTab') {
        my $label;
        if ($export_format eq 'mzIdentML') {
          $label = "Length in residues and protein coverage";
        } else {
          $label = "Length in residues";
        }
        &checkBox("$label<SUP>**</SUP>", "prot_len", "", $disabled);
      }
      unless ($export_format eq "mzIdentML") {
        &checkBox("pI<SUP>**</SUP>", "prot_pi", "", $disabled);
      }
      unless ($export_format eq "pepXML" or $export_format eq "DTASelect") {
        &checkBox("Taxonomy<SUP>**</SUP>", "prot_tax_str", "", $disabled);
        unless ($export_format eq 'mzTab') {
          &checkBox("Taxonomy ID<SUP>**</SUP>", "prot_tax_id", "", $disabled);
        }
      }
      unless ($export_format eq 'mzTab') {
        &checkBox("Protein sequence<SUP>**</SUP>", "prot_seq", "", $disabled);
      }
      if ($anyPepSumMatch and ($export_format eq "XML" or $export_format eq "CSV" or $export_format eq 'mzTab')) {
        my $can_have_emPAI = 0;
        if (not ($msresFlags & $msparser::ms_mascotresults::MSRES_DECOY)
              and $objResFile->getNumQueries() >= 100) {
          $can_have_emPAI = 1;
        }
        if ($can_have_emPAI) {
          &checkBox("emPAI", "prot_empai", "", $disabled);
        }
        if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
          &checkBox("Protein quantitation", "prot_quant", "", "");
        }
      }

      print <<'end_of_static_HTML_text_block';

      <TR>
        <TD NOWRAP COLSPAN=2><SUP>* Occasionally requires information to be retrieved from external utilities, which can be slow<BR>
        ** Always requires information to be retrieved from external utilities, which can be slow</SUP></TD>
      </TR>

end_of_static_HTML_text_block

      unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
        print "   <TR>\n";
        print "     <TD NOWRAP>\n";
        print "     <H3>Peptide Match Information</H3></TD>\n";
        print "     <TD NOWRAP VALIGN=top>\n";
        print "     <INPUT TYPE=\"checkbox\" NAME=\"peptide_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
        print "   </TR>\n";

        unless ($export_format eq 'mzTab') {
          &checkBox("Experimental Mr (Da)", "pep_exp_mr", "CHECKED", "");
        }
        unless ($export_format eq 'mzTab') {
          if ($anyPepSumMatch) {
            &checkBox("Experimental charge", "pep_exp_z", "CHECKED", "");
          } else {
            &checkBox("Experimental charge", "pep_exp_z", "", "");
          }
          &checkBox("Calculated Mr (Da)", "pep_calc_mr", "CHECKED", "");
        }
        &checkBox("Mass error (Da)", "pep_delta", "CHECKED", "");
        unless ($export_format eq 'mzTab') {
          if ($anyPepSumMatch) {
            &checkBox("Start", "pep_start", "", "");
            &checkBox("End", "pep_end", "", "");
          } else {
            &checkBox("Start", "pep_start", "CHECKED", "");
            &checkBox("End", "pep_end", "CHECKED", "");
          }
        }
        &checkBox("Number of missed cleavages", "pep_miss", "CHECKED", "");
        if ($anyPepSumMatch) {
          &checkBox("Score", "pep_score", "CHECKED", "");
          &checkBox("Homology threshold", "pep_homol", "", "");
          &checkBox("Identity threshold", "pep_ident", "", "");
          unless ($export_format eq 'mzTab') { # exported as peptide score
            &checkBox("Expectation value", "pep_expect", "CHECKED", "");
          }
        }
        unless ($export_format eq 'mzTab') { # always on in mzTab
          &checkBox("Sequence", "pep_seq", "CHECKED", "");
        }
      # enabling / disabling frame checkbox is too messy
      # because we don't want to create a summary object at this stage
        &checkBox("Frame number", "pep_frame", "", "");
        unless ($export_format eq 'mzTab') { # always on in mzTab
          &checkBox("Variable Modifications", "pep_var_mod", "CHECKED", "");
        }
        unless ($objResFile->isPMF()){
          &checkBox("Number of fragment<BR>ion matches", "pep_num_match", "", "");
          &checkBox("Query title", "pep_scan_title", "CHECKED", "");
          if ($export_format ne 'mzTab'
              and ExportDat::Util::db_type_is_mixed(ExportDat::Util::count_db_types($objResFile, $objParams))) {
            &checkBox("Type of source<BR>database (AA/NA/SL)", "pep_source", "CHECKED", "");
          }
        }
        if ($anyPepSumMatch) {
          if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
            &checkBox("Peptide quantitation", "pep_quant", "", "");
          }
          unless ($export_format eq 'mzTab') {
            &checkBox("Unassigned queries<BR>(peptide matches not<BR>assigned to protein hits)", "show_unassigned", "", "");
          }
        }

        &checkBox("Show duplicate peptides", "show_pep_dupes", "CHECKED", "");

        print "   <TR>\n";
        print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
        print "   </TR>\n";
      }
      if ($export_format eq 'mzIdentML') {
        print "   <TR>\n";
        print "     <TD NOWRAP>\n";
        print "     <H3>Optional peptide Match Information</H3></TD>\n";
        print "     <TD NOWRAP VALIGN=top>\n";
        print "     <INPUT TYPE=\"hidden\" NAME=\"peptide_master\" VALUE=1></TD>\n";
        print "   </TR>\n";

        &checkBox("Show duplicate peptides", "show_pep_dupes", "CHECKED", "");

	unless (($thisScript->param($urlParams{'mzid_version'}) eq '1.2') or ($anyXlPepSumMatch)) {
	  # Disable site analysis for 1.2. Crosslink searches can only be exported to 1.2
	  &checkBox("Export site analysis data<SUP>*</SUP>", "export_site_data", "", "");

	  print "   <TR>\n";
	  print "     <TD NOWRAP COLSPAN=2>\n";
	  print "       <SUP>* Requires extra calculations, which can be slow<BR>\n";
	  print "     </SUP></TD>\n";
	  print " </TR>\n";
	}
	print "   <TR>\n";
	print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
	print "   </TR>\n";

      }

      unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $objResFile->isPMF() or $export_format eq 'mzTab') {
        print "   <TR>\n";
        print "     <TD NOWRAP>\n";
        print "     <H3>Query Level Information</H3></TD>\n";
        print "     <TD NOWRAP VALIGN=top>\n";
        unless ($export_format eq "mzIdentML") {
          print "     <INPUT TYPE=\"checkbox\" NAME=\"query_master\" VALUE=1 onClick=\"check_slaves(this, this.form)\"></TD>\n";
        } else {
          print "     &nbsp;</TD>\n";
        }
        print "   </TR>\n";

        my $sDisabelled = "";
        unless ($export_format eq "mzIdentML") {
          $sDisabelled = "disabled";
          &checkBox("Query title,<BR>retention time*,<BR>scan number*", "query_title", "", $sDisabelled);
          &checkBox("seq(), comp(), tag(), etc.", "query_qualifiers", "", $sDisabelled);
          &checkBox("Query level<BR>search parameters", "query_params", "", $sDisabelled);
          &checkBox("MS/MS Peak lists", "query_peaks", "", $sDisabelled);
          push @query_level_information_footnotes, q{<sup>* Available for export if present in input peak lists, which depends on peak picking tool</sup>}
        }
        unless ($export_format eq "mzIdentML") {
          &checkBox("Raw peptide match data", "query_raw", "", $sDisabelled);
        } else {
          &checkBox("Matched Fragment Ions", "query_raw", "", $sDisabelled);
          &checkBox("Export data for all Queries", "query_all", "", $sDisabelled);
        }
      }
    } ## closure of unless FASTA or xiVIEW-CSV
  }  # closure of unless ($export_format eq "MascotDAT" || $export_format eq "MGF") {

  my $submission_scriptlet_cmdline = 'this.form.do_export.value=-1; this.form.submit(); return true';
  my $submission_scriptlet_data = 'this.form.do_export.value=1; this.form.submit(); return true';

  if (not @query_level_information_footnotes) {
    # Always print something to make the row height non-zero.
    push @query_level_information_footnotes, "&nbsp";
  }

  print "    <TR>\n";
  print "      <TD NOWRAP ALIGN=LEFT COLSPAN=2>\n";
  for (@query_level_information_footnotes) {
    print $_, "<BR>\n";
  }
  print "      </TD>\n";
  print "    </TR>\n";

  print <<"end_of_static_HTML_text_block";
    <TR>
      <TD NOWRAP ALIGN=LEFT COLSPAN=2>&nbsp;</TD>
    </TR>

    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=LEFT>
        <INPUT TYPE="button" VALUE="Show command line arguments" onClick="$submission_scriptlet_cmdline">
      </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <INPUT TYPE="button" VALUE="Export search results" onClick="$submission_scriptlet_data">
      </TD>
    </TR>

  </TABLE>

</FORM>

end_of_static_HTML_text_block

  &common_subs::printFooter;

  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &downloadControls()
# Print the downloading form displayed after successful caching.
# globals:
# my($myName,
# $thisScript,
# %urlParams)
###############################################################################

sub downloadControls {
  common_subs::printHeader1("Matrix Science - Mascot - Export search results");
  common_subs::printHeader2("", "Export search results", 1, $session);

  print <<"end_of_static_HTML_text_block";
  <FORM ACTION="./$myName" NAME="Download" METHOD="POST" ENCTYPE="multipart/form-data">
  <INPUT TYPE="hidden" NAME="generate_file" VALUE=1>
end_of_static_HTML_text_block

# Pass through all parameters.
  while (my($key, $value) = each %urlParams) {
    print "  <INPUT TYPE=\"hidden\" NAME=\"$key\" VALUE=\"" . common_subs::htmlEscape(scalar $thisScript->param($urlParams{$key})) . "\">\n";
  }

  my $msg = 'Exporting is ready to be started.';

  if (my $id = $thisScript->param($urlParams{'_cachewarnings'})) {
      my $session = msparser::ms_session->new($id);
      my @warnings = ();

      if ($session->isValid) {
          my $num = $session->getIntParam('num_warnings');

          for my $i (0 .. $num-1) {
              push @warnings, $session->getStringParam('warning_'.$i);
          }

          $session->destroy();
      }

      if (@warnings) {
          $msg = '<p>Caching the exported data did not succeed:</p><ul><li>';
          $msg .= join('</li><li>', map {
                  '<FONT COLOR=#FF0000>'.common_subs::htmlEscape($_).'</FONT>'
              } @warnings
          );

          $msg .= '</li></ul> <p>This means downloading may take a long time while the file is opened and parsed.</p>';

          $msg .= '<p>Exported data is ready to be downloaded.</p>';
      }
  }

  print <<"end_of_static_HTML_text_block";
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>

    <TR>
      <TD NOWRAP ALIGN=LEFT>$msg</TD>
    </TR>
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=LEFT>
        <INPUT TYPE="submit" VALUE="Download">
      </TD>
    </TR>

  </TABLE>

</FORM>

end_of_static_HTML_text_block

  common_subs::printFooter();

  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &checkBox()
# Print row of table for a checkbox option
###############################################################################

sub checkBox {

  my($description, $name, $checked, $disabled, $onclick) = @_;

  if (defined $onclick) {
    $onclick = sprintf('onClick="%s"', $onclick);
  } else {
    $onclick = '';
  }

  print <<"end_of_static_HTML_text_block";
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        $description </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <INPUT $disabled TYPE="checkbox" NAME="$name" VALUE=1 $checked $onclick>
      </TD>
    </TR>
end_of_static_HTML_text_block

}


###############################################################################
# &outputHit()
# export a hit block
# $_[0] protein object
# $_[1] hit number
# $_[2] sameset ordering within hit $_[1]
# $_[3] 1 if duplicate peptides should be filtered out
# $_[4] cache for duplicate_status_main()
# globals:
# my($objSummary, %urlParams, $thisScript, $export_format, $group_family);
###############################################################################

sub outputHit {

  my $objProtein = shift;
  my $hitNum = shift;
  my $sameset_ordering = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;

  if ($export_format eq "XML") {
    print "<hit number=\"$hitNum\">\n";
  }

  if ($group_family) {

  # The lead protein for the family
    my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
    outputProtein($sameset[0], $hitNum, 1, 1, $filter_duplicates, $dupe_state_cache);

  # Other proteins that match the same set of peptides
    if ($ShowSameSets) {
      outputProtein($_, $hitNum, 0, 1, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
    }

  # Other family members
    my $j = 1;
    while (my $objProtein = $objSummary->getNextFamilyProtein($hitNum, $j++)) {
      my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
      outputProtein($sameset[0], $hitNum, 1, $j, $filter_duplicates, $dupe_state_cache);

    # And their same-sets
      if ($ShowSameSets) {
        outputProtein($_, $hitNum, 0, $j, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
      }
    }

  # sub-sets and intersections for the entire family
    if ($ShowSubSets > 0) {
      my $j = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($hitNum, $j++, 1)) {
        my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
        outputProtein($sameset[0], $hitNum, 0, -1, $filter_duplicates, $dupe_state_cache);

      # And their same-sets
        if ($ShowSameSets) {
          outputProtein($_, $hitNum, 0, -1, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
        }
      }
    }

  } else {
  # The primary protein hit
    my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
    outputProtein($sameset[0], $hitNum, 1, 0, $filter_duplicates, $dupe_state_cache);

  # Other proteins that match the same set of peptides
    if ($ShowSameSets) {
      outputProtein($_, $hitNum, 0, 0, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
    }

  # Other proteins that match a sub-set of peptides
    if ($ShowSubSets > 0) {
      my $i = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($hitNum, $i++)) {
        outputProtein($objProtein, $hitNum, 0, 0, $filter_duplicates, $dupe_state_cache);
      }
    }

  }

  if ($export_format eq "XML") {
    print "</hit>\n";
  }

}

###############################################################################
# &outputProtein()
# export a protein block
# $_[0] protein object
# $_[1] hit or family number
# $_[2] flag set for anchor protein
# $_[3] family member number (0 if not family grouping, -1 if family subsets)
# $_[4] 1 if duplicate peptides should be filtered out
# $_[5] cache for duplicate_status_main()
# globals:
# my(%urlParams, $export_format, $thisScript, $fastaLen);
###############################################################################

sub outputProtein {

  my $objProtein = shift;
  my $hitNum = shift;
  my $primary = shift;
  my $member = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;

  if ($export_format eq "XML") {
    print "<protein accession=\"" . _mk_full_prot_acc($objProtein) . "\"";
    if ($member > 0) {
      print " member=\"" . $member . "\"";
    }
    print ">\n";
  }
  &printProteinRow($objProtein, $hitNum, $member);

  if ($export_format eq "CSV") {
    print $delimiter;
  }

  if ($thisScript->param($urlParams{'peptide_master'})) {
    printPeptideRows($objProtein, $hitNum, $primary, $member, $filter_duplicates, $dupe_state_cache);
  } else {
    if ($export_format eq "CSV") {
      print "\n";
    }
  }

# quantitation information is only output for primary protein hit
  if ($primary) {
  # and only available if we have looped through the peptides
    if ($objSummary->anyEmPAI and $thisScript->param($urlParams{'peptide_master'})) {
      if ($thisScript->param($urlParams{'prot_empai'})) {
        # TODO: mustGetProteinLen() should be used in
        # printProteinRow() too.
        my ($length) = $protData->mustGetProteinLen(
           $objProtein,
           scalar($thisScript->param($urlParams{'prot_len'})),
           scalar($thisScript->param($urlParams{'prot_mass'})),
        );
        my $emPAI = $objSummary->getProteinEmPAI($objProtein->getAccession(), $objProtein->getDB(), $length);
        if ($emPAI > 0) {
        # insert emPAI information
          if ($export_format eq "CSV") {
            print $delimiter . "\"emPAI\"" . $delimiter . round_evenly($emPAI, 2);
          } elsif ($export_format eq "XML") {
            print "<prot_empai>" . round_evenly($emPAI, 2) . "</prot_empai>\n";
          }
        } else {
          if ($export_format eq "CSV") {
            print $delimiter . $delimiter;
          }
        }
      }
    # insert quantitation summary information for the primary hit
      if ($thisScript->param($urlParams{'prot_quant'}) and $objQuantMethod) {
        if ($export_format eq "CSV") {
          print $delimiter . "\"Quantitation summary for protein\"";
        }
        my $proteinRatioType = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());
        my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
        for my $name (@ratio_names) {
          my $ratio = $objMS2Quant->getProteinRatio($objProtein->getAccession(), $objProtein->getDB(), $name);
          if ($export_format eq "CSV") {
            if ($ratio->isMissing()) {
              print $delimiter
                . '"' . &noQuotes($name) . '"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                ;
            } else {
              my ($sd, $p_value, $is_sig) = ('', '', '');
              if ($ratio->getStandardDeviation() < 0) {
                # Ignore reason why unavailable.
                $sd = 'N/A';
              } else {
                $sd = sprintf($ratio->isLogNormal ? '%s' : '[%s]', round_evenly($ratio->getStandardDeviation(), 3));
              }
              if ($ratio->getNormalityPvalue() < 0) {
                # Ignore reason why unavailable.
                $p_value = 'N/A';
              } else {
                if ($ratio->getHypothesisPvalue() < 0) {
                  # Ignore reason why unavailable.
                  $p_value = 'N/A';
                } else {
                  $p_value = sprintf($ratio->isLogNormal ? '%s' : '[%s]', round_evenly($ratio->getHypothesisPvalue(), 4, 'g'));
                  $is_sig = ($ratio->isLogNormal && $ratio->isSignificant() ? '*' : '');
                }
              }
              print $delimiter
                . '"' . &noQuotes($name) . '"'
                . $delimiter
                . '"' . round_evenly($ratio->getValue(), 3) . '"'
                . $delimiter
                . '"' . $ratio->getSampleSize() . '"'
                . $delimiter
                . '"' . $sd . '"'
                . $delimiter
                . '"' . $is_sig . '"'
                . $delimiter
                . '"' . $p_value . '"'
                ;
            }
          } elsif ($export_format eq "XML") {
            if ($ratio->isMissing()) {
              print '<quant_prot_ratio name="'
                . &noXmlTag($name)
                . '" ratio="---" n="---" sd="---" significant="" p-value=""'
                . " />\n"
                ;
            } else {
              my ($sd, $p_value, $is_sig) = ('', '', '');
              if ($ratio->getStandardDeviation() < 0) {
                # Ignore reason why unavailable.
                $sd = 'N/A';
              } else {
                $sd = sprintf($ratio->isLogNormal ? '%s' : '[%s]', round_evenly($ratio->getStandardDeviation(), 3));
              }
              if ($ratio->getNormalityPvalue() < 0) {
                # Ignore reason why unavailable.
                $p_value = 'N/A';
              } else {
                if ($ratio->getHypothesisPvalue() < 0) {
                  # Ignore reason why unavailable.
                  $p_value = 'N/A';
                } else {
                  $p_value = sprintf($ratio->isLogNormal ? '%s' : '[%s]', round_evenly($ratio->getHypothesisPvalue(), 4, 'g'));
                  $is_sig = ($ratio->isLogNormal && $ratio->isSignificant() ? '*' : '');
                }
              }
              print '<quant_prot_ratio name="'
                . &noXmlTag($name)
                . '" ratio="' . round_evenly($ratio->getValue(), 3)
                . '" n="' . $ratio->getSampleSize()
                . '" sd="' . $sd
                . '" significant="' . $is_sig
                . '" p-value="' . $p_value
                . "\" />\n"
                ;
            }
          }
        }
      }
    }
  }

  if ($export_format eq "XML") {
    print "</protein>\n";
  } else {
    print "\n";
  }

}

###############################################################################
# &printPeptideRows()
# export a peptides block
# $_[0] protein object
# $_[1] hit number
# $_[2] flag set for anchor protein
# $_[3] family member number (0 if not family grouping, -1 if family subsets)
# $_[4] 1 if duplicates should be filtered out
# $_[5] cache for duplicate_status_main()
# globals:
# my($objSummary, $export_format);
###############################################################################

sub printPeptideRows {

  my $objProtein = shift;
  my $hitNum = shift;
  my $primary = shift;
  my $member = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;

  my $isPMF = $objResFile->isPMF();

  my @selected_qp = _select_nonduplicate_protein_qp(
    $dupe_state_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  );

  my $needSpacer = 0;
  my $eol = "";   # need to suppress final EOL for CSV so as to add quant info in-line
  for (@selected_qp) {
      my ($i, $queryNum, $rank, $objPeptide) = @$_;
      #
      my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
      my $peptideIdx = $objQuery->getIndex;
      #
    print $eol;
    if ($export_format eq "CSV" && $needSpacer) {
      printProteinRow($objProtein, $hitNum, $member);
      print $delimiter;
    # print $hitNum . $delimiter x $protColumnCount;
    } else {
      $needSpacer = 1;
    }
    if ($export_format eq "XML") {
	print "<peptide query=\"$queryNum\"";
	if ($peptideIdx > -1) {
	    print " index=\"$peptideIdx\"";
	}
      if ($objPeptide->getAnyMatch() && !$isPMF) {
        print " rank=\"" . $objPeptide->getPrettyRank() . "\"";
        if ($objProtein->getPeptideIsBold($i)) {
          print " isbold=\"1\"";
        } else {
          print " isbold=\"0\"";
        }
        if ($objSummary->isPeptideUnique($queryNum, $rank)) {
          print " isunique=\"1\"";
        } else {
          print " isunique=\"0\"";
        }
      }
      print ">\n";
    }
    my $string = fmtPeptideDetails(
        $objPeptide, $peptideIdx, $queryNum, $i, $objProtein, 1, $isPMF, undef, $objSummary,
        \%db_type_query_restrictions
    );
    if ($primary
      && ($thisScript->param($urlParams{'pep_quant'})
      || $thisScript->param($urlParams{'prot_quant'}))) {
      if ($objQuantMethod) {
        $string .= fmtQuantRatioDetails($objPeptide, $queryNum, $i, $objProtein, $hitNum);
      } else {
        # Bug 11992: output quant component assignment for metabolic
        # quantitation.
        $string .= fmtQuantComponentDetails($objPeptide, $queryNum, $i, $objProtein, $hitNum);
      }
    }

    chop $string;
    if ($export_format eq "XML") {
      $string =~ s/>$delimiter</>\n</g;
    }
    if ($string) {
      print $string;
    }
    if ($export_format eq "XML") {
      print "\n</peptide>\n";
    }
    if ($export_format eq "CSV") {
      $eol = "\n";
    }
  }
}

sub _select_nonduplicate_protein_qp {
  my (
    $dupe_state_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  ) = @_;

  my @selected_qp = ();

  if ($filter_duplicates) {
    my %duplicate_groups = ();

    for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
      next if ($objProtein->getPeptideDuplicate($i) == $msparser::ms_protein::DUPE_DuplicateSameQuery);
      my $queryNum = $objProtein->getPeptideQuery($i);
      my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
      my $key = duplicate_group_main($dupe_state_cache, $objResFile, $objParams, $objSummary, $objPeptide, $quantProtocolType);
      push @{ $duplicate_groups{$key} }, $objPeptide;
    }

    for (values %duplicate_groups) {
      my $arr = sort_duplicate_group($_);
      my $master_peptide = $$arr[0];
      my ($queryNum, $rank) = ($master_peptide->getQuery(), $master_peptide->getRank());
      my $i = $objProtein->getPepNumber($queryNum, $rank);
      push @selected_qp, [$i, $queryNum, $rank, $master_peptide];
    }

    @selected_qp = sort { $$a[0] <=> $$b[0] } @selected_qp;
  } else {
    for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
      next if ($objProtein->getPeptideDuplicate($i) == $msparser::ms_protein::DUPE_DuplicateSameQuery);
      my ($queryNum, $rank) = ($objProtein->getPeptideQuery($i), $objProtein->getPeptideP($i));
      my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
      push @selected_qp, [$i, $queryNum, $rank, $objPeptide];
    }
  }

  return @selected_qp;
}

###############################################################################
# &_countProtColumns
# Used to align peptide match data in CSV output
# globals:
# my($thisScript, $anyPepSumMatch)
###############################################################################
sub _countProtColumns {
  #
  my $protColumnCount = 0;

  #prot_hit_num
  $protColumnCount++;
  #prot_family_member
  $protColumnCount++;
  #prot_acc
  $protColumnCount++;

  if ($thisScript->param($urlParams{'prot_desc'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_score'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_thresh'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_expect'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_mass'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_matches'})) {
    $protColumnCount++;
    if ($anyPepSumMatch) {
      #prot_matches_sig
      $protColumnCount++;
      #prot_sequences
      $protColumnCount++;
      #prot_sequences_sig
      $protColumnCount++;
    }
  }
  if ($thisScript->param($urlParams{'prot_cover'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_len'})) {
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_pi'})) {
    $protColumnCount++;
  }

  if ($thisScript->param($urlParams{'prot_tax_str'})){
    $protColumnCount++;
  }

  if ($thisScript->param($urlParams{'prot_tax_id'})) {
    $protColumnCount++;
  }

  if ($thisScript->param($urlParams{'prot_seq'})) {
    $protColumnCount++;
  }

  return $protColumnCount;
}

###############################################################################
# &outputUnassigned()
# export an unassigned block
# $_[0]
# globals:
# my($objSummary, $idForFatalError, $export_format);
###############################################################################

sub outputUnassigned {

  if ($export_format eq "XML") {
    print "<unassigned>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Peptide matches not assigned to protein hits\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  $objSummary->createUnassignedList($msparser::ms_mascotresults::QUERY);
  if ($objSummary->getNumberOfUnassigned() > 0) {
    my $isPMF = $objResFile->isPMF();

  # enumerate the peptides in the unassigned list
    for (my $i = 1; $i <= $objSummary->getNumberOfUnassigned(); $i++) {
      my $objPeptide = $objSummary->getUnassigned($i);
      if (defined($objPeptide)) {
        if ($export_format eq "CSV") {
          print $delimiter x _countProtColumns();
        }
        if ($export_format eq "XML") {
          print "<u_peptide query=\"" . $objPeptide->getQuery() . "\"";
          if ($objPeptide->getAnyMatch() && !$isPMF) {
            print " rank=\"" . $objPeptide->getPrettyRank() . "\"";
            if ($objSummary->getUnassignedIsBold($i)) {
              print " isbold=\"1\"";
            } else {
              print " isbold=\"0\"";
            }
            print " isunique=\"0\"";
          }
          print ">\n";
        }
	my $objQuery = new msparser::ms_inputquery($objResFile, $objPeptide->getQuery());
	my $peptideIdx = $objQuery->getIndex;
        my $string = fmtPeptideDetails(
            $objPeptide, $peptideIdx,$objPeptide->getQuery(), $i, "", 1, $isPMF, undef,
            $objSummary, \%db_type_query_restrictions
        );
        chop $string;
        if ($export_format eq "XML") {
          $string =~ s/>$delimiter</>\n</g;
        }
        if ($string) {
          print $string . "\n";
        }
        if ($export_format eq "XML") {
          print "</u_peptide>\n";
        }
      } else {
        die("Undefined peptide object ", $idForFatalError);
      }
    }
  }

  if ($export_format eq "XML") {
    print "</unassigned>\n";
  }

}

###############################################################################
# &printProteinRow()
# output fields for a protein match
# $_[0] protein object
# $_[1] hit number
# $_[2] family member number (0 if not family grouping, -1 if family subsets)
# globals:
# my($objSummary, $fileIn, %urlParams, $objParams, $export_format,
# $thisScript);
###############################################################################

sub printProteinRow {

  my $objProtein = shift;
  my $hitNum = shift;
  my $member = shift;

  my $dbNumPrefix = "";
  if ($objParams->getNumberOfDatabases > 1 && !$objProtein->isPMFMixture) {
    $dbNumPrefix = $objProtein->getDB . "::";
  }

  my $preferred_acc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession);

  my $string = "";
  if ($export_format eq "CSV") {
    $string .=  $hitNum . $delimiter;
    if ($member > 0) {
      $string .=  $member . $delimiter;
    } elsif ($member < 0) {
      $string .=  $delimiter;
    }
    $string .= "\"" . &noQuotes($dbNumPrefix . $preferred_acc) . "\"$delimiter";
  }

  if ($thisScript->param($urlParams{'prot_desc'})) {
    $string .= &formatElement($export_format, "", 1, "prot_desc",
      $protData->mustGetProteinDescription($preferred_acc, $objProtein->getDB)) . $delimiter;
  }
  if ($thisScript->param($urlParams{'prot_score'})) {
    $string .= &formatElement($export_format, "", 0, "prot_score",
      round_evenly($objProtein->getScore(), 0)) . $delimiter;
  }
  if ($thisScript->param($urlParams{'prot_thresh'})) {
    $string .= &formatElement($export_format, "", 0, "prot_thresh",
      $objSummary->getProteinThreshold(1 / $objSummary->getProbabilityThreshold)) . $delimiter;
  }
  if ($thisScript->param($urlParams{'prot_expect'})) {
    $string .= &formatElement($export_format, "", 0, "prot_expect",
      round_evenly(1 / $objSummary->getProbOfProteinBeingRandomMatch($objProtein->getScore()), 2, 'g')) . $delimiter;
  }
  if ($thisScript->param($urlParams{'prot_mass'})) {
    if ($protData->mustGetProteinMass($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB)) {
      $string .= &formatElement($export_format, "", 0, "prot_mass",
        round_evenly($protData->mustGetProteinMass($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB), 0)) . $delimiter;
    } else {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    }
  }
  if ($thisScript->param($urlParams{'prot_matches'})) {
    $string .= &formatElement($export_format, "", 0, "prot_matches",
      $objProtein->getNumDisplayPeptides(0)) . $delimiter;
    if ($anyPepSumMatch) {
      $string .= &formatElement($export_format, "", 0, "prot_matches_sig",
        $objProtein->getNumDisplayPeptides(1)) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "prot_sequences",
        $objProtein->getNumDistinctPeptides(0, $msparser::ms_protein::DPF_SEQUENCE)) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "prot_sequences_sig",
        $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE)) . $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'prot_cover'}) || $thisScript->param($urlParams{'prot_len'})) {
    my ($length) = $protData->mustGetProteinLen($objProtein, 1, scalar($thisScript->param($urlParams{'prot_mass'})));
    if ($thisScript->param($urlParams{'prot_cover'})) {
      my $coverage = "";
      if ($length > 0) {
        $coverage = $objProtein->getCoverage() * 100 / $length;
      }
      if ($coverage) {
        $string .= &formatElement($export_format, "", 0, "prot_cover",
          round_evenly($coverage, 1)) . $delimiter;
      } else {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    }
    if ($thisScript->param($urlParams{'prot_len'})) {
      if ($length) {
        $string .= &formatElement($export_format, "", 0, "prot_len", $length) . $delimiter;
      } else {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    }
  }
  if ($thisScript->param($urlParams{'prot_pi'})) {
    my $pI = $protData->getProteinpI($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    if ($pI) {
      $string .= &formatElement($export_format, "", 0, "prot_pi", $pI) . $delimiter;
    } else {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    }
  }

  if ($thisScript->param($urlParams{'prot_tax_str'}) || $thisScript->param($urlParams{'prot_tax_id'})) {
    my ($taxID, $species) = $protData->getProteinTaxonomy($preferred_acc, $objProtein->getDB);
    if ($thisScript->param($urlParams{'prot_tax_str'})) {
      if ($species) {
        $string .= &formatElement($export_format, "", 1, "prot_tax_str", $species) . $delimiter;
      } else {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    }
    if ($thisScript->param($urlParams{'prot_tax_id'})) {
      if ($taxID) {
        $string .= &formatElement($export_format, "", 0, "prot_tax_id", $taxID) . $delimiter;
      } else {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    }
  }
  if ($thisScript->param($urlParams{'prot_seq'})) {
    my ($seq, $err) = $protData->mustGetProteinSeq($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB, $shipper);
    if ($seq) {
      $string .= &formatElement($export_format, "", 0, "prot_seq", $seq) . $delimiter;
    } else {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    }
  }

  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string;
    if ($export_format eq "XML") {
      print "\n";
    }
  }

}

###############################################################################
# &_outputXlPepFamilyMembers()
# returns a formatted string with the protein family members for a crosslinked
# peptide match.
# $_[0] peptide object
# $_[1] is the summary object
# $_[2] is a reference to an array that contains the considered PSM components
# globals:
# %pepSrcFamilyMembers, $export_format, $ShowSameSets
###############################################################################
sub _outputXlPepFamilyMembers {
    my ($objPeptide, $objSummary, $psms) = @_;
    my $string = "";
    if ($objPeptide->getNumberOfLinkedPeptides() > 0) {
	my %prots = (
	    $msparser::ms_peptide::PSM_CROSSLINK_ALPHA => [],
	    $msparser::ms_peptide::PSM_CROSSLINK_BETA => [],
	    $msparser::ms_peptide::PSM_COMPLETE => []
	    );

	if ($ShowSameSets) {
	    my ($db, $psmComp) = (
		msparser::vectori->new(),
		msparser::vectori->new()
		);
	    my $accs = $objSummary->getAllProteinsWithThisPepMatch($objPeptide->getQuery(), $objPeptide->getRank(),
								   msparser::vectori->new(),
								   msparser::vectori->new(),
								   msparser::VectorString->new(),
								   msparser::VectorString->new(),
								   msparser::vectori->new(),
								   msparser::vectori->new(),
								   $db, $psmComp);

	    while (my ($idx, $acc) = each @{$accs}) {
		my $dbNumPrefix = ($objParams->getNumberOfDatabases == 1) ? "" : $db->get($idx) . "::";
		my $fullAcc = $dbNumPrefix . $protData->preferredAccession($db->get($idx), $acc);
		push @{ $prots{ $psmComp->get($idx) } }, $fullAcc;
	    }
	}
	else {
	    my $members = peptide_source_family_members(\%pepSrcFamilyMembers, $objSummary, $objPeptide->getQuery(), $objPeptide->getRank());
	    for (@$members) {
		my ($family, $db, $acc, $psmComponent) = @$_{qw(family db acc psmComponent)};
		my $dbNumPrefix = ($objParams->getNumberOfDatabases == 1) ? "" : $db . "::";
		my $fullAcc = $dbNumPrefix . $protData->preferredAccession($db, $acc);
		push @{ $prots{ $psmComponent } }, $fullAcc;
	    }
	}

	for my $psm(@{$psms}) {
	    my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '_alpha';
	    my $protsStr = join(";", @{$prots{$psm}});
	    $string .= &formatElement($export_format, "", 1, "prot_acc".$suffix, $protsStr) . $delimiter;
	}
    }
    elsif ($anyXlPepSumMatch and $export_format eq "CSV") {
	$string .= $delimiter . $delimiter;
    }
    return $string;
}
###############################################################################
# &fmtPeptideDetails()
# returns fields for a single peptide match
# $_[0] peptide object
# $_[1] query source index
# $_[2] query number
# $_[3] peptide number within protein
# $_[4] protein object (empty for unassigned list)
# $_[5] 1 to begin line with query number
# $_[6] if the results file is PMF
# $_[7] is the site analysis data (or undef if missing)
# $_[8] is the summary object
# $_[9..] is a hash of extra parameters to determine_peptide_db_type
# globals:
# my(%urlParams, $objResFile, $massDP, $thisScript,
# $export_format);
###############################################################################

sub fmtPeptideDetails {
  my ($objPeptide, $peptideIndex, $queryNum, $i, $objProtein, $printQueryNum, $isPMF,
      $siteAnalysisData, $objSummary, $db_type_query_restrictions) = @_;

  my $fmtElemIfCond = sub {
      my($condition, $need_quotes, $name, $value) = @_;

      if ($condition) {
	  return &formatElement($export_format, "", $need_quotes, $name, $value) . $delimiter;
      }
      elsif ($export_format eq "CSV") {
	  return $delimiter;
      }
      else {
	  return '';
      }
  };

  my @psms = ($anyXlPepSumMatch) ? ($msparser::ms_peptide::PSM_CROSSLINK_ALPHA, $msparser::ms_peptide::PSM_CROSSLINK_BETA) : ($msparser::ms_peptide::PSM_COMPLETE);
  my $isXlMatch = ($objPeptide->getNumberOfLinkedPeptides() > 0);
  my %isNeeded = ($msparser::ms_peptide::PSM_CROSSLINK_ALPHA, 1, $msparser::ms_peptide::PSM_CROSSLINK_BETA, 1, $msparser::ms_peptide::PSM_COMPLETE, 1);
  if (not $isXlMatch) {
      $isNeeded{$msparser::ms_peptide::PSM_CROSSLINK_BETA} = 0;
      $psms[0] = $msparser::ms_peptide::PSM_COMPLETE;
  }

  # Always output all the *_beta fields if it is a crosslinked match
  my %shouldBeOutput = ($msparser::ms_peptide::PSM_CROSSLINK_ALPHA, $isXlMatch, $msparser::ms_peptide::PSM_CROSSLINK_BETA, $isXlMatch, $msparser::ms_peptide::PSM_COMPLETE, 1);

  my $string = "";
  if ($export_format eq "CSV") {
    if ($printQueryNum) {
	$string .= $queryNum . $delimiter;
	$string .= $peptideIndex . $delimiter;
    }
    if ($objPeptide->getAnyMatch() && !$isPMF) {
      $string .= $objPeptide->getPrettyRank() . $delimiter;
      if ($printQueryNum) {
        if ($objProtein) {
          if ($objProtein->getPeptideIsBold($i)) {
            $string .= "1" . $delimiter;
          } else {
            $string .= "0" . $delimiter;
          }
          if ($objSummary->isPeptideUnique($queryNum, $objPeptide->getRank)) {
            $string .= "1" . $delimiter;
          } else {
            $string .= "0" . $delimiter;
          }
        } elsif ($objSummary->getUnassignedIsBold($i)) {
          $string .= "1" . $delimiter;
          $string .= "0" . $delimiter;
        } else {
          $string .= "0" . $delimiter;
          $string .= "0" . $delimiter;
        }
      } else {
        $string .= $delimiter . $delimiter;
      }
    } else {
      $string .= $delimiter . $delimiter . $delimiter;
    }
  }

  # BUG#13396: Export accessions of proteins in Alpha and Beta peptides
  $string .= _outputXlPepFamilyMembers($objPeptide, $objSummary, \@psms);

  $string .= &formatElement($export_format, "", 0, "pep_exp_mz",
    round_evenly($objPeptide->getObserved(), $massDP)) . $delimiter;
  if ($thisScript->param($urlParams{'pep_exp_mr'})) {
    my $mrExpt = $objResFile->getObservedMrValue($queryNum);
    if ($objPeptide->getAnyMatch) {
      $mrExpt = $objPeptide->getMrExperimental;    # only works if there's a match
    }
    $string .= &formatElement($export_format, "", 0, "pep_exp_mr",
      round_evenly($mrExpt, $massDP)) . $delimiter;
  }
  if ($thisScript->param($urlParams{'pep_exp_z'})) {
    $string .= &formatElement($export_format, "", 0, "pep_exp_z", $objPeptide->getCharge()) . $delimiter;
  }
  if ($thisScript->param($urlParams{'pep_calc_mr'})) {
    $string .= $fmtElemIfCond->($objPeptide->getAnyMatch(), 0, "pep_calc_mr", round_evenly($objPeptide->getMrCalc(), $massDP));
  }
  if ($thisScript->param($urlParams{'pep_delta'})) {
    $string .= $fmtElemIfCond->($objPeptide->getAnyMatch(), 0, "pep_delta", round_evenly($objPeptide->getDelta(), $massDP));
  }
  if ($thisScript->param($urlParams{'pep_start'})) {
    for my $psm(@psms) {
      my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '';
      if ($objProtein) {
	  my $pepStartStr = $isNeeded{$psm} ? $objProtein->getPeptideStart($i, $psm) : '';
	  $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_start".$suffix, $pepStartStr);
      } elsif ($export_format eq "CSV") {
	  $string .= $delimiter;
      }
    }
  }
  if ($thisScript->param($urlParams{'pep_end'})) {
    for my $psm(@psms) {
      my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '';
      if ($objProtein) {
	  my $pepEndStr = $isNeeded{$psm} ? $objProtein->getPeptideEnd($i, $psm) : '';
	  $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_end".$suffix, $pepEndStr);
      } elsif ($export_format eq "CSV") {
	  $string .= $delimiter;
      }
    }
  }
  if ($thisScript->param($urlParams{'pep_miss'})) {
      $string .= $fmtElemIfCond->($objPeptide->getAnyMatch(), 0, "pep_miss", $objPeptide->getMissedCleavages());
  }
  if ($thisScript->param($urlParams{'pep_score'})) {
      $string .= $fmtElemIfCond->($objPeptide->getAnyMatch(), 0, "pep_score", round_evenly($objPeptide->getIonsScore, 2));
  }
  if ($thisScript->param($urlParams{'pep_homol'})) {
    my $thr = $objSummary->getHomologyThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $objPeptide->getRank());
    if (not $objPeptide->getAnyMatch or $thr <= 0
        or ($objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant and $OmitErrTolThresholds)) {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    } else {
        $string .= &formatElement($export_format, "", 0, "pep_homol", $thr) . $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_ident'})) {
    my $dbType = $objPeptide->getIsFromLibrary()
        ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
        : $msparser::ms_mascotresults::DM_FASTA
        ;
    my $thr = $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $dbType);
    if (not $objPeptide->getAnyMatch() or $thr <= 0
        or ($objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant and $OmitErrTolThresholds)) {
      $string .= $delimiter;
    } else {
      $string .= &formatElement($export_format, "", 0, "pep_ident", $thr) . $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_expect'})) {
    if ($objPeptide->getAnyMatch()) {
      my $expect = $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $queryNum, $objPeptide->getRank());
      if ($expect > 0) {
        $string .= &formatElement($export_format, "", 0, "pep_expect",
          round_evenly($expect, 2, 'g')) . $delimiter;
      }
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }

  if ($thisScript->param($urlParams{'pep_seq'})) {
    for my $psm(@psms) {
      my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '';
      if ($objPeptide->getAnyMatch()) {
        if ($objProtein) {
          my $res = $isNeeded{$psm} ? $objProtein->getPeptideResidueBefore($i, $psm) : '';
          $res =~ s/@/-/;
          $res =~ s/\?//;
	  $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_res_before".$suffix, $res);
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
	my $pepSeq = $isNeeded{$psm} ? $objPeptide->getPeptideStr(1, $psm) : '';
	$string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_seq".$suffix, $pepSeq);
        if ($objProtein) {
          my $res = $isNeeded{$psm} ? $objProtein->getPeptideResidueAfter($i, $psm) : '';
          $res =~ s/@/-/;
          $res =~ s/\?//;
	  $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_res_after".$suffix, $res);
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } elsif ($export_format eq "CSV") {
	  for my $psm(@psms) {
	    $string .= "$delimiter$delimiter$delimiter";
	  }
	}
    }
  }
  if ($thisScript->param($urlParams{'pep_frame'})) {
    for my $psm(@psms) {
      my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '';
      if ($objProtein && $objProtein->getPeptideFrame($i, $psm) > 0) {
	my $pepFrame = $isNeeded{$psm}  ? $objProtein->getPeptideFrame($i, $psm) : '';
	$string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_frame".$suffix, $pepFrame);
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    }
  }
  if ($thisScript->param($urlParams{'pep_var_mod'})) {
    for my $psm(@psms) {
      my $suffix = ($psm == $msparser::ms_peptide::PSM_CROSSLINK_BETA) ? '_beta' : '';
      if ($objPeptide->getAnyMatch()) {
         my $pepVarMod = $isNeeded{$psm} ? $objSummary->getReadableVarMods($queryNum, $objPeptide->getRank(), 2, $psm) : '';
         $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 1, "pep_var_mod".$suffix, $pepVarMod);

	 # need to re-format string from 00000012000 to 0.000001200.0
	 # to prevent Excel from re-formatting it to 12000
	 my $tmpvar = $isNeeded{$psm} ? $objPeptide->getVarModsStr($psm) : '';
	 my $tmpsum = $isNeeded{$psm} ? $objPeptide->getSummedModsStr($psm) : '';
	 my $tmplocal = $isNeeded{$psm} ? $objPeptide->getLocalModsStr($psm) : '';
	 my $monolinks = $isNeeded{$psm} ? $objPeptide->getMonoLinkStr($psm) : '';

	 for ($tmpvar, $tmpsum, $tmplocal, $monolinks) {
	   if (/[1-9A-Za-z]/) {
	     s/(.)(.*)(.)/$1\.$2\.$3/;
	   } else {
	     $_ = "";
	   }
	 }
	 $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_var_mod_pos".$suffix, $tmpvar);
	 $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_summed_mod_pos".$suffix, $tmpsum);
	 $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_local_mod_pos".$suffix, $tmplocal);

	 if ($hasXlMethod) {
	   $string .= $fmtElemIfCond->($shouldBeOutput{$psm}, 0, "pep_monolink_mod_pos".$suffix, $monolinks);
	 }

      } elsif ($export_format eq "CSV") {
	  $string .= $delimiter . $delimiter . $delimiter;
	  if ($hasXlMethod) {
	    $string .= $delimiter;
	  }
      }
    }

    my $rank = $objPeptide->getRank();
    my $srcRank = $rank; #srcRank == rank in crosslink search
    my $srcSection = $msparser::ms_mascotresfile::SEC_PEPTIDES;
    if ($isXlMatch) {
	$srcSection = $msparser::ms_mascotresfile::SEC_CROSSLINK_PEPTIDES;
	my $looplinkedSitesAlpha = $objResFile->getSectionValueStr($srcSection, "q${queryNum}\_p${srcRank}_looplinked_sites_1");
	# Add a space at the begining of the string to prevent Excel from converting it into a time.
	$looplinkedSitesAlpha = ($looplinkedSitesAlpha ne "" and $export_format eq "CSV") ? " ".$looplinkedSitesAlpha : $looplinkedSitesAlpha;
	my $looplinkedSitesBeta = $objResFile->getSectionValueStr($srcSection, "q${queryNum}\_p${srcRank}_looplinked_sites_2");
	$looplinkedSitesBeta = ($looplinkedSitesBeta ne "" and $export_format eq "CSV") ? " ".$looplinkedSitesBeta : $looplinkedSitesBeta;
	$string .= $fmtElemIfCond->(1, 1, "pep_looplink_pos", $looplinkedSitesAlpha);
	$string .= $fmtElemIfCond->(1, 1, "pep_looplink_pos_beta", $looplinkedSitesBeta);

	my $linkedSites = $objResFile->getSectionValueStr($srcSection, "q${queryNum}\_p${srcRank}_linked_sites");
	$string .= $fmtElemIfCond->(1, 1, "pep_linked_sites", $linkedSites);
    }
    else {
	if ($hasXlMethod) {
	  my $looplinkedSites = $objResFile->getSectionValueStr($srcSection, "q${queryNum}\_p${srcRank}_looplinked_sites");
	  $looplinkedSites = ($looplinkedSites ne "" and $export_format eq "CSV") ? " ".$looplinkedSites : $looplinkedSites;
	  $string .= $fmtElemIfCond->(1, 1, "pep_looplink_pos", $looplinkedSites);

	  if (($export_format eq "CSV") and $anyXlPepSumMatch) {
	      $string .= $delimiter.$delimiter;
	  }
	}
    }
    if ($objPeptide->getAnyMatch()) {
      if ($siteAnalysisData) {
        if (exists $$siteAnalysisData{ $objPeptide->getRank }) {
          my $analysis = $$siteAnalysisData{ $objPeptide->getRank() };
          $string .= &formatElement($export_format, "", 0, "pep_var_mod_conf", round_evenly(100 * $$analysis{'confidence'}, 2) . '%' ) . $delimiter;
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    } elsif ($siteAnalysisData and $export_format eq "CSV") {
      $string .= $delimiter;
    }
  }

  if ($thisScript->param($urlParams{'pep_num_match'})) {
    $string .= $fmtElemIfCond->($objPeptide->getAnyMatch(), 0, "pep_num_match", $objPeptide->getNumIonsMatched());
  }
  if ($thisScript->param($urlParams{'pep_scan_title'})) {
  # creating query objects is much too slow
    my $title = $objResFile->getQuerySectionValueStr($queryNum, "Title");
    $title =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
    $string .= $fmtElemIfCond->(1, 1, "pep_scan_title", $title);
  }
  if ($thisScript->param($urlParams{'pep_source'})) {
    my $hitNum = 0;
    if ($objProtein) {
      $hitNum = $objProtein->getHitNumber;
    }
    my %type = (is_unmatched => 1);
    if ($db_type_query_restrictions) {
      %type = determine_peptide_db_type(
        $hitNum, $objSummary, $objPeptide, %$db_type_query_restrictions
      );
    }
    if ($type{'is_unmatched'}) {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    } else {
      my $source;
      if ($type{'is_from_library'}) {
        $source = 'SL';
      } elsif ($type{'has_nucleic_acid_accessions'}) {
        if ($type{'has_amino_acid_accessions'}) {
          $source = 'XA';
        } else {
          $source = 'NA';
        }
      } else {
        $source = 'AA';
      }
      $string .= formatElement($export_format, "", 1, "pep_source", $source) . $delimiter;
    }
  }

  return $string;

}

###############################################################################
# &fmtQuantRatioDetails()
# returns quant fields for a single peptide match
# $_[0] peptide object
# $_[1] query number
# $_[2] peptide number
# $_[3] protein object
# $_[4] hit number
# globals:
# my($objSummary, %urlParams, $objResFile, $massDP, $thisScript,
# $export_format);
###############################################################################

sub fmtQuantRatioDetails {

  my ($objPeptide,
    $queryNum,
    $i,
    $objProtein,
    $hitNum,
    ) = @_;

  my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
  my $key = msparser::ms_peptide_quant_key->new($objPeptide->getQuery(), $objPeptide->getRank());

  my @intensityValues = map {
      round_evenly($_, 6, 'g')
  } @{ $objMS2Quant->getComponentIntensities($key) };
  my @ratioValues = ();

  if (0 == @intensityValues) {
    @ratioValues = ('---') x @ratio_names;
  } else {
    for my $i (0 .. $#ratio_names) {
      my $ratio = $objMS2Quant->getPeptideRatio($key, $ratio_names[$i]);
      if ($ratio->isMissing()) {
        $ratioValues[$i] = '---';
      } elsif ($ratio->isInfinite()) {
        $ratioValues[$i] = '###';
      } else {
        $ratioValues[$i] = round_evenly($ratio->getValue(), 3);
      }
    }
  }

  my $string = "";
  if ($thisScript->param($urlParams{'pep_quant'})) {
    for my $i (0 .. $#ratio_names) {
      my $name = $ratio_names[$i];
      if ($export_format eq "CSV") {
        $string .= "\"" . &noQuotes($name) . "\"" . $delimiter;
        if ($ratioValues[$i] eq "---" || $ratioValues[$i] eq "###") {
          $string .= "\"" . $ratioValues[$i] . "\"" . $delimiter;
        } else {
          $string .= $ratioValues[$i] . $delimiter;
        }
      } elsif ($export_format eq "XML") {
        $string .= "<quant_pep_ratio name=\""
          . &noXmlTag($name)
          . "\" ratio=\""
          . $ratioValues[$i]
          . "\" />"
          . $delimiter;
      }
    }

    for (my $i = 0; $i < $objQuantMethod->getNumberOfComponents(); $i++) {
      my $name = $objQuantMethod->getComponentByNumber($i)->getName();
      my $intensity = $intensityValues[$i];
      $intensity = '---' if (not defined $intensityValues[$i] or $intensityValues[$i] eq '');
      if ($export_format eq "CSV") {
        $string .= "\"" . &noQuotes($name) . "\"" . $delimiter;
        if ($intensity eq '---') {
          $string .= '"---"' . $delimiter;
        } else {
          $string .= $intensity . $delimiter;
        }
      } elsif ($export_format eq "XML") {
        $string .= "<quant_pep_intensity name=\""
          . &noXmlTag($name)
          . "\" intensity=\""
          . $intensity
          . "\" />"
          . $delimiter;
      }
    }
  }

  return $string;

}

###############################################################################
# &fmtQuantComponentDetails()
# returns quant component assignment for a single peptide match
# $_[0] peptide object
# $_[1] query number
# $_[2] peptide number
# $_[3] protein object
# $_[4] hit number
# globals:
# my($objSummary, %urlParams, $objResFile, $massDP, $thisScript,
# $export_format);
###############################################################################

sub fmtQuantComponentDetails {
  my ($objPeptide,
    $queryNum,
    $i,
    $objProtein,
    $hitNum,
    ) = @_;

  my $string = "";
  if ($thisScript->param($urlParams{'pep_quant'})) {
    my $comp_name = $objPeptide->getComponentStr();
    if ($export_format eq "CSV") {
      $string .= "\"" . &noQuotes($comp_name) . "\"" . $delimiter;
    } elsif ($export_format eq "XML") {
      $string .= "<quant_pep_component name=\""
        . &noXmlTag($comp_name)
        . "\" />"
        . $delimiter;
    }
  }

  return $string;

}

###############################################################################
# &outputHeader()
# output header block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputHeader {
  my $j = 0;

  if ($export_format eq "XML") {
    print "<header>\n";
    $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'full');
    $j = outputHeaderWarnings($j);
    print "</header>\n";
  } elsif ($export_format eq "CSV") {
    my $line = '--------------------------------------------------------';
    if (@fileIn == 1) {
      print "\n\"Header\"$delimiter\"$line\"\n\n";
      $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'full');
      $j = outputHeaderWarnings($j);
    } else {
      my $l = $line;
      my $header_str = sprintf('results collection set of %d files', scalar(@fileIn));
      substr($l, 4, length($header_str)) = $header_str;
      print "\n\"Header\"$delimiter\"$l\"\n\n";
      $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'combined');
      $j = outputHeaderWarnings($j);
      for my $i (1 .. $objResFile->getNumberOfResfiles()) {
        my $l = $line;
        my $fileIn = $fileIn[$i-1];
        substr($l, 4, length($fileIn)) = $fileIn;
        print "\n\"Header\"$delimiter\"$l\"\n\n";
        my $resfile = $objResFile->getResfile($i);
        $j = outputHeaderOfFile($export_format, $resfile, $resfile->params, $j, 'part');
      }
    }
  }
}

sub outputHeaderOfFile {
  my ($export_format, $objResFile, $objParams, $j, $header_type) = @_;

  if ($header_type ne 'combined') {
    print &formatElement($export_format, "Search title", 1, "COM", $objParams->getCOM()) . "\n";
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = gmtime ($objResFile->getDate);
    my $timestamp = sprintf("%04d-%02d-%02dT%02d:%02d:%02dZ",
      $year+1900, $mon+1, $mday, $hour, $min, $sec);
    print &formatElement($export_format, "Timestamp", 1, "Date", $timestamp) . "\n";
    print &formatElement($export_format, "User", 1, "USERNAME", $objParams->getUSERNAME()) . "\n";
    print &formatElement($export_format, "Email", 1, "USEREMAIL", $objParams->getUSEREMAIL()) . "\n";
  }
  if ($header_type eq 'combined' or $header_type eq 'full') {
    if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
      print &formatElement($export_format, "Report URI", 1, "URI", $URI . "cgi/master_results_2.pl?fileset=" . $fileset) . "\n";
    } else {
      print &formatElement($export_format, "Report URI", 1, "URI", $URI . "cgi/master_results.pl?file=" . $fileIn[0]) . "\n";
    }
  }
  if ($header_type ne 'combined') {
    if ($objParams->getDATAURL() eq '') {
      print &formatElement($export_format, "Peak list data path", 1, "FILENAME", $objParams->getFILENAME()) . "\n";
    } else {
      print &formatElement($export_format, "Peak list data URL", 1, "DATAURL", $objParams->getDATAURL()) . "\n";
    }
    print &formatElement($export_format, "Peak list format", 1, "FORMAT", $objParams->getFORMAT()) . "\n";
    print &formatElement($export_format, "Search type", 0, "SEARCH", $objParams->getSEARCH()) . "\n";
    print &formatElement($export_format, "Mascot version", 1, "MascotVer", $objResFile->getMascotVer()) . "\n";
    my($dbString, $fastaVerString);
    if ($objParams->getNumberOfDatabases == 1) {
      $dbString = $objParams->getDB(1);
      $fastaVerString = $objResFile->getFastaVer(1);
    } else {
      my(@tempList1, @tempList2);
      for (my $i=1; $i <= $objParams->getNumberOfDatabases; $i++) {
        push @tempList1, $i . "::" . $objParams->getDB($i);
        push @tempList2, $i . "::" . $objResFile->getFastaVer($i);
      }
      $dbString = join(" ", @tempList1);
      $fastaVerString = join(" ", @tempList2);
    }
    print &formatElement($export_format, "Database", 1, "DB", $dbString) . "\n";
    print &formatElement($export_format, "Fasta file", 1, "FastaVer", $fastaVerString) . "\n";
  # wait for parser function ### debug ###
    print &formatElement($export_format, "Total sequences", 0, "NumSeqs", $objResFile->getNumSeqs()) . "\n";
    print &formatElement($export_format, "Total residues", 0, "NumResidues", $objResFile->getNumResidues()) . "\n";
    print &formatElement($export_format, "Sequences after taxonomy filter", 0, "NumSeqsAfterTax", $objResFile->getNumSeqsAfterTax()) . "\n";
    if ($objResFile->isErrorTolerant()) {
      if (my $temp = $objParams->getACCESSION()) {
        my $i = $temp =~ tr/,//;
        print &formatElement($export_format, "Number of sequences searched in error tolerant mode",
          0, "error_tolerant_num", $i + 1) . "\n";
      } elsif ($objResFile->versionGreaterOrEqual(2, 7, 101)) {
        # Bug 11120: The number is available from version 2.5 onward,
        # but it's sometimes incorrect in cluster mode. Correct number
        # is present in 2.8 and later.
        print &formatElement($export_format, "Number of sequences searched in error tolerant mode",
          0, "error_tolerant_num", $objResFile->getNumEtSeqsSearched()) . "\n";
      }
    }
  }
  if ($header_type eq 'combined' or $header_type eq 'full') {
    print &formatElement($export_format, "Number of queries", 0, "NumQueries", $objResFile->getNumQueries()) . "\n";
  }
# Display any warnings from result file header

  if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn0')){
    while ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j)){
      print &formatElement($export_format, "Warning", 1, "warning",
        $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j), number => $j) . "\n";
      $j++;
    }
  }
# Display a warning if this is an orphan manual error tolerant search
  if ($objResFile->isErrorTolerant() and not $objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY) and not $objParams->getErrTolParentFilename()) {
    print &formatElement($export_format, "Warning", 1, "warning",
      "Scores from standard search not available. Better matches may exist in other proteins.", number => $j) . "\n";
    $j++;
  }

  return $j;
}

sub outputHeaderWarnings {
  my ($j) = @_;

# Display any warnings from Parser
  for (my $k = 1; $k <= $objResFile->getNumberOfErrors(); $k++) {
    print &formatElement($export_format, "Warning", 1, "warning",
      $objResFile->getErrorString($k), number => $j) . "\n";
    $j++;
  }
# Display any warnings from quantitation
  if (@quantWarnings) {
    foreach (@quantWarnings) {
      print &formatElement($export_format, "Warning", 1, "warning",
        $_, number => $j) . "\n";
      $j++;
    }
  }
  if ($objMS2Quant) {
    my $objErr = $objMS2Quant->getErrorHandler();
    for (my $i = 1; $i <= $objErr->getNumberOfErrors(); $i++) {
      print formatElement($export_format, 'Warning', 1, 'warning',
        sprintf('%s (%s)', $objErr->getErrorString($i), $objErr->getErrorNumber($i)),
        number => $j++
      );
    }
  }

  return $j;
}

###############################################################################
# &outputDecoy()
# output decoy database search statistics
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputDecoy {

  if ($export_format eq "XML") {
    print "<decoy>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Decoy\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  my %count_types = ('PSM' => $msparser::ms_mascotresults::DS_COUNT_PSM, 'Sequence' => $msparser::ms_mascotresults::DS_COUNT_SEQUENCE);
  my %labels = ('PSM' => 'PSM', 'Sequence' => 'Sequence');
  for my $key (sort keys %count_types) {
      my $label = $labels{$key};
      print &formatElement($export_format, "${key} count in target above identity threshold",
			   0, "Num${label}sAboveIdentity", $objSummary->getNumHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_types{$key})) . "\n";
      print &formatElement($export_format, "${key} count in decoy above identity threshold",
			   0, "NumDecoy${label}sAboveIdentity", $objSummary->getNumDecoyHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_types{$key})) . "\n";
  }

  if ($objResFile->isPMF()){
    my $TF = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_SUMMARY, "h1");
    $TF = (split(/,/, $TF))[1];
    $TF = round_evenly($TF, 0);
    my $TD = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_DECOYSUMMARY, "h1");
    $TD = (split(/,/, $TD))[1];
    $TD = round_evenly($TD, 0);
    print &formatElement($export_format, "Highest score in search of target database",
      0, "HighestScoreProtein", $TF) . "\n";
    print &formatElement($export_format, "Highest score in search of decoy database",
      0, "HighestScoreDecoyProtein", $TD) . "\n";
  } elsif ($percolate != 1) {
      for my $key (sort keys %count_types) {
	my $label = $labels{$key};
        print &formatElement($export_format, "${key} count in target above homology threshold",
			       0, "Num${label}sAboveHomology", $objSummary->getNumHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_types{$key})) . "\n";
        print &formatElement($export_format, "${key} count in decoy above homology threshold",
			       0, "NumDecoy${label}sAboveHomology", $objSummary->getNumDecoyHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_types{$key})) . "\n";
      }
    }

    if ($showFDRs) {
      my $FDR;
      for my $key (sort keys %count_types) {
	my $label = $labels{$key};
	my $HF = $objSummary->getNumHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_types{$key});
	my $HD = $objSummary->getNumDecoyHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_types{$key});
	if ($HF > 0) {
	  $FDR = round_evenly(100 * $HD / $HF, 2);
	  print &formatElement($export_format, "${key} FDR using homology threshold", 0, "${label}FDRHomology", $FDR) . "\n";
	}

	my $IF = $objSummary->getNumHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_types{$key});
	my $ID = $objSummary->getNumDecoyHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_types{$key});
	if ($IF > 0) {
	  $FDR = round_evenly(100 * $ID / $IF, 2);
	  print &formatElement($export_format, "${key} FDR using identity threshold", 0, "${label}FDRIdentity", $FDR) . "\n";
	}
      }

      my ($targetCount, $decoyCount, $proteinFDR) = ExportDat::Util::calculateProteinFDR($objSummary, $objSummaryDecoy, $objResFile);
      if (defined $proteinFDR and $proteinFDR > 0) {
        print &formatElement($export_format, "Protein family member count in target", 0, "NumProteinsTarget", $targetCount) . "\n";
	print &formatElement($export_format, "Protein family member count in decoy", 0, "NumProteinsDecoy", $decoyCount) . "\n";
	print &formatElement($export_format, "Protein false discovery rate", 0, "ProteinFDR", round_evenly(100 * $proteinFDR, 2)) . "\n";
      }
  }

  if ($export_format eq "XML") {
    print "</decoy>\n";
  }

}


###############################################################################
# &outputFixedMods()
# output fixed modifications block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputFixedMods {

    if ($export_format eq "XML") {
      print "<fixed_mods>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Fixed modifications\"$delimiter\"--------------------------------------------------------\"\n\n";
      print "\"Identifier\"$delimiter";
      print "\"Name\"$delimiter";
      print "\"Delta\"$delimiter";
      print "\"Neutral loss\"\n";
    }

  my $string = "";
  my $i = 1;
  my $letter;

# $objParams->getFixedModsName() will only work with files from 2.1 and later
  if ($objParams->getFixedModsName(1)) {
    while ($objParams->getFixedModsName($i)) {
      $letter = $i;
      if ($i > 9) {
        $letter = chr($i + 55);
      }
      if ($export_format eq "CSV") {
        $string .= $letter . $delimiter;
      } elsif ($export_format eq "XML") {
        $string .= "<modification identifier=\"$letter\">$delimiter";
      }
      $string .= &formatElement($export_format, "", 1, "name", $objParams->getFixedModsName($i)) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "delta", $objParams->getFixedModsDelta($i)) . $delimiter;
      my $loss = $objParams->getFixedModsNeutralLoss($i);
      if ($loss) {
        $string .= &formatElement($export_format, "", 0, "neutral_loss", $loss) . $delimiter;
      }
      chop $string;
      $i++;
      if ($export_format eq "XML") {
        $string .= "$delimiter</modification>";
      }
      $string .= "\n";
    }
  } else {
    my @modList = split(/,/, $objParams->getMODS);
    foreach my $mod (@modList) {
      $letter = $i;
      if ($i > 9) {
        $letter = chr($i + 55);
      }
      if ($export_format eq "CSV") {
        $string .= $letter . $delimiter;
      } elsif ($export_format eq "XML") {
        $string .= "<modification identifier=\"$letter\">$delimiter";
      }
      my @modParams = &getResidueMods($mod, "N");
      unless (@modParams) {
        @modParams = &getTerminalMods($mod, "N");
      }
      $string .= &formatElement($export_format, "", 1, "name", $mod) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "delta", $modParams[1]) . $delimiter;
      chop $string;
      $i++;
      if ($export_format eq "XML") {
        $string .= "$delimiter</modification>";
      }
      $string .= "\n";
    }
  }

  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string . "\n";
  }

  if ($export_format eq "XML") {
    print "</fixed_mods>\n";
  }

}

###############################################################################
# &outputVarMods()
# output variable modifications block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputVarMods {

    if ($export_format eq "XML") {
      print "<variable_mods>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Variable modifications\"$delimiter\"--------------------------------------------------------\"\n\n";
      print "\"Identifier\"$delimiter";
      print "\"Name\"$delimiter";
      print "\"Delta\"$delimiter";
      print "\"Neutral loss(es)\"\n";
    }

  my $string = "";
  my $i = 1;
  while ($objParams->getVarModsName($i)) {
    if ($objResFile->getMonoLinkModification($i, 0)) {
      $i++;
      next;
    }
    my $letter = $i;
    if ($i > 9) {
      $letter = chr($i + 55);
    }
    if ($export_format eq "CSV") {
      $string .= $letter . $delimiter;
    } elsif ($export_format eq "XML") {
      $string .= "<modification identifier=\"$letter\">$delimiter";
    }
    $string .= &formatElement($export_format, "", 1, "name", $objParams->getVarModsName($i)) . $delimiter;
    $string .= &formatElement($export_format, "", 0, "delta", $objParams->getVarModsDelta($i)) . $delimiter;
    my $losses = $objParams->getVarModsNeutralLosses($i);
    if ($losses
      && !(scalar(@{ $losses }) == 1 && ${ $losses }[0] == 0)) {
      my $j = 1;
      foreach(@{ $losses }) {
        my $letter = $j;
        if ($j > 9) {
          $letter = chr($j + 55);
        }
        $string .= &formatElement($export_format, "", 0, "neutral_loss", $_, identifier => $letter) . $delimiter;
        $j++;
      }
    }
    chop $string;
    $i++;
    if ($export_format eq "XML") {
      $string .= "$delimiter</modification>";
    }
    $string .= "\n";
  }

  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string . "\n";
  }

  if ($export_format eq "XML") {
    print "</variable_mods>\n";
  }

}

###############################################################################
# &outputLinkers()
# output linkers block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputLinkers {

    if ($export_format eq "XML") {
      print "<linkers>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Linkers\"$delimiter\"--------------------------------------------------------\"\n\n";
      print "\"Modification identifier\"$delimiter";
      print "\"Monolink identifier\"$delimiter";
      print "\"Name\"$delimiter";
      print "\"Delta\"$delimiter\n";
    }

  my $string = "";
  my $i = 1;
    while ($objParams->getVarModsName($i)) {
	my $monoLinkIdx = 0;
	my $monoLink = $objResFile->getMonoLinkModification($i, $monoLinkIdx);
	if (not $monoLink) {
	    $i++;
	    next;
	}
	my $letter = varmod_number_to_id($i);
	if ($export_format eq "CSV") {
	    $string .= $letter;
	} elsif ($export_format eq "XML") {
	    $string .= "<modification identifier=\"$letter\">$delimiter";
	}
	while ($monoLink) {
	    my $monoLinkId = varmod_number_to_id($monoLinkIdx);
	    if ($export_format eq "CSV") {
		$string .= $delimiter.$monoLinkId . $delimiter;
	    } elsif ($export_format eq "XML") {
		$string .= "<monolink identifier=\"$monoLinkId\">$delimiter";
	    }
	    $string .= &formatElement($export_format, "", 1, "name", $monoLink->getMonoLinkTitle()) . $delimiter;
	    $string .= &formatElement($export_format, "", 0, "delta", $monoLink->getDelta($objParams->getMassType())) . $delimiter;
	    if ($export_format eq "XML") {
		$string .= "</monolink>";
	    }
	    $string .= "\n";
	    $monoLinkIdx++;
	    $monoLink = $objResFile->getMonoLinkModification($i, $monoLinkIdx);
	}
	$i++;
	if ($export_format eq "XML") {
	    $string .= "</modification>\n";
	}
	$string .= "\n";
    }
    chop $string;
    if ($export_format eq "XML") {
	$string =~ s/>$delimiter</>\n</g;
    }
    if ($string) {
	print $string . "\n";
    }

    if ($export_format eq "XML") {
	print "</linkers>\n";
    }
}

###############################################################################
# &outputParams()
# output search parameters block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputParams {

  if ($export_format eq "XML") {
    print "<search_parameters>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Search Parameters\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  print &formatElement($export_format, "Taxonomy filter", 1, "TAXONOMY", $objParams->getTAXONOMY()) . "\n";
  print &formatElement($export_format, "Enzyme", 1, "CLE", $objParams->getCLE()) . "\n";
  print &formatElement($export_format, "Maximum Missed Cleavages", 0, "PFA", $objParams->getPFA()) . "\n";
  my $local_mods = $objParams->getMODS();
  print &formatElement($export_format, "Fixed modifications", 1, "MODS", $local_mods) . "\n";

  if ($hasXlMethod) {
      print &formatElement($export_format, "Crosslinking method", 1, "CROSSLINKING", $objParams->getCROSSLINKING()) . "\n";
  }

  if ($objParams->getQUANTITATION()) {
    print &formatElement($export_format, "Quantitation method", 1, "QUANTITATION", $objParams->getQUANTITATION()) . "\n";
  } elsif ($objParams->getICAT()) {
    print &formatElement($export_format, "ICAT experiment", 0, "ICAT", $objParams->getICAT()) . "\n";
  }
  my $local_it_mods = $objParams->getIT_MODS();
  if ($objParams->getICAT()){
    my $ICATLight = $objMascotOptions->getICATLight;
    $ICATLight = "ICAT_light" unless $ICATLight;
    my $ICATHeavy = $objMascotOptions->getICATHeavy;
    $ICATHeavy = "ICAT_heavy" unless $ICATHeavy;
    my @tempMods = split(/,/, $local_it_mods);
    for (my $i = $#tempMods; $i >= 0; $i--) {
      if (($tempMods[$i] eq $ICATLight) || ($tempMods[$i] eq $ICATHeavy)) {
        splice @tempMods, $i, 1;
      }
    }
    if (@tempMods) {
      $local_it_mods = join(',', @tempMods);
    } else {
      $local_it_mods = "";
    }
  }
  print &formatElement($export_format, "Variable modifications", 1, "IT_MODS", $local_it_mods) . "\n";
  print &formatElement($export_format, "Peptide Mass Tolerance", 0, "TOL", $objParams->getTOL()) . "\n";
  print &formatElement($export_format, "Peptide Mass Tolerance Units", 0, "TOLU", $objParams->getTOLU()) . "\n";
  if ($objResFile->isPMF()){
    print &formatElement($export_format, "Peptide Charge State", 0, "CHARGE", $objParams->getCHARGE()) . "\n";
  } else {
    print &formatElement($export_format, "Fragment Mass Tolerance", 0, "ITOL", $objParams->getITOL()) . "\n";
    print &formatElement($export_format, "Fragment Mass Tolerance Units", 0, "ITOLU", $objParams->getITOLU()) . "\n";
  }
  print &formatElement($export_format, "Mass values", 0, "MASS", $objParams->getMASS()) . "\n";
  if ($objParams->getSEG() > 0){
    print &formatElement($export_format, "Protein Mass", 0, "SEG", $objParams->getSEG()) . "\n";
  }
  if (!$objResFile->isPMF() && $objParams->getINSTRUMENT()) {
    print &formatElement($export_format, "Instrument type", 1, "INSTRUMENT", $objParams->getINSTRUMENT()) . "\n";
  }
  if ($objParams->getPEP_ISOTOPE_ERROR() > -1) {
    print &formatElement($export_format, "Isotope error mode", 0, "PEP_ISOTOPE_ERROR", $objParams->getPEP_ISOTOPE_ERROR()) . "\n";
  }
  if ($objParams->getDECOY() > -1) {
    print &formatElement($export_format, "Decoy database also searched", 0, "DECOY", $objParams->getDECOY()) . "\n";
    if ($objParams->getTARGET_FDR_PERCENT() > 0) {
      print &formatElement($export_format, "Target FDR (%)", 0, "TARGET_FDR_PERCENT", $objParams->getTARGET_FDR_PERCENT()) . "\n";
    }
  }

# Print out any user parameters with defined values (USER00 to USER12 and anything beginning with underscore)
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $i)) {
    if ($nextKey =~ /^(_.*)/i || $nextKey =~ /^(user\d\d)$/i
        || $nextKey =~ /^(OPTION_.*)/i || $nextKey =~ /^(PRINT_ATTRIBUTE_.*)/i) {
      my $parameter_name = $1;
      my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
      $nextKey =~ s/\s*$//;
      if ($nextValue) {
        print &formatElement($export_format, $nextKey, 1, "user_parameter", $nextValue, name => $parameter_name) . "\n";
      }
    }
    $i++;
  }

  if ($export_format eq "XML") {
    print "</search_parameters>\n";
  }

}

###############################################################################
# &outputFormat()
# output format block
# globals:
# my($objResFile, $export_format, %urlParams, $thisScript,
# $numHits, $ignoreIonsScoreBelow, $RequireBoldRed, $mudpitSwitch, $ShowSubSets);
###############################################################################

sub outputFormat {

  if ($export_format eq "XML") {
    print "<format_parameters>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Format parameters\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  do {
    my $value = $objSummary->getProbabilityThreshold;
    if ($objSummary->getFlags2() & $msparser::ms_peptidesummary::MSPEPSUM_SL_ONLY) {
      $value = msparser::ms_peptidesummary::getSLThresholdFromMinProbability($value);
    }
    print &formatElement($export_format, "Significance threshold", 0, "sigthreshold", $value) . "\n";
  };
  if ($anyPepSumMatch and $objParams->getDECOY() > 0) {
    print &formatElement($export_format, "Target FDR (overrides sig. threshold)", 0, "TargetFDR", $targetFDR) . "\n";
    print &formatElement($export_format, "Target FDR type", 0, "TargetFDRType", $targetFDRtype) . "\n";
  }
  print &formatElement($export_format, "Max. number of hits", 0, "REPORT", $numHits) . "\n";
  print &formatElement($export_format, "Min. number of sig. unique sequences", 0, "min_num_sig_unique_seqs", $minNumSigUniqueSeqs) . "\n";
  if ($anyPepSumMatch) {
    if ($anyLibraryMatch
            or ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch)) {
      print &formatElement($export_format, "Use MudPIT protein scoring", 0, "mudpit", 1) . "\n";
    } else {
      print &formatElement($export_format, "Use MudPIT protein scoring", 0, "mudpit", 0) . "\n";
    }
    # Bug 12007: Print the value regardless of whether it's defined in the URL.
    # It's what we used to open the results file.
    print &formatElement($export_format, "Ions score cut-off", 0, "ignoreionsscorebelow", $ignoreIonsScoreBelow) . "\n";
  }
  print &formatElement($export_format, "Include same-set proteins",
    0, "show_same_sets", $ShowSameSets) . "\n";
  print &formatElement($export_format, "Include sub-set proteins",
    0, "showsubsets", $ShowSubSets) . "\n";
  if ($anyPepSumMatch) {
    print &formatElement($export_format, "Include unassigned",
      0, "show_unassigned", $thisScript->param($urlParams{'show_unassigned'}) || 0) . "\n";
    print &formatElement($export_format, "Require bold red", 0, "requireboldred", $RequireBoldRed) . "\n";
    if ($thisScript->param($urlParams{'unigene'})) {
      print &formatElement($export_format, "UniGene index", 1, "UNIGENE", $thisScript->param($urlParams{'unigene'})) . "\n";
    }
   print &formatElement($export_format, "Use homology threshold", 0, "use_homology", $use_homology) . "\n";
   print &formatElement($export_format, "Group protein families", 0, "group_family", $group_family) . "\n";
   unless ($percolate == -1) {
     print &formatElement($export_format, "Re-score using Percolator", 0, "percolate", $percolate) . "\n";
    }
  }
  if ($preferTaxonomy > 0) {
    my $tax = msparser::ms_taxonomyfile->new();
    my $taxonomy_entry;

    if ($preferTaxonomy < $tax->getNumberOfEntries()) {
        $taxonomy_entry = $tax->getEntryByNumber($preferTaxonomy);
    }

    if ($taxonomy_entry) {
      print formatElement($export_format, "Preferred taxonomy", 1, 'preferred_taxonomy', $taxonomy_entry->getTitle()), "\n";
    }
  }
  print formatElement($export_format, 'Show duplicate peptides', 0, 'show_pep_dupes', $thisScript->param($urlParams{'show_pep_dupes'}) ? 1 : 0) . "\n";

  if ($export_format eq "XML") {
    print "</format_parameters>\n";
  }

}

###############################################################################
# &outputMasses()
# output masses block
# globals:
# my($objResFile, $export_format);
###############################################################################

sub outputMasses {

  if ($export_format eq "XML") {
    print "<masses>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Mass values\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_MASSES, $i)) {
    if ($nextKey =~ /^delta/i
      || $nextKey =~ /^ignore/i
      || $nextKey =~ /^neutralloss/i
      || $nextKey =~ /^fixedmod/i
      || $nextKey =~ /^reqpep/i
      || $nextKey =~ /^pepneutral/i) {
      $i++;
      next;
    }
    print &formatElement($export_format, $nextKey, 0, "mass",
      $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey), name => $nextKey) . "\n";
    $i++;
  }

  if ($export_format eq "XML") {
    print "</masses>\n";
  }

}

###############################################################################
# &outputQuantitationHeader()
# output quantitation level details, such as normalisation bases
# globals:
# my($objMS2Quant, $export_format);
###############################################################################

sub outputQuantitationHeader {

  if (not $objMS2Quant) {
    return;
  }

  if ($export_format eq "XML") {
    print "<quantitation_header>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Quantitation header\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  if ($objMS2Quant->hasPeptideRatioNormalisation()) {
    my @names = @{ $objMS2Quant->getCombinedReportRatioNames() };
    for my $name (@names) {
      my $value = $objMS2Quant->getPeptideRatioNormalisationBase($name);
      print formatElement($export_format,
          sprintf('Normalisation base,"%s"', noQuotes($name)),
          0,
          'normalisation_base',
          $value,

          type => 'ratio',
          name => $name
      ), "\n";
    }
  }

  if ($objMS2Quant->hasIntensityNormalisation()) {
    my $qm = $objMS2Quant->getQuantitationMethod();
    for my $i (0 .. $qm->getNumberOfComponents()-1) {
      my $name = $qm->getComponentByNumber($i)->getName();
      my $value = $objMS2Quant->getIntensityNormalisationBase($name);
      print formatElement($export_format,
          sprintf('Normalisation base (%s)', $name),
          0,
          'normalisation_base',
          $value,

          type => 'component',
          name => $name
      ), "\n";
    }
  }

  if ($export_format eq "XML") {
    print "</quantitation_header>\n";
  }

}

###############################################################################
# &outputQueries()
# output queries block
# globals:
# my($objResFile, $objSummary, $objParams, $objUmodSection, $msresFlags2,
# $export_format, %urlParams, $thisScript, $numHits,
# $ignoreIonsScoreBelow, $RequireBoldRed, $mudpitSwitch, $ShowSubSets);
###############################################################################

sub outputQueries {

  my $isPMF = $objResFile->isPMF();

  if ($export_format eq "XML") {
    print "<queries>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Queries\"$delimiter\"--------------------------------------------------------\"\n\n";
    print "\"query_number\"";
    print "$delimiter\"moverz\"";
    print "$delimiter\"charge\"";
    print "$delimiter\"intensity\"";
    if ($thisScript->param($urlParams{'query_title'})) {
      print "$delimiter\"StringTitle\"";
      print "$delimiter\"Scan number range\"";
      print "$delimiter\"Retention time range\"";
    }
    if ($thisScript->param($urlParams{'query_qualifiers'})) {
      print "$delimiter\"qualifiers\"";
    }
    if ($thisScript->param($urlParams{'query_params'})) {
      print "$delimiter\"Peptide Mass Tolerance\"";
      print "$delimiter\"Peptide Mass Tolerance Units\"";
      print "$delimiter\"Variable modifications\"";
      print "$delimiter\"Instrument type\"";
    }
    if ($thisScript->param($urlParams{'query_peaks'})) {
      print "$delimiter\"TotalIonsIntensity\"";
      print "$delimiter\"NumVals\"";
      print "$delimiter\"StringIons1\"";
      print "$delimiter\"StringIons2\"";
      print "$delimiter\"StringIons3\"";
    }
    if ($thisScript->param($urlParams{'query_raw'})) {
      foreach my $field (@columns) {
        if ($field !~ /^pep/) {
          next;
        }
        if ($thisScript->param($urlParams{lc($field)})) {
	  if ($field eq 'pep_isunique') {
           print $delimiter .'pep_isunique';
           if ($anyXlPepSumMatch) {
             print $delimiter . 'prot_acc_alpha' . $delimiter . 'prot_acc_beta';
           }
         }
         elsif ($field eq "pep_seq") {
            print $delimiter . "pep_res_before" . $delimiter . $field . $delimiter . "pep_res_after";
          } elsif ($field eq "pep_var_mod") {
            print $delimiter . $field
                . $delimiter . "pep_var_mod_pos";

            print $delimiter . "pep_summed_mod_pos"
                . $delimiter . "pep_local_mod_pos"
                . $delimiter . "pep_var_mod_conf"
                ;
	    if ($hasXlMethod) {
		print $delimiter . "pep_monolink_mod_pos";
	    }
	    if ($anyXlPepSumMatch) {
	      print $delimiter .  "pep_var_mod_beta" . $delimiter . "pep_var_mod_pos_beta";
	      print $delimiter . "pep_summed_mod_pos_beta" . $delimiter
              . "pep_local_mod_pos_beta";
	      print $delimiter . "pep_mono_link_mod_pos_beta";
	    }

	    if ($hasXlMethod) {
	      print $delimiter . "pep_looplink_pos";
	    }
	    if ($anyXlPepSumMatch) {
	      print $delimiter . "pep_looplink_pos_beta";
	      print $delimiter . "pep_linked_sites";
	    }
          } elsif ($field eq "pep_query") {
            next;
          } else {
            print $delimiter . $field;
          }
        }
      }
    }
    print "\n";
  }

  # Site analysis initialisation.
  my $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
    $objUmodSection, $objParams,
  );
  my %query_level_mod_cache = ();
  # $objUmodSection could be the reduced file embedded in the results file.
  my $cs = new msparser::ms_connection_settings();
  $cs->setSessionID($sessionID);
  my $full_unimod_file = new msparser::ms_umod_configfile(
    "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
    $cs
  );
  my $query_is_sig = $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
    ? sub { Util::Peptides::above_homology($objSummary, $objSummary->getPeptide($_[0], 1)) }
    : sub { Util::Peptides::above_identity($objSummary, $objSummary->getPeptide($_[0], 1)) }
    ;
  my $max_p = $objSummary->getMaxRankValue();

  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
    my $string = "";
    $queryColumnCount = 0;
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
    if ($export_format eq "CSV") {
      $string .= $queryNum . $delimiter;
    } elsif ($export_format eq "XML") {
      $string .= "<query number=\"$queryNum\">$delimiter";
    }
    $queryColumnCount++;

    $string .= &formatElement($export_format, "", 0, "query_moverz", $objResFile->getObservedMass($queryNum) + 0) . $delimiter;
    $string .= &formatElement($export_format, "", 1, "query_charge", $objQuery->getCharge()) . $delimiter;
    my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
    if ($intensity != 0) {
      $string .= &formatElement($export_format, "", 0, "query_intensity", $intensity) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
    $queryColumnCount += 3;

  # query_title
    if ($thisScript->param($urlParams{'query_title'})) {
      if ($objQuery->getStringTitle(1)) {
        $string .= &formatElement($export_format, "", 1, "StringTitle",
          $objQuery->getStringTitle(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      my $indices = msparser::vectori->new();
      my $values = msparser::VectorString->new();
      $objQuery->getScanNumbers($indices, $values);
      if (not $values->empty()) {
        # Comma is the column separator in CSV, so separate with spaces.
        # (We don't care about indices here.)
        $string .= &formatElement($export_format, "", 1, "SCANS", join(' ', map { $values->get($_) } 0..$values->size-1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      $indices->clear(); $values->clear();
      $objQuery->getRetentionTimes($indices, $values);
      if (not $values->empty()) {
        # See above.
        $string .= &formatElement($export_format, "", 1, "RTINSECONDS", join(' ', map { $values->get($_) } 0..$values->size-1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
    }

  # query_qualifiers
    if ($thisScript->param($urlParams{'query_qualifiers'})) {
      my @qualifiers;
      if ($objQuery->getPepTolString()) {
        if ($export_format eq "XML") {
          push @qualifiers, &formatElement($export_format, "", 1, "qual_tol", $objQuery->getPepTolString());
        } else {
          push @qualifiers, $objQuery->getPepTolString();
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getSeq($j)) {
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_seq", "seq(" . $objQuery->getSeq($j) . ")", number => $j);
          } else {
            push @qualifiers, "seq(" . $objQuery->getSeq($j) . ")";
          }
        } else {
          last;
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getComp($j)) {
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_comp", "comp(" . $objQuery->getComp($j) . ")", number => $j);
          } else {
            push @qualifiers, "comp(" . $objQuery->getComp($j) . ")";
          }
        } else {
          last;
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if (my $local_tag_string = $objQuery->getTag($j)) {
          $local_tag_string =~ /([et]),(.*)/i;
          if ($1 eq "e") {
            $local_tag_string = "etag($2)";
          } else {
            $local_tag_string = "tag($2)";
          }
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_tag", $local_tag_string, number => $j);
          } else {
            push @qualifiers, $local_tag_string;
          }
        } else {
          last;
        }
      }
      if ($export_format eq "XML") {
        if (@qualifiers) {
          $string .= join($delimiter, @qualifiers) . $delimiter;
        }
      } else {
        $string .= &formatElement($export_format, "", 1, "pep_qual", join(" ", @qualifiers)) . $delimiter;
      }
      $queryColumnCount++;
    }

  # query_params
    if ($thisScript->param($urlParams{'query_params'})) {
      if ($objQuery->getPepTol() > 0) {
        $string .= &formatElement($export_format, "", 0, "query_TOL", $objQuery->getPepTol()) . $delimiter;
        $string .= &formatElement($export_format, "", 0, "query_TOLU", $objQuery->getPepTolUnits()) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter . $delimiter;
      }
      $queryColumnCount++;
      $queryColumnCount++;
      if ($objQuery->getIT_MODS(1)) {
        $string .= &formatElement($export_format, "", 1, "query_IT_MODS", $objQuery->getIT_MODS(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      if ($objQuery->getINSTRUMENT(1)) {
        $string .= &formatElement($export_format, "", 1, "query_INSTRUMENT", $objQuery->getINSTRUMENT(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
    }

  # query_peaks
    if ($thisScript->param($urlParams{'query_peaks'})) {
      $string .= &formatElement($export_format, "", 0, "TotalIonsIntensity", $objQuery->getTotalIonsIntensity() + 0) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "NumVals", $objQuery->getNumVals() + 0) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons1", $objQuery->getStringIons1()) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons2", $objQuery->getStringIons2()) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons3", $objQuery->getStringIons3()) . $delimiter;
      $queryColumnCount += 5;
    }

  # query_raw
    if ($thisScript->param($urlParams{'query_raw'})) {
      my $siteAnalysisData;
      if (&$query_is_sig($queryNum)) {
        $siteAnalysisData = ExportDat::Util::analyse_sites_in_query(
          \%query_level_mod_cache, $objResFile, $objSummary, $objParams,
          $queryNum, $modsByDelta, $full_unimod_file, $max_p,
        );
      }
      my $needSpacer = 0;
      my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
      my $peptideIdx = $objQuery->getIndex;
      for my $rank (1 .. $max_p) {
        my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
        next unless $objPeptide;
        next unless $objPeptide->getAnyMatch();

        if ($export_format eq "CSV" && $needSpacer) {
          $string .= $queryNum . $delimiter x $queryColumnCount;
        } else {
          $needSpacer = 1;
        }
        if ($export_format eq "CSV") {
          $string .= $peptideIdx . $delimiter;
        }
        if ($export_format eq "XML") {
          $string .= "<q_peptide query=\"$queryNum\"";
          if ($objPeptide->getAnyMatch() && !$isPMF) {
            $string .= " rank=\"" . $objPeptide->getPrettyRank() . "\"";
          }
          if ($peptideIdx > -1) {
            print " index=\"$peptideIdx\"";
          }
          $string .= ">\n";
        }
        $string .= fmtPeptideDetails(
            $objPeptide, $peptideIdx, $queryNum, undef, undef, 0, $isPMF, $siteAnalysisData || {},
            $objSummary, \%db_type_query_restrictions
        ) . "\n";
        if ($export_format eq "XML") {
          chop $string;
          $string .= "</q_peptide>\n";
        }
      }
    }

    chop $string;
    if ($export_format eq "XML") {
      $string .= "$delimiter</query>";
      $string =~ s/>$delimiter</>\n</g;
    }
    if ($string) {
      print $string . "\n";
    }

  }

  if ($export_format eq "XML") {
    print "</queries>\n";
  }

}

###############################################################################
# &formatElement()
# returns formatted element
#
# $_[0] "XML" or "CSV" or ...
# $_[1] CSV label, empty if none required
# $_[2] 1 if text element could contain delimiter or illegal character
# $_[3] xml element name
# $_[4] element value
# $_[5..n] optional XML attribute hash
###############################################################################

sub formatElement {

  my($format, $label, $need_quotes, $name, $value, %xml_attribs) = @_;
  my $return = "";

  if ($format eq "XML") {
    $return .= "<$name";
    for my $key (sort keys %xml_attribs) {
      $return .= " $key=\"" . &noXmlTag($xml_attribs{$key}) . "\"";
    }
    if ($value eq "") {
      $return .= " />";
    } else {
      if ($need_quotes) {
        $return .= ">" . &noXmlTag($value) . "</$name>";
      } else {
        $return .= ">$value</$name>";
      }
    }
  } elsif ($format eq "CSV") {
    if ($label) {
      $return .= "$label$delimiter";
    }
    if ($need_quotes) {
      $return .= "\"" . &noQuotes($value) . "\"";
    } else {
      $return .= $value;
    }
  }

  return $return;

}

###############################################################################
# &noXmlTag()
# returns de-tagged string
# $_[0] string which may contain XML illegal characters
###############################################################################

sub noXmlTag {

  my $temp = shift;
  return $temp if not defined($temp);

  $temp =~ s/&/&amp;/g;
  $temp =~ s/</&lt;/g;
  $temp =~ s/>/&gt;/g;
  $temp =~ s/\"/&quot;/g;
  $temp =~ s/\'/&apos;/g;

  return $temp;

}

###############################################################################
# &noQuotes()
# returns string with quotes removed
# $_[0] string which may contain quotes
###############################################################################

sub noQuotes {

  my $temp = shift;

  $temp =~ s/[\"\']/~/g;

  return $temp;

}

###############################################################################
# &getResidueMods()
# returns aminoacid_modification elements
# $_[0] Mascot modification name
# $_[1] "N" or "Y" for variable mods
###############################################################################

sub getResidueMods {

  my ($modName, $variable) = @_;

  my @retVal;
  my $treatasvariable = $variable;
  my $terminalMod = "";
  my $objMod = $objModFile->getModificationByName($modName);
  if ($objMod) {
    if ($objMod->getModificationType() == $msparser::MOD_TYPE_RESIDUE
      || $objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM_RESIDUE
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM_RESIDUE) {
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM_RESIDUE) {
        $terminalMod = "n";
        $treatasvariable = "Y";
      }
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM_RESIDUE) {
        $terminalMod = "c";
        $treatasvariable = "Y";
      }
      for (my $i = 0; $i < $objMod->getNumberOfModifiedResidues(); $i++) {
        push @retVal, $objMod->getModifiedResidue($i);
        push @retVal, sprintf("%+.*f", $massDP, round_evenly($objMod->getDelta($mass_type), $massDP));
        if ($terminalMod) {
          push @retVal, ($objMassesFile->getResidueMass($mass_type, $objMod->getModifiedResidue($i))
            + $objMod->getDelta($mass_type));
        } else {
          push @retVal, $objMod->getResidueMass($mass_type, $objMod->getModifiedResidue($i));
        }
        push @retVal, $treatasvariable;
        push @retVal, $terminalMod;
        push @retVal, $modName;
      # print "<!--\n";
      # print $objMod->getResidueMass($mass_type, $objMod->getModifiedResidue($i)) . "\n";
      # print $objMod->getDelta($mass_type) . "\n";
      # print $objMod->isResidueModified($objMod->getModifiedResidue($i)) . "\n";
      # print $objMod->getCTerminusMass($mass_type) . "\n";
      # print $objMassesFile->getResidueMass($mass_type, $objMod->getModifiedResidue($i)) . "\n";
      # print $objMassesFile->isResidueModified($objMod->getModifiedResidue($i)) . "\n";
      # print $objMassesFile->getCtermDelta($mass_type) . "\n";
      # print $objMassesFile->getCterminalMass($mass_type) . "\n";
      # print "-->\n";
      }
    }
  } else {
  # mod not found in file, extract values from result file
    unless ($modName =~ /\(c-term\)/i
      || $modName =~ /\(n-term\)/i
      || $modName =~ /\(protein\)/i) {
      (my $specificity) = $modName =~ /\((.*)\)/;
    # special cases
      if ($specificity =~ /([cn])-term (.+)/i) {
        $terminalMod = lc($1);
        $specificity = $2;
        $treatasvariable = "Y";
      }
      if ($specificity eq "") {
      # must be ICAT_light or ICAT_heavy
        $specificity = "C";
      } elsif ($specificity eq "camC") {
      # Pyro-cmC
        $specificity = "C";
      }
      my @spec_list = split(//, $specificity);
      foreach my $this_spec (@spec_list) {
        my ($mass_diff, $mass_total);
        if ($treatasvariable eq "Y") {
          while (my($key, $value) = each %vmString) {
            if ($value eq $modName) {
              $mass_diff = $vmMass{$key};
              $mass_total = $mass_diff + $objMassesFile->getResidueMass($mass_type, uc $this_spec);
            }
          }
        } else {
          $mass_total = $groupMasses{lc($this_spec)};
          $mass_diff = $mass_total - $objMassesFile->getResidueMass($mass_type, uc $this_spec);
        }
        push @retVal, $this_spec;
        push @retVal, sprintf("%+.*f", $massDP, round_evenly($mass_diff, $massDP));
        push @retVal, $mass_total;
        push @retVal, $treatasvariable;
        push @retVal, $terminalMod;
        push @retVal, $modName;
      }
    }
  }
  return @retVal;

}

###############################################################################
# &getTerminalMods()
# returns terminal_modification elements
# $_[0] Mascot modification name
# $_[1] "N" or "Y" for variable mods
###############################################################################

sub getTerminalMods {

  my ($modName, $variable) = @_;

  my @retVal;
  my $treatasvariable = $variable;
  my $protein_terminus = "";
  my $objMod = $objModFile->getModificationByName($modName);
  if ($objMod) {
    if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM
      || $objMod->getModificationType() == $msparser::MOD_TYPE_N_PROTEIN
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_PROTEIN) {
        $treatasvariable = "Y";
        $protein_terminus = "n";
      }
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
        $treatasvariable = "Y";
        $protein_terminus = "c";
      }
      my $term_end = "n";
      my $term_mass = $objMassesFile->getNterminalMass($msparser::MASS_TYPE_MONO);
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM
        || $objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
        $term_end = "c";
        $term_mass = $objMassesFile->getCterminalMass($mass_type);
      }
      push @retVal, $term_end;
      push @retVal, sprintf("%+.*f", $massDP, round_evenly($objMod->getDelta($mass_type), $massDP));
      push @retVal, ($term_mass + $objMod->getDelta($mass_type));
      push @retVal, $treatasvariable;
      push @retVal, $protein_terminus;
      push @retVal, $modName;
    }
  } else {
  # mod not found in file, try to guess values
    if ($modName =~ /\(c-term\)/i
      || $modName =~ /\(n-term\)/i
      || $modName =~ /\(protein\)/i) {
      (my $specificity) = $modName =~ /\((.*)\)/;
      my ($this_spec, $term_end);
    # special cases
      if ($specificity =~ /protein/i) {
        $treatasvariable = "Y";
        $this_spec = "N_term";   # assumption
        $term_end = "n";         # assumption
        $protein_terminus = "n"; # assumption
      } elsif ($specificity =~ /c-term/i) {
        $this_spec = "C_term";
        $term_end = "c";
      } elsif ($specificity =~ /n-term/i) {
        $this_spec = "N_term";
        $term_end = "n";
      }
      my ($mass_diff, $mass_total);
      if ($treatasvariable eq "Y") {
        while (my($key, $value) = each %vmString) {
          if ($value eq $modName) {
            $mass_diff = $vmMass{$key};
            if ($this_spec eq 'C_term') {
              $mass_total = $mass_diff + $objMassesFile->getCterminalMass($mass_type);
            } else {
              $mass_total = $mass_diff + $objMassesFile->getNterminalMass($mass_type);
            }
          }
        }
      } else {
        $mass_total = $groupMasses{lc($this_spec)};
        if ($this_spec eq "C_term") {
          $mass_diff = $mass_total - $objMassesFile->getCterminalMass($mass_type);
        } else {
          $mass_diff = $mass_total - $objMassesFile->getNterminalMass($mass_type);
        }
      }
      push @retVal, $term_end;
      push @retVal, sprintf("%+.*f", $massDP, round_evenly($mass_diff, $massDP));
      push @retVal, $mass_total;
      push @retVal, $treatasvariable;
      push @retVal, $protein_terminus;
      push @retVal, $modName;
    }
  }
  return @retVal;

}

###############################################################################
# &fasta()
# output protein seqs
# my($objSummary, $objResFile, $objParams globals);
###############################################################################
sub fasta {

  my $thisHit = 1;
  my $j = 1;
  my $k = 1;
  my $objProtein = $objSummary->getHit($thisHit);
  my $objDatabases = $objDatFile->getDatabases();
  my $nDBs = $objParams->getNumberOfDatabases();
  my @dbHasAnyHit = (0) x $nDBs+1; #All indices are 1-based
  my $shouldBeExported = sub {
    my $i = shift;
    # return true if we want to export everything or current DB type is the one we want.
    return (not defined($fastaOutputType) or ($fastaOutputType  == $objResFile->getDatabaseType($i)) );
  };

  while (defined($objProtein)) {
    if( _isFromSeqDB($objProtein) ) {
      # export the hit
      my $dbIdx = $objProtein->getDB();
      if ( $shouldBeExported->($dbIdx) ) {
        my $preferredAcc = $protData->preferredAccession($dbIdx, $objProtein->getAccession());
        my ($seq, $err) = $protData->mustGetProteinSeq($objProtein->getAccession(), 0, $dbIdx, 'TRUE'); # frame = 0
        my $desc = $protData->mustGetProteinDescription($preferredAcc, $dbIdx);
        if ($seq) {
          my @chunks = grep {$_} split /(.{60})/, $seq; # Limit sequence line to 60 chars
          $seq = join("\n", @chunks);
        }
        else {
          print "#";
          $seq = "#" . $err;
        }
        print ">" .$objProtein->getAccession();
        if ($desc) {
          print " ".$desc;
        }
        print "\n";
        print $seq."\n";

        $dbHasAnyHit[$dbIdx] = 1;
      }
    }

    if($ShowSameSets) {
      $objProtein = $objSummary->getNextSimilarProtein($thisHit, $k++);
    }
    unless($objProtein and $ShowSameSets) {
      $k = 1;
      $objProtein = $objSummary->getNextFamilyProtein($thisHit, $j++);
      unless($objProtein) {
        $thisHit++;
        $objProtein = $objSummary->getHit($thisHit);
        $j = 1;
      }
    }
  }

  for my $dbIdx (1 .. $nDBs) {
    if(not $dbHasAnyHit[$dbIdx] and $shouldBeExported->($dbIdx) ) {
      print "#". $objParams->getDB($dbIdx). " has no protein hits.\n";
    }
  }
}

###############################################################################
# &xiVIEW_CSV()
# output crosslinked results
# my($objSummary, $objResFile, $objParams globals);

###############################################################################
sub xiVIEW_CSV {

    if(not $anyXlPepSumMatch) {
        return;
    }

    my $header = "Id,Protein1,PepPos1,PepSeq1,LinkPos1,Protein2,PepPos2,PepSeq2,LinkPos2,Score,CrossLinkerModMass,Rank,Charge,ScanId,PeakListFileName,FragmentTolerance,ExpMz,CalcMz,PassThreshold\n";
    print $header;

    my $max_p = $objSummary->getMaxRankValue();
    # loop through all matched queries
    for my $queryNum (1 .. $objResFile->getNumQueries()) {
        # Only print the rank 1 match.
        for my $rank (1) {
            my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
            next if not $objPeptide;
            next if not $objPeptide->getAnyMatch();
            # Only print if the match is significant.
            my $threshold = $objSummary->getPeptideThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold);
            next if $objPeptide->getIonsScore() < $threshold;
            _printXLinkInfo($objPeptide);
        }
    }
}

###############################################################################
# &_printXLinkInfo
# checks that at least one of the peptide matches of a given protein is from a sequence database
# my($objSummary $objResFile globals);
###############################################################################
sub _printXLinkInfo {
    my ($pep) = @_;

    my $intactLinks = $pep->getIntactLinks(); # Contains up to 1 element in Mascot 2.7
    if ($intactLinks->size() == 0) {
      return;
    }

    my $q = $pep->getQuery();
    my $p = $pep->getRank();
    my $id = "q" . $q ."_p" . $p . "_";
    my $countLinks = 0;
    my $alphaLength = $pep->getPeptideLength($msparser::ms_peptide::PSM_CROSSLINK_ALPHA);

    my $charge = $pep->getCharge();
    my ($resfile_location) = $fileIn[0] =~ /([^\/\\]*)\.dat$/;
    # Use the the name of the exported .mgf file as the peak list filename.
    my $fileName = "$resfile_location.mgf";
    my $itol = $objParams->getITOL();
    my $itolU = $objParams->getITOLU();

    # xiVIEW only supports Da and ppm.
    # XXX This same code is in peptide_view.pl.
    if ($itolU eq 'Da') {
        # OK
    } elsif ($itolU eq 'mmu') {
        $itol /= 1000;
        $itolU = 'Da';
    } elsif ($itolU eq 'ppm') {
        # OK
    } else {
        # Should never come here...
        die "Cannot export to xiVIEW with ITOLU=$itolU";
    }

    # xiVIEW expects both experimental and calculated precursor values
    # to be m/z, not relative mass.
    my $mrExp = $objResFile->getObservedMass($q);
    my $mrCalc = $pep->getMrCalc() + $charge * $elementalMasses{'hydrogen'};
    $mrCalc /= $charge if $charge != 0;

    #my $series = $pep->getSeriesUsedStr(); #TODO:
    #my $isDecoy = $objSearchParams->getDECOY();#TODO:
    my $passThreshold = 'TRUE'; # we only export significant matches

    # xiVIEW ScanID seems to be 0-based, so we must report $q-1.
    my $pepInfo = $p . "," . $charge . "," . ($q-1) . ",\"" . $fileName . "\"," . $itol . " " . $itolU . "," . $mrExp . ",".$mrCalc . "," . $passThreshold;

    my $isAlpha = sub {
        my $x = shift;
        return ($x < $alphaLength + 2);
    };

    my ($start, $end, $pre, $post, $frame, $multiplicity, $db, $psmComp) = (
        msparser::vectori->new(),
        msparser::vectori->new(),
        msparser::VectorString->new(),
        msparser::VectorString->new(),
        msparser::vectori->new(),
        msparser::vectori->new(),
        msparser::vectori->new(),
        msparser::vectori->new()
    );
    my $confidenceScore = round_evenly($pep->getIonsScore(), 2);
    my $accs = $objSummary->getAllProteinsWithThisPepMatch($pep->getQuery(), $pep->getRank(),
                                                        $start, $end, $pre, $post, $frame,
                                                        $multiplicity, $db, $psmComp);
    my %prots = (
        $msparser::ms_peptide::PSM_CROSSLINK_ALPHA => [],
        $msparser::ms_peptide::PSM_CROSSLINK_BETA => [],
        $msparser::ms_peptide::PSM_COMPLETE => []
    );

    my %startProt = (
        $msparser::ms_peptide::PSM_CROSSLINK_ALPHA => [],
        $msparser::ms_peptide::PSM_CROSSLINK_BETA => [],
        $msparser::ms_peptide::PSM_COMPLETE => []
    );

    while (my ($idx, $acc) = each @{$accs}) {
        push @{ $prots{ $psmComp->get($idx) } }, $acc;
        push @{ $startProt{ $psmComp->get($idx) } }, $start->get($idx);
    }

    my ($modTypeComp, $modDeltasComp, $peptideComp, undef) = ExportDat::Util::make_mods_data($pep, $objResFile, $objModFile, $mass_type);

    # Build the representation of the modified peptide
    # by appending the delta mod to the aminoacid letter
    my $buildPepStr = sub {
      my $psm = shift;

      my $pepStr = "";
      my $N = $#{$peptideComp->{$psm}};
      for my $i ( 1 ..  $N-1) {
	# xiVIEW position numbering is 1..N, where N is the
	# sequence length. We must encode N-terminal (0) and
	# C-terminal (N+1) positions as 1 and N, respectively.
	my $termDelta = 0;
	if ($i == 1) {
	  $termDelta = ($modTypeComp->{$psm}[0] ne 'I') ? $modDeltasComp->{$psm}[0] : 0;
	}
	elsif ($i == $N-1) {
	  $termDelta = ($modTypeComp->{$psm}[$N] ne 'I') ? $modDeltasComp->{$psm}[$N] : 0;
	}

	my $delta = $termDelta;
	# If this position is an intact link, don't add the mod symbol
	# to the sequence. Intact links are encoded in the LinkPos1
	# and LinkPos2 columns.
	if ($modTypeComp->{$psm}[$i] ne 'I') {
	  $delta += $modDeltasComp->{$psm}[$i];
	}
	my $aa = $peptideComp->{$psm}[$i];
	if($delta > 0) {
	  # If it's a monolink, varmod or fixed mod, mod symbol should be the mod
	  # delta up to a limit number of digits after the decimal
	  # point.
	  $aa .= sprintf("%.6f", $delta);
	}
	$pepStr .= $aa;
      }
      return $pepStr;
    };

    for my $i (0 .. $intactLinks->size()-1) {
        my $linkSite = $intactLinks->get($i);
        my $linkFrom = $linkSite->getLinkFrom();
	my $pepSeqFrom = $buildPepStr->($linkFrom);
        my $siteFrom = min( max($linkSite->getSiteFrom(), 1), length($pepSeqFrom) );
        my $linkTo = $linkSite->getLinkTo();
        my $pepSeqTo = $buildPepStr->($linkTo);
        my $siteTo = min( max($linkSite->getSiteTo(), 1), length($pepSeqTo) );
        my $protTo = join(";", @{$prots{$linkTo}});
        my $pepPosTo = join(";",@{$startProt{$linkTo}});

        my $modIdx = $linkSite->getVarModIdxFrom();
        my $delta = $objResFile->params()->getVarModsDelta($modIdx);
        $countLinks++;
        print $id . $countLinks . "," . join(";", @{$prots{$linkFrom}}) . ","
            . join(";",@{$startProt{$linkFrom}})
            . ",", $pepSeqFrom . ","
            . $siteFrom . "," . $protTo
            . ",". $pepPosTo . ","
            . $pepSeqTo . "," . $siteTo . "," . $confidenceScore . ",". $delta . "," . $pepInfo . "\n";
    }
}

###############################################################################
# &_isFromProtSeq()
# checks that at least one of the peptide matches of a given protein is from a sequence database
# my($objSummary globals);
###############################################################################
sub _isFromSeqDB {
  my $prot = shift;

  my $n = $prot->getNumPeptides();
  for my $i (1 .. $n) {
    my $q = $prot->getPeptideQuery($i);
    my $p = $prot->getPeptideP($i);
    my $pep= $objSummary->getPeptide($q, $p);
    return 1 unless $pep->getIsFromLibrary();
  }
  return 0;
}
###############################################################################
# &pepXML()
# output ISB pepXML elements
# my($objParams);
###############################################################################

sub pepXML {

  my %query_level_mod_cache = ();

# msms_run_summary
  print "<msms_run_summary";
  print " base_name=\"\"";
  print " raw_data_type=\"\"";
  print " raw_data=\"\"";
  # optional msManufacturer
  # optional msModel
  # optional msIonization
  # optional msMassAnalyzer
  # optional msDetector
  print ">\n";

# sample_enzyme
  print "<sample_enzyme";
  print " name=\"" . &noXmlTag($objParams->getCLE()) . "\"";
  print " description=\"\"";
  my %enzymeDetails;
  eval { %enzymeDetails = getEnzymeDef($objParams->getCLE()) };
  if (my $err = $@) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $err;
      common_subs::multipart_final($MIME_boundary);
    }
    die $err;
  }

  if ($enzymeDetails{'semispecific'}) {
    print " fidelity=\"semispecific\"";
  } elsif ($objParams->getCLE() eq "None") {
    print " fidelity=\"nonspecific\"";
  } else {
    print " fidelity=\"specific\"";
  }
  print " independent=\"" . $enzymeDetails{'independent'} . "\"";
  if ($objParams->getCLE() eq "None") {
    print " />\n";
  } else {
  # specificities
    print ">\n";
    for my $i (0 .. $enzymeDetails{'num_spec'}-1) {
      print "<specificity";
      if (defined $enzymeDetails{'spec_cleavage_end'}[$i]) {
        print " sense=\"" . $enzymeDetails{'spec_cleavage_end'}[$i] . "\"";
      } else {
        print " sense=\"\"";
      }
      print " min_spacing=\"1\"";
      if (defined $enzymeDetails{'spec_cleavage'}[$i]) {
        print " cut=\"" . $enzymeDetails{'spec_cleavage'}[$i] . "\"";
      } else {
        print " cut=\"\"";
      }
      if (defined $enzymeDetails{'spec_restrict'}[$i]) {
        print " no_cut=\"" . $enzymeDetails{'spec_restrict'}[$i] . "\"";
      } else {
        print " no_cut=\"\"";
      }
      print " />\n";
    }
    print "</sample_enzyme>\n";
  }

# search_summary
  print "<search_summary";
  if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
    print " base_name=\"" . $URI . "cgi/master_results_2.pl?fileset=" . $fileset. "\"";
  } else {
    print " base_name=\"" . $URI . "cgi/master_results.pl?file=" . $fileIn[0] . "\"";
  }
  print " search_engine=\"MASCOT\"";
  print " precursor_mass_type=\"" . lc($objParams->getMASS()) . "\"";
  print " fragment_mass_type=\"" . lc($objParams->getMASS()) . "\"";
  print " out_data_type=\"\"";
  print " out_data=\"\"";
  print " search_id=\"1\">\n";

# search_database
# NB pepXML does not support multiple databases, so just provide details for first
# and list any others in search_summary/parameter elements
  print "<search_database";
  my $databasePath = &noXmlTag($objResFile->getFastaPath(1));
  print " local_path=\"$databasePath\"";
  # optional URL
  print " database_name=\"" . &noXmlTag($objParams->getDB(1)) . "\"";
  # optional orig_database_url
  # optional database_release_date
  print " database_release_identifier=\"" . &noXmlTag($objResFile->getFastaVer(1)) . "\"";
  # wait for parser function ### debug ###
  print " size_in_db_entries=\"" . $objResFile->getNumSeqs . "\"";
  print " size_of_residues=\"" . $objResFile->getNumResidues . "\"";
  if ($objResFile->getDatabaseType(1) == $msparser::NA) {
    print " type=\"NA\"";
  } else {
    print " type=\"AA\"";
  }
  print " />\n";

# enzymatic_search_constraint
  print "<enzymatic_search_constraint";
  print " enzyme=\"" . &noXmlTag($objParams->getCLE()) . "\"";
  print " max_num_internal_cleavages=\"" . $objParams->getPFA() . "\"";
  my $min_number_termini = 2;
  if ($enzymeDetails{'semispecific'}) {
    $min_number_termini = 1;
  } elsif ($objParams->getCLE() eq "None") {
    $min_number_termini = 0;
  }
  print " min_number_termini=\"$min_number_termini\" />\n";

# sequence_search_constraint
# output at query level

# aminoacid_modifications
# fixed
  my(@res_mods, @term_mods, $iCount);
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;
    while ($objParams->getFixedModsName($iCount)) {
      push @res_mods, &getResidueMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @res_mods, &getResidueMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;
  while ($objParams->getVarModsName($iCount)) {
    push @res_mods, &getResidueMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  while (@res_mods) {
    print "<aminoacid_modification";
    print " aminoacid=\"" . shift(@res_mods) . "\"";
    print " massdiff=\"" . shift(@res_mods) . "\"";
    print " mass=\"" . shift(@res_mods) . "\"";
    print " variable=\"" . shift(@res_mods) . "\"";
    my $terminalMod = shift(@res_mods);
    if ($terminalMod) {
      print " peptide_terminus=\"$terminalMod\"";
    }
    # optional symbol
    print " binary=\"N\"";
    print " description=\"" . &noXmlTag(shift(@res_mods)) . "\" />\n";
  }

# terminal_modifications
# fixed
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;
    while ($objParams->getFixedModsName($iCount)) {
      push @term_mods, &getTerminalMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @term_mods, &getTerminalMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;
  while ($objParams->getVarModsName($iCount)) {
    push @term_mods, &getTerminalMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  while (@term_mods) {
    print "<terminal_modification";
    print " terminus=\"" . shift(@term_mods) . "\"";
    print " massdiff=\"" . shift(@term_mods) . "\"";
    print " mass=\"" . shift(@term_mods) . "\"";
    print " variable=\"" . shift(@term_mods) . "\"";
    # optional symbol
    print " protein_terminus=\"" . shift(@term_mods) . "\"";
    print " description=\"" . &noXmlTag(shift(@term_mods)) . "\" />\n";
  }

# parameter
  if($objParams->getNumberOfDatabases > 1) {
    # Bug 12416: Note that we only allow pepXML export if the search is
    # all AA or all NA. Mixed case or SL are not supported.
    for (my $i=2; $i <= $objParams->getNumberOfDatabases; $i++) {
      print "<parameter name=\"local_path_$i\" value=\"" . &noXmlTag($objResFile->getFastaPath($i)) . "\" />\n";
      print "<parameter name=\"database_name_$i\" value=\"" . &noXmlTag($objParams->getDB($i)) . "\" />\n";
      print "<parameter name=\"database_release_identifier_$i\" value=\"" . &noXmlTag($objResFile->getFastaVer($i)) . "\" />\n";
      # wait for parser function ### debug ###
      print "<parameter name=\"size_in_db_entries_$i\" value=\"" . $objResFile->getNumSeqs . "\" />\n";
      print "<parameter name=\"size_of_residues_$i\" value=\"" . $objResFile->getNumResidues . "\" />\n";
    }
  }
  my $index = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $index)) {
    my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
    $nextKey =~ s/\s*$//;
    if ($nextValue) {
      print "<parameter name=\"" . &noXmlTag($nextKey) . "\" value=\"" . &noXmlTag($nextValue) . "\" />\n";
    }
    $index++;
  }

# end of search_summary
  print "</search_summary>\n";

  my $print_prot_desc = $thisScript->param($urlParams{'prot_desc'});
  my $print_prot_mass = $thisScript->param($urlParams{'prot_mass'});
  my $print_prot_pI = $thisScript->param($urlParams{'prot_pi'});

  my $max_p = $objSummary->getMaxRankValue();

# loop through all matched queries
  my $string = "";
  my $needClosingTag;
  for my $queryNum (1 .. $objResFile->getNumQueries()) {
    $needClosingTag = 0;
    for my $rank (1 .. $max_p) {
      my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
      if (defined($objPeptide)) {
        my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
        if ($objPeptide->getAnyMatch()) {
          my $srcRank = $objSummary->getSrcRank($queryNum, $rank);
          my $srcSection = $objSummary->getSrcSection($queryNum, $rank);
          my $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank");
          my($left, $right) = split(/;/, $tempString, 2);
          my @protein_list = $right =~ /"(.+?)"/g;
          $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_db");
          my @db_list;
          if ($tempString) {
            @db_list = $tempString =~ /(\d\d)/g;
          } else {
            for (my $i = 0; $i <= $#protein_list; $i++) {
              $db_list[$i] = "01";
            }
          }
          $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_terms");
          my($peptide_prev_aa, $peptide_next_aa) = $tempString =~ /^(.),(.)/;
          if ($rank == 1) {
          # spectrum_query
            $needClosingTag = 1;
            print "<spectrum_query";
            print " spectrum=\"" . &noXmlTag($objQuery->getStringTitle(1)) . "\"";
            print " start_scan=\"0\"";
            print " end_scan=\"0\"";
            print " precursor_neutral_mass=\"" . $objPeptide->getMrExperimental . "\"";
            print " assumed_charge=\"" . $objResFile->getObservedCharge($queryNum) . "\"";
            my $repeatSearchString = $objResFile->getRepeatSearchString($queryNum);
            $repeatSearchString =~ s/^[0-9.]+\s*//;
            $repeatSearchString =~ s/from\(.+?\)\s*//;
            $repeatSearchString =~ s/title\(.+?\)\s*//;
            $repeatSearchString =~ s/query\(.+?\)\s*//;
            $repeatSearchString =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
            $repeatSearchString =~ s/^\s*//;
            $repeatSearchString =~ s/\s*$//;
            if ($repeatSearchString) {
              print " search_specification=\"" . &noXmlTag($repeatSearchString) . "\"";
            }
            print " index=\"$queryNum\">\n";
          # search_result
            print "<search_result";
            print " search_id=\"1\">\n";
          }

        # search_hit
          print "<search_hit";
          print " hit_rank=\"$rank\"";
          print " peptide=\"" . $objPeptide->getPeptideStr() . "\"";
          if ($peptide_prev_aa) {
          # only available for 2.1 and later
            print " peptide_prev_aa=\"$peptide_prev_aa\"";
            print " peptide_next_aa=\"$peptide_next_aa\"";
          }
          if ($db_list[0] > 1) {
            print " protein=\"" . &noXmlTag(($db_list[0] + 0) . "::" . $protein_list[0]) . "\"";
          } else {
            print " protein=\"" . &noXmlTag($protein_list[0]) . "\"";
          }
          print " num_tot_proteins=\"" . scalar(@protein_list) . "\"";
          print " num_matched_ions=\"" . $objPeptide->getNumIonsMatched() . "\"";
          # optional tot_num_ions
          print " calc_neutral_pep_mass=\"" . $objPeptide->getMrCalc . "\"";
          print " massdiff=\""
            . sprintf("%+.*f", $massDP, round_evenly($objPeptide->getDelta, $massDP))
            . "\"";
          # optional num_tol_term
          print " num_missed_cleavages=\"" . $objPeptide->getMissedCleavages . "\"";
          print " is_rejected=\"0\"";
          my $prot_desc = "";
          if ($print_prot_desc) {
            $prot_desc = &noXmlTag($protData->mustGetProteinDescription($protein_list[0], $db_list[0]));
          }
          if ($prot_desc) {
            print " protein_descr=\"$prot_desc\"";
          }
          my $frame = 0;
          if (($print_prot_pI || $print_prot_mass) && ExportDat::Util::db_is_NA($objResFile, $objSummary, $db_list[0])) {
            my $objProtein = $objSummary->getProtein($protein_list[0], $db_list[0]);
            if ($objProtein) {
              $frame = $objProtein->getFrame();
            }
          }
          if ($print_prot_pI) {
            my $pI = $protData->getProteinpI($protein_list[0], $frame, $db_list[0]);
            if ($pI) {
              print " calc_pI=\"$pI\"";
            }
          }
          my $protein_mw = 0;
          if ($print_prot_mass) {
            $protein_mw = $protData->mustGetProteinMass($protein_list[0], $frame, $db_list[0]);
          }
          if ($protein_mw) {
            print " protein_mw=\""
              . round_evenly($protein_mw, 0)
              . "\"";
          }
          print ">\n";

          if (scalar(@protein_list) > 1) {
          # alternative_protein
            for (my $prot_num = 1; $prot_num <= $#protein_list; $prot_num++) {
              print "<alternative_protein";
              if ($db_list[$prot_num] > 1) {
                print " protein=\"" . &noXmlTag(($db_list[$prot_num] + 0) . "::" . $protein_list[$prot_num]) . "\"";
              } else {
                print " protein=\"" . &noXmlTag($protein_list[$prot_num]) . "\"";
              }
              if ($print_prot_desc) {
              # only look for description in result file for alternative proteins
                if ($tempString = &noXmlTag($objSummary->getProteinDescription($protein_list[$prot_num], $db_list[$prot_num]))) {
                  print " protein_descr=\"$tempString\"";
                }
              }
              # optional num_tol_term
              if ($print_prot_mass) {
              # only look for mass in result file for alternative proteins
                if ($tempString = $objSummary->getProteinMass($protein_list[$prot_num], $db_list[$prot_num])) {
                  print " protein_mw=\""
                    . round_evenly($tempString, 0)
                    . "\"";
                }
              }
              print " />\n";
            }
          }

          my $varModsStr = $objPeptide->getVarModsStr();
          my $summedModsStr = $objPeptide->getSummedModsStr() eq '' ? '0' x length($varModsStr) : $objPeptide->getSummedModsStr();
          my $local_mods = lookup_query_level_mods(\%query_level_mod_cache, $objResFile, $objPeptide, $objUmodSection) || {};
          my $psmComponent = $msparser::ms_peptide::PSM_COMPLETE;
          if ($varModsStr =~ /[1-9A-X]/ or $summedModsStr =~ /[1-9A-X]/ or ($local_mods and $$local_mods{$psmComponent}) ) {
          # modification_info
          # grab any mod found in error tolerant search
            $tempString =
                $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_et_mods");
            if ($tempString) {
              ($vmMass{'X'}, undef, $vmString{'X'}) = split(/,/, $tempString, 3);
            }
            print "<modification_info";
          # N term
            if (substr($varModsStr, 0, 1) =~ /[1-9A-X]/ or substr($summedModsStr, 0, 1) =~ /[1-9A-X]/ or $$local_mods{$psmComponent}{0}){
              my $mass = $objMassesFile->getNterminalMass($mass_type);
              for my $m (substr($varModsStr, 0, 1), substr($summedModsStr, 0, 1)) {
                $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
              }
              if ($$local_mods{$psmComponent}{0}) {
                $mass += $$local_mods{$psmComponent}{0}{'delta'};
              }
              print " mod_nterm_mass=\"" . $mass . "\"";
            }
          # C term
            if (substr($varModsStr, -1, 1) =~ /[1-9A-X]/ or substr($summedModsStr, -1, 1) or $$local_mods{$psmComponent}{length($varModsStr)-1}){
              my $mass = $objMassesFile->getCterminalMass($mass_type);
              for my $m (substr($varModsStr, -1, 1), substr($summedModsStr, -1, 1)) {
                $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
              }
              if ($$local_mods{$psmComponent}{length($varModsStr)-1}) {
                $mass += $$local_mods{$psmComponent}{length($varModsStr)-1}{'delta'};
              }
              print " mod_cterm_mass=\"" . $mass . "\"";
            }
            # optional modified_peptide
            print ">\n";

          # residues
            for (my $i = 1; $i < (length($varModsStr) - 1); $i++){
              if (substr($varModsStr, $i, 1) =~ /[1-9A-X]/ or substr($summedModsStr, $i, 1) =~ /[1-9A-X]/ or $$local_mods{$psmComponent}{$i}) {
                my $mass = $objMassesFile->getResidueMass($mass_type, uc substr($objPeptide->getPeptideStr(), $i - 1, 1));
                for my $m (substr($varModsStr, $i, 1), substr($summedModsStr, $i, 1)) {
                  $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
                }
                if ($$local_mods{$psmComponent}{$i}) {
                  $mass += $$local_mods{$psmComponent}{$i}{'delta'};
                }
                print "<mod_aminoacid_mass";
                print " position=\"$i\"";
                print " mass=\"" . $mass . "\" />\n";
              }
            }

            print "</modification_info>\n";
          }

        # search_score
        my $dbType = $objPeptide->getIsFromLibrary()
            ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
            : $msparser::ms_mascotresults::DM_FASTA
            ;
          print "<search_score name=\"ionscore\" value=\""
            . round_evenly($objPeptide->getIonsScore, 2) . "\" />\n";
          print "<search_score name=\"identityscore\" value=\""
            . $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $dbType) . "\" />\n";
          if (0 < $objSummary->getHomologyThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $objPeptide->getRank())) {
            print "<search_score name=\"homologyscore\" value=\""
              . $objSummary->getHomologyThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $objPeptide->getRank()) . "\" />\n";
          }
          print "<search_score name=\"expect\" value=\""
            . round_evenly($objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $queryNum, $objPeptide->getRank()), 2, 'g') . "\" />\n";
          # optional type
          print "</search_hit>\n";

        } else {
          if ($rank == 1) {
            print "<spectrum_query";
            print " spectrum=\"" . &noXmlTag($objQuery->getStringTitle(1)) . "\"";
            print " start_scan=\"0\"";
            print " end_scan=\"0\"";
            print " precursor_neutral_mass=\"" . $objResFile->getObservedMrValue($queryNum) . "\"";
            print " assumed_charge=\"" . $objResFile->getObservedCharge($queryNum) . "\"";
            my $repeatSearchString = $objResFile->getRepeatSearchString($queryNum);
            $repeatSearchString =~ s/^[0-9.]+\s*//;
            $repeatSearchString =~ s/from\(.+?\)\s*//;
            $repeatSearchString =~ s/title\(.+?\)\s*//;
            $repeatSearchString =~ s/query\(.+?\)\s*//;
            $repeatSearchString =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
            $repeatSearchString =~ s/^\s*//;
            $repeatSearchString =~ s/\s*$//;
            if ($repeatSearchString) {
              print " search_specification=\"" . &noXmlTag($repeatSearchString) . "\"";
            }
            print " index=\"$queryNum\" />\n";
          }
          last;
        }
      } else {
        last;
      }
    }
    if ($needClosingTag) {
      print "</search_result>\n";
      print "</spectrum_query>\n";
    }
  }

  print "</msms_run_summary>\n";
  print "</msms_pipeline_analysis>\n";

}

###############################################################################
# &getEnzymeDef()
# $_[0] enzyme title
# returns a hash:
#   - 'independent' (1 or 0)
#   - 'semispecific' (1 or 0)
#   - 'num_spec' (number of specificities)
#   - 'spec_cleavage' (array of cleavage strings for each specificity)
#   - 'spec_restrict' (array of restriction strings for each specificity)
#   - 'spec_cleavage_end' (array of C/N for each specificity)
###############################################################################

sub getEnzymeDef {
  my ($title) = @_;

# get enzyme cleavage rules
  open(my $fh, "../config/enzymes") or die "cannot open enzymes file: $! ($^E)";

  my $title_re = qr{^Title:\Q$title\E\s*$}i;

  my $found = 0;
  my $independent = 0;
  my $semispecific = 0;
  my (@cleavage, @restrict, @cleaveEnd);

  while (<$fh>) {
    next unless $_ =~ $title_re;

    # We've found the correct title line; now let's read the attributes.
    $found = 1;

    while (<$fh>) {
      last if /^\*/; # end of definition

      chomp;
      s/\s*$//;  # delete trailing white space

      if (/^Cleavage:(\w+)/i){
        $cleavage[0] = $1;
      } elsif (/^Restrict:(\w+)/i){
        $restrict[0] = $1;
      } elsif (/^Restrict:/i){
        # ignore
      } elsif (/^(\w)term$/i){
        $cleaveEnd[0] = $1;
      } elsif (/^Cleavage\[(\d+)\]:(\w+)/i){
        $cleavage[$1] = $2;
      } elsif (/^Restrict\[(\d+)\]:(\w+)/i){
        $restrict[$1] = $2;
      } elsif (/^Restrict\[(\d+)\]:/i){
        # ignore
      } elsif (/^(\w)term\[(\d+)\]/i){
        $cleaveEnd[$2] = $1;
      } elsif (/^Independent:(\d)/i){
        $independent = $1;
      } elsif (/^SemiSpecific:(\d)/i){
        $semispecific = $1;
      } else {
        die "Don't understand the enzyme file format (line $.): $_";
      }

      next;
    }

    last;
  }

  close($fh) or die "cannot close enzymes file: $! ($^E)";

  if (not $found) {
    die "could not find enzyme definition for $title";
  }

# Sanity checks
  unless (defined($cleavage[0]) && $cleavage[0]) {
    die "invalid enzyme definition for $_[0]";
  }
  if (@cleaveEnd != @cleavage || @restrict > @cleavage) {
    die "invalid enzyme definition for $_[0]";
  }
  foreach (@cleavage) {
    unless ($_) {
      die "invalid enzyme definition for $_[0]";
    }
  }
  foreach (@cleaveEnd) {
    unless ($_ eq "N" || $_ eq "C") {
      die "invalid enzyme definition for $_[0]";
    }
  }

  return (
    independent => $independent,
    semispecific => $semispecific,
    num_spec => scalar(@cleavage),
    spec_cleavage => \@cleavage,
    spec_restrict => \@restrict,
    spec_cleavage_end => \@cleaveEnd,
  );
}

###############################################################################
# &DTASelect()
# output Scripps DTASelect file
# my($objParams, $objResFile, $objSummary, $msresFlags);
###############################################################################

sub DTASelect {

  my $sameset_ordering = $protData->sameset_ordering;

# pre-calculate total intensity values
  my @totalIntensity;
  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
    $totalIntensity[$queryNum] = $objQuery->getTotalIonsIntensity();
  }

# terminal_modifications
# fixed
  my(@term_mods, $iCount);
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;
    while ($objParams->getFixedModsName($iCount)) {
      push @term_mods, &getTerminalMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @term_mods, &getTerminalMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;
  while ($objParams->getVarModsName($iCount)) {
    push @term_mods, &getTerminalMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  my $fixProtC = 0;
  my $fixProtN = 0;
  my $fixPepC = 0;
  my $fixPepN = 0;
  while (@term_mods) {
    my $terminus = shift(@term_mods);
    my $massdiff = shift(@term_mods);
    my $mass = shift(@term_mods);
    my $variable = shift(@term_mods);
    my $protein_terminus = shift(@term_mods);
    my $description = shift(@term_mods);
    if ($variable eq "N") {
      if ($protein_terminus eq "c") {
        $fixProtC = $massdiff;
      } elsif ($protein_terminus eq "n") {
        $fixProtN = $massdiff;
      } elsif ($terminus eq "c") {
        $fixPepC = $massdiff;
      } elsif ($terminus eq "n") {
        $fixPepN = $massdiff
      }
    }
  }

# output header lines
  print  "DTASelect v1.9\n";
  print  "\n";
  print  "\n";
  print  "S\t" . $objResFile->getFastaVer(1) . "\t$fixProtC\t$fixProtN\t$fixPepC\t$fixPepN\tfalse\ttrue\n";

# aminoacid_modifications
# fixed
  my @res_mods;
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;
    while ($objParams->getFixedModsName($iCount)) {
      push @res_mods, &getResidueMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @res_mods, &getResidueMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;
  while ($objParams->getVarModsName($iCount)) {
    push @res_mods, &getResidueMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  # 'use warnings' complains if # is inside qw().
  my @symbols = (qw(* @ $ % & ~ ? / + = 1 2 3 4 5 6 7 8 9), '#');
  my %symLink;
  while (@res_mods) {
    my $aminoacid = shift(@res_mods);
    my $massdiff = shift(@res_mods);
    my $mass = shift(@res_mods);
    my $variable = shift(@res_mods);
    my $terminalMod = shift(@res_mods);
    my $description = shift(@res_mods);
    if ($variable eq "N") {
      print "SM\t$massdiff\t$aminoacid\n";
    } else {
      my $symbol = shift(@symbols);
      if (not defined $symbol) {
        common_subs::fatal("DTASelect cannot handle this many variable mods", __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
      }
      ${ $symLink{$description} }{$aminoacid} = $symbol;
      print "DM\t$massdiff\t$symbol\t$aminoacid\t0.0\t0.0\n";
    }
  }

  print  "Source\tMascot\t" . $objResFile->getMascotVer() . "\n";
  print  "Format\tMascot\n";

# start of MAIN LOOP to enumerate the protein hits
  my $numL = 0;
  my $numD = 0;
  my $thisHit = 1;
  my $objProtein = $objSummary->getHit($thisHit);
  while (defined($objProtein)) {
    my @sameset = common_subs::compileSameset($objSummary, $objProtein,
      $msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS,
      $$sameset_ordering{$thisHit}
    );

    &outputL($sameset[0]);
    $numL++;
    &outputD($sameset[0], \@totalIntensity, \%symLink);
    $numD++;
    if ($msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS) {
      for (@sameset[1 .. $#sameset]) {
        &outputL($_);
        $numL++;
        &outputD($_, \@totalIntensity, \%symLink);
        $numD++;
      }
    # $ShowSubSets hard coded to 1 for DTASelect
      $i = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($thisHit, $i++)) {
        if ($objProtein->getScore() > 0) {
          &outputL($objProtein);
          $numL++;
          &outputD($objProtein, \@totalIntensity, \%symLink);
          $numD++;
        }
      }
    }

    $thisHit++;
    $objProtein = $objSummary->getHit($thisHit);

  }
# end of MAIN LOOP to enumerate the protein hits

  print  "C\t$numL\t" . $objResFile->getNumQueries() . "\t$numD\n";

}

###############################################################################
# &outputL()
# output DTASelect L line
# $_[0] $objProtein
# my($objParams, $objSummary);
###############################################################################

sub outputL {

  my $objProtein = shift;

  my ($Length, $Description, $MW, $pI);
  my $preferred_acc = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  if ($thisScript->param($urlParams{'prot_desc'})) {
    $Description = $protData->mustGetProteinDescription($preferred_acc, $objProtein->getDB);
  }
  $Description = "no description" unless ($Description);
  if ($thisScript->param($urlParams{'prot_mass'})) {
    $MW = $protData->mustGetProteinMass($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  }
  $MW = 0 unless ($MW);
  if ($thisScript->param($urlParams{'prot_len'})) {
    $Length = getProteinLen($objResFile, $objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  } elsif ($MW) {
    $Length = int($MW / 110);
  }
  $Length = 0 unless ($Length);
  if ($thisScript->param($urlParams{'prot_pi'})) {
    $pI = $protData->getProteinpI($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  }
  $pI = 0 unless ($pI);

  my $dbNumPrefix = "";
  if ($objParams->getNumberOfDatabases > 1 && $objProtein->getDB > 1) {
    $dbNumPrefix = $objProtein->getDB . "::";
  }

  print  join("\t",
    "L",
    $dbNumPrefix . $preferred_acc,
    $Length,
    $MW,
    $pI,
    $Description,
    "U\n");

}

###############################################################################
# &outputD()
# output DTASelect D line
# $_[0] $objProtein
# $_[1] reference to @totalIntensity
# $_[2] reference to %symLink
# my($objResFile, $objSummary, $fileIn);
###############################################################################

sub outputD {

  my ($objProtein, $totalIntensity_ref, $symLink_ref) = @_;

# Enumerate the peptides that matched this protein
  for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
    my $queryNum = $objProtein->getPeptideQuery($i);
    my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
    my $ionSeriesString = $objPeptide->getSeriesUsedStr();
    my $numScoreSeries = $ionSeriesString =~ tr/2//;
    my $unique2locus = "true";
    if ($objPeptide->getNumProteins() > 0) {
      $unique2locus = "false";
    }

    my $local_pep = $objPeptide->getPeptideStr();

  # use the variable mods string to interpolate variable mod symbols into peptide string
  # Note: neither summed nor query-level modifications can be exported here
    my $varModsString = $objPeptide->getVarModsStr();
    unless ($varModsString eq "0" x length($varModsString)) {
      for (my $i = length($varModsString) - 2; $i > 0; $i--) {
        if (substr($varModsString, $i, 1) =~ /[1-9A-X]/) {
          my $symbol = ${ ${ $symLink_ref }{$vmString{substr($varModsString, $i, 1)}} }{substr($local_pep, $i - 1, 1)};
          substr($local_pep, $i - 1, 1) = substr($local_pep, $i - 1, 1) . $symbol;
        }
      }
    }

    my $aa_before = $objProtein->getPeptideResidueBefore($i);
    my $aa_after = $objProtein->getPeptideResidueAfter($i);
 #  unless ($aa_before eq "?" || $aa_after eq "?") {
      $aa_before =~ s/@/-/;
      $aa_after =~ s/@/-/;
      $local_pep = "$aa_before.$local_pep.$aa_after";
 #  }

    my $dbType = $objPeptide->getIsFromLibrary()
        ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
        : $msparser::ms_mascotresults::DM_FASTA
        ;

    print  join("\t",
      "D",
      "$fileIn[0].$queryNum." . $objPeptide->getCharge(),
      $fileIn[0] =~ /^(.*).dat/,
      round_evenly($objPeptide->getIonsScore, 2),
      round_evenly($objSummary->getProbOfPepBeingRandomMatch($objPeptide->getIonsScore(), $queryNum), 1),
      ($objPeptide->getMrExperimental + $elementalMasses{'hydrogen'} - $elementalMasses{'electron'}),
      ($objPeptide->getMrCalc() + $elementalMasses{'hydrogen'} - $elementalMasses{'electron'}),
      round_evenly(${ $totalIntensity_ref }[$queryNum], 1),
      $objPeptide->getPrettyRank(),
      $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $objSummary->getProbabilityThreshold, $dbType),
      round_evenly($objPeptide->getNumIonsMatched() / $numScoreSeries / (length($objPeptide->getPeptideStr()) - 1), 3),
      $local_pep,
      ($objProtein->getPeptideStart($i) - 1),
      "0",
      $unique2locus,
      "1",
      "U\n");

  }

}

###############################################################################
# &mascotDAT($thisScript, $isCGI, $fileIn, $objMascotOptions)
# download raw Mascot result file
###############################################################################

sub mascotDAT {
  my ($thisScript, $isCGI, $fileIn, $objMascotOptions, $MIME_boundary) = @_;

  # The Mascot data directory can be anywhere, so it's best to look at
  # the actual directory the input filepath points to and see if it's
  # the same directory as the Mascot data directory. (See bug #11057.)

  my $mergeAbsRelPath = sub {
    # Bug 11388: Mascot output files are written to a path given by:
    # InterFileBasePath/InterFileRelPath/yyyymmdd/Fnnnnnn.dat
    # If InterFileRelPath is a relative path, use this sub to join InterFileBasePath and InterFileRelPath
    # without following symbolic links to avoid this:
    # InterFileBasePath =/usr/local/mascot/data -> /somewhere/actual_data
    # InterFileRelPath=../data
    # If following symlinks, InterFileBasePath/InterFileRelPath is : /somewhere/data
    my ($absPath, $relPath) = @_;

    my @absDirs = File::Spec->splitdir($absPath);
    my @relDirs = File::Spec->splitdir($relPath);

    while($relDirs[0] eq '..') {
      shift(@relDirs);
      pop(@absDirs);
    }
    return File::Spec->canonpath( File::Spec->catfile( (@absDirs, @relDirs), '' ) );
  };

  my ($last_bit) = $fileIn =~ m#([/\\]\d{8}[/\\]F[a-zA-Z\d]+\.dat)$#;

  if (not defined $last_bit) {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $parentdir = Cwd::abs_path(
    substr($fileIn, 0, length($fileIn) - length($last_bit))
  );

  if ($parentdir eq '') {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $datadir = do {
    my $base = $objMascotOptions->getInterFileBasePath(),
    # Despite the relativity implied by the name, InterFileRelPath can
    # actually be an absolute pathname.
    my $rel = $objMascotOptions->getInterFileRelPath();

    if ($rel =~ m#^[.]#) {
      Cwd::abs_path( $mergeAbsRelPath->($base, $rel) )
    } else {
      Cwd::abs_path($rel)
    }
  };

  if ($datadir eq '') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header(
        -charset => "utf-8",
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Unable to determine the location of the Mascot data directory\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  }

  # Bug #11354: if InterFileBasePath or fileIn are absolute pathnames,
  # and the paths are correct, then Cwd::abs_path() retains the case of
  # the input strings. On Linux, which has case-sensitive filesystems,
  # we can be sure that if the two strings are equal, the paths are the
  # same. However, on Windows, two paths with different cases can be the same
  # path; for example, one could start with C:/, while the other starts
  # with c:/. We must force the strings into the same case before comparison.

  if ($^O =~ /Win/) {
    $parentdir = uc($parentdir);
    $datadir = uc($datadir);
  }

  if ($parentdir ne $datadir) {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $filepath = common_subs::decompress($fileIn);

  if ($filepath eq '') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header(
        -charset => "utf-8",
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Results file $fileIn does not exist\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  }

  open(my $fh, '<', $filepath) or do {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header(
        -charset => "utf-8",
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Unable to open results file for reading: $! ($^E)\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  };

  if ($isCGI or defined $MIME_boundary) {
    my ($filename) = $last_bit =~ m#[/\\](F[a-zA-Z\d]+.dat)$#;

    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', $filename);
    } else {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => $filename);
    }
  }

  print while <$fh>;

  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }

  return 0;
}

sub _daily_only {
  my ($thisScript, $isCGI, $MIME_boundary) = @_;

  if (defined $MIME_boundary) {
    common_subs::multipart_end($MIME_boundary);
    common_subs::multipart_start($MIME_boundary);
  } elsif ($isCGI) {
    print $thisScript->header(
      -charset => "utf-8",
      -type => 'text/plain',
      -X_Content_Type_Options => 'nosniff',
      -Content_Disposition => 'inline;filename=error.txt',
    );
  }
  print "You can only download result files from daily folders under the Mascot data directory\n";
  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }
  return 1;
}

###############################################################################
# &exportMGF()
# download MGF Peak List
# my($objResFile, $fileIn)
###############################################################################

sub exportMGF {
  my ($thisScript, $MIME_boundary) = @_;

  my ($fileIn) = $fileIn[0] =~ /([^\/\\]*)\.dat$/;

  if ($isCGI or defined $MIME_boundary) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', "$fileIn.mgf");
    } else {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "$fileIn.mgf");
    }
  }

  my $something_was_printed = 0;
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $i)) {
    if ($nextKey =~ /^(?:_.*)/i || $nextKey =~ /^(?:user\d\d)$/i
        || $nextKey =~ /^(OPTION_.*)/i || $nextKey =~ /^(PRINT_ATTRIBUTE_.*)/i) {
      my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
      $nextKey =~ s/\s*$//;
      if (defined $nextValue and $nextValue ne '') {
        print $nextKey.'='.$nextValue."\n";
        $something_was_printed = 1;
      }
    }
    $i++;
  }

  print "\n" if $something_was_printed;

  # List of parameters that can be embedded as query level parameters:
  # CHARGE
  # COMP
  # ETAG
  # INSTRUMENT
  # IT_MODS
  # LOCUS
  # PEPMASS
  # RAWFILE
  # RAWSCANS
  # RTINSECONDS
  # SCANS
  # SEQ
  # TAG
  # TITLE
  # TOL
  # TOLU

  if ($objResFile->anyMSMS() || $objResFile->anySQ() || $objResFile->anyTag()) {
      if($mgfOutputOrder == 1) { #Input order
	  _print_ions_to_mgf_by_idx($objResFile);
      }
      else { # Query order
	  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
	      _print_ions_to_mgf($objResFile, ($queryNum));
	  }
      }
  } else {
      # PMF
      _print_pmf_to_mgf($objResFile);
  }

  if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
  }

}

sub _print_ions_to_mgf_by_idx {
    my ($objResFile) = @_;
    use constant { IDX => 0, QUERY => 1 };

    my $nQueries = $objResFile->getNumQueries;
    # indices = [[index, query_number], ...]
    my @indices = map[ length $objResFile->getQuerySectionValueStr($_, "index")  ? $objResFile->getQuerySectionValueStr($_, "index") + 0 : -1, $_], 1 .. $nQueries;
    # In case the input indices are not available (-1) @sortedIndices is identical to @indices
    my @sortedIndices = sort {$a->[IDX] <=> $b->[IDX]} @indices;

    my @queryNums;
    my $queryNum;
    my $prevIdx;
    if ($nQueries > 0) {
        $prevIdx = $sortedIndices[0]->[IDX];
	$queryNum = $sortedIndices[0]->[QUERY];
	push @queryNums, $queryNum;
    }

    my $i = 1;
    while ($i < $nQueries){
	$queryNum = $sortedIndices[$i]->[QUERY];
	my $idx = $sortedIndices[$i]->[IDX];

	if ( ($idx == $prevIdx) && ($idx > -1) ){
	    push @queryNums, $queryNum;
	}
	else{
	    _print_ions_to_mgf($objResFile, @queryNums);
	    $prevIdx = $idx;
	    @queryNums = ($queryNum);
	}
	$i++;
    }
    _print_ions_to_mgf($objResFile, @queryNums);
}

sub _print_ions_to_mgf {
    my ($objResFile, @queryNums) = @_;
    if (!@queryNums) {
	return;
    }

    # It assumes that the list of query numbers come from the same chimera spectrum
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNums[0]);

    print "BEGIN IONS\n";
    if ($objQuery->getCharge() ne '') {
        print "CHARGE=" . $objQuery->getCharge() . "\n";
    }
    _print_mgf_indexed_array($objQuery, 'getComp', 'COMP', 20);
    if ($objQuery->getINSTRUMENT(1) ne '') {
        print "INSTRUMENT=" . $objQuery->getINSTRUMENT(1) . "\n";
    }
    if ($objQuery->getIT_MODS(1)) {
	print "IT_MODS=" . $objQuery->getIT_MODS(1) . "\n";
    }
    if ($objQuery->getLocus() ne '') {
        print "LOCUS=" . $objQuery->getLocus . "\n";
    }

    my $fmtCharge = sub {
      my $charge = shift;
      if (($charge == 0) or ($charge == -100)) {
	return $charge;
      }
      my $sign = ($charge>0) ? '+' : '-';
      return abs($charge) . $sign;
    };

    # An ambiguous charge is a list of charge states, each one has 2 characters
    my $isAmbiguousCharge = (length($objQuery->getCharge()) > 2);
    for my $queryNum(@queryNums){
	my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
	my $charge = $objResFile->getObservedCharge($queryNum) + 0;
	print "PEPMASS=";
	my $line = $objResFile->getObservedMass($queryNum);
	if ($intensity > 0) {
	    $line .= " $intensity";
	}
	if (not $isAmbiguousCharge and $charge != 0){
	    unless ($intensity > 0) {
		$line .= " 0";
	    }
	    $line .= " " . $fmtCharge->($charge);
	}
	print "$line\n";

	my $ionMobility = $objResFile->getQuerySectionValueDouble($queryNum, "ion_mobility");
	if ($ionMobility > 0) {
	  print "ION_MOBILITY=$line $ionMobility\n";
	}

    }

    if ($objQuery->getRawfile() ne '') {
        print "RAWFILE=" . $objQuery->getRawfile() . "\n";
    }
    _print_mgf_sparse_array($objQuery, 'getRawScans', 'RAWSCANS');
    _print_mgf_sparse_array($objQuery, 'getRetentionTimes', 'RTINSECONDS');
    _print_mgf_sparse_array($objQuery, 'getScanNumbers', 'SCANS');
    _print_mgf_indexed_array($objQuery, 'getSeq', 'SEQ', 20);

    for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getTag($j)) {
	    $objQuery->getTag($j) =~ /([et]),(.*)/i;
	    if (lc($1) eq "e") {
		print "ETAG=$2\n";
	    } else {
		print "TAG=$2\n";
	    }
        } else {
	    last;
        }
    }
    if ($objQuery->getStringTitle(1) ne '') {
        print "TITLE=" . $objQuery->getStringTitle(1) . "\n";
    }
    if ($objQuery->getPepTol()) {
        print "TOL=" . $objQuery->getPepTol() . "\n";
    }
    if ($objQuery->getPepTolUnits()) {
        print "TOLU=" . $objQuery->getPepTolUnits() . "\n";
    }

    my $peakStr = $objQuery->getStringIons1();
    # no way to represent ions(b- or ions(y- in MGF format, so just output values as unassigned
    if ($objQuery->getStringIons2()) {
        $peakStr .= "," . $objQuery->getStringIons2();
    }
    if ($objQuery->getStringIons3()) {
        $peakStr .= "," . $objQuery->getStringIons3();
    }
    if ($peakStr) {
        $peakStr =~ s/:/ /g;
        $peakStr =~ s/,/\n/g;
        print "$peakStr\n";
    }
    print "END IONS\n\n";
}

sub _print_pmf_to_mgf {
    my ($objResFile) = @_;
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
      my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
      print $objResFile->getObservedMass($queryNum);
      if ($intensity != 0) {
        print " $intensity\n";
      } else {
        print "\n";
      }
    }
}

sub _print_mgf_sparse_array {
  my ($objQuery, $fun, $param) = @_;

  my $indices = msparser::vectori->new();
  my $values = msparser::VectorString->new();
  $objQuery->$fun($indices, $values);

  for my $i (0 .. $indices->size-1) {
    my $idx = $indices->get($i);
    my $value = $values->get($i);
    if ($idx == -1) {
      print "$param=$value\n";
    } else {
      print "$param\[$idx\]=$value\n";
    }
  }
}

sub _print_mgf_indexed_array {
  my ($objQuery, $fun, $param, $max) = @_;

  for my $i (1 .. $max) {
    my $str = $objQuery->$fun($i);
    return if $str eq '';
    print "$param\[$i\]=$str\n";
  }
}

###############################################################################
# &display_args()
# display command line arguments for copy and paste
###############################################################################

sub display_args {

  &common_subs::printHeader1("Matrix Science - Mascot - Export search results");
  &common_subs::printHeader2("", "Export search results", 1, $session);

  print <<'end_of_static_HTML_text_block';
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
    <TR>
      <TD NOWRAP ALIGN=LEFT><h2>Command line arguments equivalent to current form settings</h2></TD>
    </TR>
    <TR>
      <TD ALIGN=LEFT><tt>
end_of_static_HTML_text_block

  my $args = ExportDat::Util::cmdline_arguments($thisScript, \%urlParams);

  for (sort keys %$args) {
    # Bug 12001: filter out sessionID from command-line arguments.
    next if $_ eq 'sessionid';
    if ($$args{$_} =~ /\s/) {
      print lc($_) . "=\"" . $$args{$_} . "\" ";
    } else {
      print lc($_) . "=" . $$args{$_} . " ";
    }
  }

  # Bug 11639: Below is code for outputting a properly encoded string to
  # be used as part of a URL. We decided to back out from displaying it,
  # since it can be confusing to less technically skilled users.

=pod
  # In most cases, you could just concatenate the command-line string using
  # '&', but this won't work with string arguments that allow spaces, such as
  # _quant_pep_threshold_type. So we output a URL-encoded string as
  # a convenience.
  #
  # However, browsers don't break words with semicolon or ampersand, so
  # in order to prevent silly display (one long line), we need to introduce
  # optional breakpoints. The <wbr> hint works in all browsers since IE6,
  # except in IE8 in pure IE8 mode. See:
  #
  # http://www.cs.tut.fi/~jkorpela/html/nobr.html#wbr
  #
  # The solution is to output both <wbr> and an empty inline element like
  # <a> or <span> with the right CSS class. The trick has no effect on how
  # other browsers render the URL, so there's no need for IE8-specific
  # CSS targeting.
  #
  # Additionally, put do_export=1 and generate_file=1 in the parameter list,
  # so that if you use e.g. wget to fetch it, file download starts
  # immediately.

  my @url_params = Util::Hash::query_array({ %$args, do_export => 1, generate_file => 1 }, 1);

  print <<'end_of_static_HTML_text_block';
      </tt></TD>
    </TR>
    <TR>
      <TD>&nbsp;</TD>
    </TR>
    <TR>
      <TD NOWRAP ALIGN=LEFT><h2>Arguments encoded as a URL parameter list</h2></TD>
    </TR>
    <TR>
      <TD ALIGN=LEFT><tt style=".wbr-hack:after { content: '\00200B'; }">
end_of_static_HTML_text_block

  print join('&<wbr><span class="wbr-hack"></span>', @url_params);
=cut

  print <<'end_of_static_HTML_text_block';
      </tt></TD>
    </TR>
  </TABLE>
end_of_static_HTML_text_block

  # Bug 11639: Link to the corresponding help text on how to execute the
  # command.

  print <<'end_of_static_HTML_text_block';
  <p style="margin-top: 1em">Instructions for running
  <code>export_dat_2.pl</code> using these arguments can be found in the help
  section
  <a href="../help/export_help.html#COMMAND">Export search results &ndash; Command Line Execution</a>.
  </p>
end_of_static_HTML_text_block

  &common_subs::printFooter;

  print "</BODY>\n";
  print "</HTML>\n";

}





###############################################################################
# mzTab()
# output PSI mzTab metadata (MTD), protein table (PRH, PRT) and peptide table
# (PEH, PEP)
###############################################################################

sub mzTab {
  my $vecFixed = ExportDat::Util::make_mod_vector($objResFile, $objModFile, $objXlModFile, 'FixedMod');

# make hash for fast retrieval of fixed mods
  my %fixedMods = ExportDat::Util::make_fixed_mods_map($vecFixed, $mass_type);

  # We export only two kinds of mzTab 1.0 files:
  #   - Complete Identification
  #   - Complete Quantification
  # Quantification export is selected when the results file has a quantitation
  # method and the ms_ms2quantitation object has been initialised without
  # error. Otherwise the file is exported in Identification mode.

  # We can't print out the header block until we know exactly which
  # modifications the peptide and protein tables contain. This because the
  # 'mod' header in MTD must list all modifications mentioned in either data
  # table. The peptide table might not contain all modifications given as
  # search parameters, and it might contain error tolerant mods and library
  # mods. Thus, we need to iterate over proteins and peptides first, pool the
  # data table rows, and then print out everything at the very end.
  #
  # Also, we need to iterate over peptide data before constructing the protein
  # rows, because we need to map all variable modifications from the peptides
  # onto the protein sequence.

  my $sameset_ordering = $protData->sameset_ordering;

  my $showPRH = $thisScript->param($urlParams{'protein_master'});
  my $protDesc = $thisScript->param($urlParams{'prot_desc'});
  my $protMass = $thisScript->param($urlParams{'prot_mass'});
  my $protTaxStr = $thisScript->param($urlParams{'prot_tax_str'});
  my $protCover = $thisScript->param($urlParams{'prot_cover'});
  my $protEmPAI = $thisScript->param($urlParams{'prot_empai'});
  my $protpI = $thisScript->param($urlParams{'prot_pi'});
  my $protQuant = $thisScript->param($urlParams{'prot_quant'});

  my $showPEH = $thisScript->param($urlParams{'peptide_master'});
  my $pepDelta = $thisScript->param($urlParams{'pep_delta'});
  my $pepMiss = $thisScript->param($urlParams{'pep_miss'});
  my $pepScore = $thisScript->param($urlParams{'pep_score'});
  my $pepHomThr = $thisScript->param($urlParams{'pep_homol'});
  my $pepIdThr = $thisScript->param($urlParams{'pep_ident'});
  my $pepFrame = $thisScript->param($urlParams{'pep_frame'});
  my $pepNumMatch = $thisScript->param($urlParams{'pep_num_match'});
  my $pepScanTitle = $thisScript->param($urlParams{'pep_scan_title'});
  my $pepQuant = $thisScript->param($urlParams{'pep_quant'});

  my $filter_duplicates = 1;
  if ($thisScript->param($urlParams{'show_pep_dupes'})) {
    $filter_duplicates = 0;
  }

  my $is_decoy = $objParams->getDECOY > 0 && $ShowDecoyReport;
  my $is_error_tolerant = $objResFile->isErrorTolerant();

  if (not $protQuant and not $pepQuant) {
    # Export Identification only.
    $objQuantMethod = undef;
    $objMS2Quant = undef;
  } else {
    # If either quantitation data has been selected, we must export both.
    # This is required by the specification. The user can choose not to
    # export the peptide tables if he does not want to see peptide ratios.
    $pepQuant = $protQuant = 1;
  }

  my ($ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs) = mzTabExperimentStructure(
      \@fileIn, $objResFile, $objParams, $objSummary, $objQuantMethod,
      $objMS2Quant, $msresFlags, $numHits, $UniGeneFile,
      $ignoreIonsScoreBelow, $minPepLen, $minNumSigUniqueSeqs, $msresFlags2, $ShowSubSets,
  );

  my %query_mod_and_dupe_cache = ();

  # We need to collect protein data even if $showPEH == 0, because the
  # peptide rows still report protein accessions.
  my ($protein_rows, $peptide_map) = collectMzTabProteins(
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $ShowSameSets, $ShowSubSets, $reportType,
    $protDesc, $protMass, $protTaxStr, $protCover, $protEmPAI, $protpI, $protQuant,
    \%query_mod_and_dupe_cache, $filter_duplicates, $sameset_ordering
  );

  my $indexIsQueryNumber = 1;
  my $objQueryTmp = new msparser::ms_inputquery($objResFile, 1);
  if ($objQueryTmp->getIndex > -1) {
    $indexIsQueryNumber = 0;
  } elsif ($objParams->getFORMAT eq "mzData (.XML)") {
    $indexIsQueryNumber = 0;
  } elsif ($objParams->getFORMAT eq "mzML (.mzML)") {
    $indexIsQueryNumber = 0;
  }

  # Note that we must collect peptide information even if $showPRH == 0
  # and $showPEH == 0, because we need the modification tally.
  my ($peptide_rows, $protein_mods, $mod_tally, $any_retention_time) = collectMzTabPeptides(
      $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
      $msresFlags2, $objMS2Quant, $objUmodSection,
      $objFullUmodFile, $sessionID, $indexIsQueryNumber, $is_decoy,
      $is_error_tolerant, $OmitErrTolThresholds, $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr,
      $pepFrame, $pepNumMatch, $pepScanTitle, $pepQuant, $massDP, \%vmMass,
      \%vmString, $percolate, \%fixedMods, \%query_mod_and_dupe_cache,
      $peptide_map
  );

  my @database_map = ();

  if ($UniGeneFile) {
    # We know the UniGene file exists, because it has been validated before.
    my $mtime = (stat($UniGeneFile))[9];
    $database_map[0] = {
      full_name => $UniGeneFile,
      abbr_name => $UniGeneFile,
      version => _fmtdate_iso8601($mtime),
    };
  }

  for my $i (1 .. $objParams->getNumberOfDatabases()) {
    # The specification recommends using MIRIAM registered names, but that's
    # up to the user who created and named the database definition.
    my $name = $objParams->getDB($i);

    $database_map[$i] = {
      full_name => sprintf('%s (%d entries)', $name, $objResFile->getNumSeqs($i)),
      abbr_name => $name,
      version => $objResFile->getFastaVer($i),
    };
  }

  printMzTabMTD(
    $thisScript, $URI, \%urlParams, $ms_run_resfiles, $comp_idx_to_assay,
    $study_variable_defs, $objSummary, $objQuantMethod, $objMS2Quant,
    $objUmodSection, $objFullUmodFile, $indexIsQueryNumber, $reportType,
    $libraryMode, $protTaxStr, \%taxNames, $pepDelta, \@quantWarnings,
    $mod_tally, $any_retention_time, $showPRH, $showPEH, $protQuant,
    $pepQuant, $percolate,
  );

  my $num_ms_runs = $#$ms_run_resfiles;
  my $num_study_variables = scalar(@$study_variable_defs);

  if ($showPRH) {
    print "\n";

    printMzTabPRHandPRT(
      $protein_rows, $objResFile->isPMF, $objMS2Quant, $protQuant,
      $num_ms_runs, $num_study_variables, $protMass, $protEmPAI, $protpI, $reportType,
      $protein_mods, \@database_map,
    );
  }

  if ($showPEH) {
    if ($objQuantMethod) {
      print "\n";

      # Only export a Peptides section in Quantification files.
      printMzTabPEHandPEP(
        $peptide_rows, $objMS2Quant, $num_ms_runs, $num_study_variables,
        $is_decoy, $is_error_tolerant, $percolate, $libraryMode,
        \@database_map, $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr,
        $pepFrame, $pepNumMatch, $pepScanTitle, $pepQuant
      );
    }

    print "\n";

    # Always export the PSM section if peptide output is desired, even in
    # Quantification files. We treat PSMs as if they were peptides (for
    # simplicity), so these two sections are nearly identical. The main
    # difference is that the PSM table contains data about start and end
    # residues and calculated mass, while the Peptides table contains
    # quantitation ratios.
    printMzTabPSHandPSM(
      $peptide_rows, $is_decoy, $is_error_tolerant, $percolate, $libraryMode,
      \@database_map, $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr, $pepFrame,
      $pepNumMatch, $pepScanTitle
    );
  }
}

sub mzTabExperimentStructure {
  my (
      $fileIn, $objResFile, $objParams, $objSummary, $objQuantMethod,
      $objMS2Quant, $msresFlags, $numHits, $UniGeneFile,
      $ignoreIonsScoreBelow, $minPepLen, $minNumSigUniqueSeqs, $msresFlags2, $ShowSubSets,
  ) = @_;

  # An index from resfile (ms_run) index to resfile and params is needed
  # for reporting ms_run-specific protein scores (which is a required column).
  my @ms_run_resfiles = (
      # combined resfile at index 0
      { resfile => $objResFile, summary => $objSummary, params => $objParams }
  );

  for my $i (1 .. @$fileIn) {
    my $resfile = $objResFile;
    my $params = $objParams;
    my $summary = $objSummary;

    if (@$fileIn > 1) {
      # If there is more than one ms_run, we must open a summary for each one.
      # It's possible that some of the individual cache files do not exist, but
      # as this is a rather rare use case, we'll ignore the problem and simply
      # suppress keepalive to prevent output corruption.

      my $resfile = $objResFile->getResfile($i);
      $params = $resfile->params;
      $objResFile->resetKeepAlive(0, '', 1);

      if ($reportType eq 'concise') {
        my $pepsum_params = msparser::ms_mascotresults_params->new;
        $pepsum_params->setFlags($msresFlags);
        $pepsum_params->setMinProbability($objSummary->getProbabilityThreshold);
        $pepsum_params->setMaxHitsToReport($numHits);
        $pepsum_params->setMinNumSigUniqueSequences($minNumSigUniqueSeqs);
        $summary = new msparser::ms_proteinsummary($resfile, $pepsum_params);
      } else {
        my $pepsum_params = msparser::ms_mascotresults_params->new(
            $msresFlags,
            $objSummary->getProbabilityThreshold,
            $numHits,
            $UniGeneFile,
            $ignoreIonsScoreBelow,
            $minPepLen,
            "",
            $msresFlags2,
        );
        $pepsum_params->setMinNumSigUniqueSequences($minNumSigUniqueSeqs);
        # Note: for consistency, must set target FDR = 0 and use the
        # combined report's minProbability.
        #
        # TODO: won't work with ET decoy, but that's not yet supported
        # with quantitation.
        common_subs::setTargetFDRandTypeParam($pepsum_params, 0, 'psm');
        $summary = new msparser::ms_peptidesummary($resfile, $pepsum_params);
        $summary->setSubsetsThreshold($ShowSubSets);
      }
    }

    $ms_run_resfiles[$i] = {
        filepath => $$fileIn[$i-1],
        resfile => $resfile,
        params => $params,
        summary => $summary,
    };
  }

  if (not $objQuantMethod or not $objMS2Quant) {
    # There is no need to report assay information in Identification-only
    # files.
    return (\@ms_run_resfiles, [], []);
  }

  # The relationship between samples, runs and assays is somewhat subtle.
  # If there is more than one results file (combined export), we assume each
  # file is an ms_run and a technical replicate. This is the most common
  # case for merging files and if it is not the case, we would need more
  # input from the user to determine the relationships. Further, we know
  # the quantitation method component definitions must be the same in all
  # of the files. Thus, if the quantitation method contains C components, we
  # assume there are C biological samples. If there are K files, we assume
  # there are K technical replicates of each sample, making K*C assays.
  #
  # In both protocols, with K input files, C channels and R ratios, we have
  # (for 1 <= k <= K and 1 <= c <= C and 1 <= r <= R)
  #
  #    assay[c+C*(k-1)]-sample_ref   = sample[c]
  #    assay[c+C*(k-1)]-ms_run_ref   = ms_run[k]
  #    study_variable[r]-ms_run_refs = ms_run[1], ..., ms_run[K]
  #    study_variable[r]-assay_refs  = assay[c+C(*k-1)] for some c's and all k
  #
  # For example, iTRAQ 4-plex with ratios 115/114, 116/114 and 117/114
  # and one file has (K = 1, C = 4, R = 3)
  #
  #   assay[c]-sample_ref           = sample[c] for c = 1,2,3,4
  #   assay[c]-ms_run_ref           = ms_run[1] for c = 1,2,3,4
  #   study_variable[r]-ms_run_refs = ms_run[1] for r = 1,2,3
  #   study_variable[1]-assay_refs  = assay[1], assay[2]
  #   study_variable[2]-assay_refs  = assay[1], assay[3]
  #   study_variable[3]-assay_refs  = assay[1], assay[4]
  #
  # In 18O Multiplex with two files, two heavy and one light component, and
  # quantitation ratio heavy/light = (18O1 + 18O2)/18O0, the definitions
  # are (K = 2, C = 3, R = 1)
  #
  #   assay[c]-sample_ref           = sample[1] for c = 1,4
  #   assay[c]-sample_ref           = sample[2] for c = 2,5
  #   assay[c]-sample_ref           = sample[3] for c = 3,6
  #   assay[c]-ms_run_ref           = ms_run[1] for c = 1,2,3
  #   assay[c]-ms_run_ref           = ms_run[2] for c = 4,5,6
  #   study_variable[1]-ms_run_refs = ms_run[1], ms_run[2]
  #   study_variable[1]-assay_refs  = assay[1], assay[2], assay[3], assay[4], assay[5], assay[6]
  #
  # Here assay[1] and assay[4] are 18O0, assay[2] and assay[5] are 18O1
  # and assay[3] and assay[6] are 18O2.
  #
  # Note 1: mzTab 1.0 does not provide any mechanism for defining the ratio
  # formula. We export the ratio name in the free-text field
  # study_variable[]-description and the peptide ratio formula in a custom
  # field. For example, iTRAQ 4-plex above has
  #
  #   MTD study_variable[1]-description  115/114
  #   MTD custom [,,study_variable[1]-peptide_ratio_formula, assay[2]/assay[1]]
  #
  # where the ratio name defined in the quantitation method is in quotes.
  # The two-file 18O Multiplex above has
  #
  #   MTD study_variable[1]-description  heavy/light
  #   MTD custom [,,study-variable[1]-peptide_ratio_formula, (assay[2] + assay[5] + assay[3] + assay[6])/(assay[1] + assay[4])]
  #
  # because both assay[2] and assay[5] are 18O1 etc.
  #
  # For practical reasons, when we export a Peptides section, we treat peptides
  # as if they were PSMs. This means a peptide only ever has component
  # intensities from one ms_run, so in the above formula,
  #   either assay[4] = assay[5] = assay[6] = null
  #   or     assay[1] = assay[2] = assay[3] = null.
  # Both simplify the formula to its usual form.
  #
  # Note 2: protein ratios are complicated functions of peptide ratios,
  # e.g. a geometric mean. The mzTab 1.0 specification does not provide
  # any way to describe how protein ratios are calculated. Instead, we
  # export a custom metadata field, e.g.:
  #
  #   MTD custom [,,study-variable[1]-protein_ratio_type, median]

  my @comp_names = ();

  for my $i (0 .. $objQuantMethod->getNumberOfComponents()-1) {
    my $comp = $objQuantMethod->getComponentByNumber($i);
    my $comp_name = '';
    if ($comp->haveName()) {
      $comp_name = $comp->getName();
    }
    push @comp_names, $comp_name;
  }

  my @comp_idx_to_assay = ();
  my %comp_name_to_assay = ();
  my $num_assays = 0;

  for my $i (1 .. @$fileIn) {
    for my $j (1 .. @comp_names) {
      $comp_name_to_assay{$comp_names[$j-1]}{$i}
        = $comp_idx_to_assay[$i][$j]
        = ++$num_assays;
    }
  }

  my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
  my @study_variable_defs = ();

  for my $i (1 .. @ratio_names) {
    my $ratio_def = $objMS2Quant->getCombinedReportRatio($ratio_names[$i-1]);
    my $ratio_name = '';
    if ($ratio_def->haveName()) {
      $ratio_name = $ratio_def->getName();
    }

    my %assay_indices = ();
    my @numerators = ();
    my @denominators = ();
    for my $j (0 .. $ratio_def->getNumberOfNumerators()-1) {
      my $numerator = $ratio_def->getNumeratorByNumber($j);
      next if not $numerator->haveName();
      for my $k (1 .. @$fileIn) {
        my $assay_idx = $comp_name_to_assay{$numerator->getName()}{$k};
        next if not defined $assay_idx;
        push @numerators, sprintf('assay[%d]', $assay_idx);
        $assay_indices{$assay_idx} = 1;
      }
    }
    for my $j (0 .. $ratio_def->getNumberOfDenominators()-1) {
      my $denominator = $ratio_def->getDenominatorByNumber($j);
      next if not $denominator->haveName();
      for my $k (1 .. @$fileIn) {
        my $assay_idx = $comp_name_to_assay{$denominator->getName()}{$k};
        next if not defined $assay_idx;
        push @denominators, sprintf('assay[%d]', $assay_idx);
        $assay_indices{$assay_idx} = 1;
      }
    }
    my $numerator = '1';
    if (@numerators == 0) {
      # Keep '1'
    } elsif (@numerators == 1) {
      $numerator = $numerators[0];
    } else {
      $numerator = sprintf('(%s)', join(' + ', @numerators));
    }
    my $denominator = '';
    if (@denominators == 0) {
      # Keep ''
    } elsif (@denominators == 1) {
      $denominator = sprintf('/%s', $denominators[0]);
    } else {
      $denominator = sprintf('/(%s)', join(' + ', @denominators));
    }

    push @study_variable_defs, {
        idx => $i,
        ratio_name => $ratio_name,
        description => $ratio_name,
        formula => $numerator.$denominator,
        assay_refs => join(', ', map { sprintf('assay[%d]', $_) } sort { $a <=> $b } keys %assay_indices),
    };
  }

  return (
      \@ms_run_resfiles,
      \@comp_idx_to_assay,
      \@study_variable_defs,
  );
}

sub str_or_null {
    my $str = defined($_[0]) ? $_[0] : '';
    $str =~ s/\s/ /sg;
    return $str =~ m#^\s*$# ? 'null' : $str;
}
sub printTabbed { print join("\t", map { str_or_null($_) } @_), "\n" }
sub printMTD { printTabbed('MTD', $_[0], join('', @_[1..$#_])) }

sub _fmtdate_iso8601 {
  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($_[0]);
  return sprintf('%4d-%02d-%02dT%02d:%02d:%02d',  $year+1900, $mon+1, $mday, $hour, $min, $sec);
}

# We print metadata headers in logical clumps, as apart from a few exceptions,
# the mzTab 1.0 specification does not require a fixed order. Some headers
# are not output, because they are not relevant here, e.g. sample processing
# details, publication, tissue, cell_type and disease.
sub printMzTabMTD {
  my (
    $thisScript, $URI, $urlParams, $ms_run_resfiles, $comp_idx_to_assay,
    $study_variable_defs, $objSummary, $objQuantMethod, $objMS2Quant,
    $objUmodSection, $objFullUmodFile, $indexIsQueryNumber, $reportType,
    $libraryMode, $protTaxStr, $taxNames, $pepDelta, $quantWarnings,
    $mod_tally, $any_retention_time, $showPRH, $showPEH, $protQuant,
    $pepQuant, $percolate,
  ) = @_;

  my $MISSING = 'null';

  my @custom = ();

  printMTD('mzTab-version', '1.0');
  printMTD('mzTab-mode', 'Complete');

  if ($objQuantMethod) {
    printMTD('mzTab-type', 'Quantification');
  } else {
    printMTD('mzTab-type', 'Identification');
  }

  for my $i (1 .. $#$ms_run_resfiles) {
    my $run_date = _fmtdate_iso8601($$ms_run_resfiles[$i]{'resfile'}->getDate());
    push @custom, sprintf('[,,Search date[%d], %s]', $i, $run_date);
  }

  do {
    my $export_date = _fmtdate_iso8601(time);
    push @custom, sprintf('[,,Export date, %s]', $export_date);
  };

  my $NEWT_vocabulary_used = 0;

  # Each input file $i becomes ms_run[$i] and sample[$i]. We can't know,
  # without more information from the user, whether the input files refer
  # to the same sample (technical replicates) or a different sample.
  # The mzTab 1.0 specification deliberately leaves the definition of
  # sample open to interpretation, so it is not actually incorrect to treat
  # technical replicates as different samples.
  do {
    my %file_formats = (
        'Mascot generic' => { param => 'MS:1001062', name => 'Mascot MGF format' },
        'Sequest (.DTA)' => { param => 'MS:1001063', name => 'DTA format' },
        #'Finnigan (.ASC)' => { param => '99999', name => 'TODO: Finnigan List file' }, # see mzIdentML output
        'Micromass (.PKL)' => { param => 'MS:1000565', name => 'Micromass PKL format' },
        'PerSeptive (.PKS)' => { param => 'MS:1001245', name => 'PerSeptive PKS format' },
        'Sciex API III' => { param => 'MS:1001246', name => 'Sciex API III format' },
        'Bruker (.XML)' => { param => 'MS:1001247', name => 'Bruker XML format' },
        'mzData (.XML)' => { param => 'MS:1000564', name => 'PSI mzData format' },
        'mzML (.mzML)' => { param => 'MS:1000584', name => 'mzML format' },
    );

    my @com = ();
    my @contact_name = ();
    my @contact_email = ();
    my @instrument = ();
    my @species_header = ();
    my @species_excluded = ();
    my @filename = ();
    my @file_format = ();
    my @id_format = ();

    for my $i (1 .. $#$ms_run_resfiles) {
      my $resfile = $$ms_run_resfiles[$i]{'resfile'};
      my $params = $$ms_run_resfiles[$i]{'params'};

      if ($params->getCOM ne '') {
        push @com, $params->getCOM;
      } else {
        push @com, 'untitled search';
      }

      my $has_contact = 0;

      if ($params->getUSERNAME) {
        $contact_name[$i] = $params->getUSERNAME;
      }

      # (contact-affiliation[] omitted)
      # (publication[] omitted)

      if ($params->getUSEREMAIL) {
        $contact_email[$i] = $params->getUSEREMAIL;
      }

      if ($params->getINSTRUMENT) {
        $instrument[$i] = $params->getINSTRUMENT;
      }

      my $taxonomy = new msparser::ms_taxonomyfile;
      my @header = ();
      my @excluded = ();

      if ($resfile->getTaxonomy($taxonomy)) {
        if ($taxonomy->getNumberOfEntries == 0 ) {
          push @header, { id => 1, name => 'all' };
        } else {
          my $choice = $taxonomy->getEntryByNumber(0);
          if ($choice->getNumberOfIncludeTaxonomies > 0) {
            for my $i (0 .. $choice->getNumberOfIncludeTaxonomies()-1) {
              my $id = $choice->getIncludeTaxonomy($i);
              my $tax_title = ExportDat::TaxonomyData::getTaxonomyNameFromId($id, $taxNames);
              push @header, { id => $id, name => $tax_title };
            }
          }

          if ($choice->getNumberOfExcludeTaxonomies > 0) {
            # No way to specify excluded taxa in the standard species[k] header,
            # so we need to print a custom header.
            for my $i (0 .. $choice->getNumberOfExcludeTaxonomies()-1) {
              my $id = $choice->getExcludeTaxonomy($i);
              my $tax_title = ExportDat::TaxonomyData::getTaxonomyNameFromId($id, $taxNames);
              push @excluded, { id => $id, name => $tax_title };
            }
          }
        }
      }

      $species_header[$i] = \@header;
      $species_excluded[$i] = \@excluded;

      if (my $filename = $params->getFILENAME()) {
        $filename =~ s/\\/\//g;
        $filename[$i] = 'file:///' . $filename;
      } elsif (my $url = $params->getDATAURL()) {
        $filename[$i] = $url;
      } else {
        # ms_run[]-location is mandatory, so if it's not present, we must
        # report it as missing.
        $filename[$i] = $MISSING;
      }

      if (my $cv = $file_formats{$params->getFORMAT}) {
        $file_format[$i] = sprintf('[MS, %s, %s, ]', $$cv{'param'}, $$cv{'name'});
      } elsif ($resfile->isPMF or $resfile->isSQ) {
        $file_format[$i] = sprintf('[MS, %s, %s, ]', 'MS:1001369', 'text format');
      } else {
        # A bit nonsensical, but we really don't know what the format is.
        $file_format[$i] = sprintf('[MS, %s, %s, ]', 'MS:1000560', 'mass spectrometer file format');
      }

      if ($resfile->isPMF or $resfile->isSQ) {
        $id_format[$i] = '[MS, MS:1000775, single peak list nativeID format, ]';
      } elsif ($indexIsQueryNumber) {
        $id_format[$i] = '[MS, MS:1001528, Mascot query number, ]';
      } elsif ($params->getFORMAT eq "mzML (.mzML)") {
        $id_format[$i] = '[MS, MS:1001530, mzML unique identifier, ]';
      } elsif ($params->getFORMAT eq "mzData (.XML)") {
        $id_format[$i] = '[MS, MS:1000777, spectrum identifier nativeID format, ]';
      } else {
        $id_format[$i] = '[MS, MS:1000774, multiple peak list nativeID format, ]';
      }

      # (ms_run[]-fragmentation_method omitted)
    }

    # The MTD section offers both description and title. Description is
    # mandatory, while title is not, so we'll put the search title in the
    # description field.

    # Bug 12552: PRIDE requires title to be present. The PRIDE
    # Submission Tool prints a warning if mzTab-ID isn't present. The
    # latter is optional in mzTab 1.0 and has no particular syntax, so
    # let's print the search title in all three fields.

    if (@com == 1) {
      printMTD('mzTab-ID', $com[0]);
      printMTD('title', $com[0]);
      printMTD('description', $com[0]);
    } else {
      my $str = sprintf('Combined export of %d files: %s', scalar(@com), joinand(@com));
      printMTD('mzTab-ID', $str);
      printMTD('title', $str);
      printMTD('description', $str);
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      if (defined $contact_name[$i]) {
        printMTD('contact['.$i.']-name', $contact_name[$i]);
      }
      if (defined $contact_email[$i]) {
        printMTD('contact['.$i.']-email', $contact_email[$i]);
      }
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      # Bug 12552: mzTab 1.0 requires a running index for analyzer
      # lines. Numbering restarts at every instrument. (We only have
      # one analyzer per instrument, which is the INSTRUMENT parameter
      # from the results file.)
      my $j = 1;
      if (defined $instrument[$i]) {
        printMTD('instrument['.$i.']-analyzer['.$j.']', sprintf('[,,%s,]', $instrument[$i]));
        ++$j;
      }
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      my $run_str = "ms_run[$i]";
      printMTD($run_str.'-location', $filename[$i]);
      printMTD($run_str.'-format', $file_format[$i]);
      printMTD($run_str.'-id_format', $id_format[$i]);
    }

    # sample[]-species[]

    for my $i (1 .. $#$ms_run_resfiles) {
      my $excluded = $species_excluded[$i];

      for my $j (1 .. $#$excluded) {
        my $s = $$excluded[$j];
        push @custom, sprintf(
          '[,, sample[%d]-species_excluded[%d], [NEWT, %d, %s, ]]',
          $i, $j, $$s{'id'}, defined($$s{'name'}) ? $$s{'name'} : ''
        );
        ++$NEWT_vocabulary_used;
      }

      my $header = $species_header[$i];

      for my $j (1 .. $#$header) {
        my $s = $$header[$j];
        # NEWT = NCBI taxonomy IDs.
        printMTD(sprintf('sample[%d]-species[%d]', $i, $j),
            sprintf('[NEWT, %d, %s, ]',
                $$s{'id'}, defined($$s{'name'}) ? $$s{'name'} : ''
            )
        );
        ++$NEWT_vocabulary_used;
      }
    }
  };

  # (sample_processing[] omitted)
  # (sample[]-tissue[] omitted)
  # (sample[]-cell_type[] omitted)
  # (sample[]-disease[] omitted)
  # (sample[]-description omitted)
  # (sample[]-custom[] omitted)

  #printMTD('contact[2]-name', 'Matrix Science');
  #printMTD('contact[2]-affiliation', '64 Baker Street, London W1U 7GB, UK');
  #printMTD('contact[2]-email', 'support@matrixscience.com');

  # (uri[] omitted; it's the URI to the source data, not to the Mascot report.)

  do {
    # Index number of processing software.
    my $i = 1;

    if ($libraryMode eq 'integrated' or $libraryMode eq 'onlylibrary') {
      # Version isn't saved in the results file, so print none.
      printMTD('software['.$i.']', '[MS, MS:1002750, NIST MSPepSearch, ]');

      # Bug 12552: software[]-setting[] lines must have a second
      # running index.
      my $k = 1;

      # XXX should be a list of cmdlines for combined files
      for my $j (1 .. $$ms_run_resfiles[0]{'params'}->getNumberOfDatabases()) {
        next unless $$ms_run_resfiles[0]{'resfile'}->getDatabaseType($j) == $msparser::SL;
        my $cmdline = $$ms_run_resfiles[0]{'resfile'}->getSectionValueStr(
          $msparser::ms_mascotresfile::SEC_HEADER,
          sprintf('sl_exec_command%d', $j)
        );
        next unless $cmdline;
        printMTD('software['.$i.']-setting['.$k.']', $cmdline);
        ++$k;
      }

      ++$i;
    }

    # XXX should be a list of versions for combined files
    printMTD('software['.$i.']', '[MS, MS:1001207, Mascot, ', $$ms_run_resfiles[0]{'resfile'}->getMascotVer(), ']');

    # However, all the important parameters between the combined results files
    # must be the same, so it's fine to print out only one set of parameters.
    printMTDsearchParameters('software['.$i.']-setting', $$ms_run_resfiles[0]{'params'});

    ++$i;

    if ($percolate == 1) {
      printMTD('software['.$i.']', '[MS, MS:1001490, Percolator, ', ']');
      # The Percolator command line used for this results file is not saved
      # anywhere. We could report it as it appears in mascot.dat, but that
      # could have changed since.
      #printMTDpercolatorParameters('software['.$i.']-setting', );
      ++$i;
    }

    printMTD('software['.$i.']', '[MS, MS:1001478, Mascot Parser, ', $$ms_run_resfiles[0]{'resfile'}->getMSParserVersion(), ' (export_dat_2.pl)]');
    printMTDformatParameters('software['.$i.']-setting', $thisScript, $urlParams);
  };

  for (['fixed', 'fixed_mod', '[MS, MS:1002453, No fixed modifications searched,]'],
       ['variable', 'variable_mod', '[MS, MS:1002454, No variable modifications searched,]']) {
    my ($tally_type, $param_name, $cv_empty) = @$_;
    my $i = 1;

    for my $id (sort keys %{ $$mod_tally{$tally_type} || {} }) {
      my $mod = $$mod_tally{$tally_type}{$id};
      my $mtd = sprintf('%s[%d]', $param_name, $i);
      printMTD($mtd, $$mod{'cv_element'});
      if (defined $$mod{'position'}) {
        printMTD($mtd.'-position', $$mod{'position'});
      }
      if (defined $$mod{'site'}) {
        printMTD($mtd.'-site', $$mod{'site'});
      }
      ++$i;
    }

    if ($i == 1) {
      # No mods, so report it.
      printMTD(sprintf("%s[1]", $param_name), $cv_empty);
    }
  }

  if ($objQuantMethod) {
    my $protocol = lc($objQuantMethod->getProtocol()->getType());

    # The mzTab 1.0 specification recommends outputting quantitation methods
    # as CV parameters (e.g. [MS, MS:1001837, iTRAQ quantitation analysis]),
    # but we can't infer it from the Mascot quantitation method -- the latter
    # is far more generic. The best we can do is output a user parameter.
    printMTD('quantification_method',
        sprintf('[,,"%s", "%s"]',
            $objQuantMethod->getName(), $objQuantMethod->getDescription()
        )
    );

    if ($showPRH and $protQuant) {
      printMTD('protein-quantification_unit', '[MS, MS:1001134, protein ratio, ]');
    }

    if ($showPEH and $pepQuant) {
      printMTD('peptide-quantification_unit', '[MS, MS:1001132, peptide ratio, ]');
    }

    my ($method_unimods, $method_chemmods) = _collect_mod_types($objQuantMethod, $objUmodSection);
    my @comp_mod_cache = ();

    my @assay_structure = ();
    my @assay_defs = ();

    for my $k (1 .. $#$ms_run_resfiles) {
      for my $i (1 .. $objQuantMethod->getNumberOfComponents()) {
        my $component = $objQuantMethod->getComponentByNumber($i-1);
        my $comp_name = '';
        if ($component->haveName()) {
          $comp_name = $component->getName();
        }

        my $assay_id = sprintf('assay[%d]', $$comp_idx_to_assay[$k][$i]);

        # As above, it's impossible to infer the right CV terms for
        # quantification_reagent. We have to use a user parameter.
        push @assay_defs, [$assay_id.'-quantification_reagent', '[,,'.$component->getName().',]'];

        my @mods = ();

        if ($comp_mod_cache[$i]) {
          @mods = @{ $comp_mod_cache[$i] };
        } else {
          @mods = _determine_component_quant_mods(
            $protocol, $method_unimods, $method_chemmods, $component,
            $objUmodSection
          );
          # Cache just so that we don't need to re-determine as many times
          # as there are resfiles.
          $comp_mod_cache[$i] = \@mods;
        }

        for my $j (1 .. @mods) {
          my $mod = $mods[$j-1];
          push @assay_defs, [$assay_id.'-quantification_mod['.$j.']', $$mod{'cv'}];
          if (defined $$mod{'position'}) {
            push @assay_defs, [$assay_id.'-quantification_mod['.$j.']-position', $$mod{'position'}];
          }
          if (defined $$mod{'site'}) {
            push @assay_defs, [$assay_id.'-quantification_mod['.$j.']-site', $$mod{'site'}];
          }
        }

        # At least one field should exist for each sample referred to here,
        # whether it is sample[i]-description or sample[i]-species[].
        my $sample_desc;
        if ($#$ms_run_resfiles > 1) {
          $sample_desc = sprintf('Component %d in file %d', $i, $k);
        } else {
          $sample_desc = sprintf('Component %d', $i);
        }
        push @assay_structure, [sprintf('sample[%d]-description', $i), ''];

        push @assay_structure, [$assay_id.'-sample_ref', sprintf('sample[%d]', $i)];
        push @assay_structure, [$assay_id.'-ms_run_ref', sprintf('ms_run[%d]', $k)];
      }
    }

    printMTD(@$_) for @assay_structure;
    printMTD(@$_) for @assay_defs;

    # Study variables always use information from all resfiles.
    my $sv_ms_run_refs = join(', ', map { sprintf('ms_run[%d]', $_) } 1 .. $#$ms_run_resfiles);
    my $protein_ratio_type = $objQuantMethod->getProteinRatioType();

    for my $sv (@$study_variable_defs) {
      my $sv_id = sprintf('study_variable[%d]', $$sv{'idx'});
      printMTD($sv_id.'-description', $$sv{'description'});
      printMTD($sv_id.'-assay_refs', $$sv{'assay_refs'});
      push @custom, sprintf('[,,study-variable[%d]-peptide_ratio_formula, %s]', $$sv{'idx'}, $$sv{'formula'});
      push @custom, sprintf('[,,study-variable[%d]-protein_ratio_type, %s]', $$sv{'idx'}, $protein_ratio_type);
    }
  }

  if ($objParams->getDECOY > 0
          and not $ShowDecoyReport
          and $reportType ne 'protein'
          and $reportType ne 'concise') {
    my @cv;

    my @count_types = (
        $msparser::ms_mascotresults::DS_COUNT_PSM,
        $msparser::ms_mascotresults::DS_COUNT_SEQUENCE,
    );

    for my $count_type (@count_types) {
      my $hits_above_id = $objSummary->getNumHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_type);
      my $hits_above_hom = $objSummary->getNumHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_type);
      my $decoy_above_id = $objSummary->getNumDecoyHitsAboveIdentity(1 / $objSummary->getProbabilityThreshold, $count_type);
      my $decoy_above_hom = $objSummary->getNumDecoyHitsAboveHomology(1 / $objSummary->getProbabilityThreshold, $count_type);

      my $FDR = 0;
      if ($use_homology) {
        if ($hits_above_hom) {
          $FDR = $decoy_above_hom / $hits_above_hom;
        }
      } else {
        if ($hits_above_id) {
          $FDR = $decoy_above_id / $hits_above_id;
        }
      }

      if ($FDR > 0) {
        if ($count_type == $msparser::ms_mascotresults::DS_COUNT_SEQUENCE) {
          push @cv, sprintf('[MS, MS:1001364, peptide sequence-level global FDR, %s]', round_evenly($FDR, 4));
        } else {
          push @cv, sprintf('[MS, MS:1002350, PSM-level global FDR, %s]', round_evenly($FDR, 4));
        }
      }
    }

    my ($targetCount, $decoyCount, $proteinFDR) = ExportDat::Util::calculateProteinFDR($objSummary, $objSummaryDecoy, $objResFile);

    if (defined $proteinFDR and $proteinFDR > 0) {
      push @cv, sprintf('[MS, MS:1001214, protein-level global FDR, %s]', round_evenly($proteinFDR, 4));
    }

    if (@cv) {
      printMTD('false_discovery_rate', join('|', @cv));
    }
  }

  if ($showPRH) {
    printMTD('protein_search_engine_score[1]', '[MS, MS:1001171, Mascot:score,]');
  }

  if ($showPEH) {
    if ($objQuantMethod) {
      printMTD('peptide_search_engine_score[1]', '[MS, MS:1001172, Mascot:expectation value,]');
    }
    printMTD('psm_search_engine_score[1]', '[MS, MS:1001172, Mascot:expectation value,]');
  }

  my $UO_vocabulary_used = 0;

  if ($showPEH) {
    if ($any_retention_time) {
      ++$UO_vocabulary_used;
      printMTD('colunit-peptide', sprintf('retention_time=[UO, %s, %s,]', 'UO:0000010', 'second'));
    }

    if ($pepDelta) {
      ++$UO_vocabulary_used;
      printMTD('colunit-peptide', sprintf('opt_%s_%s=[UO, %s, %s,]', 'MS:1001976', 'delta_M', 'UO:0000221', 'dalton'));
    }
  }

  do {
    my $i = 1;

    printMTD('cv['.$i.']-label', 'MS');
    printMTD('cv['.$i.']-full_name', 'Proteomics Standards Initiative Mass Spectrometry Vocabularies');
    printMTD('cv['.$i.']-version', '4.1.30');
    printMTD('cv['.$i.']-url', 'http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo');

    ++$i;

    printMTD('cv['.$i.']-label', 'UNIMOD');
    printMTD('cv['.$i.']-full_name', 'UNIMOD');
    printMTD('cv['.$i.']-url', 'http://www.unimod.org/obo/unimod.obo');

    ++$i;

    if ($UO_vocabulary_used) {
      printMTD('cv['.$i.']-label', 'UO');
      printMTD('cv['.$i.']-full_name', 'UNIT-ONTOLOGY');
      printMTD('cv['.$i.']-url', 'http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/phenotype/unit.obo');

      ++$i;
    }

    if ($NEWT_vocabulary_used) {
      printMTD('cv['.$i.']-label', 'NEWT');
      printMTD('cv['.$i.']-full_name', 'NEWT Uniprot Taxonomy Database');
      printMTD('cv['.$i.']-url', 'http://www.ebi.ac.uk/ontology-lookup/browse.do?ontName=NEWT');

      ++$i;
    }
    # NB: There is no OBO definition for CHEMMOD.
  };

  # custom headers
  do {
  # Display any warnings from result file header
    my $warning_idx = 1;

    if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn0')){
      my $j = 0;
      while ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j)){
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++,
          $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j++)
        );
      }
    }

  # Display a warning if this is an orphan manual error tolerant search
  if ($objResFile->isErrorTolerant() and not $objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY) and not $objParams->getErrTolParentFilename()) {
      push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++,
        "Scores from standard search not available. Better matches may exist in other proteins."
      );
    }

  # Display any warnings from Parser
    for my $k (1 .. $objResFile->getNumberOfErrors()) {
      push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++,
        $objResFile->getErrorString($k)
      );
    }

  # Display any warnings from quantitation
    if (@$quantWarnings) {
      foreach (@$quantWarnings) {
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++,
          $_
        );
      }
    }
    if ($objMS2Quant) {
      my $objErr = $objMS2Quant->getErrorHandler();
      for my $i (1 .. $objErr->getNumberOfErrors()) {
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++,
          sprintf('%s (%s)', $objErr->getErrorString($i), $objErr->getErrorNumber($i)),
        );
      }
    }

    if (my $fileset = $thisScript->param($$urlParams{'fileset'})) {
      push @custom, '[,,Report URI, ' . $URI . 'cgi/master_results_2.pl?fileset=' . $fileset . ']';
    } else {
      push @custom, '[,,Report URI, ' . $URI . 'cgi/master_results_2.pl?file=' . $fileIn[0] . ']';
    }

    my $i = 1;

    for (@custom) {
      printMTD('custom['.($i++).']', $_);
    }
  };

  return ($indexIsQueryNumber);
}

sub _determine_component_quant_mods {
  my ($protocol, $method_unimods, $method_chemmods, $component,
      $objUmodSection) = @_;

  # Many possible modification types:
  # - iTRAQ has modfile names iTRAQ4plex or iTRAQ8plex.
  # - TMT has modfile names TMT6plex or TMT2plex.
  # - DiLeu has modfile name DiLeu4plex.
  # - 18O has modfile names Label:18O(1) or Label:18O(2).
  # - SILAC K+6 R+6 has local definition 13C6_NL_K and 13C6_NL_R.
  # - IPTL has local definitions IMID_2H(0), Succ_2H(4)_NL or IMID_2H(4)_NL
  #   or modfile name Succinyl.
  #
  # Since the Reporter protocol does not specify reagents, we have to make
  # the following guesses:
  # - If it's a modfile (Unimod) mod, use
  #     [UNIMOD, UNIMOD:$id, $title,]
  # - If it's a local definition (unlikely), use
  #     [CHEMMOD, CHEMMOD:$formula, $title,]
  #
  # Since it's theoretically possible (but probably practically meaningless)
  # to mix and match, we export the full union of all found reagents.
  #
  # The Multiplex protocol defines mods (= reagents), which can be unlabelled,
  # labelled with modfile name, or labelled with a local definition.
  # - The standard 18O has no mods in component 18O0, Label:18O(1) in 18O1 and
  #   Label:18O(2) in 18O2, and no other mods in any component.
  # - The standard SILAC K+6 R+6 method has no mods in light component
  #   and two local definitions on heavy, and no other mods.
  # - The standard IPTL method has two local definitions on n-label and
  #   one modfile and one local definition on c-label, and no other mods.
  #
  # Thus, we assume that all mods in Multiplex components are part of the
  # reagent.
  #
  # There is no standard chemical formula syntax for local definitions in
  # mzTab 1.0. We will use our own: IMID_2H(4)_NL, for example, becomes
  #   CHEMMOD:+(2H)4 -H4
  # where parentheses are used for disambiguating isotopes.

  my @mods = ();

  if ($protocol eq 'reporter') {
    @mods = _make_quant_mod_defs($method_unimods, $method_chemmods);
  } else { # multiplex
    my ($unimod_mods, $chemmods) = _collect_mod_types($component, $objUmodSection);
    @mods = _make_quant_mod_defs($unimod_mods, $chemmods);
  }

  return @mods;
}

sub _make_quant_mod_defs {
  my ($unimod_mods, $chemmods) = @_;

  my @mods = ();

  for (sort { $a <=> $b } keys %$unimod_mods) {
    my $mod = $$unimod_mods{$_};
    push @mods, {
      %$mod,
      cv => sprintf('[UNIMOD, UNIMOD:%d, %s,]', $_, $$mod{'title'}),
    };
  }

  for (sort keys %$chemmods) {
    my $mod = $$chemmods{$_};
    push @mods, {
      %$mod,
      cv => sprintf('[CHEMMOD, CHEMMOD:%s, %s,]', $_, $$mod{'title'}),
    };
  }

  return @mods;
}


sub _collect_mod_types {
  my ($obj, $ms_umod_configfile) = @_;

  my %unimod_mods = ();
  my %chemmods = ();

  for my $i (0 .. $obj->getNumberOfModificationGroups()-1) {
    my $modgroup = $obj->getModificationGroupByNumber($i);

    for my $j (0 .. $modgroup->getNumberOfModFiles()-1) {
      my $modfile_name = $modgroup->getModFile($j);
      my ($mod_name) = $modfile_name =~ /(.*)\s+\((?:.*)\)/;
      my $mod = $ms_umod_configfile->getModificationByName($mod_name);
      next if not $mod;
      my $id = $mod->getRecordID();
      next if $unimod_mods{$id};
      my ($position, $site) = _unimod_position_and_site($mod, $modfile_name);
      $unimod_mods{$id} = {
        title => $mod->getTitle(),
        position => $position,
        site => $site,
      };
    }

    for my $j (0 .. $modgroup->getNumberOfLocalDefinitions()-1) {
      my $def = $modgroup->getLocalDefinition($j);
      my $comp = $def->getDelta();
      my (@plus, @minus);
      for my $k (0 .. $comp->getNumberOfElements()-1) {
        my $elt = $comp->getElement($k);
        next if not $elt->haveSymbol();
        my $symbol = $elt->getSymbol();
        my $num = 1;
        if ($elt->haveNumber()) {
          $num = 0 + $elt->getNumber();
        }
        my $arr = $num > 0 ? \@plus : \@minus;
        if ($symbol =~ /\d/) {
          push @$arr, sprintf('(%s)%d', $symbol, abs($num));
        } else {
          push @$arr, sprintf('%s%d', $symbol, abs($num));
        }
      }
      next if (not @plus and not @minus);
      if ($def->getNumberOfSpecificities() == 0) {
        $chemmods{join(' ', join('', @plus), join('', @minus))} = {
          title => $def->getTitle(),
        };
      } else {
        # We can only export the first specificity...
        my $spec = $def->getSpecificity(0);
        $chemmods{join(' ', join('', @plus), join('', @minus))} = {
          title => $def->getTitle(),
          position => $spec->havePosition() ? $spec->getPosition() : undef,
          site => $spec->haveSite() ? $spec->getSite() : undef,
        };
      }
    }
  }

  return (\%unimod_mods, \%chemmods);
}

sub printMzTabPRHandPRT {
  my ($protein_rows, $isPMF, $objMS2Quant, $protQuant, $num_ms_runs,
      $num_study_variables, $protMass, $protEmPAI, $protpI, $reportType, $protein_mods,
      $database_map) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    accession
    description
    taxid
    species
    database
    database_version
    search_engine
    best_search_engine_score[1]
    ambiguity_members
    modifications
    protein_coverage
  );

  if ($protQuant and $objMS2Quant) {
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_stdev_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_std_error_study_variable[%d]', $i);
    }
  }

  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('search_engine_score[1]_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_psms_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_peptides_distinct_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_peptides_unique_ms_run[%d]', $i);
  }

  if ($protQuant and $objMS2Quant) {
    my $n = $objMS2Quant->getQuantitationMethod()->getNumberOfComponents();
    for my $i (1 .. $n*$num_ms_runs) {
      push @column_data_names_pre, sprintf('protein_abundance_assay[%d]', $i);
    }
  }

  my @MS_available = ();

  push @MS_available, ['MS:1001301', 'protein rank'];

  if (not $isPMF) {
      push @MS_available, ['MS:1001592', 'family member protein'];
  }

  if ($protEmPAI) {
    push @MS_available, ['MS:1001905', 'emPAI value'];
  }

  if ($protpI) {
    push @MS_available, ['MS:1000862', 'isoelectric point'];
  }

  if ($reportType eq 'concise') {
    push @MS_available, ['MS:1001172', 'Mascot:expectation value'];
  }

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  for my $i (1 .. $num_study_variables) {
    for (qw(_sample_size _is_normal _is_significant)) {
      push @column_names, sprintf('opt_study_variable[%d]%s', $i, $_);
      push @opt_data_names, sprintf('study_variable[%d]%s', $i, $_);
    }
    # BUG#13302: Add two optional columns for the protein ratio numerator and denominator
    # only when ratio type is 'weighted'. Calculating it for ratio types 'mean' and 'average'
    # here would lead to numerical differences with the ratios calculated within Parser.
    my $ratio_type = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());
    if ($ratio_type eq 'weighted') {
      for (qw(_numerator _denominator)) {
	push @column_names, sprintf('opt_study_variable[%d]%s', $i, $_);
	push @opt_data_names, sprintf('study_variable[%d]%s', $i, $_);
      }
    }
  }

  for (@MS_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }
  if ($protMass) {
      push @column_names, 'opt_global_mass';
      push @opt_data_names, 'mass';
  }
  if ($objParams->getNumberOfDatabases() != 1) {
    push @column_names, 'opt_global_ambiguity_member_databases';
    push @opt_data_names, 'ambiguity_member_databases';
  }

  # Optional columns all come before reliability.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PRH', @column_names);

  for my $row (@$protein_rows) {
    my @data = ();

    my $db = $$row{'database'};
    my $db_entry = $$database_map[$db];
    $$row{'database'} = $$db_entry{'full_name'};
    $$row{'database_version'} = $$db_entry{'version'};

    $$row{'search_engine'} = '[MS, MS:1001207, Mascot, ]';

    my $mods = $$protein_mods{$db}{$$row{'accession'}};

    if ($mods) {
      my %m = ();

      for my $idx (keys %$mods) {
        for (@{ $$mods{$idx} }) {
          next if not $$_{'mod_id'};
          my $id = sprintf('%s%s', $$_{'mod_id'},
            defined($$_{'neutral_loss'}) ? '|'.$$_{'neutral_loss'} : ''
          );
          $m{$idx}{$id}++;
        }
      }

      my @m = ();

      # Ambiguity in modification position is not allowed, but ambiguity in
      # modification at a specified position is. Since there is no ambiguity
      # in position at peptide level, there is no ambiguity at protein level.

      for my $idx (sort { $a <=> $b } keys %m) {
        push @m, map { sprintf('%d-%s', $idx, $_) } sort keys %{ $m{$idx} };
      }

      if (@m) {
        $$row{'modifications'} = join(',', @m);
      } else {
        $$row{'modifications'} = '0';
      }
    }

    my (@amb_acc, @amb_db);

    for (sort { $$a{'accession'} cmp $$b{'accession'} } @{ $$row{'ambiguity_members'} }) {
      push @amb_acc, $$_{'accession'};
      my $db_entry = $$database_map[$$_{'db'}];
      push @amb_db, $$db_entry{'abbr_name'};
    }

    $$row{'ambiguity_members'} = join(',', @amb_acc);
    $$row{'opt'}{'ambiguity_member_databases'} = join(',', @amb_db);

    for (@column_data_names_pre) {
      push @data, $$row{$_};
    }

    for (@opt_data_names) {
      push @data, $$row{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PRT', @data);
  }
}

sub printMzTabPEHandPEP {
  my ($peptide_rows, $objMS2Quant, $num_ms_runs, $num_study_variables,
      $is_decoy, $is_error_tolerant, $percolate, $libraryMode, $database_map,
      $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch,
      $pepScanTitle, $pepQuant) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    sequence
    accession
    unique
    database
    database_version
    search_engine
    best_search_engine_score[1]
    modifications
    retention_time
    retention_time_window
    charge
    mass_to_charge
  );

  if ($pepQuant and $objMS2Quant) {
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_stdev_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_std_error_study_variable[%d]', $i);
    }
  }

  # Bug 12186: In a spectral library search, whether integrated or SL-only,
  # there is still only one search_engine_score, and that's the expect value.
  # Raw Mascot and MSPepSearch scores are output in their own columns.
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('search_engine_score[1]_ms_run[%d]', $i);
  }

  if ($pepQuant and $objMS2Quant) {
    my $n = $objMS2Quant->getQuantitationMethod()->getNumberOfComponents();
    for my $i (1 .. $n*$num_ms_runs) {
      push @column_data_names_pre, sprintf('peptide_abundance_assay[%d]', $i);
    }
  }

  push @column_data_names_pre, qw(
    spectra_ref
  );

  for my $i (1 .. @fileIn)  {
    push @column_data_names_pre, sprintf('opt_ms_run[%d]_RAWFILE', $i);
  }

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  my @opt_available = ();

  push @opt_available, ['MS:1000797', 'peak list scans'];
  push @opt_available, ['MS:1000798', 'peak list raw scans'];

  push @opt_available, ['MS:1001528', 'Mascot query number'];

  if ($percolate == 1) {
    push @opt_available, (
      ['MS:1001491', 'percolator:Q value'],
      ['MS:1001493', 'percolator:PEP'],
    );
  }

  if ($is_decoy) {
    push @opt_available, ['MS:1002217', 'decoy peptide'];
  }
  if ($is_error_tolerant) {
    push @opt_available, ['MS:1002341', 'second-pass peptide identification'];
  }

  if ($pepDelta) {
    push @opt_available, ['MS:1001976', 'delta M'];
  }

  if ($pepScore) {
    if ($libraryMode eq 'onlylibrary') {
      push @opt_available, ['MS:1002754', 'MSPepSearch:score'];
    } elsif ($libraryMode eq 'integrated') {
      push @opt_available, ['MS:1001171', 'Mascot:score'];
      push @opt_available, ['MS:1002754', 'MSPepSearch:score'];
    } else {
      push @opt_available, ['MS:1001171', 'Mascot:score'];
    }
  }

  # Bug 12186: In a spectral library search, the thresholds are encoded as
  # Mascot thresholds, since we calculate them ourselves.
  if ($pepIdThr) {
    push @opt_available, ['MS:1001371', 'Mascot:identity threshold'];
  }

  if ($pepHomThr) {
    push @opt_available, ['MS:1001370', 'Mascot:homology threshold'];
  }

  if ($pepFrame) {
    # NB: This term was obsoleted in PSI-MS CV for a while but became
    # current again in 3.54.0 due to our request.
    push @opt_available, ['MS:1001024', 'translation frame'];
  }

  if ($pepNumMatch) {
    push @opt_available, ['MS:1001121', 'number of matched peaks'];
  }

  if ($pepScanTitle) {
    push @opt_available, ['MS:1000796', 'spectrum title'];
  }

  for (@opt_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }

  if ($pepMiss) {
    push @column_names, 'opt_global_missed_cleavages';
    push @opt_data_names, 'missed_cleavages';
  }

  # Reliability comes after all optional columns.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PEH', @column_names);

  for my $row (@$peptide_rows) {
    my @data = ();
    my %row_data = %$row;

    my $db = $row_data{'database'};
    my $db_entry = $$database_map[$db];
    $row_data{'database'} = $$db_entry{'full_name'};
    $row_data{'database_version'} = $$db_entry{'version'};
    if (defined $row_data{'retention_time'}) {
      $row_data{'retention_time_window'} = join('|', $row_data{'retention_time'}, $row_data{'retention_time'});
    }
    $row_data{'mass_to_charge'} = $row_data{'exp_mass_to_charge'};

    # For peptides, the modifications field must contain only variable
    # modifications.
    $row_data{'modifications'} = _format_mztab_peptide_modifications(@{ $row_data{'.varmods'} });

    for (@column_data_names_pre) {
      push @data, $row_data{$_};
    }

    for (@opt_data_names) {
      push @data, $row_data{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PEP', @data);
  }
}

sub _format_mztab_peptide_modifications {
  my @modifications = @_;

  my @ids = ();

  for (sort { $$a{'idx'} <=> $$b{'idx'} } @modifications) {
    next if not $$_{'mod_id'};
    my $id = sprintf('%d-%s', $$_{'idx'}, $$_{'mod_id'});
    if (defined $$_{'neutral_loss'}) {
      $id .= '|' . $$_{'neutral_loss'};
    }
    push @ids, $id;
  }

  if (not @ids) {
    return '0';
  }

  return join(',', @ids);
}

sub printMzTabPSHandPSM {
  my ($peptide_rows, $is_decoy, $is_error_tolerant, $percolate, $libraryMode,
      $database_map, $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr, $pepFrame,
      $pepNumMatch, $pepScanTitle) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    sequence
    PSM_ID
    accession
    unique
    database
    database_version
    search_engine
    search_engine_score[1]
    modifications
    spectra_ref
    retention_time
    charge
    exp_mass_to_charge
    calc_mass_to_charge
    pre
    post
    start
    end
  );

  for my $i (1 .. @fileIn)  {
    push @column_data_names_pre, sprintf('opt_ms_run[%d]_RAWFILE', $i);
  }

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  my @opt_available = ();

  push @opt_available, ['MS:1000797', 'peak list scans'];
  push @opt_available, ['MS:1000798', 'peak list raw scans'];

  push @opt_available, ['MS:1001528', 'Mascot query number'];

  if ($percolate == 1) {
    push @opt_available, (
      ['MS:1001491', 'percolator:Q value'],
      ['MS:1001493', 'percolator:PEP'],
    );
  }

  if ($is_decoy) {
    push @opt_available, ['MS:1002217', 'decoy peptide'];
  }
  if ($is_error_tolerant) {
    push @opt_available, ['MS:1002341', 'second-pass peptide identification'];
  }

  if ($pepDelta) {
    push @opt_available, ['MS:1001976', 'delta M'];
  }

  if ($pepScore) {
    if ($libraryMode eq 'onlylibrary') {
      push @opt_available, ['MS:1002754', 'MSPepSearch:score'];
    } elsif ($libraryMode eq 'integrated') {
      push @opt_available, ['MS:1001171', 'Mascot:score'];
      push @opt_available, ['MS:1002754', 'MSPepSearch:score'];
    } else {
      push @opt_available, ['MS:1001171', 'Mascot:score'];
    }
  }

  if ($pepIdThr) {
    push @opt_available, ['MS:1001371', 'Mascot:identity threshold'];
  }

  if ($pepHomThr) {
    push @opt_available, ['MS:1001370', 'Mascot:homology threshold'];
  }

  if ($pepFrame) {
    # NB: This term was obsoleted in PSI-MS CV for a while but became
    # current again in 3.54.0 due to our request.
    push @opt_available, ['MS:1001024', 'translation frame'];
  }

  if ($pepNumMatch) {
    push @opt_available, ['MS:1001121', 'number of matched peaks'];
  }

  if ($pepScanTitle) {
    push @opt_available, ['MS:1000796', 'spectrum title'];
  }

  for (@opt_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }

  if ($pepMiss) {
    push @column_names, 'opt_global_missed_cleavages';
    push @opt_data_names, 'missed_cleavages';
  }

  # Reliability comes after all optional columns.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PSH', @column_names);

  for my $row (@$peptide_rows) {
    my @data = ();
    my %row_data = %$row;

    my $db = $row_data{'database'};
    my $db_entry = $$database_map[$db];
    $row_data{'database'} = $$db_entry{'full_name'};
    $row_data{'database_version'} = $$db_entry{'version'};
    $row_data{'search_engine_score[1]'} = $row_data{'best_search_engine_score[1]'};

    # For PSMs, the modifications field must contain all fixed and variable
    # modifications.
    $row_data{'modifications'} = _format_mztab_peptide_modifications(@{ $row_data{'.fixedmods'} }, @{ $row_data{'.varmods'} });

    for (@column_data_names_pre) {
      push @data, $row_data{$_};
    }

    for (@opt_data_names) {
      push @data, $row_data{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PSM', @data);
  }
}

sub printMTDsearchParameters {
  my ($software_id, $objParams) = @_;

  # Bug 12552: software[]-settings[] lines must have a second running index.
  my $k = 1;

  for my $i (1 .. $objParams->getNumberOfDatabases) {
    printMTD($software_id.'['.$k.']', 'DB', '=', $objParams->getDB($i));
    ++$k;
  }

  # Parameters not in this list are omitted entirely.
  my @param_names = (
    ['ACCESSION', ''],
    ['CHARGE', ''],
    ['CLE', ''],
    ['DATAURL', ''],
    ['DECOY', -1],
    ['ERRORTOLERANT', ''],
    ['FILENAME', ''],
    ['FORMAT', ''],
    ['FORMVER', ''],
    ['ICAT', '', 0],
    ['INSTRUMENT', ''],
    ['IT_MODS', ''],
    ['MASS', ''],
    ['MODS', ''],
    ['MULTI_SITE_MODS', '', 0],
    ['OVERVIEW', '', 0],
    ['PEP_ISOTOPE_ERROR', -1],
    ['PFA', -1],
    ['PRECURSOR', -1],
    ['QUANTITATION', '', 'none'],
    ['QUE', ''],
    ['REPORT', 0],
    ['REPTYPE', ''],
    ['SEARCH', ''],
    ['SEG', -1],
    ['SUBCLUSTER', -1],
    ['TAXONOMY', ''],
    ['TARGET_FDR_PERCENT', 0],
    ['TOL', 0],
    ['TOLU', ''],
  );

  if (not $objResFile->isPMF) {
    push @param_names, (
        ['ITOL', ''],
        ['ITOLU', ''],
    );
  }

  PARAM: for (sort { $$a[0] cmp $$b[0] } @param_names) {
    my ($param_name, @skiplist) = @$_;
    my $f = 'get'.$param_name;
    my $val = $objParams->$f();
    next if not defined $val;
    for (@skiplist) {
      next PARAM if $val eq $_;
    }
    printMTD($software_id.'['.$k.']', $param_name, '=', $val);
    ++$k;
  }

  for (0 .. 12) {
    my $val = $objParams->getUSERField($i);
    next if $val eq '';
    printMTD($software_id.'['.$k.']', sprintf('USER%02d', $_), '=', $val);
    ++$k;
  }

  if (my $str = $objParams->getAllUSERParams()) {
    for (split /[\r\n]+/, $str) {
      printMTD($software_id.'['.$k.']', $_);
      ++$k;
    }
  }

}

sub printMTDformatParameters {
  my ($settings_id, $thisScript, $urlParams) = @_;

  # Bug 12552: software[]-settings[] lines must have a second running index.
  my $k = 1;

  my $args = ExportDat::Util::cmdline_arguments($thisScript, $urlParams);
  delete $$args{'sessionid'};

  for my $arg (sort keys %$args) {
    my $value = $$args{$arg};

    if ($value =~ /\s/) {
      printMTD($settings_id.'['.$k.']', sprintf("%s='%s'", $arg, $value));
      ++$k;
    } else {
      printMTD($settings_id.'['.$k.']', sprintf("%s=%s", $arg, $value));
      ++$k;
    }
  }
}

sub _mk_protein_id { _mk_protein_id_str($_[0]->getDB(), $_[0]->getAccession()) }
sub _mk_protein_id_str { sprintf('%d::%s', $_[0], $_[1]) }

sub collectMzTabProteins {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $ShowSameSets, $ShowSubSets, $reportType,
    $protDesc, $protMass, $protTaxStr, $protCover, $protEmPAI, $protpI, $protQuant,
    $query_mod_and_dupe_cache, $filter_duplicates, $sameset_ordering
  ) = @_;

  my $isPMF = $objResFile->isPMF;

  my @protein_rows = ();
  # Peptide map is a mapping from q,p to list of protein data (accession,
  # start, end). We use it later to decide which peptide matches to export
  # and under which accessions to report them.
  my %peptide_map = ();

  for my $i (1 .. $objSummary->getNumberOfHits()) {
    my $objProtein = $objSummary->getHit($i);
    next if $isPMF and $objProtein->isPMFMixture();

    my @sameset = common_subs::compileSameset(
      $objSummary, $objProtein, $ShowSameSets, $$sameset_ordering{$i}
    );

    my $data = collectMzTabProteinData(
      $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
      $objMS2Quant, $sameset[0], $reportType, $protDesc, $protMass,
      $protTaxStr, $protCover, $protEmPAI, $protpI, $query_mod_and_dupe_cache,
      $filter_duplicates, \%peptide_map,
    );

    $$data{'opt'}{'MS:1001301'} = $i; # rank

    if (not $isPMF) {
        $$data{'opt'}{'MS:1001592'} = 1; # member number
    }

    # Other proteins that match the same set of peptides
    $$data{'ambiguity_members'} = [ map {
      { accession => $_->getAccession(), db => $_->getDB() }
    } @sameset[1 .. $#sameset] ];

    push @protein_rows, $data;

    my %family_superset_map = ();
    $family_superset_map{ _mk_protein_id($sameset[0]) } = $data;

    # We can use the same overall loop over family members and samesets
    # in both kinds of grouping (family and conventional). In the latter
    # case, there are simply no family proteins.

    # Other family members
    my $k = 1;
    while (my $objProtein = $objSummary->getNextFamilyProtein($i, $k++)) {
      my @sameset = common_subs::compileSameset(
        $objSummary, $objProtein, $ShowSameSets, $$sameset_ordering{$i},
      );

      my $data = collectMzTabProteinData(
        $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs,
        $objSummary, $objMS2Quant, $sameset[0], $reportType,
        $protDesc, $protMass, $protTaxStr, $protCover, $protEmPAI, $protpI,
        $query_mod_and_dupe_cache, $filter_duplicates, \%peptide_map,
      );

      $$data{'opt'}{'MS:1001301'} = $i; # rank

      if (not $isPMF) {
          $$data{'opt'}{'MS:1001592'} = $k; # member number
      }

      $family_superset_map{ _mk_protein_id($sameset[0]) } = $data;

    # And their same-sets
      $$data{'ambiguity_members'} = [ map {
        { accession => $_->getAccession(), db => $_->getDB() }
      } @sameset[1 .. $#sameset] ];

      push @protein_rows, $data;
    }

    # sub-sets and intersections for the entire family

    # mzTab 1.0 forbids exporting subset and intersection proteins, so we
    # never do so here. The loop could be of use in future.
    if (0 and $ShowSubSets > 0) {
      my $k = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($i, $k++, 1)) {
        # We need the full sameset to filter out proteins from
        # getSimilarProteins().
        my @sameset = common_subs::compileSameset(
          $objSummary, $objProtein, 1, $$sameset_ordering{$i},
        );
        my %is_sameset = map { key_acc($_) => 1 } @sameset;

        my @supersets = ExportDat::Util::get_supersets($sameset[0], \%is_sameset);
        my @sameset_data = map {
          { accession => $_->getAccession(), db => $_->getDB() }
        } @sameset;

        for (@supersets) {
          my $superset = $family_superset_map{ _mk_protein_id_str(@$_{qw(db acc)}) };
          next if not $superset;

          # The intersection protein appears once under each parent (superset)
          # protein.
          push @{ $$superset{'ambiguity_members'} }, @sameset_data;
        }
      }
    }
  }

  return (\@protein_rows, \%peptide_map);
}

sub _determine_protein_hit_reliability {
  my ($objProtein) = @_;

  my $n = $objProtein->getNumDistinctPeptides(1);

  if ($n >= 2) {
    return 1; # high
  }

  if ($n == 1) {
      return 2; # medium
  }

  return 3; # low
}

sub _determine_peptide_match_reliability {
  my ($objSummary, $objPeptide, $OmitErrTolThresholds) = @_;

  if ($OmitErrTolThresholds and $objPeptide->getIsFromErrorTolerant()) {
      return 3; # no negotiation
  }

  my $score = $objPeptide->getIonsScore();
  my $thr = $objSummary->getPeptideThreshold($objPeptide->getQuery(), 1 / $objSummary->getProbabilityThreshold, $objPeptide->getRank);

  if ($score > $thr) {
    return 1; # high
  }

  return 3; # low
}

sub _calculate_protein_quant_ratio {
  my ($ratio, $ratio_def, $objMS2Quant, $component_name_to_component_index) = @_;

  my $average = sub {
    my @ratios = @{$_[0]};

    my $num = $ratios[0][0];
    my $den = $ratios[0][1];
    my $n = scalar(@ratios);
    # a/b + c/d = a*d+c*b / b*c
    for my $i (1 .. $n-1) {
      $num = $num * $ratios[$i][1] + $ratios[$i][0] * $den;
      $den = $den * $ratios[$i][1];
    }
    $den *= $n;
    return ($num, $den);
  };

  my $median = sub {
    my @ratios = @{$_[0]};

    my $n = $#ratios; #array size - 1
    my @sorted = sort { $$a[0]/$$a[1] <=> $$b[0]/$$b[1] } @ratios;
    my $mid1 = int($n/2);
    my $mid2 = int($n/2 +0.5);
    my $num = $sorted[$mid1][0] * $sorted[$mid2][1] + $sorted[$mid2][0] * $sorted[$mid1][1];
    my $den = $sorted[$mid1][1] * $sorted[$mid2][1];
    $den *= 2;
    return ($num, $den);
  };

  my $weighted = sub {
    my @ratios = @{$_[0]};

    my $num = 0;
    my $den = 0;
    for my $i (0 .. $#ratios) {
      $num += $ratios[$i][0];
      $den += $ratios[$i][1];
    }
    return ($num, $den);
  };

  my $ratio_type = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());

  # quantKeys is the set of peptide keys used to calculate the protein ratio:
  # It doesn't include outliers.
  my $quantKeys = $ratio->getActiveKeys();

  my @ratios;
  for my $j (0 .. $quantKeys->size()-1) {
    my $intensities = $objMS2Quant->getComponentIntensities($quantKeys->get($j));

    my $totalNumer = 0;
    for my $i (0 .. $ratio_def->getNumberOfNumerators()-1) {
      my $numer = $ratio_def->getNumeratorByNumber($i);
      next unless $numer->haveName();
      my $idx = $component_name_to_component_index->{$numer->getName()};
      next unless defined $idx;
      my $coeff = $numer->getCoefficient() || 1.0; #defaults to 1.0
      $totalNumer += $coeff * ($$intensities[$idx] // 0);
    }
    my $totalDenom = 0;
    for my $i (0 .. $ratio_def->getNumberOfDenominators()-1) {
      my $denom = $ratio_def->getDenominatorByNumber($i);
      next unless $denom->haveName();
      my $idx = $component_name_to_component_index->{$denom->getName()};
      next unless defined $idx;
      my $coeff = $denom->getCoefficient() || 1.0;
      $totalDenom += $coeff * ($$intensities[$idx] // 0);
    }
    push @ratios, [$totalNumer, $totalDenom];
  }
  my $num = 0;
  my $den = 0;
  if (@ratios) {
    if ($ratio_type eq 'average') { ($num, $den) = $average->(\@ratios); }
    elsif ($ratio_type eq 'median') { ($num, $den) = $median->(\@ratios); }
    elsif ($ratio_type eq 'weighted') { ($num, $den) = $weighted->(\@ratios); }
  }
  return ($num, $den);
}

# Mandatory protein columns:
#   accession
#   description (*)
#   taxid (*)
#   species (*)
#   database
#   [database_version]
#   [search_engine]
#   best_search_engine_score[1]
#   search_engine_score[1]_ms_run[i]
#   reliability
#   num_psms_ms_run[k]
#   num_peptides_distinct_ms_run[i]
#   num_peptides_unique_ms_run[i]
#   [ambiguity_members]
#   modifications
#   (uri)
#   (go_terms)
#   protein_coverage (*)
#
# Terms in parentheses are always null. 'database_version' and
# 'search_engine' are set later, as is 'ambiguity_members'.
# Note that best_search_engine_score[] and search_engine_score[]-ms_run[i]
# always have index 1, because we only export Mascot data.
#
# For 'modifications', we need to collect the start and end positions of each
# peptide match and later, once we've looped over peptide matches, construct
# the protein modification list. Also, 'database' is an index, which is mapped
# to the database ID later.
#
# Terms marked with an asterisk (*) may be null, unless the user has opted
# to fetch missing data from ms-getseq.exe and ms-gettaxonomy.exe.
#
# If quantitation is in use, the following optional columns will be
# exported:
#   protein_abundance_assay[k]
#   protein_abundance_study_variable[j]
#   protein_abundance_stdev_study_variable[j]
#   protein_abundance_std_error_study_variable[j]
#
# k is an index over component names in order as given by $objMS2Quant,
# while j is an index over ratio names.
sub collectMzTabProteinData {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $objProtein, $reportType, $protDesc, $protMass, $protTaxStr,
    $protCover, $protEmPAI, $protpI, $query_mod_and_dupe_cache,
    $filter_duplicates, $peptide_map,
  ) = @_;

  my $objResFile = $$ms_run_resfiles[0]{'resfile'};
  my $objParams = $$ms_run_resfiles[0]{'params'};

  my %data = ();

  my $preferredAcc = $data{'accession'} = $protData->preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  my $dbIdx = $data{'database'} = $objProtein->getDB();

  if ($protDesc) {
    $data{'description'} = $protData->mustGetProteinDescription($preferredAcc, $dbIdx);
  }

  if ($protTaxStr) {
    ($data{'taxid'}, $data{'species'}) = $protData->getProteinTaxonomy($objResFile, $objSummary, $objParams, $preferredAcc, $dbIdx);
  }

  # "Best" means the combined file score if there are multiple results files.
  # NB: we should really have a protein-specific CV term here; MS:1001171 is
  # the score for PSMs in PSI-MS CV 3.59 (2014-01-22).
  $data{'best_search_engine_score[1]'} = round_evenly($objProtein->getScore(), 0);

  # It is rather meaningless to report Mascot protein scores for each ms_run
  # individually.
  for my $i (1 .. $#$ms_run_resfiles) {
    my $summary = $$ms_run_resfiles[$i]{'summary'};
    my $protein = $summary->getProtein($objProtein->getAccession(), $objProtein->getDB());
    if ($protein) {
      $data{'search_engine_score[1]_ms_run['.$i.']'} = round_evenly($protein->getScore(), 0);

      # "count of the total significant PSMs"
      $data{'num_psms_ms_run['.$i.']'} = $protein->getNumDisplayPeptides(1);

      # "count of the number of different peptide sequences ... above the
      # significance threshold"
      $data{'num_peptides_distinct_ms_run['.$i.']'} = $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE);

      # Uniqueness means:
      # "if ambiguity members have been reported, the count MUST be derived from
      # the number of peptides that can be uniquely mapped to the group of
      # accessions"
      # In terms of Mascot, this means the number of significant PSMs marked
      # unique to the protein hit, as this is the set of matches that defines
      # the sameset proteins (the ambiguity members).
      $data{'num_peptides_unique_ms_run['.$i.']'} = $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE | $msparser::ms_protein::DPF_UNIQUE);
    } else {
      $data{'search_engine_score[1]_ms_run['.$i.']'} = 'null';
      $data{'num_psms_ms_run['.$i.']'} = 'null';
      $data{'num_peptides_distinct_ms_run['.$i.']'} = 'null';
      $data{'num_peptides_unique_ms_run['.$i.']'} = 'null';
    }
  }

  $data{'reliability'} = _determine_protein_hit_reliability(
    $objProtein,
  );

  my @selected_qp = _select_nonduplicate_protein_qp(
    $query_mod_and_dupe_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  );

  my @quant_intensity = ();

  for (@selected_qp) {
    my ($pepnum, $query, $rank) = @$_;
    my $startpos  = $objProtein->getPeptideStart($pepnum);
    my $endpos    = $objProtein->getPeptideEnd($pepnum);
    my $pre       = $objProtein->getPeptideResidueBefore($pepnum);
    my $post      = $objProtein->getPeptideResidueAfter($pepnum);

    push @{ $$peptide_map{$query}{$rank} }, {
      accession => $preferredAcc,
      database => $objProtein->getDB(),
      start => $startpos,
      end => $endpos,
      pre => $pre,
      post => $post,
      pepnum => $pepnum,
    };

    if ($objMS2Quant) {
      my $arr = $objMS2Quant->getComponentIntensities(msparser::ms_peptide_quant_key->new($query, $rank));
      if ($arr and @$arr) {
        # We know that by definition, the component intensities of a PSM
        # can only come from one results file.
        my (undef, undef, $src_resfile_idx) = $objResFile->getSrcQueryAndFileIdForMultiFile($query);
        # We assume component intensities are in the same order as in the
        # quantitation method and that the assays are defined in the same
        # order in mzTabExperimentStructure().
        for my $i (1 .. @$arr) {
          my $assay_idx = $$comp_idx_to_assay[$src_resfile_idx][$i];
          $quant_intensity[$assay_idx] += $$arr[$i-1];
        }
      }
    }
  }

  if ($reportType eq 'concise') {
    my $expect = $objSummary->getProteinExpectationValue($objProtein->getScore());
    $data{'opt'}{'MS:1001172'} = round_evenly($expect, 3, 'g');
  }

  my ($length, $mass);

  if ($protCover) {
      # User gives explicit request to fetch protein length.
      ($length, $mass) = $protData->mustGetProteinLen(
          $objProtein,
          1,
          $protMass
      );
  } elsif ($protEmPAI) {
      # Form has no indication that emPAI requires length, so estimate
      # it if necessary.
      ($length, $mass) = $protData->mustGetProteinLen(
          $objProtein,
          0,
          $protMass
      );
  }

  if ($protCover) {
      if ($length and $length > 0) {
	  $data{'protein_coverage'} = round_evenly($objProtein->getCoverage() / $length, 3);
      }
  }

  if ($protEmPAI) {
      my $value = $objSummary->getProteinEmPAI($objProtein->getAccession(), $objProtein->getDB(), $length);
      if ($value >= 0) {
	  $data{'opt'}{'MS:1001905'} = round_evenly($value, 2);
      }
  }

  if ($protMass) {
      if (not defined $mass) {
	  $mass = $protData->mustGetProteinMass($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
      }

      if ($mass) {
	  $data{'opt'}{'mass'} = round_evenly($mass, 0);
      }
  }

  if ($protpI) {
      $data{'opt'}{'MS:1000862'} = $protData->getProteinpI($objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  }

  if ($objMS2Quant) {
    for my $assay_idx (1 .. @quant_intensity) {
      # (@quant_intensity is 1-based, so no index error here.)
      if (defined $quant_intensity[$assay_idx]) {
	# BUG#13302: Set 'protein_abundance_assay' to null
	# We use opt_study_variable[\d]_[numerator|denominator] instead
        #$data{'protein_abundance_assay['.$assay_idx.']'} = round_evenly($quant_intensity[$assay_idx], 6, 'g');
	$data{'protein_abundance_assay['.$assay_idx.']'} = 'null';
      } else {
        $data{'protein_abundance_assay['.$assay_idx.']'} = 'null';
      }
    }

    my %component_name_to_component_index;
    do {
      my $qm = $objMS2Quant->getQuantitationMethod();
      for my $i (0 .. $qm->getNumberOfComponents()-1) {
	my $comp = $qm->getComponentByNumber($i);
	if ($comp->haveName()) {
	  $component_name_to_component_index{$comp->getName()} = $i;
	}
      }
    };

    my $ratio_type = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());
    for my $sv (@$study_variable_defs) {
      my $ratio = $objMS2Quant->getProteinRatio($objProtein->getAccession(), $objProtein->getDB(), $$sv{'ratio_name'});
      my $idx = $$sv{'idx'};

      if ($ratio->isMissing()) {
        $data{'protein_abundance_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_stdev_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_std_error_study_variable['.$idx.']'} = 'null';

        $data{'opt'}{'study_variable['.$idx.']_sample_size'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_is_normal'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_is_significant'} = 'null';

	if ($ratio_type eq 'weighted') {
	  $data{'opt'}{'study_variable['.$idx.']_numerator'} = 'null';
	  $data{'opt'}{'study_variable['.$idx.']_denominator'} = 'null';
	}
      } else {
        $data{'protein_abundance_study_variable['.$idx.']'} = round_evenly($ratio->getValue(), 3);
        $data{'protein_abundance_stdev_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_std_error_study_variable['.$idx.']'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_sample_size'} = $ratio->getSampleSize();

        my $stdev = $ratio->getStandardDeviation();
        my $stderr = $ratio->getStandardError();

        if ($stdev > 0) {
          $data{'protein_abundance_stdev_study_variable['.$idx.']'} = round_evenly($stdev, 3);
        } else {
          # Ignore reason why unavailable.
        }

        if ($stderr > 0) {
          $data{'protein_abundance_std_error_study_variable['.$idx.']'} = round_evenly($stderr, 3);
        } else {
          # Ignore reason why unavailable.
        }

        if ($ratio->getNormalityPvalue() > 0) {
          $data{'opt'}{'study_variable['.$idx.']_is_normal'} = $ratio->isLogNormal() ? 1 : 0;
          if ($ratio->getHypothesisPvalue() > 0) {
            $data{'opt'}{'study_variable['.$idx.']_is_significant'} = $ratio->isLogNormal() && $ratio->isSignificant() ? 1 : 0;
          }
        }

	if ($ratio_type eq 'weighted') {
	  my $ratio_def = $objMS2Quant->getCombinedReportRatio($$sv{'ratio_name'});
	  my ($num, $den) = _calculate_protein_quant_ratio($ratio, $ratio_def, $objMS2Quant, \%component_name_to_component_index);
	  $data{'opt'}{'study_variable['.$idx.']_numerator'} = round_evenly($num, 6, 'g');
	  $data{'opt'}{'study_variable['.$idx.']_denominator'} = round_evenly($den, 6, 'g');
	}

      } # !else ($ratio->isMissing())
    } # ! for my $sv
  }

  return \%data;
}


# Mandatory Peptide columns:
#   sequence
#   accession
#   unique
#   database
#   [database_version]
#   [search_engine]
#   best_search_engine_score[1]
#   modifications
#   retention_time
#   retention_time_window
#   charge
#   mass_to_charge
#   search_engine_score[1]_ms_run[k]
#   spectra_ref
#
# Mandatory PSM columns (not already listed above):
#   PSM_ID
#   search_engine_score[1]
#   exp_mass_to_charge
#   calc_mass_to_charge
#   pre
#   post
#   start
#   end
#
# Optional in both:
#   reliability
#   (uri)
#
# Columns in parentheses ('uri') are null. 'database_version'
# and 'search_engine' are filled in later. The mzTab 1.0 specification requires
# that each peptide match in each protein hit be reported, even when the
# peptide is shared between hits, so there may be several nearly identical
# output rows which only differ by the 'accession' column.
#
# To simplify matters, we assume Peptide = PSM. In Quantification mode,
# we output both the Peptides and PSM table; in Identification mode, only
# the PSM table. Both tables use the same source data collected here.
#
# If quantitation is in use, the following optional columns will be
# exported in the Peptides table:
#   peptide_abundance_study_variable[k]
#   (peptide_abundance_stdev_study_variable[k])
#   (peptide_abundance_std_error_study_variable[k])
#   peptide_abundance_assay[j]
#
# k is an index over ratio names in order as given by $objMS2Quant, while
# j is an index over assays (quantitation components).
# The standard deviation and standard error columns are always null, since we
# don't have those for peptide ratios.
sub collectMzTabPeptides {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $msresFlags2, $objMS2Quant, $objUmodSection,
    $objFullUmodFile, $sessionID, $indexIsQueryNumber, $is_decoy,
    $is_error_tolerant, $OmitErrTolThresholds, $pepDelta, $pepMiss, $pepScore, $pepHomThr, $pepIdThr,
    $pepFrame, $pepNumMatch, $pepScanTitle, $pepQuant, $massDP, $vmMass,
    $vmString, $percolate, $fixedMods, $query_mod_and_dupe_cache, $peptide_map
  ) = @_;

  my $isPMF = $objResFile->isPMF;

  my ($modsByDelta, $full_unimod_file, $query_is_sig);

  if (not $isPMF) {
    # Site analysis initialisation.
    $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
      $objUmodSection, $objParams,
    );
    # $objUmodSection could be the reduced file embedded in the results
    # file.
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
    $full_unimod_file = new msparser::ms_umod_configfile(
      "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
      $cs
    );

    $query_is_sig = $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
      ? sub { Util::Peptides::above_homology($objSummary, $objSummary->getPeptide($_[0], 1)) }
      : sub { Util::Peptides::above_identity($objSummary, $objSummary->getPeptide($_[0], 1)) }
      ;
  }

  my $export_quant = 0;

  if ($pepQuant and $objMS2Quant) {
    $export_quant = 1;
  }

  # Now collect all peptide matches
  my @peptide_rows = ();
  my %protein_mods = ();

  my %mod_name_lookup = (); # cache for Unimod objects
  my %mod_tally = (); # set of all Unimod and Chemmod identifiers found
  my $has_retention_time = 0;
  my $max_rank_value = $objSummary->getMaxRankValue();

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless (exists($$peptide_map{$query}) or $isPMF);

    my %query_data = ();
    my %query_data_opt = ();

    # Bug 13819: Add new fields for the scan number(s) and raw scan number(s).
    my $objQuery = new msparser::ms_inputquery($objResFile, $query);
    $query_data_opt{'MS:1000797'} = $objQuery->getScanNumbers();
    $query_data_opt{'MS:1000798'} = $objQuery->getRawScans();

    $query_data_opt{'MS:1001528'} = $query;
    $query_data{'exp_mass_to_charge'} = round_evenly($objResFile->getObservedMass($query), $massDP);

    if (my @values = ExportDat::Util::collect_indexed_MGF_values($objQuery, 'getRetentionTimes')) {
      # XXX We can only use the first value.
      $query_data{'retention_time'} = $values[0];
      ++$has_retention_time;
    }

    if ($objQuery->getIT_MODS(1)) {
        #print "            <userParam name=\"IT_MODS\" value=\"" . noXmlTag($objQuery->getIT_MODS(1)) . "\" />\n";
    }
    if ($objQuery->getINSTRUMENT(1)) {
        #print "            <userParam name=\"INSTRUMENT\" value=\"" . noXmlTag($objQuery->getINSTRUMENT(1)) . "\" />\n";
    }
    if ($objQuery->getRULES()) {
        #print "            <userParam name=\"RULES\" value=\"" . $objQuery->getRULES() . "\" />\n";
    }

    if ($pepScanTitle) {
        $query_data_opt{'MS:1000796'} = $objQuery->getStringTitle(1);
    }

    my $siteAnalysisData;
    # NB: Site analysis is not included. The mzTab 1.0 specification provides
    # a mechanism for reporting variable modification site probability, but
    # only one site at a time. Our probabilities are for the whole set of
    # observed modifications.
=pod
    if (not $isPMF) {
      if (&$query_is_sig($query)) {
        $siteAnalysisData = analyse_sites_in_query(
          $objSummary, $objParams, $queryNum, $modsByDelta, $full_unimod_file,
          $objSummary->getMaxRankValue(),
        );
      }
    }
=cut

    for my $rank (1 .. $max_rank_value) {
      my $objPeptide = $objSummary->getPeptide($query, $rank);
      my $pepstr     = $objPeptide->getPeptideStr;
      next if $pepstr eq '';
      next if not exists $$peptide_map{$query}{$rank}; # unassigned

      my $protein_map = $$peptide_map{$query}{$rank};

      my %base_data = %query_data;
      my %base_opt = %query_data_opt;

      $base_data{'PSM_ID'} = 100 * $query + $rank;
      $base_data{'sequence'} = $pepstr;
      $base_data{'unique'} = $objSummary->isPeptideUnique($query, $rank) ? 1 : 0;

      for my $i (1 .. $#$ms_run_resfiles) {
          $base_data{'search_engine_score[1]_ms_run['.$i.']'} = 'null';
      }

      if ($objPeptide->getIsFromLibrary()) {
        $base_data{'search_engine'} = '[MS, MS:1002750, NIST MSPepSearch, ]';
      } else {
        $base_data{'search_engine'} = '[MS, MS:1001207, Mascot, ]';
      }

      my (undef, $src_query_idx, $src_resfile_idx) = $$ms_run_resfiles[0]{'resfile'}->getSrcQueryAndFileIdForMultiFile($query);

      my $scanID = ExportDat::Util::determine_scan_id($indexIsQueryNumber, $$ms_run_resfiles[0]{'params'}, $objQuery, $src_query_idx);
      $base_data{'spectra_ref'} = sprintf('ms_run[%d]:%s', $src_resfile_idx, $scanID);
      # Bug 13819: Add a new field for the raw file identifier.
      $base_data{sprintf('opt_ms_run[%d]_RAWFILE', $src_resfile_idx)} = $objQuery->getRawfile();

      if ($isPMF) {
          $base_data{'best_search_engine_score[1]'} = 'null';
          $base_data{'reliability'} = 'null';
      } else {
          # NB: We export expect value as the search engine score, because the
          # plain score is misleading without the associated significance
          # thresholds.

          # Each PSM comes from one file only, and the presence of other files
          # in a combined results file does not affect PSM score. So we only
          # need to report one search_engine_score[].

          my $expect = $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $query, $objPeptide->getRank());

          if ($expect > 0) {
            $base_data{'best_search_engine_score[1]'}
              = $base_data{'search_engine_score[1]_ms_run['.$src_resfile_idx.']'}
              = round_evenly($expect, 2, 'g');
          }

          $base_data{'reliability'} = _determine_peptide_match_reliability($objSummary, $objPeptide, $OmitErrTolThresholds);
      }

      if (not $isPMF) {
          if ($pepScore) {
            if ($objPeptide->getIsFromLibrary()) {
              $base_opt{'MS:1002754'} = round_evenly($objPeptide->getIonsScore(), 0);
            } else {
              $base_opt{'MS:1001171'} = round_evenly($objPeptide->getIonsScore(), 2, 'g');
            }
          }

          if ($objPeptide->getIsFromErrorTolerant() and $OmitErrTolThresholds) {
              # No thresholds.
          } else {
              if ($pepHomThr) {
                  my $thr = $objSummary->getHomologyThreshold($query, 1 / $objSummary->getProbabilityThreshold, $objPeptide->getRank());
                  if ($thr >= 0) {
                      $base_opt{'MS:1001370'} = round_evenly($thr, 0);
                  }
              }
              if ($pepIdThr) {
                  my $dbType = $objPeptide->getIsFromLibrary()
                      ? $msparser::ms_mascotresults::DM_SPECTRAL_LIBRARY
                      : $msparser::ms_mascotresults::DM_FASTA
                      ;
                  my $thr = $objSummary->getPeptideIdentityThreshold($query, 1 / $objSummary->getProbabilityThreshold, $dbType);
                  if ($thr >= 0) {
                      $base_opt{'MS:1001371'} = round_evenly($thr, 0);
                  }
              }
          }
      }

      $base_data{'calc_mass_to_charge'} = round_evenly($objPeptide->getMrCalc(), $massDP);

      if ($pepDelta) {
          $base_opt{'MS:1001976'} = round_evenly($objPeptide->getDelta(), $massDP); # in Daltons
      }

      if ($pepNumMatch) {
          $base_opt{'MS:1001121'} = $objPeptide->getNumIonsMatched();
      }

      my ($fixed_mods, $variable_mods) = _collect_mzTab_modification_data(
        $$ms_run_resfiles[0], $objSummary, $objUmodSection, $objFullUmodFile,
        $objPeptide,
        exists($$siteAnalysisData{$rank}) ? $$siteAnalysisData{$rank} : undef,
        $vmMass, $vmString, $fixedMods, $query_mod_and_dupe_cache,
        \%mod_name_lookup, \%mod_tally,
      );

      # Note: .varmods and .fixedmods are not the final column names. We keep
      # the two types separate here, because the same source data is used for
      # both PSM and PEP rows.
      $base_data{'.varmods'} = $variable_mods;
      $base_data{'.fixedmods'} = $fixed_mods;

      my $pepCharge = $objPeptide->getCharge();
      $pepCharge = 1 if $pepCharge == 0;

      $base_data{'charge'} = $pepCharge;

      if ($export_quant) {
        my $key = msparser::ms_peptide_quant_key->new($query, $rank);
        for my $sv (@$study_variable_defs) {
          my $ratio = $objMS2Quant->getPeptideRatio($key, $$sv{'ratio_name'});
          my $idx = $$sv{'idx'};

          $base_data{'peptide_abundance_stdev_study_variable['.$idx.']'} = 'null';
          $base_data{'peptide_abundance_std_error_study_variable['.$idx.']'} = 'null';

          if ($ratio->isMissing()) {
            $base_data{'peptide_abundance_study_variable['.$idx.']'} = 'null';
          } else {
            $base_data{'peptide_abundance_study_variable['.$idx.']'} = round_evenly($ratio->getValue(), 3);
          }
        }
        my $intensities = $objMS2Quant->getComponentIntensities($key);
        # See protein abundance (intensity) reporting in
        # collectMzTabProteinData().
        if ($intensities and @$intensities) {
          for my $i (1 .. @$intensities) {
            my $assay_idx = $$comp_idx_to_assay[$src_resfile_idx][$i];
            if (defined $$intensities[$i-1]) {
              $base_data{'peptide_abundance_assay['.$assay_idx.']'} = round_evenly($$intensities[$i-1], 3);
            } else {
              # No need to set; will be reported 'null'.
            }
          }
        }
      }

      if ($percolate == 1) {
        my ($pep, $q) = $objPeptide->getPercolatorScores();
        $base_opt{'MS:1001491'} = $q;
        $base_opt{'MS:1001493'} = $pep;
      }

      if ($is_decoy) {
        $base_opt{'MS:1002217'} = 1;
      }
      if ($is_error_tolerant) {
        $base_opt{'MS:1002341'} = $objPeptide->getIsFromErrorTolerant() ? 1 : 0;
      }

      if ($pepMiss) {
	$base_opt{'missed_cleavages'} = $objPeptide->getMissedCleavages();
      }

      # And now we repeat once per accession:

      for my $protein_context (@$protein_map) {
        my %data = %base_data;
        $data{'opt'} = { %base_opt };

        $data{'accession'} = my $acc = $$protein_context{'accession'};
        $data{'database'} = my $db = $$protein_context{'database'};

        # For the PSM section:
        $data{'start'} = $$protein_context{'start'};
        $data{'end'} = $$protein_context{'end'};
        $data{'pre'} = $$protein_context{'pre'};
        $data{'post'} = $$protein_context{'post'};

        if ($pepFrame) {
          my $objProtein = $objSummary->getProtein($acc, $db);
          my $frame = $objProtein->getPeptideFrame($$protein_context{'pepnum'});
          $data{'opt'}{'MS:1001024'} = $frame > 0 ? $frame : 'null';
        }

        push @peptide_rows, \%data;

        # The mzTab 1.0 specification requires that all modifications
        # (fixed, variable, quantitation) are reported for PSMs, but only
        # variable modifications for proteins and peptides. So, we
        # completely skip processing @$fixed_mods here.

        my $start_idx = $$protein_context{'start'};

        for my $mod (@$variable_mods) {
          next if $$mod{'idx'} == 0 or $$mod{'idx'} == 2 + length($pepstr);
          my $idx = $$mod{'idx'} + $start_idx;
          push @{ $protein_mods{$db}{$acc}{$idx} }, $mod;
        }
      }
    }
  }

  return (\@peptide_rows, \%protein_mods, \%mod_tally, $has_retention_time);
}

sub _collect_mzTab_modification_data {
  my ($ms_run_resfile, $objSummary, $objUmodSection, $objFullUmodFile,
      $objPeptide, $siteAnalysisData, $vmMass, $vmString, $fixedMods,
      $query_mod_and_dupe_cache, $mod_name_lookup, $mod_tally) = @_;

  my @fixed = ();

  my @pep_str = split //, $objPeptide->getPeptideStr();

  if ($objPeptide->getIsFromLibrary()) {
    # Don't report fixed mods for library matches.
  } else {
    if ($$fixedMods{'n_term'}) {
      push @fixed, {
        idx => 0,
        name_str => $$fixedMods{'n_term'}{'name'},
        delta => $$fixedMods{'n_term'}{'delta'},
      };
    }

    for my $i (1 .. @pep_str) {
      my $residue = $pep_str[$i-1];
      if ($$fixedMods{lc $residue}) {
        push @fixed, {
          idx => $i,
          name_str => $$fixedMods{lc $residue}{'name'},
          delta => $$fixedMods{lc $residue}{'delta'},
        };
      }
    }

    if ($$fixedMods{'c_term'}) {
      push @fixed, {
        idx => scalar(@pep_str),
        name_str => $$fixedMods{'c_term'}{'name'},
        delta => $$fixedMods{'c_term'}{'delta'},
      };
    }
  }

  my @variable = ();

  my @var_mods = split //, $objPeptide->getVarModsStr();
  my @nl = split //, $objPeptide->getPrimaryNlStr();
  my @summed_mods = split //, $objPeptide->getSummedModsStr();
  my $len = 2 + @pep_str;
  my $query = $objPeptide->getQuery();
  my $rank = $objPeptide->getRank();

  my %varmod_index_map = do {
      my @indices = (0..9, 'A'..'W');
      map { $indices[$_] => $_ } 0 .. $#indices
  };

  my %modfile_lookup = ();

  for my $i (0 .. $len) {
    my @mod_ids = ();
    my @neutral_loss = ();

    if (defined $var_mods[$i] and $var_mods[$i] ne '0') {
      my $id = $var_mods[$i];
      push @mod_ids, $id;

      if (defined $nl[$i] and $nl[$i] ne '0') {
        my $nl_id = $varmod_index_map{$nl[$i]};
        my @losses = @{ $$ms_run_resfile{'params'}->getVarModsNeutralLosses($id) };
        if ($nl_id-1 < @losses) {
          my $m = $losses[$nl_id-1];
          if ($m != 0) {
            push @neutral_loss, sprintf('[MS, MS:1001524, fragment neutral loss, %+f]', $m);
          }
        }
      }
    }

    if (defined $summed_mods[$i] and $summed_mods[$i] ne '0') {
      push @mod_ids, $summed_mods[$i];
    }

    for my $j (0 .. $#mod_ids) {
      my $mod_id = $mod_ids[$j];
      my ($mod_name_str, $mod_delta);

      if ($mod_id eq 'X') {
        my $srcSection = $objSummary->getSrcSection($query, $rank);
        my $srcRank = $objSummary->getSrcRank($query, $rank);
        my $et_mods_str = $$ms_run_resfile{'resfile'}->getSectionValueStr($srcSection, "q$query\_p$srcRank\_et_mods");
        if ($et_mods_str) {
          ($mod_delta, undef, $mod_name_str) = split(/,/, $et_mods_str, 3);

          if ($mod_name_str =~ /^NA_/) {
            # We can't export these in mzTab 1.0. The peptide match will be
            # marked error tolerant in an optional column, and that's the best
            # we can do.
            undef $mod_name_str;
          }
        }
      } elsif ($mod_id eq 'Y') {
        # Do nothing; spectral library mods are processed separately.
      } else {
        $mod_delta = $$vmMass{$mod_id};
        $mod_name_str = $$vmString{$mod_id};
      }

      next if (not $mod_name_str and not $mod_delta);

      push @variable, {
        idx => $i,
        name_str => $mod_name_str,
        delta => $mod_delta,
        neutral_loss => $neutral_loss[$j],
      };
    }
  }

  if (not exists $$mod_tally{'fixed'}) {
    $$mod_tally{'fixed'} = {};
  }

  if (not exists $$mod_tally{'variable'}) {
    $$mod_tally{'variable'} = {};
  }

  for ([$$mod_tally{'fixed'}, \@fixed], [$$mod_tally{'variable'}, \@variable]) {
    my ($tally, $mods) = @$_;

    for (@$mods) {
      my $mod_name_str = $$_{'name_str'};
      my $mod_delta = $$_{'delta'};

      my $umod_title;

      if ($mod_name_str) {
        ($umod_title) = $mod_name_str =~ /(.*)\s+\((.*)\)/;
      }

      my ($final_id, $cv_element, $position, $site);

      if ($umod_title) {
        if (not exists $$mod_name_lookup{$umod_title}) {
          $$mod_name_lookup{$umod_title} = $objUmodSection->getModificationByName($umod_title);
            # could be undef if the modification doesn't exist in the file
        }

        if (defined(my $mod = $$mod_name_lookup{$umod_title})) {
          ($position, $site) = _unimod_position_and_site($mod, $mod_name_str);

          if ($mod->getRecordID() eq '') {
            # Unimod section from Mascot 2.2 or earlier; use the CHEMMOD
            # syntax.
          } else {
            $final_id = 'UNIMOD:' . $mod->getRecordID();
            $cv_element = sprintf('[UNIMOD, %s, %s,]', $final_id, $umod_title);
          }
        }
      }

      if (not defined $final_id) {
        if ($mod_delta) {
          $final_id = 'CHEMMOD:' . sprintf('%+f', $mod_delta);
          $cv_element = sprintf('[CHEMMOD, %s, %s,]', $final_id, $umod_title ? $umod_title : $mod_name_str);
        } else {
          $_ = undef;
          next;
        }
      }

      my $tally_id = join('.', $cv_element, map { $_ // '<undef>' } $position, $site);

      $_ = {
        idx => $$_{'idx'},
        neutral_loss => $$_{'neutral_loss'},
        mod_id => $final_id,
      };

      if (not exists $$tally{$tally_id}) {
        $$tally{$tally_id} = {
          cv_element => $cv_element,
          position => $position,
          site => $site,
        };
      }
    }
  }

  my $psmComponent = $msparser::ms_peptide::PSM_COMPLETE;
  my $local_mods = lookup_query_level_mods($query_mod_and_dupe_cache, $$ms_run_resfile{'resfile'}, $objPeptide, $objUmodSection);
  my @local_nl = split //, $objPeptide->getLocalModsNlStr();

  if ($local_mods) {
    my $tally = $$mod_tally{'variable'};

    for my $i (keys %{ $$local_mods{$psmComponent} }) {
      my $mod_name_str = $$local_mods{$psmComponent}{$i}{'name'};
      my ($umod_title) = $mod_name_str =~ /(.*)\s+\((.*)\)/;

      if (not exists $$mod_name_lookup{$umod_title}) {
        $$mod_name_lookup{$umod_title} = $objUmodSection->getModificationByName($umod_title);
      }

      if (not $$mod_name_lookup{$umod_title}) {
        next;
      }

      my $unimod_obj = $$mod_name_lookup{$umod_title};

      my $unimod_id = $unimod_obj->getRecordID();
      my ($position, $site) = _unimod_position_and_site($unimod_obj, $mod_name_str);

      my ($final_id, $cv_element);

      if ($unimod_id eq '') {
        # Mascot 2.2 or earlier.
        $final_id = 'CHEMMOD:' . sprintf('%+f', $$local_mods{$psmComponent}{$i}{'delta'});
        $cv_element = sprintf('[CHEMMOD, %s, %s,]', $final_id, $umod_title);
      } else {
        $final_id = 'UNIMOD:' . $unimod_id;
        $cv_element = sprintf('[UNIMOD, %s, %s,]', $final_id, $umod_title);
      }

      my $nl_value = undef;

      if (defined($local_nl[$i]) and $local_nl[$i] ne 0) {
        if (not exists $modfile_lookup{$mod_name_str}) {
          my $modfile_mod = msparser::ms_modification->new;
          if ($modfile_mod->getFromUnimod($mod_name_str, $objUmodSection)) {
            $modfile_lookup{$mod_name_str} = $modfile_mod;
          } else {
            $modfile_lookup{$mod_name_str} = undef;
          }
        }

        if ($modfile_lookup{$mod_name_str}) {
          my $nl_id = $varmod_index_map{$local_nl[$i]};
          my $nl_arr = $modfile_lookup{$mod_name_str}->getNeutralLoss($objParams->getMassType());
          if ($nl_arr and $nl_id-1 < @$nl_arr and $$nl_arr[$nl_id-1] > 0) {
            $nl_value = sprintf('[MS, MS:1001524, fragment neutral loss, %+f]', $$nl_arr[$nl_id-1]);
          }
        }
      }

      my $tally_id = join('.', $cv_element, map { $_ // '<undef>' } $position, $site);

      push @variable, {
        idx => $i,
        neutral_loss => $nl_value,
        mod_id => $final_id,
      };

      if (not exists $$tally{$tally_id}) {
        $$tally{$tally_id} = {
          cv_element => $cv_element,
          position => $position,
          site => $site,
        };
      }
    }
  }

  # Bug 12186: The mzTab 1.0 specification makes only one mention of
  # spectral libraries, not relevant here. There's nothing about whether
  # library modifications are fixed or variable. We'll assume they are
  # variable.

  if (my $lib_mods_str = $objSummary->getLibraryModString($objPeptide->getQuery, $objPeptide->getRank)) {
    my $tally = $$mod_tally{'variable'};

    for (split /,/, $lib_mods_str) {
      my ($pos, $mod_id) = split /:/;

      my $delta = $objParams->getLibraryModDelta($mod_id);
      my $umod_title = $objParams->getLibraryModName($mod_id);

      # $umod_title is typically a Unimod title, but it could be any other
      # string, empty or malformed, depending on library quality.
      if (not exists $$mod_name_lookup{$umod_title}) {
        $$mod_name_lookup{$umod_title} = $objUmodSection->getModificationByName($umod_title);

        if (not $$mod_name_lookup{$umod_title}) {
          # Try the full unimod.xml file. In a varied library with many
          # different modifications, it isn't likely the modification is in
          # the Unimod section of the file.
          $$mod_name_lookup{$umod_title} = $objFullUmodFile->getModificationByName($umod_title);
        }
      }

      if (not $$mod_name_lookup{$umod_title}) {
        next;
      }

      my $unimod_obj = $$mod_name_lookup{$umod_title};

      my $unimod_id = $unimod_obj->getRecordID();
      my ($final_id, $cv_element, $position, $site);

      if ($unimod_id eq '') {
        # ID could be missing if there is no modification with this name in
        # Unimod.
        if ($delta eq '') {
          # Bad luck -- no idea what the delta is either. The best we can do
          # is report this as an unknown modification.
          $cv_element = '[MS, MS:1001460, unknown modification, ]';
        } else {
          $final_id = 'CHEMMOD:' . sprintf('%+f', $delta);
          $cv_element = sprintf('[CHEMMOD, %s, %s,]', $final_id, $umod_title);
        }
      } else {
        $final_id = 'UNIMOD:' . $unimod_id;
        $cv_element = sprintf('[UNIMOD, %s, %s,]', $final_id, $umod_title);

        # We don't know the specificity, so the best we can do is report
        # it as Anywhere.
        $position = 'Anywhere';

        if ($pos == 0) {
          $site = 'N-term';
        } elsif ($pos == $len-1) {
          $site = 'C-term';
        } else {
          $site = $pep_str[$pos - 1];
        }
      }

      my $tally_id = join('.', $cv_element, map { $_ // '<undef>' } $position, $site);

      # Library modifications don't have neutral losses, so we're done.

      # Note that $final_id could be undefined if this is an unknown
      # modification. mzTab 1.0 doesn't allow unknowns at protein level, so
      # such terms are skipped. (The CV namespace must be one of CHEMMOD,
      # UNIMOD or MOD, and none of them contain a term for unknown
      # modification.)
      push @variable, {
        idx => $pos,
        mod_id => $final_id,
      };

      if (not exists $$tally{$tally_id}) {
        $$tally{$tally_id} = {
          cv_element => $cv_element,
          position => $position,
          site => $site,
        };
      }
    }
  }

  my @fixed_final = grep { defined($_) } @fixed;
  my @variable_final = sort { $$a{'idx'} <=> $$b{'idx'} } grep { defined($_) } @variable;

  return (\@fixed_final, \@variable_final);
}

sub _unimod_position_and_site {
  my ($unimod_obj, $mod_name_str) = @_;

  my $spec_id = $unimod_obj->findSpecGroup($mod_name_str);
  my ($position, @site);

  if ($spec_id != -1) {
    for (0 .. $unimod_obj->getNumberOfSpecificities()-1) {
      my $spec = $unimod_obj->getSpecificity($_);
      if ($spec->getSpecGroup() == $spec_id) {
        if ($spec->havePosition()) {
          # We assume the specification group has the same position for all
          # sites.
          $position = $spec->getPosition();
        }
        if ($spec->haveSite()) {
          push @site, $spec->getSite();
        }
      }
    }
  }

  return ($position, join('', @site));
}

###############################################################################
# &set_option()
# set a non-optional option
# $_[0] parameter name
# $_[1] optional parameter value; 1 by default
###############################################################################

sub set_option {

  my $param = shift;
  my $value = shift;
  $value = 1 unless defined $value;
  $urlParams{$param} = $param;
  $thisScript->param(-name=>$param, -value=>$value);

}

###############################################################################
# &_percolate()
# run Percolator, if required
###############################################################################
sub _percolate {
 my ($resfile, $use_proteins, $use_rt, $MIME_boundary) = @_;

 my $bin_createpip = join('/', MascotDat->bindir, MascotDat->bin_createpip);
 my $bin_percolator = join('/', MascotDat->bindir, MascotDat->bin_percolator);

 if (not -x $bin_createpip or not -x $bin_percolator) {
   common_subs::fatal("Unable to find Percolator executables", __LINE__, $myName, $idForFatalError, $MIME_boundary);
 }

 my %filenames = MascotResults::percolator_filenames($resfile);

 my $keepalive_pip = undef;
 my $keepalive_perc = undef;

 if (defined $MIME_boundary) {
   # Bug 13372: Use the same CSV columns as in openResultsCollectionSet().
   # Unfortunately, seconds passed is not known here.
   my $N = $resfile->getNumberOfResfiles;
   $keepalive_pip = "perc=<!--,Creating Percolator input file,%p,0,$N-->\n";
   $keepalive_perc = "perc=<!--,Percolating,%p,0,$N-->\n";
 }

 run_ms_createpip(
    undef,
    $bin_createpip,
    $resfile->getFileName,
    $filenames{'pipname'},
    $keepalive_pip,
    $use_rt
 );

 run_percolator(
    undef,
    $bin_percolator,
    MascotResults::percolator_flags($resfile, $use_rt),
    $keepalive_perc,
 );
}

###############################################################################
#  my $pepsumWillCache = maybeRedirectIfPeptideSummaryCachingNeeded(
#    $redirectIsAllowed, $objResFile, $pepsum_params, $desc, $line
#  );
#
# Check whether ms_peptidesummary or decoy cache needs to be created. If no,
# returns false. If yes, call maybeRedirectToCaching(). If it doesn't exit,
# return yes to indicate that ms_peptidesummary constructor will cache if
# called with these parameters.
################################################################################
sub maybeRedirectIfPeptideSummaryCachingNeeded {
    my ($redirectIsAllowed, $objResFile, $pepsum_params, $desc, $line) = @_;

    my $pepsumWillCache = 0;

    if (msparser::ms_peptidesummary::willCreateCache($objResFile, $pepsum_params)) {
        if (WARN_ABOUT_PEPSUM_PARAMS) {
            warn sprintf("will create %s (line %d):\n%s\n",
                $desc, $line, join("\n",
                    PeptideSummary::Util::stringify_params($pepsum_params)
                ));
        }
        maybeRedirectToCaching($redirectIsAllowed, 'build:peptidesummary');
        # No redirect, so set a flag signalling that caching will be done.
        $pepsumWillCache = 1;
    }

    # Bug 6529: Check whether decoy cache exists.
    # XXX: Same logic as in PeptideSummary::needs_caching().
    if ($objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_DECOYPEPTIDES)
           or $objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_ERRTOLDECOYPEPTIDES)) {
        if ($pepsum_params->getFlags() & $msparser::ms_mascotresults::MSRES_DECOY) {
            # Decoy report; no caching needed.
        } else {
            my $decoy_params = msparser::ms_mascotresults_params->new($pepsum_params);
            $decoy_params->setFlag($msparser::ms_mascotresults::MSRES_DECOY, 1);

            if (msparser::ms_peptidesummary::willCreateCache($objResFile, $decoy_params)) {
                if (WARN_ABOUT_PEPSUM_PARAMS) {
                    warn sprintf("will create %s (line %d):\n%s\n",
                        $desc, $line, join("\n",
                            PeptideSummary::Util::stringify_params($decoy_params)
                        ));
                }
                maybeRedirectToCaching($redirectIsAllowed, 'build:peptidesummary');
                # No redirect, so set a flag signalling that caching will be done.
                $pepsumWillCache = 1;
            }
        }
    }

    return $pepsumWillCache;
}

###############################################################################
# &maybeRedirectToCaching($redirectIsAllowed, $sub, %extra_params)
# Redirect to nph-cache_families.pl if either the resfile or peptidesummary
# objects need to be cached, or if anything needs percolation. No redirections
# are done if we're running outside CGI.
# $_[0] the subresource to use (sub:resultsfile or sub:peptidesummary)
# $_[1..n] additional parameters overwriting CGI parameters
###############################################################################
sub maybeRedirectToCaching {
  return unless $isCGI;
  my ($redirectIsAllowed, $sub, %extra_params) = @_;
  return unless $redirectIsAllowed;
  my $uri = URI->new_abs(MascotDat->uri_cache_families, Request->cgidir.'/');
  if (%extra_params) {
    # We are free to modify the global CGI state here, because we know we
    # will exit afterwards.
    for (keys %extra_params) {
      $thisScript->param($_, $extra_params{$_});
    }
  }
  # The query string is non-empty, because at least the 'file' parameter
  # has to exist to get this far.
  my $separator = $thisScript->query_string =~ m#\&# ? '&' : ';';
  $uri->query($thisScript->query_string.$separator.'sub='.common_subs::urlEscape($sub).$separator.'c.from='.common_subs::urlEscape($myName));
  print $thisScript->redirect(
    -uri => $uri->as_string,
    -status => '303 See Other',
  );
  exit;
}
